unit ConvertItUnit;

interface

{$IFDEF VCL}
uses
  Classes, Controls, StdCtrls, Menus, Forms, ExtCtrls;
{$ELSE}
uses
  SysUtils, Classes, QGraphics, QControls,
  QForms, QStdCtrls, QMenus, QDialogs, QExtCtrls;
{$ENDIF}

type
  TForm1 = class(TForm)
    ConvTypes: TListBox;
    ConvResults: TListBox;
    ConvValue: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    procedure FormShow(Sender: TObject);
    procedure ConvTypesClick(Sender: TObject);
    procedure ConvValueChange(Sender: TObject);
    procedure ConvFamiliesChange(Sender: TObject);
  private
    FButtonCount: Integer;
    function AddButton(Caption: string; ATag: Integer): TButton;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

uses
  ConvUtils, StdConvs, EuroConv;

procedure TForm1.FormShow(Sender: TObject);
var
  LFamilies: TConvFamilyArray;
  I: Integer;
  AButton: TButton;
begin
  GetConvFamilies(LFamilies);
  for I := 0 to Length(LFamilies) - 1 do
  begin
    AButton := AddButton(ConvFamilyToDescription(LFamilies[I]),
        Integer(LFamilies[I]));
    if I = 0 then ConvFamiliesChange(AButton);
  end;
end;

procedure TForm1.ConvFamiliesChange(Sender: TObject);
var
  LFamily: TConvFamily;
  LTypes: TConvTypeArray;
  I: Integer;
begin
  LFamily := TConvFamily((Sender as TButton).Tag);
  with ConvTypes, Items do
  begin
    BeginUpdate;
    Clear;
    GetConvTypes(LFamily, LTypes);
    for I := 0 to Length(LTypes) - 1 do
      AddObject(ConvTypeToDescription(LTypes[I]), TObject(LTypes[I]));
    ItemIndex := 0;
    EndUpdate;
  end;
  ConvTypesClick(Sender);
end;

procedure TForm1.ConvTypesClick(Sender: TObject);
begin
  ConvValueChange(Sender);
end;

procedure TForm1.ConvValueChange(Sender: TObject);
var
  LValue: Double;
  LBaseType, LTestType: TConvType;
  I: Integer;
begin
  with ConvResults, Items do
  try
    BeginUpdate;
    Clear;
    try
      LValue := StrToFloatDef(ConvValue.Text, 0);
      if ConvTypes.ItemIndex <> -1 then
      begin
        LBaseType := TConvType(ConvTypes.Items.Objects[ConvTypes.ItemIndex]);
        for I := 0 to ConvTypes.Items.Count - 1 do
        begin
          LTestType := TConvType(ConvTypes.Items.Objects[I]);
          Add(Format('%n %s', [Convert(LValue, LBaseType, LTestType),
                               ConvTypeToDescription(LTestType)]));
        end;
      end
      else
        Add('No base type');
    except
      Add('Cannot parse value');
    end;
  finally
    EndUpdate;
  end;
end;

function TForm1.AddButton(Caption: string; ATag: Integer): TButton;
var
  Button: TButton;
  ATop: Integer;
begin
  Button := TButton.Create(Self);
  Button.Parent := Self;
  ATop := 16 + FButtonCount * 32;
  Button.SetBounds(432, ATop, 105, 25);
  Button.Caption := '&' + Caption;
  Button.Tag := ATag;
  Button.OnClick := ConvFamiliesChange;
  Inc(FButtonCount);
  Result := Button;
end;

end.
