unit Chatmain;

interface

uses
  SysUtils, Classes, QControls, QStdCtrls, QExtCtrls, QForms, Sockets,
  Threads, QTypes, SocketUtils;

type
  TForm1 = class(TForm)
    TcpClient1: TTcpClient;
    TcpServer1: TTcpServer;
    lblRemoteHost: TLabel;
    edtRemoteHost: TEdit;
    lblRemotePort: TLabel;
    edtRemotePort: TEdit;
    lblLocalPort: TLabel;
    edtLocalPort: TEdit;
    btnActivateServer: TButton;
    lbxRecv: TListBox;
    edtSend: TEdit;
    lblSend: TLabel;
    btnSend: TButton;
    btnClear: TButton;
    procedure btnSendClick(Sender: TObject);
    procedure TcpServer1Accept(Sender: TObject; ClientSocket: TCustomIpClient);
    procedure btnActivateServerClick(Sender: TObject);
    procedure btnClearClick(Sender: TObject);
    procedure lbxRecvDblClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure TcpClient1Error(Sender: TObject; SocketError: Integer);
  private
    procedure ApplicationIdle(Sender: TObject; var Done: Boolean);
    { Private declarations }
  public
    { Public declarations }
  end;

  // you must create your own thread to synch
  // writing to a gui component
  TClientDataThread = class(TThread)
  private
  public
    ListBuffer :TStringList;
    TargetList :TStrings;
    procedure synchAddDataToControl;
    constructor Create(CreateSuspended: Boolean);
    procedure Execute; override;
    procedure Terminate;
  end;

var
  Form1: TForm1;
  //DataThread: TClientDataThread;

implementation

{$R *.xfm}

//------------- TClientDataThread impl -----------------------------------------
constructor TClientDataThread.Create(CreateSuspended: Boolean);
begin
  inherited Create(CreateSuspended);
  FreeOnTerminate := True;
  ListBuffer := TStringList.Create;
end;

procedure TClientDataThread.Terminate;
begin
  ListBuffer.Free;
  inherited;
end;

procedure TClientDataThread.Execute;
begin
  Synchronize(synchAddDataToControl);
end;

procedure TClientDataThread.synchAddDataToControl;
begin
 TargetList.AddStrings(ListBuffer);
end;
//------------- end TClientDataThread impl -------------------------------------


procedure TForm1.btnActivateServerClick(Sender: TObject);
begin
  TcpServer1.LocalPort := edtLocalPort.Text;
  TcpServer1.Active := True;
end;

procedure TForm1.btnSendClick(Sender: TObject);
begin
  TcpClient1.RemoteHost := edtRemoteHost.Text;
  TcpClient1.RemotePort := edtRemotePort.Text;
  try
    if TcpClient1.Connect then
      TcpClient1.Sendln(edtSend.Text);
  finally
    TcpClient1.Disconnect;
  end;
end;

procedure TForm1.TcpServer1Accept(Sender: TObject;
  ClientSocket: TCustomIpClient);
var
  s: string;
  DataThread: TClientDataThread;
begin
  // create thread
  DataThread:= TClientDataThread.Create(True);
  // set the TagetList to the gui list that you
  // wish to synch with.
  DataThread.TargetList := lbxRecv.Items;

  // Load the Threads ListBuffer
  DataThread.ListBuffer.Add('*** Connection Accepted ***');
  DataThread.ListBuffer.Add('Remote Host: '
     + ClientSocket.LookupHostName(ClientSocket.RemoteHost)
     + ' (' + ClientSocket.RemoteHost + ')');
  DataThread.ListBuffer.Add('===== Begin message =====');
  s := ClientSocket.Receiveln;
  while s <> '' do
  begin
    DataThread.ListBuffer.Add(s);
    s := ClientSocket.Receiveln;
  end;
  DataThread.ListBuffer.Add('===== End of message =====');

  // Call Resume which will execute and synch the
  // ListBuffer with the TargetList
  DataThread.Resume;
end;

procedure TForm1.btnClearClick(Sender: TObject);
begin
  lbxRecv.Clear;
end;

procedure TForm1.lbxRecvDblClick(Sender: TObject);
begin
  btnClear.Click;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  Application.OnIdle := ApplicationIdle;
  btnActivateServerClick(Self);
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  TcpServer1.Active := False;
end;

procedure TForm1.TcpClient1Error(Sender: TObject; SocketError: Integer);
begin
  lbxRecv.Items.Add(Format('TcpClient Error: %d', [SocketError]));
  lbxRecv.Items.Add(SocketErrorDesc(SocketError));
end;

procedure TForm1.ApplicationIdle(Sender: TObject; var Done: Boolean);
begin
  // periodical network update is essencial for WDOSX mode applications
  Sockets.UpdateNetwork;
end;

end.
