unit Dlg1Main;

interface

uses
  SysUtils, Types, Classes, Variants, QTypes, QGraphics, QControls, QForms, 
  QDialogs, QStdCtrls;

type
  TForm1 = class(TForm)
    edtDefaultExt: TEdit;
    btnOpen: TButton;
    btnSave: TButton;
    SaveDialog1: TSaveDialog;
    OpenDialog1: TOpenDialog;
    btnInitialDir: TButton;
    edtInitialDir: TEdit;
    lbxEventsLog: TListBox;
    edtFileName: TEdit;
    lblInitialDir: TLabel;
    lblFileName: TLabel;
    lblDefaultExt: TLabel;
    edtTitleOpen: TEdit;
    edtTitleSave: TEdit;
    lblTitleOpen: TLabel;
    lblTitleSave: TLabel;
    cbxPathMustExist: TCheckBox;
    cbxFileMustExist: TCheckBox;
    cbxOverwritePrompt: TCheckBox;
    btnClear: TButton;
    lblEventsLog: TLabel;
    procedure btnOpenClick(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
    procedure btnInitialDirClick(Sender: TObject);
    procedure DialogCanClose(Sender: TObject; var CanClose: Boolean);
    procedure lbxEventsLogClear(Sender: TObject);
    procedure DialogShow(Sender: TObject);
    procedure DialogClose(Sender: TObject);
  private
    procedure AddDialogHistory(Dialog: TOpenDialog; FileName: string);
    procedure SetProperties(Dialog: TOpenDialog);
    { Private-Deklarationen }
  public
    { Public-Deklarationen }
  end;

var
  Form1: TForm1;

implementation

{$R *.xfm}

procedure TForm1.btnOpenClick(Sender: TObject);
var
  FileName: string;
begin
  SetProperties(OpenDialog1);
  OpenDialog1.Title := edtTitleOpen.Text;

  if OpenDialog1.Execute then
  begin
    FileName := OpenDialog1.FileName;
    ShowMessage('OK ' + FileName);
    AddDialogHistory(OpenDialog1, FileName);
  end
  else
    ShowMessage('Cancel');
end;

procedure TForm1.btnSaveClick(Sender: TObject);
var
  FileName: string;
begin
  SetProperties(SaveDialog1);
  SaveDialog1.Title := edtTitleSave.Text;

  if SaveDialog1.Execute then
  begin
    FileName := SaveDialog1.FileName;
    ShowMessage('OK ' + FileName);
    AddDialogHistory(SaveDialog1, FileName);
  end
  else
    ShowMessage('Cancel');
end;

procedure TForm1.btnInitialDirClick(Sender: TObject);
var
  Root: string;
  Directory: string;
begin
  Root := edtInitialDir.Text;
  if SelectDirectory('Select directory', Root, Directory) then
    edtInitialDir.Text := Directory;
end;

procedure TForm1.AddDialogHistory(Dialog: TOpenDialog; FileName: string);
var
  Strings: TStrings;
begin
  FileName := ExtractFileName(FileName);
  if Dialog.HistoryList.IndexOf(FileName) < 0 then
  begin
    Strings := TStringList.Create;
    try
      Strings.Assign(Dialog.HistoryList);
      Strings.Add(FileName);
      Dialog.HistoryList := Strings;
    finally
      Strings.Free;
    end;
  end;
end;

procedure TForm1.lbxEventsLogClear(Sender: TObject);
begin
  lbxEventsLog.Clear;
end;

procedure TForm1.SetProperties(Dialog: TOpenDialog);
begin
  Dialog.InitialDir := edtInitialDir.Text;
  Dialog.FileName := edtFileName.Text;
  Dialog.DefaultExt := edtDefaultExt.Text;

  if cbxPathMustExist.Checked then
    Dialog.Options := Dialog.Options + [ofPathMustExist]
  else
    Dialog.Options := Dialog.Options - [ofPathMustExist];

  if cbxFileMustExist.Checked then
    Dialog.Options := Dialog.Options + [ofFileMustExist]
  else
    Dialog.Options := Dialog.Options - [ofFileMustExist];

  if cbxOverwritePrompt.Checked then
    Dialog.Options := Dialog.Options + [ofOverwritePrompt]
  else
    Dialog.Options := Dialog.Options - [ofOverwritePrompt];
end;

procedure TForm1.DialogCanClose(Sender: TObject;
  var CanClose: Boolean);
begin
  lbxEventsLog.Items.Add('OnCanClose');
end;

procedure TForm1.DialogShow(Sender: TObject);
begin
  lbxEventsLog.Items.Add('OnShow');
end;

procedure TForm1.DialogClose(Sender: TObject);
begin
  lbxEventsLog.Items.Add('OnClose');
end;

end.
