program Client;

uses
  Crt, Windows, SysUtils, WDosSock, Threads;

label
  Final;

var
  C: Char;
  ToRead: DWORD;
  ServerPort: Integer;
  SockAddr: TSockAddr;
  ServerIP: u_long;
  HSocket: Integer;
  ConnTimeout: Integer;
  I: Integer;
  H: PHostEnt;

begin
  // Check Params
  ServerPort := 0;
  try
    if (ParamCount <> 2) then
      raise exception.Create('');
    ServerPort := StrToInt(Paramstr(2));
    if (ServerPort < 0) or (ServerPort > 65535) then
      raise exception.Create('');
  except
    WriteLn('Usage:   Client [IP | HostName] [Port]');
    WriteLn('Example: Client smtp.fibertel.com.ar 25');
    Halt(1);
  end;

  // Resolve Server IP
  ServerIP := inet_addr(PChar(ParamStr(1)));
  if ServerIP = INADDR_ANY then
  begin
    H := gethostbyname(PChar(ParamStr(1)));
    // Doesn't seem to be an IP. Let's try to resolve it with DNS.
    ServerIP := u_long(PInAddr(H^.h_addr^)^);
    if ServerIP = INADDR_NONE then
    begin
      WriteLn('Could not resolve host name "' + ParamStr(1) + '".');
      goto Final;
    end;
  end;

  // Create socket to connect to the Server
  HSocket := socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
  if HSocket <= 0 then
  begin
    WriteLn('Error creating socket: ', WSAGetLastError);
    goto Final;
  end;

  // Set timeout to 10 seconds
  ConnTimeout := 10000;
  SetSockOpt(HSocket, SOL_SOCKET, SO_CONNTIMEOUT, ConnTimeout,
    SizeOf(ConnTimeout));

  // Establish connection
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := ServerIP;
  SockAddr.sin_port := Swap(ServerPort);
  Write('Connecting to ' + inet_ntoa(SockAddr.sin_addr) + ':'
      + IntToStr(ServerPort) + '... ');
  if Connect(HSocket, SockAddr, SizeOf(SockAddr)) <> 0 then
  begin
    // Could not connect
    WriteLn('Error ' + IntToStr(WSAGetLastError));
    CloseSocket(HSocket);
    goto Final;
  end;
  WriteLn('Ok');

  // Loop until we are disconnected. When the user types something, we send it
  // to the server. If the server sends something, we show it on screen.
  while ioctlsocket(HSocket, FIONREAD, @ToRead) = 0 do
  begin
    // Any data to read from the server?
    if ToRead = 0 then
    begin
      // No data. At least, check to see if the server closed
      // the connection.
      if recv(HSocket, c, 0, 0) < 0 then
        Break;
    end
    else
    begin
      // Yes, there is!. Print it on screen.
      for i := 1 to ToRead do
      begin
        recv(HSocket, c, 1, 0);
        Write(c);
      end;
    end;
    // Something to send?
    if KeyPressed then
    begin
      c := ReadKey;
      if c = #27 then
        break;
      if c = #13 then
      begin
        c := #10;
        send(HSocket, c, 1, 0);
        c := #13;
      end;
      send(HSocket, c, 1, 0);
    end;
    Sleep(0);
  end;
  CloseSocket(HSocket);

Final:

end.

