program WebDemo;

{$APPTYPE CONSOLE}

uses
  Crt,
  SysUtils,
{$IFDEF WIN}
  WinSock,
{$ELSE}
  WDosSock,
  WDosInterrupts,
{$ENDIF}
  Threads,
{$IFNDEF WIN}
  SocketInfos in '..\common\SocketInfos.pas',
{$ENDIF}
  WebSrv in 'WEBSRV.PAS',
  WebFS in 'WEBFS.PAS';

type
  ESocketError = class(Exception)
  end;

var
  MustRestart: Boolean;
  QuitServer: Boolean;

procedure HomePage(Server: TWebServer; Request: TWebRequest; Response:
  TWebResponse);
begin
  Response.Content :=
    '<HTML><TITLE>Dynamic Page</TITLE>' +
    '<H1>This is a Dynamic Page</H1>' +
    'Today it''s ' + DateTimeToStr(Now) + '<br>' +
    '</HTML>';
end;

procedure WebAdmin(Server: TWebServer; Request: TWebRequest; Response:
  TWebResponse);
begin
  Response.Content :=
    '<HTML><TITLE>Web Administration</TITLE>' +
    '<H1>Web Administration</H1>' +
    '<a href="/restart">Restart the Server</a><br />' +
    '<a href="/quit">Quit the Server</a>' +
    '</HTML>';
end;

procedure DoReset(Server: TWebServer; Request: TWebRequest; Response:
  TWebResponse);
begin
  QuitServer := True;
  MustRestart := True;
  Response.Content :=
    '<HTML><TITLE>Restart Server</TITLE>' +
    '<H1>Restart Server</H1>' +
    'Server is restarting...<BR><BR>' +
    '<a href="/">Go back to Home Page</a>' +
    '</HTML>';
end;

procedure DoQuit(Server: TWebServer; Request: TWebRequest; Response:
  TWebResponse);
begin
  QuitServer := True;
  MustRestart := False;
  Response.Content :=
    '<HTML><TITLE>Quit Server</TITLE>' +
    '<H1>Quit Server</H1>' +
    'HTTP Server is ending... <BR><BR>' +
    '<a href="/">Go back to Home Page</a>' +
    '</HTML>';
end;

procedure Run;
var
  Server: TWebServer;
  LocalPort: Word;
begin
  ClrScr;
{$IFNDEF WIN}
  // display all current socket infos
  DisplaySocketInfos;
{$ENDIF}
  // Bring up a Web Server
  Write('Starting Web Server... ');
  Server := TWebServer.Create;
  try
{$IFDEF WIN}
    LocalPort := 8080;
{$ELSE}
    LocalPort := 80;
{$ENDIF}
    if not Server.Start(INADDR_ANY, Swap(LocalPort), '.\',
        '/pages/index.htm') then
    begin
      WriteLn('Error ', WSAGetLastError, ' starting Web Server.');
      Halt(1);
    end;
    WriteLn('Started.');
    Writeln('Port: ', LocalPort);
    Writeln('Press <Esc> to quit ...');

    // Add Web Server Actions
    Server.AddAction('/index', @HomePage);
    Server.AddAction('/webadmin', @WebAdmin);
    Server.AddAction('/restart', @DoReset);
    Server.AddAction('/quit', @DoQuit);
    StartWebFS(Server, '/filesystem');

    // Loop until user presses <Esc>
    QuitServer := False;
    repeat
{$IFNDEF WIN}
      UpdateNetwork;
{$ENDIF}
      Sleep(0);
    until QuitServer or (KeyPressed and (ReadKey = #27));
  finally
    Server.Stop;
    Sleep(500);
    // Close everything
    Server.Free;
  end;
end;

{$IFDEF WIN}
var
  WSAData: TWSAData;

procedure Startup;
var
  ErrorCode: Integer;
begin
  ErrorCode := WSAStartup($0101, WSAData);
  if ErrorCode <> 0 then
    raise ESocketError.Create('WSAStartup');
end;

procedure Cleanup;
var
  ErrorCode: Integer;
begin
  ErrorCode := WSACleanup;
  if ErrorCode <> 0 then
    raise ESocketError.Create('WSACleanup');
end;
{$ENDIF}

begin
  try
    MustRestart := False;
{$IFDEF WIN}
    Startup;
    try
{$ENDIF}
      Run;
{$IFDEF WIN}
    finally
      Cleanup;
    end;
{$ENDIF}
{$IFNDEF WIN}
    if MustRestart then Reboot(True);
{$ENDIF}
  except
    on e: Exception do
      Writeln(e.Message);
  end;
end.

