unit WebSrv;

interface

uses
  Windows, SysUtils, Classes,
{$IFDEF WIN}
  WinSock,
{$ELSE}
  WDosSock,
  WDosNetwork,
  WDosClock,
{$ENDIF}
  Threads;

type
  { 4 Byte IP address}
  PIPAddress = ^TIPAddress;
  TIPAddress = Cardinal;


  TWebMethodType = (wmNone, wmGet, wmPost);

  TWebServer = class;

  TWebRequest = class
  private
    FSocket: LongInt;
    FURL: AnsiString;
    FQuery: AnsiString;
    FServer: TWebServer;
    FHeader: AnsiString;
    FMethod: AnsiString;
    FContent: AnsiString;
    FPathInfo: AnsiString;
    FMethodType: TWebMethodType;
    FRootPathInfo: string;
    function ReadHeader: Boolean;
    function GetContentLength: Integer;
    function GetContentType: AnsiString;
    function GetQueryFields(Value: AnsiString): AnsiString;
    function GetHeaderFields(Value: AnsiString): AnsiString;
    function GetRequestContentType: string;
  public
    constructor Create(Server: TWebServer; HSocket: Integer);
    function ReadContent: Boolean;
    function SaveContent(Stream: TStream): Boolean;
    property Content: AnsiString read FContent;
    property ContentType: AnsiString read GetContentType;
    property ContentLength: Integer read GetContentLength;
    property QueryFields[Name: AnsiString]: AnsiString read GetQueryFields;
    property HeaderFields[Name: AnsiString]: AnsiString read GetHeaderFields;
    property PathInfo: AnsiString read FPathInfo;
    property Query: AnsiString read FQuery;
    property URL: AnsiString read FURL;
    property Method: AnsiString read FMethod;
    property MethodType: TWebMethodType read FMethodType;
    property RequestContentType: string read GetRequestContentType;
    property RootPathInfo: string read FRootPathInfo;
  end;

  TWebResponse = class
  private
    FSent: Boolean;
    FSocket: LongInt;
    FServer: TWebServer;
    FStatusCode: Integer;
    FContent: AnsiString;
    FHeaderSent: Boolean;
    FLocation: AnsiString;
    FContentLength: Integer;
    FContentType: AnsiString;
  protected
    procedure SendHeader;
  public
    constructor Create(Server: TWebServer; HSocket: Integer);
    property Sent: Boolean read FSent;
    function SendResponse: Boolean;
    function SendString(s: Ansistring): Boolean;
    function SendFile(FileName: Ansistring): Boolean;
    procedure Redirect(URL: AnsiString);
    property Content: AnsiString read FContent write FContent;
    property ContentType: AnsiString read FContentType write FContentType;
    property ContentLength: Integer read FContentLength write FContentLength;
    property StatusCode: Integer read FStatusCode write FStatusCode;
    property Location: AnsiString read FLocation write FLocation;
  end;

  TWebProc = procedure(Server: TWebServer; Request: TWebRequest; Response:
    TWebResponse);

  TActionStruct = record
    Proc: TWebProc;
    PathInfo: string;
  end;

  TWebConnection = record
    HSocket: LongInt;
    HTask: Cardinal;
    Server: TWebServer;
  end;

  TWebServer = class
  private
    FSocket: LongInt;
    FWorkTask: Cardinal;
    FRootPath: AnsiString;
    FDefaultPage: AnsiString;
    FActions: array[1..100] of TActionStruct;
    FConnections: array[1..100] of TWebConnection;
  protected
    procedure Work; virtual;
    function AddConnection(HSocket: LongInt): Boolean; virtual;
    procedure ProcessAction(Request: TWebRequest; Response: TWebResponse);
      virtual;
  public
    constructor Create;
    destructor Destroy; override;
    function Start(Address: TIPAddress; Port: WORD; RootPath, DefaultPage:
      AnsiString): Boolean;
    function AddAction(Action: AnsiString; Proc: TWebProc): Boolean;
    procedure Stop;
    property RootPath: AnsiString read FRootPath;
    property DefaultPage: AnsiString read FDefaultPage;
  end;

function ParseParamByNumber(Texto: AnsiString; NumParam: Integer; const
    Separador: AnsiString): string;
function GoodDir(s: string): string;

implementation

type
  TContType = record
    Ext: AnsiString;
    ContentType: AnsiString;
  end;

const
  CR = #13;
  LF = #10;
  CRLF = CR + LF;

  ContTypes: array[1..10] of TContType = (
    (Ext: '.TXT'; ContentType: 'text/plain'),
    (Ext: '.DOC'; ContentType: 'application/msword'),
    (Ext: '.BAT'; ContentType: 'text/plain'),
    (Ext: '.INI'; ContentType: 'text/plain'),
    (Ext: '.HTM'; ContentType: 'text/html'),
    (Ext: '.EXE'; ContentType: 'application/octet-stream'),
    (Ext: '.COM'; ContentType: 'application/octet-stream'),
    (Ext: '.GIF'; ContentType: 'image/gif'),
    (Ext: '.JPG'; ContentType: 'image/jpeg'),
    (Ext: '.BMP'; ContentType: 'image/bmp')
    );

var
  ServerList: TList;

{
*************************** common functions ***********************************
}

function ParseParamByNumber(Texto: AnsiString; NumParam: Integer; const
  Separador: AnsiString): string;
var
  n, i: Integer;
begin
  n := 1;
  Result := '';
  while n <> NumParam do
  begin
    i := Pos(Separador, Texto);
    if i = 0 then
      Exit;
    Texto := TrimLeft(Copy(Texto, i + Length(Separador), Length(Texto)));
    Inc(n);
  end;
  i := Pos(Separador, Texto);
  if i > 0 then
    Texto := Copy(Texto, 1, i - 1);
  Result := Trim(Texto);
end;

const
  Numeros: set of char = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.',
    '-'];

function IVal(s: string): LongInt;
var
  i: Byte;
  Code: integer;
  Resu: LongInt;
  ss: string;
begin
  s := Trim(s);
  ss := '';
  for i := 1 to Length(s) do
  begin
    if s[i] in numeros then
      ss := ss + s[i]
    else
      break;
  end;
  Val(ss, resu, code);
  IVal := resu;
end;

function StrRight(s: string; cuantos: Integer): string;
begin
  StrRight := Copy(s, Length(s) - Cuantos + 1, 255);
end;

function GoodDir(s: string): string;
begin
  if s = '' then
    GoodDir := ''
  else
  begin
    s := Trim(s);
    if StrRight(s, 1) = '\' then
      GoodDir := s
    else
      GoodDir := s + '\';
  end;
end;

function GetTimeCounter: Integer;
begin
{$IFDEF WIN}
  Result := GetTickCount;
{$ELSE}
  Result := GetMSCounter;
{$ENDIF}
end;

{
*************************** WebSrv functions ***********************************
}

procedure WebConnectionTask(Param: Pointer); forward;
procedure WebServerTask(Param: Pointer); forward;

function Select(ASocket: TSocket; ReadReady, WriteReady,
    ExceptFlag: PBoolean; TimeOut: Integer): Boolean;
var
  ReadFds: TFDset;
  ReadFdsptr: PFDset;
  WriteFds: TFDset;
  WriteFdsptr: PFDset;
  ExceptFds: TFDset;
  ExceptFdsptr: PFDset;
  tv: timeval;
  Timeptr: PTimeval;
begin
  if Assigned(ReadReady) then
  begin
    ReadFdsptr := @ReadFds;
    FD_ZERO(ReadFds);
    FD_SET(ASocket, ReadFds);
  end
  else
    ReadFdsptr := nil;
  if Assigned(WriteReady) then
  begin
    WriteFdsptr := @WriteFds;
    FD_ZERO(WriteFds);
    FD_SET(ASocket, WriteFds);
  end
  else
    WriteFdsptr := nil;
  if Assigned(ExceptFlag) then
  begin
    ExceptFdsptr := @ExceptFds;
    FD_ZERO(ExceptFds);
    FD_SET(ASocket, ExceptFds);
  end
  else
    ExceptFdsptr := nil;
  if TimeOut >= 0 then
  begin
    tv.tv_sec := TimeOut div 1000;
    tv.tv_usec :=  1000 * (TimeOut mod 1000);
    Timeptr := @tv;
  end
  else
    Timeptr := nil;
  try
{$IFDEF WIN}
    Result := WinSock.select(ASocket + 1, ReadFdsptr, WriteFdsptr,
        ExceptFdsptr, Timeptr) > 0;
{$ELSE}
    Result := WDosSock.select(ASocket + 1, ReadFdsptr, WriteFdsptr,
        ExceptFdsptr, Timeptr) > 0;
{$ENDIF}
  except
    Result := False;
  end;
  if Assigned(ReadReady) then
    ReadReady^ := FD_ISSET(ASocket, ReadFds);
  if Assigned(WriteReady) then
    WriteReady^ := FD_ISSET(ASocket, WriteFds);
  if Assigned(ExceptFlag) then
    ExceptFlag^ := FD_ISSET(ASocket, ExceptFds);
end;


function Web_GetParam(const Params, ParamName, Default: AnsiString): AnsiString;
var
  i: Integer;
  S: AnsiString;
begin
  Result := Default;
  for i := 1 to 1000 do
  begin
    S := ParseParamByNumber(Params, i, '&');
    if S = '' then
      Exit;
    if UpperCase(ParseParamByNumber(S, 1, '=')) = UpperCase(ParamName) then
    begin
      Result := ParseParamByNumber(S, 2, '=');
      Exit;
    end;
  end;
end;

function Web_GetContentType(FileName: AnsiString): AnsiString;
var
  i: Integer;
  Ext: AnsiString;
begin
  Ext := UpperCase(ExtractFileExt(FileName));
  for i := Low(ContTypes) to High(ContTypes) do
  begin
    if ContTypes[i].Ext = Ext then
    begin
      Result := ContTypes[i].ContentType;
      Exit;
    end;
  end;
  Result := 'application/octet-stream';
end;

{ TBebRequest }

constructor TWebRequest.Create(Server: TWebServer; HSocket: Integer);
begin
  inherited Create;
  FHeader := '';
  FContent := '';
  FServer := Server;
  FSocket := HSocket;
  ReadHeader;
end;

function TWebRequest.GetQueryFields(Value: AnsiString): AnsiString;
begin
  Result := Web_GetParam(FQuery, Value, '');
end;

function TWebRequest.GetHeaderFields(Value: AnsiString): AnsiString;
var
  i: Integer;
  HeaderLine: AnsiString;
begin
  Value := UpperCase(Value);
  for i := 1 to 1000 do
  begin
    HeaderLine := ParseParamByNumber(FHeader, i, CRLF);
    if HeaderLine = '' then
      break;
    if UpperCase(ParseParamByNumber(HeaderLine, 1, ':')) = Value then
    begin
      Result := ParseParamByNumber(HeaderLine, 2, ':');
      Exit;
    end;
  end;
  Result := '';
end;

function TWebRequest.GetContentType: AnsiString;
begin
  Result := GetHeaderFields('Content-Type');
end;

function TWebRequest.GetContentLength: Integer;
begin
  Result := IVal(GetHeaderFields('Content-Length'));
end;

function TWebRequest.SaveContent(Stream: TStream): Boolean;
var
  Timeout: Int64;
  Buffer: string;
  Len: Integer;
  Offset: Integer;
begin
  // Receive Content, with a timeout of 10 seconds between chunks
  Offset := 0;
  FContent := '';
  Timeout := GetTimeCounter + 10000;
  while (Offset < ContentLength) and (GetTimeCounter < Timeout)
    and (recv(FSocket, buffer, 0, 0) = 0) do
  begin
{$IFDEF WIN}
    IOCtlSocket(FSocket, FIONREAD, Len);
{$ELSE}
    IOCtlSocket(FSocket, FIONREAD, @Len);
{$ENDIF}
    if Len > 0 then
    begin
      SetLength(Buffer, Len);
      Len := recv(FSocket, Buffer[1], Len, 0);
      Stream.WriteBuffer(Buffer[1], Len);
      Inc(Offset, Len);
      Timeout := GetTimeCounter + 10000;
    end;
    Yield;
  end;
  // Done
  Result := (Offset = ContentLength);
end;

function TWebRequest.ReadContent: Boolean;
var
  Stream: TStream;
begin
  // If we've already read the content, we're done.
  if FContent <> '' then
  begin
    Result := True;
    Exit;
  end;
  Stream := TMemoryStream.Create;
  try
    Result := SaveContent(Stream);
    if Result then
    begin
      Stream.Position := 0;
      SetLength(FContent, Stream.Size);
      Stream.Read(FContent[1], Stream.Size);
    end;
  finally
    Stream.Free;
  end;
end;

function TWebRequest.ReadHeader: Boolean;
var
  c: char;
  L: LongInt;
  i: Integer;
  Timeout: Int64;
  Line: AnsiString;
begin
  // Get Header by reading from the socket.
 //	WriteLn('Reading header');
  Result := False;
  FHeader := '';
  Timeout := GetTimeCounter + 5000;
  while not Result and (GetTimeCounter < Timeout)
    and (recv(FSocket, c, 0, 0) = 0) do
  begin
{$IFDEF WIN}
    ioctlsocket(FSocket, FIONREAD, L);
{$ELSE}
    ioctlsocket(FSocket, FIONREAD, @L);
{$ENDIF}
    if L > 0 then
    begin
      recv(FSocket, c, 1, 0);
      FHeader := FHeader + c;
      Result := Pos(CRLF + CRLF, FHeader) <> 0;
      Timeout := GetTimeCounter + 5000;
    end
    else
    begin
      Yield;
    end;
  end;
  if not Result then
    Exit;
  //    WriteLn('Content-Type  : ', ContentType);
  //    WriteLn('Content-Length: ', IntToStr(ContentLength));

      // Extract basic information
  Line := ParseParamByNumber(FHeader, 1, CRLF);
  FMethod := ParseParamByNumber(Line, 1, ' ');
  FURL := ParseParamByNumber(Line, 2, ' ');

  // Check method type
  if UpperCase(FMethod) = 'GET' then
    FMethodType := wmGet
  else if UpperCase(FMethod) = 'POST' then
    FMethodType := wmPost
  else
  begin
    FMethodType := wmNone;
    Exit;
  end;

  // Check URL, PathInfo and Query
  i := Pos('?', FURL);
  if i = 0 then
  begin
    FQuery := '';
    FPathInfo := FURL;
  end
  else
  begin
    FQuery := Copy(FURL, i + 1, Length(FURL));
    FPathInfo := Copy(FURL, 1, i - 1);
  end;
  //	WriteLn(Method, ',', URL, ',', PathInfo, ',', Query);

      // Check for default Page
  if FPathInfo = '/' then
    FPathInfo := FServer.DefaultPage;

  // extract RootPathInfo
  FRootPathInfo := FPathInfo;
  if Length(FRootPathInfo) > 1 then
  begin
    Delete(FRootPathInfo, 1, 1);
    I := Pos('/', FRootPathInfo);
    if I > 0 then
      FRootPathInfo := Copy(FRootPathInfo, 1, I - 1);
  end;
  FRootPathInfo := '/' + FRootPathInfo;
end;

function TWebRequest.GetRequestContentType: string;
begin
  Result := ContentType;
end;

{ TWebResponse }

constructor TWebResponse.Create(Server: TWebServer; HSocket: Integer);
begin
  inherited Create;
  FSent := False;
  FContent := '';
  FServer := Server;
  FSocket := HSocket;
  FLocation := '';
  FStatusCode := 200;
  FHeaderSent := False;
  FContentLength := -1;
  FContentType := 'text/html';
end;

procedure TWebResponse.SendHeader;
var
  Str: AnsiString;
begin
  Str :=
    'HTTP/1.1 ' + IntToStr(StatusCode) + ' ' + CRLF +
    'Server: FreePascal Free Server/1.1 by Diego De Marco' + CRLF +
    'Content-Type: ' + ContentType + CRLF;
  if FContentLength >= 0 then
    Str := Str + 'Content-Length: ' +
      IntToStr(FContentLength) + CRLF;
  if FLocation <> '' then
    Str := Str + 'Location: ' + FLocation + CRLF;
  Str := Str + CRLF;
  send(FSocket, Str[1], Length(Str), 0);
  FHeaderSent := True;
end;

function TWebResponse.SendResponse: Boolean;
begin
  if FContent = '' then
  begin
    if not FHeaderSent then
      SendHeader;
    Result := True;
  end
  else
  begin
    if not FHeaderSent then
      FContentLength := Length(FContent);
    Result := SendString(Content);
  end;
end;

function TWebResponse.SendString(s: Ansistring): Boolean;
begin
  if Length(s) = 0 then
    Result := True
  else
  begin
    if not FHeaderSent then
      SendHeader;
    Result := send(FSocket, s[1], Length(s), 0) = Length(s);
  end;
end;

function TWebResponse.SendFile(FileName: Ansistring): Boolean;
const
  CHUNK_SIZE = 2048;
var
  H: Integer;
  Buffer: AnsiString;
begin
  if FHeaderSent then
  begin
    Result := False;
    Exit;
  end;
  H := FileOpen(FileName, fmOpenRead);
  if H < 0 then
    Result := False
  else
  begin
    ContentType := Web_GetContentType(FileName);
    ContentLength := FileSeek(H, 0, 2);
    SendHeader;
    FileSeek(H, 0, 0);
    Result := False;
    repeat
      SetLength(Buffer, CHUNK_SIZE);
      SetLength(Buffer, FileRead(H, Buffer[1], Length(Buffer)));
      if Buffer = '' then
      begin
        Result := True;
        Break;
      end;
      if send(FSocket, Buffer[1], Length(Buffer), 0) <> Length(Buffer) then
        Break;
    until False;
    FileClose(H);
  end;
end;

procedure TWebResponse.Redirect(URL: AnsiString);
begin
  StatusCode := 302;
  Location := URL;
end;

{ TWebServer }

constructor TWebServer.Create;
var
  i: Integer;
begin
  inherited Create;
  FRootPath := '';
  FDefaultPage := '';
  FSocket := 0;
  FillChar(FConnections, SizeOf(FConnections), 0);
  for i := Low(FActions) to High(FActions) do
  begin
    FActions[i].PathInfo := '';
    @FActions[i].Proc := nil;
  end;
  ServerList.Add(Self);
end;

destructor TWebServer.Destroy;
var
  i: Integer;
begin
  ServerList.Remove(Self);
  if FSocket <> 0 then
  begin
    TerminateTask(FWorkTask);
    CloseSocket(FSocket);
  end;
  for i := Low(FConnections) to High(FConnections) do
  begin
    if FConnections[i].HSocket <> 0 then
    begin
      TerminateTask(FConnections[i].HTask);
      CloseSocket(FConnections[i].HSocket);
      FConnections[i].HSocket := 0;
    end;
  end;
  inherited Destroy;
end;

function TWebServer.Start(Address: TIPAddress; Port: WORD;
  RootPath, DefaultPage: AnsiString): Boolean;
var
  Addr: TSockAddr;
begin
  Result := False;
  // Open Server Socket to accept incoming connections
  Addr.sin_family := PF_INET;
  Addr.sin_port := Port;
  Addr.sin_addr.S_addr := Address;
  FSocket := Socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
  if (FSocket <= 0) or (Bind(FSocket, Addr, SizeOf(Addr)) <> 0)
    or (Listen(FSocket, 10) <> 0) then
  begin
    if FSocket <> 0 then
      CloseSocket(FSocket);
    FSocket := 0;
    Exit;
  end;
  // Done
  FRootPath := RootPath;
  FDefaultPage := DefaultPage;
  FWorkTask := CreateTask(WebServerTask, Self, 32768);
  Result := True;
end;

procedure TWebServer.Stop;
begin
  // don't accept any more page requests
  if FSocket <> 0 then
  begin
    TerminateTask(FWorkTask);
    CloseSocket(FSocket);
  end;
  FSocket := 0;
end;

function TWebServer.AddConnection(HSocket: LongInt): Boolean;
var
  i: Integer;
begin
  for i := Low(FConnections) to High(FConnections) do
  begin
    if FConnections[i].HSocket = 0 then
    begin
      FConnections[i].HSocket := HSocket;
      FConnections[i].HTask := CreateTask(WebConnectionTask, @FConnections[i],
        8192);
      FConnections[i].Server := Self;
      Result := True;
      Exit;
    end;
  end;
  //	WriteLn('No free connections');
  Result := False;
end;

function TWebServer.AddAction(Action: AnsiString; Proc: TWebProc): Boolean;
var
  i: Integer;
begin
  for i := Low(FActions) to High(FActions) do
  begin
    if @FActions[i].Proc = nil then
    begin
      FActions[i].PathInfo := Action;
      FActions[i].Proc := Proc;
      Result := True;
      Exit;
    end;
  end;
  Result := False;
end;

procedure TWebServer.ProcessAction(Request: TWebRequest; Response:
  TWebResponse);
var
  i: Integer;
  Path: AnsiString;
  PathInfo: string;
begin
  // Check for a handler registered for this request.
  PathInfo := Uppercase(Request.RootPathInfo);
  for i := Low(FActions) to High(FActions) do
  begin
    if (Uppercase(FActions[i].PathInfo) = PathInfo)
      and Assigned(@FActions[i].Proc) then
    begin
      //			WriteLn('Handler Registered');
      FActions[i].Proc(Self, Request, Response);
      Exit;
    end;
  end;

  // There was no handler. Check to see if it's a file.
  Path := Copy(Request.PathInfo, 2, Length(Request.PathInfo));
  for i := 1 to Length(Path) do
    if Path[i] = '/' then
      Path[i] := '\';
  Path := GoodDir(RootPath) + Path;
  if FileExists(Path) then
  begin
    //		WriteLn('Request is a file');
    Response.SendFile(Path);
    Exit;
  end;

  // There was no handler or file. Show Error 404 "Not Found".
//	WriteLn('File not found');
  Response.StatusCode := 404;
  Response.Content :=
    '<HTML><TITLE>404 NOT FOUND</TITLE><H1><B>404 NOT FOUND</B></H1>' +
    'The file ' + Request.PathInfo + ' is not on this server.<br>' +
    'Please try again using another URL.' +
    '</HTML>';
end;

procedure TWebServer.Work;
var
  Addr: TSockAddr;
  HSock, L: LongInt;
  ReadReady, ExceptFlag: Boolean;
  WaitForConnection: Boolean;
begin
  if FSocket = 0 then
    Exit;
  if Select(FSocket, @ReadReady, nil, @ExceptFlag, 0) then
    WaitForConnection := ReadReady and not ExceptFlag
  else
    WaitForConnection := False;

  if WaitForConnection then
  begin
    // Accept next incomming connection
    L := SizeOf(Addr);
{$IFDEF WIN}
    HSock := Accept(FSocket, @Addr, @L);
{$ELSE}
    HSock := Accept(FSocket, Addr, L);
{$ENDIF}
    if HSock > 0 then
      AddConnection(HSock);
  end;
end;

procedure WebServerTask(Param: Pointer);
begin
  repeat
    TWebServer(Param).Work;
{$IFNDEF WIN}
    Net_Work_Low;
{$ENDIF}
    Yield;
  until False;
end;

procedure WebConnectionTask(Param: Pointer);
var
  P: ^TWebConnection;
  Request: TWebRequest;
  Response: TWebResponse;
begin
  P := Param;
  // Process incoming request
 //	Write('Processing... ');
  Request := TWebRequest.Create(P.Server, P.HSocket);
  Response := TWebResponse.Create(P.Server, P.HSocket);
  //	WriteLn(Request.FHeader);
  try
    if Request.MethodType <> wmNone then
    begin
      P.Server.ProcessAction(Request, Response);
      if not Response.Sent then
        Response.SendResponse;
    end;
  finally
    Request.Free;
    Response.Free;
    CloseSocket(P.HSocket);
    P.HSocket := 0;
  end;
end;

initialization
  // Creamos la lista de Servers
  ServerList := TList.Create;
  
finalization
  // Quitamos los Servers
  while (ServerList.Count > 0) do
  begin
    TWebServer(ServerList[0]).Free;
  end;
  // Eliminamos la lista de servers
  ServerList.Free;
end.

