
{*******************************************************}
{                                                       }
{ DWPL Internet Application Runtime (NetCLX)            }
{ HttpClients.pas                                       }
{                                                       }
{ for Delphi 6, 7                                       }
{ Version 2.6.0                                         }
{ Copyright (c) 2003 by The DWP-Group                   }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - ICS - Internet Component Suite                    }
{     Copyright (C) 1997-2002 by Franois PIETTE        }
{     e-mail: francois.piette@overbyte.be               }
{             francois.piette@swing.be                  }
{             francois.piette@rtfm.be                   }
{             francois.piette@pophost.eunet.be          }
{     www: http://overbyte.delphicenter.com             }
{     Realeased under the terms of the Legal Issues     }
{     from the ICS - Internet Component Suite           }
{     license agreement.                                }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}

{*_* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Author:       Franois PIETTE
Creation:     November 23, 1997
Version:      1.47
Description:  THttpCli is an implementation for the HTTP protocol
              RFC 1945 (V1.0), and some of RFC 2068 (V1.1)
Credit:       This component was based on a freeware from by Andreas
              Hoerstemeier and used with his permission.
              andy@hoerstemeier.de http://www.hoerstemeier.com/index.htm
EMail:        http://www.overbyte.be        http://www.rtfm.be/fpiette
              francois.piette@overbyte.be   francois.piette@rtfm.be
                                            francois.piette@pophost.eunet.be
Support:      Use the mailing list twsocket@elists.org
              Follow "support" link at http://www.overbyte.be for subscription.
Legal issues: Copyright (C) 1997-2002 by Franois PIETTE
              Rue de Grady 24, 4053 Embourg, Belgium. Fax: +32-4-365.74.56
              <francois.piette@overbyte.be>

              This software is provided 'as-is', without any express or
              implied warranty.  In no event will the author be held liable
              for any  damages arising from the use of this software.

              Permission is granted to anyone to use this software for any
              purpose, including commercial applications, and to alter it
              and redistribute it freely, subject to the following
              restrictions:

              1. The origin of this software must not be misrepresented,
                 you must not claim that you wrote the original software.
                 If you use this software in a product, an acknowledgment
                 in the product documentation would be appreciated but is
                 not required.

              2. Altered source versions must be plainly marked as such, and
                 must not be misrepresented as being the original software.

              3. This notice may not be removed or altered from any source
                 distribution.

              4. You must register this software by sending a picture postcard
                 to the author. Use a nice stamp and mention your name, street
                 address, EMail address and any comment you like to say.

Quick Reference:
HTTP component can retrieve documents or files using HTTP protocol; that is
connect to a HTTP server also known as a webserver. It can also trigger a
CGI/ISAPI/NSAPI script and post data using either GET or POST method.
Syntax of an URL: protocol://[user[:password]@]server[:port]/path
Path can include data: question mark followed by URL encoded data.
This implementation of HTTP component works only synchonous (blocking).
If you need asynchonous (non-blocking) mode you must put it into a thread
and call the events by synchronization.
Methods:
    Get         Synchronous, blocking Get, blocks until finished.
    Post        Synchronous, blocking Post, blocks until finished.
    Head        Synchronous, blocking Head, blocks until finished.
    Abort       Immediately close communication.

 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

{
  $Header: /cvsroot/dwpl/dwpl/patch60/source/HttpClients.pas,v 1.2 2003/10/07 19:10:22 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: HttpClients.pas,v $
  Revision 1.2  2003/10/07 19:10:22  iwache
  To class THttpClient property ReceiveTimeout added.
  Default value is 10000 i.e. 10 seconds.

  Revision 1.1  2003/09/21 19:53:41  iwache
  New NetCLX units added.

  ----------------------------------------------------------------------------
}

unit HttpClients;

interface

uses
  Windows, SysUtils, Classes,
{$IFDEF WIN}
  WinSock,
{$ENDIF WIN}
{$IFDEF WDOSX}
  WDosSock,
{$ENDIF WDOSX}
  SocketUtils, Sockets;

const
  DefaultProxyPort = '80';
  DReceiveTimeout = 10000; // max. wait delay for response chunks

  HTTP_RCV_BUF_SIZE = 8193;
  HTTP_SND_BUF_SIZE = 8193;

  httperrNoError = 0;
  httperrBusy = 1;
  httperrNoData = 2;
  httperrAborted = 3;
  httperrOverflow = 4;

type
  EHttpException = class(Exception)
    ErrorCode: Word;
    constructor Create(const Msg: string; ErrCode: Word);
  end;

  THttpEncoding = (encUUEncode, encBase64, encMime);
  THttpRequest = (httpAbort, httpGET, httpPOST, httpHEAD);
  THttpState = (httpReady, httpNotConnected, httpConnected,
    httpDnsLookup, httpDnsLookupDone,
    httpWaitingHeader, httpWaitingBody, httpAborting);

  TOnCommand = procedure(Sender: TObject; var S: string) of object;
  TDocDataEvent = procedure(Sender: TObject; Buffer: Pointer;
      Len: Integer) of object;
  TCookieRcvdEvent = procedure(Sender: TObject; const Data: string;
      var Accept: Boolean) of object;
  THttpRequestDone = procedure(Sender: TObject; RqType: THttpRequest;
      Error: Word) of object;
  TBeforeHeaderSendEvent = procedure(Sender: TObject; const Method: string;
      Headers: TStrings) of object;
  TDataAvailable = procedure(Sender: TObject; Error: word) of object;
  TObjectMethod = procedure of object;


  THttpClient = class(TComponent)
  private
    FReceiveTimeout: Integer;
  protected
    FCtrlSocket: TTcpClient;
    FState: THttpState;
    FLocalAddr: string;
    FHostName: string;
    FTargetHost: string;
    FPort: string;
    FProxy: string;
    FProxyPort: string;
    FUsername: string;
    FPassword: string;
    FProxyUsername: string;
    FProxyPassword: string;
    FLocation: string;
    FConnected: Boolean;
    FDnsResult: string;
    FSendBuffer: array[0..HTTP_SND_BUF_SIZE - 1] of Char;
    FRequestType: THttpRequest;
    FReceiveBuffer: array[0..HTTP_RCV_BUF_SIZE - 1] of Char;
    FReceiveLen: Integer;
    FLastResponse: string;
    FHeaderLineCount: Integer;
    FBodyLineCount: Integer;
    FURL: string;
    FPath: string;
    FDocName: string;
    FSender: string;
    FReference: string;
    FAgent: string;
    FAccept: string;
    FAcceptLanguage: string;
    FModifiedSince: TDateTime; { Warning ! Use GMT date/Time }
    FNoCache: Boolean;
    FStatusCode: Integer;
    FReasonPhrase: string;
    FContentLength: LongInt;
    FContentType: string;
    FDoAuthor: TStringList;
    FContentPost: string;
    FContentRangeBegin: string;
    FContentRangeEnd: string;
    FAcceptRanges: string;
    FCookie: string;
    FLocationFlag: Boolean;
    FHeaderEndFlag: Boolean;
    FRcvdHeader: TStrings;
    FRcvdStream: TStream; { If assigned, will recv the answer }
    FRcvdCount: LongInt; { Number of rcvd bytes for the body }
    FSentCount: LongInt;
    FSendStream: TStream; { Contains the data to send         }
    FReqStream: TMemoryStream;
    FRequestDoneError: Integer;
    FCtrlSocketLastError: Integer;
    FNext: TObjectMethod;
    FOnStateChange: TNotifyEvent;
    FOnSessionConnected: TNotifyEvent;
    FOnRequestHeaderBegin: TNotifyEvent;
    FOnRequestHeaderEnd: TNotifyEvent;
    FOnHeaderBegin: TNotifyEvent;
    FOnHeaderEnd: TNotifyEvent;
    FOnHeaderData: TNotifyEvent;
    FOnDocBegin: TNotifyEvent;
    FOnDocEnd: TNotifyEvent;
    FOnDocData: TDocDataEvent;
    FOnSendBegin: TNotifyEvent;
    FOnSendEnd: TNotifyEvent;
    FOnSendData: TDocDataEvent;
    FOnTrace: TNotifyEvent;
    FOnCommand: TOnCommand;
    FOnCookie: TCookieRcvdEvent;
    FOnDataAvailable: TDataAvailable;
    FOnRequestDone: THttpRequestDone;
    FOnLocationChange: TNotifyEvent;
    FOnSocketError: TSocketErrorEvent;
    FOnBeforeHeaderSend: TBeforeHeaderSendEvent;
    procedure SocketErrorTransfer(Sender: TObject; SocketError: Integer);
    procedure SendRequest(const method, Version: string);
    procedure GetHeaderLineNext; virtual;
    procedure GetBodyLineNext; virtual;
    procedure SendCommand(const Cmd: string); virtual;
    procedure Login; virtual;
    procedure Logout; virtual;
    procedure InternalClear; virtual;
    procedure SocketDNSLookupDone(Sender: TObject; Error: Word); virtual;
    procedure SocketSessionClosed(Sender: TObject; Error: Word); virtual;
    procedure SocketSessionConnected(Sender: TObject; Error: Word); virtual;
    procedure SendData(Sender: TObject; Error: Word); virtual;
    procedure SocketDataAvailable(Sender: TObject; Error: Word); virtual;
    procedure LocationSessionClosed(Sender: TObject; Error: Word); virtual;
    procedure DoRequest(Rq: THttpRequest); virtual;
    procedure StateChange(NewState: THttpState); virtual;
    procedure TriggerStateChange; virtual;
    procedure TriggerCookie(const Data: string;
      var bAccept: Boolean); virtual;
    procedure TriggerSessionConnected; virtual;
    procedure TriggerBeforeHeaderSend(const Method: string;
      Headers: TStrings); virtual;
    procedure TriggerRequestHeaderBegin; virtual;
    procedure TriggerRequestHeaderEnd; virtual;
    procedure TriggerHeaderBegin; virtual;
    procedure TriggerHeaderEnd; virtual;
    procedure TriggerDocBegin; virtual;
    procedure TriggerDocData(Data: Pointer; Len: Integer); virtual;
    procedure TriggerDocEnd; virtual;
    procedure TriggerSendBegin; virtual;
    procedure TriggerSendData(Data: Pointer; Len: Integer); virtual;
    procedure TriggerSendEnd; virtual;
    procedure TriggerRequestDone; virtual;
    procedure SetReady; virtual;
    procedure AdjustDocName; virtual;
    procedure CheckSocketDataAvailable; virtual;
  public
    constructor Create(Aowner: TComponent); override;
    destructor Destroy; override;
    procedure Get;
    procedure Post;
    procedure Head;
    procedure Abort;

    property CtrlSocket: TTcpClient read FCtrlSocket;
    property State: THttpState read FState;
    property LastResponse: string read FLastResponse;
    property ContentLength: LongInt read FContentLength;
    property ContentType: string read FContentType;
    property RcvdCount: LongInt read FRcvdCount;
    property SentCount: LongInt read FSentCount;
    property StatusCode: Integer read FStatusCode;
    property ReasonPhrase: string read FReasonPhrase;
    property DnsResult: string read FDnsResult;
    property AuthorizationRequest: TStringList read FDoAuthor;
    property DocName: string read FDocName;
    property Location: string read FLocation
      write FLocation;
    property RcvdStream: TStream read FRcvdStream
      write FRcvdStream;
    property SendStream: TStream read FSendStream
      write FSendStream;
    property RcvdHeader: TStrings read FRcvdHeader;
    property Hostname: string read FHostname;
  published
    property URL: string read FURL
      write FURL;
    property LocalAddr: string read FLocalAddr
      write FLocalAddr;
    property Proxy: string read FProxy
      write FProxy;
    property ProxyPort: string read FProxyPort
      write FProxyPort;
    property Sender: string read FSender
      write FSender;
    property Agent: string read FAgent
      write FAgent;
    property Accept: string read FAccept
      write FAccept;
    property AcceptLanguage: string read FAcceptLanguage
      write FAcceptLanguage;
    property Reference: string read FReference
      write FReference;
    property Username: string read FUsername
      write FUsername;
    property Password: string read FPassword
      write FPassword;
    property ProxyUsername: string read FProxyUsername
      write FProxyUsername;
    property ProxyPassword: string read FProxyPassword
      write FProxyPassword;
    property NoCache: Boolean read FNoCache
      write FNoCache;
    property ModifiedSince: TDateTime read FModifiedSince
      write FModifiedSince;
    property Cookie: string read FCookie
      write FCookie;
    property ContentTypePost: string read FContentPost
      write FContentPost;
    property ContentRangeBegin: string read FContentRangeBegin
      write FContentRangeBegin;
    property ContentRangeEnd: string read FContentRangeEnd
      write FContentRangeEnd;
    property AcceptRanges: string read FAcceptRanges;
    property OnTrace: TNotifyEvent read FOnTrace
      write FOnTrace;
    property OnSessionConnected: TNotifyEvent read FOnSessionConnected
      write FOnSessionConnected;
    property OnHeaderData: TNotifyEvent read FOnHeaderData
      write FOnHeaderData;
    property OnCommand: TOnCommand read FOnCommand
      write FOnCommand;
    property OnHeaderBegin: TNotifyEvent read FOnHeaderBegin
      write FOnHeaderBegin;
    property OnHeaderEnd: TNotifyEvent read FOnHeaderEnd
      write FOnHeaderEnd;
    property OnRequestHeaderBegin: TNotifyEvent read FOnRequestHeaderBegin
      write FOnRequestHeaderBegin;
    property OnRequestHeaderEnd: TNotifyEvent read FOnRequestHeaderEnd
      write FOnRequestHeaderEnd;
    property OnDocBegin: TNotifyEvent read FOnDocBegin
      write FOnDocBegin;
    property OnDocData: TDocDataEvent read FOnDocData
      write FOnDocData;
    property OnDocEnd: TNotifyEvent read FOnDocEnd
      write FOnDocEnd;
    property OnSendBegin: TNotifyEvent read FOnSendBegin
      write FOnSendBegin;
    property OnSendData: TDocDataEvent read FOnSendData
      write FOnSendData;
    property OnSendEnd: TNotifyEvent read FOnSendEnd
      write FOnSendEnd;
    property OnStateChange: TNotifyEvent read FOnStateChange
      write FOnStateChange;
    property OnRequestDone: THttpRequestDone read FOnRequestDone
      write FOnRequestDone;
    property OnLocationChange: TNotifyEvent read FOnLocationChange
      write FOnLocationChange;
    property OnCookie: TCookieRcvdEvent read FOnCookie
      write FOnCookie;
    property OnSocketError: TSocketErrorEvent read FOnSocketError write
        FOnSocketError;
    property OnBeforeHeaderSend: TBeforeHeaderSendEvent read FOnBeforeHeaderSend
      write FOnBeforeHeaderSend;
    property ReceiveTimeout: Integer read FReceiveTimeout write FReceiveTimeout
      default DReceiveTimeout;
  end;

{ Syntax of an URL: protocol://[user[:password]@]server[:port]/path }
procedure ParseURL(const URL: string; var Proto, User, Pass, Host,
  Port, Path: string);
function Posn(const s, t: string; count: Integer): Integer;
procedure ReplaceExt(var FName: string; const newExt: string);
function EncodeLine(Encoding: THttpEncoding;
  SrcData: PChar; Size: Integer): string;
function EncodeStr(Encoding: THttpEncoding; const Value: string): string;
function RFC1123_Date(aDate: TDateTime): string;
function UrlEncode(S: string): string;
function UrlDecode(S: string): string;
function IsDigit(Ch: Char): Boolean;
function IsXDigit(Ch: char): Boolean;
function XDigit(Ch: char): Integer;
function htoin(value: PChar; len: Integer): Integer;
function htoi2(value: PChar): Integer;

implementation

{$IFDEF WDOSX}
uses
  WDosClock;
{$ENDIF}

const
  bin2uue: string =
    '`!"#$%&''()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_';
  bin2b64: string =
    'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
  uue2bin: string =
    ' !"#$%&''()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_ ';
  b642bin: string =
    '~~~~~~~~~~~^~~~_TUVWXYZ[\]~~~|~~~ !"#$%&''()*+,-./0123456789~~~~~~:;<=>?@ABCDEFGHIJKLMNOPQRS';
  linesize = 45;
type
  TCharSet = set of Char;

const
  UriProtocolSchemeAllowedChars: TCharSet = ['a'..'z', '0'..'9', '+', '-', '.'];

function GetBaseUrl(const Url: string): string; forward;

function GetTimeCounter: Integer;
begin
{$IFDEF WDOSX}
  Result := GetMSCounter;
{$ENDIF}
{$IFDEF WIN}
  Result := GetTickCount;
{$ENDIF}
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

constructor EHttpException.Create(const Msg: string; ErrCode: Word);
begin
  inherited Create(Msg);
  ErrorCode := ErrCode;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
{ We cannot use Delphi own function because the date must be specified in   }
{ english and Delphi use the current language.                              }

function RFC1123_Date(aDate: TDateTime): string;
const
  StrWeekDay: string = 'MonTueWedThuFriSatSun';
  StrMonth: string = 'JanFebMarAprMayJunJulAugSepOctNovDec';
var
  Year, Month, Day: Word;
  Hour, Min, Sec, MSec: Word;
  DayOfWeek: Word;
begin
  DecodeDate(aDate, Year, Month, Day);
  DecodeTime(aDate, Hour, Min, Sec, MSec);
  DayOfWeek := ((Trunc(aDate) - 2) mod 7);
  Result := Copy(StrWeekDay, 1 + DayOfWeek * 3, 3) + ', ' +
    Format('%2.2d %s %4.4d %2.2d:%2.2d:%2.2d',
    [Day, Copy(StrMonth, 1 + 3 * (Month - 1), 3),
    Year, Hour, Min, Sec]);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

constructor THttpClient.Create(Aowner: TComponent);
begin
  inherited create(AOwner);
  FCtrlSocket := TTcpClient.Create(Self);
  FProxyPort := DefaultProxyPort;
  FContentPost := 'application/x-www-form-urlencoded';
  FAccept := 'image/gif, image/x-xbitmap, ' +
    'image/jpeg, image/pjpeg, */*';
  FAgent := 'Mozilla/3.0 (compatible)';
  FDoAuthor := TStringlist.Create;
  FCtrlSocket.OnError := SocketErrorTransfer;
  FRcvdHeader := TStringList.Create;
  FReqStream := TMemoryStream.Create;
  FState := httpReady;
  FLocalAddr := '0.0.0.0';
  FReceiveTimeout := DReceiveTimeout;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

destructor THttpClient.Destroy;
begin
  FDoAuthor.Free;
  FCtrlSocket.Free;
  FRcvdHeader.Free;
  FReqStream.Free;
  inherited Destroy;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.SocketErrorTransfer(Sender: TObject; SocketError: Integer);
begin
  FCtrlSocketLastError := SocketError;
  if (Assigned(FOnSocketError)) then
     { Substitute Self for subcomponent's Sender. }
     FOnSocketError(Self, SocketError);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.SetReady;
begin
  StateChange(httpReady);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.StateChange(NewState: THttpState);
begin
  if FState <> NewState then
  begin
    FState := NewState;
    TriggerStateChange;
    if NewState = httpReady then
      TriggerRequestDone;
  end;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerStateChange;
begin
  if Assigned(FOnStateChange) then
    FOnStateChange(Self);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerCookie(const Data: string; var bAccept: Boolean);
begin
  if Assigned(FOnCookie) then
    FOnCookie(Self, Data, bAccept);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerSessionConnected;
begin
  if Assigned(FOnSessionConnected) then
    FOnSessionConnected(Self);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerDocBegin;
begin
  if Assigned(FOnDocBegin) then
    FOnDocBegin(Self);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerDocEnd;
begin
  if Assigned(FOnDocEnd) then
    FOnDocEnd(Self);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerDocData(Data: Pointer; Len: Integer);
begin
  if Assigned(FOnDocData) then
    FOnDocData(Self, Data, Len);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerSendBegin;
begin
  if Assigned(FOnSendBegin) then
    FOnSendBegin(Self);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerSendEnd;
begin
  if Assigned(FOnSendEnd) then
    FOnSendEnd(Self);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerSendData(Data: Pointer; Len: Integer);
begin
  if Assigned(FOnSendData) then
    FOnSendData(Self, Data, Len);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerHeaderBegin;
begin
  FHeaderEndFlag := TRUE;
  if Assigned(FOnHeaderBegin) then
    FOnHeaderBegin(Self);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerHeaderEnd;
begin
  FHeaderEndFlag := FALSE;
  if Assigned(FOnHeaderEnd) then
    FOnHeaderEnd(Self);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerBeforeHeaderSend(
  const Method: string;
  Headers: TStrings);
begin
  if Assigned(FOnBeforeHeaderSend) then
    FOnBeforeHeaderSend(Self, Method, Headers);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerRequestHeaderBegin;
begin
  if Assigned(FOnRequestHeaderBegin) then
    FOnRequestHeaderBegin(Self);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerRequestHeaderEnd;
begin
  if Assigned(FOnRequestHeaderEnd) then
    FOnRequestHeaderEnd(Self);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.TriggerRequestDone;
begin
  if Assigned(FOnRequestDone) then
    FOnRequestDone(Self, FRequestType, FRequestDoneError);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure ReplaceExt(var FName: string; const newExt: string);
var
  I: Integer;
begin
  I := Posn('.', FName, -1);
  if I <= 0 then
    FName := FName + '.' + newExt
  else
    FName := Copy(FName, 1, I) + newExt;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.Abort;
var
  bFlag: Boolean;
begin
  if FState = httpReady then
  begin
    if FCtrlSocket.Active then
      FCtrlSocket.Close; { This should never occurs ! }
    Exit;
  end;

  bFlag := (FState = httpDnsLookup);
  StateChange(httpAborting);

  if bFlag then
  begin
    try
(*
      FCtrlSocket.CancelDnsLookup;
*)
    except
      { Ignore any exception }
    end;
  end;

  FStatusCode := 404;
  FReasonPhrase := 'Connection aborted on request';
  FRequestDoneError := httperrAborted;

  if bFlag then
    SocketSessionClosed(Self, 0)
  else
    FCtrlSocket.Close;
  StateChange(httpReady);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.Login;
var
  Error: Word;
begin
  FDnsResult := '';
  StateChange(httpDnsLookup);
  FCtrlSocket.LocalHost := FLocalAddr; {bb}
  try
    with FCtrlSocket do
    begin
      FDnsResult := LookupHostAddr(FHostName);
      if FDnsResult = '0.0.0.0' then
        Error := WSAHOST_NOT_FOUND
      else
        Error := 0;
    end;
  except
    on E: Exception do
    begin
      FStatusCode := 404;
      FReasonPhrase := E.Message;
      FConnected := FALSE;
      StateChange(httpReady);
      Exit;
    end;
  end;
  SocketDNSLookupDone(FCtrlSocket, Error);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.SocketDNSLookupDone(Sender: TObject; Error: Word);
begin
  if Error <> 0 then
  begin
    if FState = httpAborting then
      Exit;
    FRequestDoneError := Error;
    FStatusCode := 404;
    FReasonPhrase := 'can''t resolve hostname to IP address';
    SocketSessionClosed(Sender, Error);
  end
  else
  begin
    StateChange(httpDnsLookupDone); { 19/09/98 }
    FCtrlSocket.RemoteHost := FDnsResult;
    FCtrlSocket.RemotePort := FPort;
    FCtrlSocket.LocalHost := FLocalAddr; {bb}
    try
      FCtrlSocket.Connect;
    except
      FRequestDoneError := FCtrlSocketLastError;
      FStatusCode := 404;
      FReasonPhrase := 'can''t connect: ' +
        SocketErrorDesc(FCtrlSocketLastError) +
        ' (Error #' + IntToStr(FCtrlSocketLastError) + ')';
      FCtrlSocket.Close;
      SocketSessionClosed(Sender, FCtrlSocketLastError);
      Exit;
    end;
    SocketSessionConnected(FCtrlSocket, FCtrlSocketLastError);
  end;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.SocketSessionConnected(Sender: TObject; Error: Word);
begin
  if Error <> 0 then
  begin
    FRequestDoneError := Error;
    FStatusCode := 404;
    FReasonPhrase := SocketErrorDesc(Error) +
      ' (Error #' + IntToStr(Error) + ')';
    SocketSessionClosed(Sender, Error);
    Exit;
  end;

  FConnected := TRUE;
  StateChange(httpConnected);
  TriggerSessionConnected;

  FNext := GetHeaderLineNext;
  StateChange(httpWaitingHeader);

  try
    case FRequestType of
      httpPOST:
        begin
          SendRequest('POST', '1.0');
          TriggerSendBegin;
          SendData(FCtrlSocket, 0);
        end;
      httpHEAD:
        begin
          SendRequest('HEAD', '1.0');
        end;
      httpGET:
        begin
          SendRequest('GET', '1.0');
        end;
    end;
  except
    Logout;
    Exit;
  end;
  CheckSocketDataAvailable;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.CheckSocketDataAvailable;
var
  Timeout: Int64;
begin
  Timeout := GetTimeCounter + FReceiveTimeout;
  while GetTimeCounter < Timeout do
  begin
    // Check if we are still online
    if FCtrlSocket.ConnectionAborted then Break;

    // Check if we got a relocation
    if not FCtrlSocket.Connected and FLocationFlag then Exit;

    // Any data to read from the server?
    while FCtrlSocket.WaitForData do
    begin
      SocketDataAvailable(FCtrlSocket, 0);
      Timeout := GetTimeCounter + FReceiveTimeout;
    end;

    // keep network alive
    Sockets.UpdateNetwork;
  end;
  Logout;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.Logout;
begin
  FCtrlSocket.Close;
  FConnected := FALSE;
  SocketSessionClosed(FCtrlSocket, FCtrlSocketLastError);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.SendCommand(const Cmd: string);
const
  CRLF: string[2] = #13#10;
var
  Buf: string;
begin
  Buf := Cmd;
  if Assigned(FOnCommand) then
    FOnCommand(Self, Buf);
  if Length(Buf) > 0 then
    FReqStream.Write(Buf[1], Length(Buf));
  FReqStream.Write(CRLF[1], 2);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.SendRequest(const Method, Version: string);
var
  Headers: TStrings;
  N: Integer;
begin
  Headers := TStringList.Create;
  try
    FReqStream.Clear;
    TriggerRequestHeaderBegin;
    Headers.Add(method + ' ' + FPath + ' HTTP/' + Version);
    if FSender <> '' then
      Headers.Add('From: ' + FSender);
    if FAccept <> '' then
      Headers.Add('Accept: ' + FAccept);
    if FReference <> '' then
      Headers.Add('Referer: ' + FReference);
    if FAcceptLanguage <> '' then
      Headers.Add('Accept-Language: ' + FAcceptLanguage);
    if (method = 'POST') and (FContentPost <> '') then
      Headers.Add('Content-Type: ' + FContentPost);
    if FAgent <> '' then
      Headers.Add('User-Agent: ' + FAgent);
    Headers.Add('Host: ' + FTargetHost);
    if FNoCache then
      Headers.Add('Pragma: no-cache');
    if method = 'POST' then
      Headers.Add('Content-Length: ' + IntToStr(SendStream.Size));
    if FModifiedSince <> 0 then
      Headers.Add('If-Modified-Since: ' +
        RFC1123_Date(FModifiedSince) + ' GMT');
    if FUsername <> '' then
      Headers.Add('Authorization: Basic ' +
        EncodeStr(encBase64, FUsername + ':' + FPassword));
    if (FProxy <> '') and (FProxyUsername <> '') then
      Headers.Add('Proxy-Authorization: Basic ' +
        EncodeStr(encBase64, FProxyUsername + ':' + FProxyPassword));
    if FCookie <> '' then
      Headers.Add('Cookie: ' + FCookie);
    if (FContentRangeBegin <> '') or (FContentRangeEnd <> '') then
    begin {JMR!! Added this line!!!}
      Headers.Add('Range: bytes=' + FContentRangeBegin + '-' + FContentRangeEnd);
        {JMR!! Added this line!!!}
      FContentRangeBegin := ''; {JMR!! Added this line!!!}
      FContentRangeEnd := ''; {JMR!! Added this line!!!}
    end; {JMR!! Added this line!!!}
    FAcceptRanges := '';

    {SendCommand('UA-pixels: 1024x768'); }
    {SendCommand('UA-color: color8'); }
    {SendCommand('UA-OS: Windows 95'); }
    {SendCommand('UA-CPU: x86'); }
    {SendCommand('User-Agent: Mozilla/3.0 (compatible)');}{; MSIE 3.01; Update a; Windows 95)');}
    {SendCommand('Proxy-Connection: Keep-Alive'); }

    TriggerBeforeHeaderSend(Method, Headers);
    for N := 0 to Headers.Count - 1 do
      SendCommand(Headers[N]);

    TriggerRequestHeaderEnd;
    SendCommand('');
    FReqStream.Seek(0, soFromBeginning);
    FCtrlSocket.SendStream(FReqStream);
    FReqStream.Clear;
  finally
    Headers.Free;
  end;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.GetBodyLineNext;
var
  Len: Integer;
  Data: Pointer;
const
  CRLF: string[2] = #13#10;
begin
  if FBodyLineCount = 0 then
    TriggerDocBegin;
  Inc(FBodyLineCount);

  Len := Length(FLastResponse);
  if Len > 0 then
    Data := @FLastResponse[1]
  else
    Data := @Len;
  FRcvdCount := FRcvdCount + Len;

  if Assigned(FRcvdStream) then
    FRcvdStream.WriteBuffer(Data^, Len);
  TriggerDocData(Data, Len);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.GetHeaderLineNext;
var
  proto: string;
  user: string;
  pass: string;
  port: string;
  Host: string;
  Path: string;
  Field: string;
  Data: string;
  nSep: Integer;
  tmpInt: LongInt;
  bAccept: Boolean;
  DocExt: string;
begin
  if FHeaderLineCount = 0 then
    TriggerHeaderBegin;
  Inc(FHeaderLineCount);

  { Some server send HTML document without header ! I don't know if it is  }
  { legal, but it exists (AltaVista Discovery does that).                  }
  if UpperCase(Copy(FLastResponse, 1, 6)) = '<HTML>' then
  begin { 15/09/98 }
    if FContentType = '' then
      FContentType := 'text/html';
    StateChange(httpWaitingBody);
    FNext := GetBodyLineNext;
    TriggerHeaderEnd;
    GetBodyLineNext;
    Exit;
  end;

  if FLastResponse = '' then
  begin
    if FLocationFlag then
    begin
      TriggerHeaderEnd;
      FReceiveLen := 0;
      FHeaderLineCount := 0;
      FBodyLineCount := 0;
      FCtrlSocket.Close;
      Exit;
    end;
    { FContentLength = -1 when server doesn't send a value }
    if FContentLength = 0 then
    begin
      TriggerHeaderEnd;
      SetReady;
      Exit;
    end;
    DocExt := lowercase(ExtractFileExt(FDocName));
    if (DocExt = '.exe') or (DocExt = '') then
    begin
      if FContentType = 'text/html' then
        ReplaceExt(FDocName, 'htm');
    end;

    StateChange(httpWaitingBody);
    FNext := GetBodyLineNext;
    TriggerHeaderEnd;
    if FReceiveLen > 0 then
    begin
      SetLength(FLastResponse, FReceiveLen);
      Move(FReceiveBuffer, FLastResponse[1], FReceiveLen);
      GetBodyLineNext;
      FReceiveLen := 0;
    end;
//    if FStatusCode >= 400 then { 01/11/01 }
//      FCtrlSocket.Close;
    Exit;
  end;

  FRcvdHeader.Add(FLastResponse);

  nSep := pos(':', FLastResponse);
  if (Copy(FLastResponse, 1, 8) = 'HTTP/1.0') or
    (Copy(FLastResponse, 1, 8) = 'HTTP/1.1') then
  begin
    FStatusCode := StrToInt(Copy(FLastResponse, 10, 3));
    FReasonPhrase := Copy(FLastResponse, 14, Length(FLastResponse));
  end
  else if nSep > 0 then
  begin
    Field := LowerCase(Copy(FLastResponse, 1, nSep - 1));
    { Skip spaces }
    Inc(nSep);
    while (nSep < Length(FLastResponse)) and
      (FLastResponse[nSep] = ' ') do
      Inc(nSep);
    Data := Copy(FLastResponse, nSep, Length(FLastResponse));
    if Field = 'location' then
    begin { Change the URL ! }
      { URL with relocations:                 }
      { http://www.webcom.com/~wol2wol/       }
      { http://www.purescience.com/delphi/    }
      { http://www.maintron.com/              }
      { http://www.infoseek.com/AddURL/addurl }
      { http://www.micronpc.com/              }
      { http://www.amazon.com/                }
      { http://count.paycounter.com/?fn=0&si=44860&bt=msie&bv=5&    }
      { co=32&js=1.4&sr=1024x768&re=http://www.thesite.com/you.html }
      FLocationFlag := TRUE;
      if Proxy <> '' then
      begin
        { We are using a proxy }
        if Data[1] = '/' then
        begin
          { Absolute location }
          ParseURL(FPath, proto, user, pass, Host, port, Path);
          if Proto = '' then
            Proto := 'http';
          FLocation := Proto + '://' + Host + Data;
          FPath := FLocation;

          if (user <> '') and (pass <> '') then
          begin
            { save user and password given in location @@@}
            FUsername := user;
            FPassword := pass;
          end;
        end
        else if Copy(Data, 1, 7) <> 'http://' then
        begin
          { Relative location }
          FPath := GetBaseUrl(FPath) + Data;
          if Proto = '' then
            Proto := 'http';
          FLocation := Proto + '://' + FHostName + '/' + FPath;
        end
        else
        begin
          ParseURL(Data, proto, user, pass, Host, port, Path);
          if port <> '' then
            FPort := port
          else
            FPort := '80';

          if (user <> '') and (pass <> '') then
          begin
            { save user and password given in location @@@}
            FUsername := user;
            FPassword := pass;
          end;

          if (Proto <> '') and (Host <> '') then
          begin
            { We have a full relocation URL }
            FTargetHost := Host;
            FLocation := Proto + '://' + Host + Path;
            FPath := FLocation;
          end
          else
          begin
            if Proto = '' then
              Proto := 'http';
            if FPath = '' then
              FLocation := Proto + '://' + FTargetHost + '/' + Host
            else if Host = '' then
              FLocation := Proto + '://' + FTargetHost + FPath
            else
              FTargetHost := Host;
          end;
        end;
      end
        { We are not using a proxy }
      else
      begin
        if Data[1] = '/' then
        begin
          { Absolute location }
          FPath := Data;
          if Proto = '' then
            Proto := 'http';
          FLocation := Proto + '://' + FHostName + FPath;
        end
        else if Copy(Data, 1, 7) <> 'http://' then
        begin
          { Relative location }
          FPath := GetBaseUrl(FPath) + Data;
          if Proto = '' then
            Proto := 'http';
          FLocation := Proto + '://' + FHostName + '/' + FPath;
        end
        else
        begin
          ParseURL(Data, proto, user, pass, FHostName, port, FPath);
          if port <> '' then
            FPort := port
          else
            FPort := '80';

          if (user <> '') and (pass <> '') then
          begin
            { save user and password given in location @@@}
            FUsername := user;
            FPassword := pass;
          end;

          if (Proto <> '') and (FHostName <> '') then
          begin
            { We have a full relocation URL }
            FTargetHost := FHostName;
            if FPath = '' then
            begin
              FPath := '/';
              FLocation := Proto + '://' + FHostName;
            end
            else
              FLocation := Proto + '://' + FHostName + FPath;
          end
          else
          begin
            if Proto = '' then
              Proto := 'http';
            if FPath = '' then
            begin
              FLocation := Proto + '://' + FTargetHost + '/' + FHostName;
              FHostName := FTargetHost;
              FPath := FLocation; { 26/11/99 }
            end
            else if FHostName = '' then
            begin
              FLocation := Proto + '://' + FTargetHost + FPath;
              FHostName := FTargetHost;
            end
            else
              FTargetHost := FHostName;
          end;
        end;
      end;
    end
    else if Field = 'content-length' then
      FContentLength := StrToInt(Trim(Data))
    else if Field = 'content-range' then
    begin {JMR!! Added this line!!!}
      tmpInt := Pos('-', Data) + 1; {JMR!! Added this line!!!}
      FContentRangeBegin := Copy(Data, 7, tmpInt - 8); {JMR!! Added this line!!!}
      FContentRangeEnd := Copy(Data, tmpInt, Pos('/', Data) - tmpInt);
        {JMR!! Added this line!!!}
    end {JMR!! Added this line!!!}
    else if Field = 'accept-ranges' then
      FAcceptRanges := Data
    else if Field = 'content-type' then
      FContentType := LowerCase(Data)
    else if Field = 'www-authenticate' then
      FDoAuthor.add(Data)
    else if Field = 'set-cookie' then
    begin
      bAccept := TRUE;
      TriggerCookie(Data, bAccept);
    end
      {   else if Field = 'date' then }
      {   else if Field = 'mime-version' then }
      {   else if Field = 'pragma' then }
      {   else if Field = 'allow' then }
      {   else if Field = 'server' then }
      {   else if Field = 'content-encoding' then }
      {   else if Field = 'expires' then }
      {   else if Field = 'last-modified' then }
  end
  else { Ignore  all other responses }
    ;

  if Assigned(FOnHeaderData) then
    FOnHeaderData(Self);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.InternalClear;
begin
  FRcvdHeader.Clear;
  FRequestDoneError := 0;
  FDocName := '';
  FStatusCode := 0;
  FRcvdCount := 0;
  FSentCount := 0;
  FHeaderLineCount := 0;
  FBodyLineCount := 0;
  FContentLength := -1;
  FContentType := '';
  FLocation := FURL;
  FCtrlSocketLastError := 0;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.DoRequest(Rq: THttpRequest);
var
  Proto, User, Pass, Host, Port, Path: string;
  LoginDone: Boolean;
begin
  if FState <> httpReady then
    raise EHttpException.Create('HTTP component is busy', httperrBusy);

  if (Rq = httpPOST) and (not Assigned(FSendStream)) then
    raise EHttpException.Create('HTTP component has nothing to post',
      httperrNoData);

  { Now clear all internal state variables }
  StateChange(httpNotConnected);
  FRequestType := Rq;
  InternalClear;

  { parse url and proxy to FHostName, FPath and FPort }
  if FProxy <> '' then
  begin
    ParseURL(FURL, Proto, User, Pass, Host, Port, Path);
    FTargetHost := Host;
    FPath := FURL;
    FDocName := Path;
    if User <> '' then
      FUserName := User;
    if Pass <> '' then
      FPassword := Pass;
    { We need to remove usercode/Password from the URL given to the proxy }
    { but preserve the port                                               }
    if Port <> '' then
      Port := ':' + Port;
    if Proto = '' then
      FPath := 'http://' + Host + Port + Path
    else
      FPath := Proto + '://' + Host + Port + Path;
    ParseURL(FProxy, Proto, User, Pass, Host, Port, Path);
    if Port = '' then
      Port := ProxyPort;
  end
  else
  begin
    ParseURL(FURL, Proto, User, Pass, Host, Port, FPath);
    FTargetHost := Host;
    FDocName := FPath;
    if User <> '' then
      FUserName := User;
    if Pass <> '' then
      FPassword := Pass;
    if Port = '' then
      Port := '80';
  end;
  if Proto = '' then
    Proto := 'http';
  if FPath = '' then
    FPath := '/';

  AdjustDocName;

  FHostName := host;
  FPort := Port;

  { Ask to connect. When connected, we go to SocketSeesionConnected. }
  repeat
    Login;
    LoginDone := True;
    if FLocationFlag then
    begin
      LocationSessionClosed(FCtrlSocket, FCtrlSocketLastError);
      // Restart the login procedure
      LoginDone := False;
    end;
  until LoginDone;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.AdjustDocName;
var
  I: Integer;
begin
  I := Pos('?', FDocName);
  if I > 0 then
    FDocName := Copy(FDocName, 1, I - 1);

  if (FDocName = '') or (FDocName[Length(FDocName)] = '/') then
    FDocName := 'document.htm'
  else
  begin
    if FDocName[Length(FDocName)] = '/' then
      SetLength(FDocName, Length(FDocName) - 1);
    FDocName := Copy(FDocName, Posn('/', FDocName, -1) + 1, 255);
    I := Pos('?', FDocName);
    if I > 0 then
      FDocName := Copy(FDocName, 1, I - 1);
  end;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.LocationSessionClosed(Sender: TObject; Error: Word);
var
  Proto, User, Pass, Host, Port, Path: string;
  RealLocation: string;
  I: Integer;
begin
  { Remove any bookmark from the URL }
  I := Pos('#', FLocation);
  if I > 0 then
    RealLocation := Copy(FLocation, 1, I - 1)
  else
    RealLocation := FLocation;

  { Parse the URL }
  ParseURL(RealLocation, Proto, User, Pass, Host, Port, Path);
  FDocName := Path;
  AdjustDocName;
  FConnected := FALSE;
  FLocationFlag := FALSE;
  { When relocation occurs doing a POST, new relocated page has to be GET }
  if FRequestType = httpPOST then
    FRequestType := httpGET;
  { Trigger the location changed event }
  if Assigned(FOnLocationChange) then
    FOnLocationChange(Self);
  { Clear header from previous operation }
  FRcvdHeader.Clear;
  { Clear status variables from previous operation }
  FHeaderLineCount := 0;
  FBodyLineCount := 0;
  FContentLength := -1;
  FContentType := '';
  FStatusCode := 0;
  { Adjust for proxy use  (Fixed: Nov 10, 2001) }
  if FProxy <> '' then
    FPort := ProxyPort;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.SocketSessionClosed(Sender: TObject; Error: Word);
begin
  if Error <> 0 then { WM 15 sep 2002 }
    FRequestDoneError := Error; { WM 15 sep 2002 }
  FConnected := FALSE;
  if FHeaderEndFlag then
  begin
    { TriggerHeaderEnd has not been called yet }
    TriggerHeaderEnd;
    if FLocationFlag then { 28/10/01 }
      LocationSessionClosed(Self, 0);
    Exit;
  end;
  if FBodyLineCount > 0 then
    TriggerDocEnd;
  SetReady; {StateChange(httpReady);}
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.SocketDataAvailable(Sender: TObject; Error: Word);
var
  Len: Integer;
  I: Integer;
begin
  I := sizeof(FReceiveBuffer) - FReceiveLen - 1;
  if I <= 0 then
    raise EHttpException.Create('HTTP line too long', httperrOverflow);

  Len := FCtrlSocket.ReceiveBuf(FReceiveBuffer[FReceiveLen], I);

  if FRequestType = httpAbort then
    Exit;

  if Len <= 0 then
    Exit;

  FReceiveBuffer[FReceiveLen + Len] := #0;
  FReceiveLen := FReceiveLen + Len;

  if FState = httpWaitingBody then
  begin
    if FReceiveLen > 0 then
    begin
      SetLength(FLastResponse, FReceiveLen);
      Move(FReceiveBuffer, FLastResponse[1], FReceiveLen);
      if Assigned(FNext) then
        FNext
      else
        SetReady;
    end;
    FReceiveLen := 0;
    Exit;
  end;

  while FReceiveLen > 0 do
  begin
    I := Pos(#10, FReceiveBuffer);
    if I <= 0 then
      break;
    if I > FReceiveLen then
      break;

    if (I > 1) and (FReceiveBuffer[I - 2] = #13) then
      FLastResponse := Copy(FReceiveBuffer, 1, I - 2)
    else
      FLastResponse := Copy(FReceiveBuffer, 1, I - 1);

    FReceiveLen := FReceiveLen - I;
    if FReceiveLen > 0 then
      Move(FReceiveBuffer[I], FReceiveBuffer[0], FReceiveLen + 1);

    if FState in [httpWaitingHeader, httpWaitingBody] then
    begin
      if Assigned(FNext) then
        FNext
      else
        SetReady;
    end
    else
    begin
      if Assigned(FOnDataAvailable) then
        FOnDataAvailable(Self, Error);
    end;
  end;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

procedure THttpClient.SendData(Sender: TObject; Error: Word);
var
  Len: Integer;
begin
  while True do
  begin
    Len := FSendStream.Read(FSendBuffer, SizeOf(FSendBuffer));
    if Len <= 0 then
    begin
      TriggerSendEnd;
      Break;
    end
    else
    begin
      FSentCount := FSentCount + Len;
      TriggerSendData(@FSendBuffer, Len);
      FCtrlSocket.SendBuf(FSendBuffer, Len);
      Sleep(0);
    end;
  end;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
{ This will start the Get process and wait until terminated (blocking)      }

procedure THttpClient.Get;
begin
  DoRequest(httpGet);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
{ This will start the Head process and wait until terminated (blocking)     }

procedure THttpClient.Head;
begin
  DoRequest(httpHEAD);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
{ This will start the Post process and wait until terminated (blocking)     }

procedure THttpClient.Post;
begin
  DoRequest(httpPOST);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

function GetBaseUrl(const Url: string): string;
var
  I: Integer;
begin
  I := 1;
  while (I <= Length(Url)) and (Url[I] <> '?') do
    Inc(I);
  Dec(I);
  while (I > 0) and (not (Url[I] in ['/', ':'])) do
    Dec(I);
  Result := Copy(Url, 1, I);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
{ Syntax of an URL: protocol://[user[:password]@]server[:port]/path         }

procedure ParseURL(const url: string; var Proto, User, Pass, Host,
    Port, Path: string);
var
  p, q, i: Integer;
  s: string;
  CurPath: string;
begin
  CurPath := Path;
  proto := '';
  User := '';
  Pass := '';
  Host := '';
  Port := '';
  Path := '';

  if Length(url) < 1 then
    Exit;

  { Handle path beginning with "./" or "../".          }
  { This code handle only simple cases !               }
  { Handle path relative to current document directory }
  if (Copy(url, 1, 2) = './') then
  begin
    p := Posn('/', CurPath, -1);
    if p > Length(CurPath) then
      p := 0;
    if p = 0 then
      CurPath := '/'
    else
      CurPath := Copy(CurPath, 1, p);
    Path := CurPath + Copy(url, 3, Length(url));
    Exit;
  end
    { Handle path relative to current document parent directory }
  else if (Copy(url, 1, 3) = '../') then
  begin
    p := Posn('/', CurPath, -1);
    if p > Length(CurPath) then
      p := 0;
    if p = 0 then
      CurPath := '/'
    else
      CurPath := Copy(CurPath, 1, p);

    s := Copy(url, 4, Length(url));
    { We could have several levels }
    while TRUE do
    begin
      CurPath := Copy(CurPath, 1, p - 1);
      p := Posn('/', CurPath, -1);
      if p > Length(CurPath) then
        p := 0;
      if p = 0 then
        CurPath := '/'
      else
        CurPath := Copy(CurPath, 1, p);
      if (Copy(s, 1, 3) <> '../') then
        break;
      s := Copy(s, 4, Length(s));
    end;

    Path := CurPath + Copy(s, 1, Length(s));
    Exit;
  end;

  p := pos('://', url);
  q := p;
  if p <> 0 then
  begin
    S := Copy(url, 1, p - 1);
    for i := 1 to Length(S) do
    begin
      if not (S[i] in UriProtocolSchemeAllowedChars) then
      begin
        q := i;
        Break;
      end;
    end;
    if q < p then
    begin
      p := 0;
      proto := 'http';
    end;
  end;
  if p = 0 then
  begin
    if (url[1] = '/') then
    begin
      { Relative path without protocol specified }
      proto := 'http';
      p := 1;
      if (Length(url) > 1) and (url[2] <> '/') then
      begin
        { Relative path }
        Path := Copy(url, 1, Length(url));
        Exit;
      end;
    end
    else if lowercase(Copy(url, 1, 5)) = 'http:' then
    begin
      proto := 'http';
      p := 6;
      if (Length(url) > 6) and (url[7] <> '/') then
      begin
        { Relative path }
        Path := Copy(url, 6, Length(url));
        Exit;
      end;
    end
    else if lowercase(Copy(url, 1, 7)) = 'mailto:' then
    begin
      proto := 'mailto';
      p := pos(':', url);
    end;
  end
  else
  begin
    proto := Copy(url, 1, p - 1);
    inc(p, 2);
  end;
  s := Copy(url, p + 1, Length(url));

  p := pos('/', s);
  q := pos('?', s);
  if (q > 0) and ((q < p) or (p = 0)) then
    p := q;
  if p = 0 then
    p := Length(s) + 1;
  Path := Copy(s, p, Length(s));
  s := Copy(s, 1, p - 1);

  p := Posn(':', s, -1);
  if p > Length(s) then
    p := 0;
  q := Posn('@', s, -1);
  if q > Length(s) then
    q := 0;
  if (p = 0) and (q = 0) then
  begin { no user, password or port }
    Host := s;
    Exit;
  end
  else if q < p then
  begin { a port given }
    Port := Copy(s, p + 1, Length(s));
    Host := Copy(s, q + 1, p - q - 1);
    if q = 0 then
      Exit; { no user, password }
    s := Copy(s, 1, q - 1);
  end
  else
  begin
    Host := Copy(s, q + 1, Length(s));
    s := Copy(s, 1, q - 1);
  end;
  p := pos(':', s);
  if p = 0 then
    User := s
  else
  begin
    User := Copy(s, 1, p - 1);
    Pass := Copy(s, p + 1, Length(s));
  end;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

function EncodeStr(Encoding: THttpEncoding; const Value: string): string;
begin
  Result := EncodeLine(Encoding, @Value[1], Length(Value));
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

function EncodeLine(
  Encoding: THttpEncoding;
  SrcData: PChar;
  Size: Integer): string;
var
  Offset: Integer;
  Pos1: Integer;
  Pos2: Integer;
  I: Integer;
begin
  SetLength(Result, Size * 4 div 3 + 4);
  FillChar(Result[1], Size * 4 div 3 + 2, #0);

  if Encoding = encUUEncode then
  begin
    Result[1] := Char(((Size - 1) and $3F) + $21);
    Size := ((Size + 2) div 3) * 3;
  end;
  Offset := 2;
  Pos1 := 0;
  Pos2 := 0;
  case Encoding of
    encUUEncode: Pos2 := 2;
    encBase64, encMime: Pos2 := 1;
  end;
  Result[Pos2] := #0;

  while Pos1 < Size do
  begin
    if Offset > 0 then
    begin
      Result[Pos2] := Char(ord(Result[Pos2]) or
        ((ord(SrcData[Pos1]) and
        ($3F shl Offset)) shr Offset));
      Offset := Offset - 6;
      Inc(Pos2);
      Result[Pos2] := #0;
    end
    else if Offset < 0 then
    begin
      Offset := Abs(Offset);
      Result[Pos2] := Char(ord(Result[Pos2]) or
        ((ord(SrcData[Pos1]) and
        ($3F shr Offset)) shl Offset));
      Offset := 8 - Offset;
      Inc(Pos1);
    end
    else
    begin
      Result[Pos2] := Char(ord(Result[Pos2]) or
        ((ord(SrcData[Pos1]) and $3F)));
      Inc(Pos2);
      Inc(Pos1);
      Result[Pos2] := #0;
      Offset := 2;
    end;
  end;

  case Encoding of
    encUUEncode:
      begin
        if Offset = 2 then
          Dec(Pos2);
        for i := 2 to Pos2 do
          Result[i] := bin2uue[ord(Result[i]) + 1];
      end;
    encBase64, encMime:
      begin
        if Offset = 2 then
          Dec(Pos2);
        for i := 1 to Pos2 do
          Result[i] := bin2b64[ord(Result[i]) + 1];
        while (Pos2 and 3) <> 0 do
        begin
          Inc(Pos2);
          Result[Pos2] := '=';
        end;
      end;
  end;
  SetLength(Result, Pos2);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
{ Find the count'th occurence of the s string in the t string.              }
{ If count < 0 then look from the back                                      }

function Posn(const s, t: string; Count: Integer): Integer;
var
  i, h, Last: Integer;
  u: string;
begin
  u := t;
  if Count > 0 then
  begin
    Result := Length(t);
    for i := 1 to Count do
    begin
      h := Pos(s, u);
      if h > 0 then
        u := Copy(u, h + 1, Length(u))
      else
      begin
        u := '';
        Inc(Result);
      end;
    end;
    Result := Result - Length(u);
  end
  else if Count < 0 then
  begin
    Last := 0;
    for i := Length(t) downto 1 do
    begin
      u := Copy(t, i, Length(t));
      h := Pos(s, u);
      if (h <> 0) and ((h + i) <> Last) then
      begin
        Last := h + i - 1;
        Inc(count);
        if Count = 0 then
          break;
      end;
    end;
    if Count = 0 then
      Result := Last
    else
      Result := 0;
  end
  else
    Result := 0;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

function IsDigit(Ch: Char): Boolean;
begin
  Result := (ch in ['0'..'9']);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

function IsXDigit(Ch: char): Boolean;
begin
  Result := (ch in ['0'..'9']) or (ch in ['a'..'f']) or (ch in ['A'..'F']);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

function XDigit(Ch: char): Integer;
begin
  if ch in ['0'..'9'] then
    Result := ord(Ch) - ord('0')
  else
    Result := (ord(Ch) and 15) + 9;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

function htoin(value: PChar; len: Integer): Integer;
var
  i: Integer;
begin
  Result := 0;
  i := 0;
  while (i < len) and (Value[i] = ' ') do
    i := i + 1;
  while (i < len) and (isxDigit(Value[i])) do
  begin
    Result := Result * 16 + xdigit(Value[i]);
    i := i + 1;
  end;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

function htoi2(value: PChar): Integer;
begin
  Result := htoin(value, 2);
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

function UrlEncode(S: string): string;
var
  I: Integer;
begin
  Result := '';
  for I := 1 to Length(S) do
  begin
    if S[I] in ['0'..'9', 'A'..'Z', 'a'..'z'] then
      Result := Result + S[I]
    else
      Result := Result + '%' + IntToHex(Ord(S[I]), 2);
  end;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

function UrlDecode(S: string): string;
var
  I: Integer;
  Ch: Char;
begin
  Result := '';
  I := 1;
  while (I <= Length(S)) and (S[I] <> '&') do
  begin
    Ch := S[I];
    if Ch = '%' then
    begin
      Ch := chr(htoi2(@S[I + 1]));
      Inc(I, 2);
    end
    else if Ch = '+' then
      Ch := ' ';
    Result := Result + Ch;
    Inc(I);
  end;
end;

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

end.

