
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Library (NetCLX)           }
{ WebSrv.pas                                            }
{                                                       }
{ for Delphi 6, 7                                       }
{ Version 2.6.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   www:      http://www.dwp42.org                      }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/patch60/source/WebSrv.pas,v 1.3 2003/11/16 20:54:03 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WebSrv.pas,v $
  Revision 1.3  2003/11/16 20:54:03  iwache
  Bug in function TWebRequest.ReadContent fixed.
  This function failed while reading Octet type
  binary files.

  Revision 1.2  2003/10/07 19:06:16  iwache
  TWebRequest.ReadHeader rewritten. Works now much faster.

  Revision 1.1  2003/09/21 19:53:41  iwache
  New NetCLX units added.

  ----------------------------------------------------------------------------
}


unit WebSrv;

interface

uses
  Windows, SysUtils, Classes, Contnrs, Sockets,
{$IFDEF WDOSX}
  WDosClock,
{$ENDIF}
  Threads;


type
  TWebMethodType = (wmNone, wmGet, wmPost);

  TWebServer = class;

  TWebRequest = class
  private
    FSocket: TCustomIpClient;
    FURL: string;
    FQuery: string;
    FServer: TWebServer;
    FHeader: string;
    FMethod: string;
    FContent: string;
    FPathInfo: string;
    FRootPathInfo: string;
    FMethodType: TWebMethodType;
    FRequestAccept: String;
    FRequestAcceptEncoding: String;
    FRequestAcceptLanguage: String;
    FRequestConnection: String;
    FRequestContentLength: string;
    FContentLength: Integer;
    FRequestContentType: String;
    FRequestHost: String;
    FRequestReferer: String;
    FRequestUserAgent: String;
    FProtocolVersion: string;
    FServerPort: string;
    FHeaderLines: TStrings;
    function ReadHeader: Boolean;
    function GetQueryFields(Value: string): string;
    function GetHeaderFields(Value: string): string;
  public
    constructor Create(Server: TWebServer; Socket: TCustomIpClient);
    function ReadContent: Boolean;
    function SaveContent(Stream: TStream): Boolean;
    destructor Destroy; override;
    property Content: string read FContent;
    property ContentLength: Integer read FContentLength;
    property QueryFields[Name: string]: string read GetQueryFields;
    property HeaderFields[Name: string]: string read GetHeaderFields;
    property PathInfo: string read FPathInfo;
    property Query: string read FQuery;
    property URL: string read FURL;
    property Method: string read FMethod;
    property MethodType: TWebMethodType read FMethodType;
    property RootPathInfo: string read FRootPathInfo;
    property RequestAccept: string read FRequestAccept;
    property RequestAcceptEncoding: string read FRequestAcceptEncoding;
    property RequestAcceptLanguage: string read FRequestAcceptLanguage;
    property RequestConnection: string read FRequestConnection;
    property RequestContentLength: string read FRequestContentLength;
    property RequestContentType: string read FRequestContentType;
    property RequestHost: string read FRequestHost;
    property RequestReferer: string read FRequestReferer;
    property RequestUserAgent: string read FRequestUserAgent;
    property ProtocolVersion: string read FProtocolVersion;
    property ServerPort: string read FServerPort;
  end;

  TWebResponse = class
  private
    FSent: Boolean;
    FSocket: TCustomIpClient;
    FServer: TWebServer;
    FStatusCode: Integer;
    FContent: string;
    FHeaderSent: Boolean;
    FLocation: string;
    FContentLength: Integer;
    FContentType: string;
    FStatusText: string;
    FHeaders: string;
  protected
    procedure SendHeader;
  public
    constructor Create(Server: TWebServer; Socket: TCustomIpClient);
    property Sent: Boolean read FSent;
    function SendResponse: Boolean;
    function SendString(s: string): Boolean;
    function SendFile(FileName: string): Boolean;
    procedure Redirect(URL: string);
    property Content: string read FContent write FContent;
    property ContentType: string read FContentType write FContentType;
    property ContentLength: Integer read FContentLength write FContentLength;
    property StatusCode: Integer read FStatusCode write FStatusCode;
    property Location: string read FLocation write FLocation;
    property StatusText: string read FStatusText write FStatusText;
    property Headers: string read FHeaders write FHeaders;
  end;

  TWebProcedure = procedure(Server: TWebServer; Request: TWebRequest; Response:
    TWebResponse);

  TAction = class
  private
    FWebProcedure: TWebProcedure;
    FPathInfo: string;
  public
    property WebProcedure: TWebProcedure read FWebProcedure write FWebProcedure;
    property PathInfo: string read FPathInfo write FPathInfo;
  end;

  TWebServer = class
  private
    FTcpServer: TTcpServer;
    FActive: Boolean;
    FRemoteHost: string;
    FLocalPort: string;
    FRootPath: string;
    FDefaultPage: string;
    FActions: TObjectList;
    procedure SetActive(Value: Boolean);
    procedure TcpServerAccept(Sender: TObject; ClientSocket: TCustomIpClient);
  protected
    procedure ProcessAction(Request: TWebRequest; Response: TWebResponse);
      virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Start;
    procedure Stop;
    procedure AddAction(PathInfo: string; WebProcedure: TWebProcedure);
    property RootPath: string read FRootPath write FRootPath;
    property DefaultPage: string read FDefaultPage write FDefaultPage;
    property RemoteHost: string read FRemoteHost write FRemoteHost;
    property LocalPort: string read FLocalPort write FLocalPort;
    property Active: Boolean read FActive write SetActive;
  end;

  TClientDataThread = class(TThread)
  private
    FSocket: TCustomIpClient;
    FWebServer: TWebServer;
  public
    constructor Create(CreateSuspended: Boolean);
    procedure Execute; override;
    property Socket: TCustomIpClient read FSocket write FSocket;
    property WebServer: TWebServer read FWebServer write FWebServer;
  end;

function ParseParamByNumber(Texto: AnsiString; NumParam: Integer; const 
    Separador: AnsiString): string;

function GoodDir(s: string): string;

implementation

type
  TContType = record
    Ext: string;
    ContentType: string;
  end;

const
  ContTypes: array[1..10] of TContType = (
    (Ext: '.TXT'; ContentType: 'text/plain'),
    (Ext: '.DOC'; ContentType: 'application/msword'),
    (Ext: '.BAT'; ContentType: 'text/plain'),
    (Ext: '.INI'; ContentType: 'text/plain'),
    (Ext: '.HTM'; ContentType: 'text/html'),
    (Ext: '.EXE'; ContentType: 'application/octet-stream'),
    (Ext: '.COM'; ContentType: 'application/octet-stream'),
    (Ext: '.GIF'; ContentType: 'image/gif'),
    (Ext: '.JPG'; ContentType: 'image/jpeg'),
    (Ext: '.BMP'; ContentType: 'image/bmp')
    );

function ParseParamByNumber(Texto: AnsiString; NumParam: Integer; const 
    Separador: AnsiString): string;
var
  n, i: Integer;
begin
  n := 1;
  Result := '';
  while n <> NumParam do
  begin
    i := Pos(Separador, Texto);
    if i = 0 then
      Exit;
    Texto := TrimLeft(Copy(Texto, i + Length(Separador), Length(Texto)));
    Inc(n);
  end;
  i := Pos(Separador, Texto);
  if i > 0 then
    Texto := Copy(Texto, 1, i - 1);
  Result := Trim(Texto);
end;

function Web_GetParam(const Params, ParamName, Default: string): string;
var
  i: Integer;
  S: string;
begin
  Result := Default;
  for i := 1 to 1000 do
  begin
    S := ParseParamByNumber(Params, i, '&');
    if S = '' then
      Exit;
    if UpperCase(ParseParamByNumber(S, 1, '=')) = UpperCase(ParamName) then
    begin
      Result := ParseParamByNumber(S, 2, '=');
      Exit;
    end;
  end;
end;

function Web_GetContentType(FileName: string): string;
var
  i: Integer;
  Ext: string;
begin
  Ext := UpperCase(ExtractFileExt(FileName));
  for i := Low(ContTypes) to High(ContTypes) do
  begin
    if ContTypes[i].Ext = Ext then
    begin
      Result := ContTypes[i].ContentType;
      Exit;
    end;
  end;
  Result := 'application/octet-stream';
end;

function GetTimeCounter: Integer;
begin
{$IFDEF WDOSX}
  Result := GetMSCounter;
{$ENDIF}
{$IFDEF WIN}
  Result := GetTickCount;
{$ENDIF}
end;

function StrRight(s: string; cuantos: Integer): string;
begin
  StrRight := Copy(s, Length(s) - Cuantos + 1, 255);
end;

function GoodDir(s: string): string;
begin
  if s = '' then
    GoodDir := ''
  else
  begin
    s := Trim(s);
    if StrRight(s, 1) = '\' then
      GoodDir := s
    else
      GoodDir := s + '\';
  end;
end;

{ TWebRequest }

constructor TWebRequest.Create(Server: TWebServer; Socket: TCustomIpClient);
begin
  inherited Create;
  FHeader := '';
  FContent := '';
  FServer := Server;
  FSocket := Socket;
  FServerPort := Server.LocalPort;
  FHeaderLines := TStringList.Create;
  ReadHeader;
  if MethodType = wmPost then
    ReadContent;
end;

destructor TWebRequest.Destroy;
begin
  FHeaderLines.Free;
  inherited Destroy;
end;

function TWebRequest.GetQueryFields(Value: string): string;
begin
  Result := Web_GetParam(FQuery, Value, '');
end;

function TWebRequest.GetHeaderFields(Value: string): string;
var
  I, Col: Integer;
  HeaderLine: string;
begin
  for I := 0 to FHeaderLines.Count - 1 do
  begin
    HeaderLine := FHeaderLines[I];
    if HeaderLine = '' then Continue;
    Col := Pos(':', HeaderLine) + 1;
    if StrLIComp(@HeaderLine[1], PChar(Value + ':'), Length(Value) + 1) = 0 then
    begin
      Result := Trim(Copy(HeaderLine, Col, Length(HeaderLine)));
      Result := AnsiDequotedStr(Result, '"');
      Result := AnsiDequotedStr(Result, '''');
      Exit;
    end;
  end;
  Result := '';
end;

function TWebRequest.SaveContent(Stream: TStream): Boolean;
var
  Timeout: Int64;
  Buffer: string;
  Len: Integer;
  Offset: Integer;
begin
  // Receive Content, with a timeout of 10 seconds between chunks
  Offset := 0;
  FContent := '';
  Timeout := GetTimeCounter + 10000;

  while (Offset {Length(FContent)} < ContentLength)
      and (GetTimeCounter < Timeout) do
  begin
    // Check if we are still online
    if FSocket.ConnectionAborted then Break;

    // Any data to read from the socket?
    while FSocket.WaitForData do
    begin
      SetLength(Buffer, 512);
      Len := FSocket.ReceiveBuf(Pointer(Buffer)^, 512);
      Stream.WriteBuffer(Pointer(Buffer)^, Len);
      Inc(Offset, Len);
      Timeout := GetTimeCounter + 10000;
    end;
    Sleep(0);
  end;
  // Done
  Result := (Offset = ContentLength);
end;

function TWebRequest.ReadContent: Boolean;
var
  Stream: TStream;
begin
  // If we've already read the content, we're done.
  if FContent <> '' then
  begin
    Result := True;
    Exit;
  end;
  Stream := TMemoryStream.Create;
  try
    Result := SaveContent(Stream);
    if Result then
    begin
      Stream.Position := 0;
      SetLength(FContent, Stream.Size);
      Stream.Read(FContent[1], Stream.Size);
    end;
  finally
    Stream.Free;
  end;
end;

function TWebRequest.ReadHeader: Boolean;
var
  I: Integer;
  Timeout: Int64;
  Line: string;
  RcvdLine: string;
  Col: Integer;
  Buffer: string;
  Len: Integer;
  HeaderLen: Integer;
begin
  // Get Header by reading from the socket.
  Result := False;
  FHeader := '';
  Timeout := GetTimeCounter + 5000;

  while not Result and (GetTimeCounter < Timeout) do
  begin
    // Check if we are still online
    if FSocket.ConnectionAborted then Break;

    // Any data to read from the socket?
    while FSocket.WaitForData do
    begin
      SetLength(Buffer, 1024);
      Len := FSocket.PeekBuf(Buffer[1], Length(Buffer));
      SetLength(Buffer, Len);
      HeaderLen := Pos(CRLF + CRLF, FHeader + Buffer) + Length(CRLF + CRLF) - 1;
      if HeaderLen <> 0 then
      begin
        FSocket.ReceiveBuf(Buffer[1], HeaderLen - Length(FHeader));
        Result := True;
      end
      else
        FSocket.ReceiveBuf(Buffer[1], Len);
      FHeader := FHeader + Buffer;
      if Result then Break;
      Timeout := GetTimeCounter + 5000;
    end;
    Sleep(0);
  end;
  if not Result then Exit;

  // Extract basic information
  Line := ParseParamByNumber(FHeader, 1, CRLF);
  FMethod := ParseParamByNumber(Line, 1, ' ');
  FURL := ParseParamByNumber(Line, 2, ' ');
  FProtocolVersion := ParseParamByNumber(Line, 3, ' ');

  FHeaderLines.Text := FHeader;
  for I := 0 to FHeaderLines.Count - 1 do
  begin
    RcvdLine := FHeaderLines[I];
    if RcvdLine = '' then Continue;
    Col := Pos(':', RcvdLine) + 1;
    if StrLIComp(@RcvdLine[1], 'content-type:', 13) = 0 then
      FRequestContentType := Trim(Copy(RcvdLine, Col, Length(RcvdLine)))
    else if StrLIComp(@RcvdLine[1], 'content-length:', 15) = 0 then
    begin
      FRequestContentLength := Trim(Copy(RcvdLine, Col, Length(RcvdLine)));
      FContentLength := StrToIntDef(FRequestContentLength, 0);
    end
    else if StrLIComp(@RcvdLine[1], 'Accept:', 7) = 0 then
        FRequestAccept:= Trim(Copy(RcvdLine, Col, Length(RcvdLine)))
    else if StrLIComp(@RcvdLine[1], 'Referer:', 8) = 0 then
        FRequestReferer := Trim(Copy(RcvdLine, Col, Length(RcvdLine)))
    else if StrLIComp(@RcvdLine[1], 'Accept-Language:', 16) = 0 then
        FRequestAcceptLanguage := Trim(Copy(RcvdLine, Col, Length(RcvdLine)))
    else if StrLIComp(@RcvdLine[1], 'Accept-Encoding:', 16) = 0 then
        FRequestAcceptEncoding := Trim(Copy(RcvdLine, Col, Length(RcvdLine)))
    else if StrLIComp(@RcvdLine[1], 'User-Agent:', 11) = 0 then
        FRequestUserAgent := Trim(Copy(RcvdLine, Col, Length(RcvdLine)))
    else if StrLIComp(@RcvdLine[1], 'Host:', 5) = 0 then
        FRequestHost := Trim(Copy(RcvdLine, Col, Length(RcvdLine)))
    else if StrLIComp(@RcvdLine[1], 'Connection:', 11) = 0 then
        FRequestConnection := Trim(Copy(RcvdLine, Col, Length(RcvdLine)));
  end;

  // Check method type
  if UpperCase(FMethod) = 'GET' then
    FMethodType := wmGet
  else if UpperCase(FMethod) = 'POST' then
    FMethodType := wmPost
  else
  begin
    FMethodType := wmNone;
    Exit;
  end;

  // Check URL, PathInfo and Query
  I := Pos('?', FURL);
  if I = 0 then
  begin
    FQuery := '';
    FPathInfo := FURL;
  end
  else
  begin
    FQuery := Copy(FURL, I + 1, Length(FURL));
    FPathInfo := Copy(FURL, 1, I - 1);
  end;

  // Check for default Page
  if FPathInfo = '/' then
    FPathInfo := FServer.DefaultPage;

  // extract RootPathInfo
  FRootPathInfo := FPathInfo;
  if Length(FRootPathInfo) > 1 then
  begin
    Delete(FRootPathInfo, 1, 1);
    I := Pos('/', FRootPathInfo);
    if I > 0 then
      FRootPathInfo := Copy(FRootPathInfo, 1, I - 1);
  end;
  FRootPathInfo := '/' + FRootPathInfo;
end;

{ TWebResponse }

constructor TWebResponse.Create(Server: TWebServer; Socket: TCustomIpClient);
begin
  inherited Create;
  FSent := False;
  FContent := '';
  FServer := Server;
  FSocket := Socket;
  FLocation := '';
  FStatusCode := 200;
  FStatusText := 'OK';
  FHeaderSent := False;
  FContentLength := -1;
  FContentType := 'text/html';
end;

procedure TWebResponse.SendHeader;
var
  Str: string;
begin
  Str := 'HTTP/1.1 ' + IntToStr(StatusCode) + ' ' + StatusText + CRLF +
      'Server: Delphi DWPL Server/1.1 by Team-DWP' + CRLF;
  if FHeaders <> '' then
    Str := Str + Headers
  else
  begin
    Str := Str + 'Content-Type: ' + ContentType + CRLF;
    if FContentLength >= 0 then
      Str := Str + 'Content-Length: ' +
        IntToStr(FContentLength) + CRLF;
    if FLocation <> '' then
      Str := Str + 'Location: ' + FLocation + CRLF;
    Str := Str + CRLF;
  end;
  FSocket.SendBuf(Pointer(Str)^, Length(Str));
  FHeaderSent := True;
end;

function TWebResponse.SendResponse: Boolean;
begin
  if FContent = '' then
  begin
    if not FHeaderSent then
      SendHeader;
    Result := True;
  end
  else
  begin
    if not FHeaderSent then
      FContentLength := Length(FContent);
    Result := SendString(Content);
  end;
end;

function TWebResponse.SendString(S: string): Boolean;
begin
  if Length(S) = 0 then
    Result := True
  else
  begin
    if not FHeaderSent then
      SendHeader;
    Result := (FSocket.SendBuf(S[1], Length(S)) = Length(S));
  end;
end;

function TWebResponse.SendFile(FileName: string): Boolean;
const
  CHUNK_SIZE = 2048;
var
  H: Integer;
  Buffer: string;
begin
  if FHeaderSent then
  begin
    Result := False;
    Exit;
  end;
  H := FileOpen(FileName, fmOpenRead);
  if H < 0 then
    Result := False
  else
  begin
    ContentType := Web_GetContentType(FileName);
    ContentLength := FileSeek(H, 0, 2);
    SendHeader;
    FileSeek(H, 0, 0);
    Result := False;
    repeat
      SetLength(Buffer, CHUNK_SIZE);
      SetLength(Buffer, FileRead(H, Buffer[1], Length(Buffer)));
      if Buffer = '' then
      begin
        Result := True;
        Break;
      end;
      if FSocket.SendBuf(Buffer[1], Length(Buffer)) <> Length(Buffer) then
        Break;
    until False;
    FileClose(H);
  end;
end;

procedure TWebResponse.Redirect(URL: string);
begin
  StatusCode := 302;
  Location := URL;
end;

{ TWebServer }

constructor TWebServer.Create;
begin
  inherited Create;
  FRemoteHost := '';
  FLocalPort := '80';
  FRootPath := '.\';
  FDefaultPage := '/index.htm';

  FTcpServer := TTcpServer.Create(nil);
  FTcpServer.OnAccept := TcpServerAccept;
  FActions := TObjectList.Create(True);
end;

destructor TWebServer.Destroy;
begin
  Stop;
  FActions.Free;
  FTcpServer.Free;
  inherited Destroy;
end;

procedure TWebServer.Start;
begin
  if not Active then
  begin
    FTcpServer.RemoteHost := FRemoteHost;
    FTcpServer.LocalPort := FLocalPort;
    FTcpServer.Open;
    FActive := FTcpServer.Active;
  end;
end;

procedure TWebServer.Stop;
begin
  if Active then
  begin
    FTcpServer.Close;
    FActive := FTcpServer.Active;
  end;
end;

procedure TWebServer.AddAction(PathInfo: string; WebProcedure: TWebProcedure);
var
  Action: TAction;
begin
  Action := TAction.Create;
  Action.PathInfo := Uppercase(PathInfo);
  Action.WebProcedure := WebProcedure;
  FActions.Add(Action);
end;

procedure TWebServer.ProcessAction(Request: TWebRequest; Response:
  TWebResponse);
var
  I: Integer;
  Path: string;
  Action: TAction;
  PathInfo: string;
begin
  // Check for a handler registered for this request.
  PathInfo := Uppercase(Request.RootPathInfo);
  for I := 0 to FActions.Count -1 do
  begin
    Action := FActions[I] as TAction;
    if (Action.PathInfo = PathInfo) and Assigned(Action.WebProcedure) then
    begin
      Action.WebProcedure(Self, Request, Response);
      Exit;
    end;
  end;

  // There was no handler. Check to see if it's a file.
  Path := Copy(Request.PathInfo, 2, Length(Request.PathInfo));
  for i := 1 to Length(Path) do
    if Path[i] = '/' then
      Path[i] := '\';
  Path := GoodDir(RootPath) + Path;
  if FileExists(Path) then
  begin
    Response.SendFile(Path);
    Exit;
  end;

  // There was no handler or file. Show Error 404 "Not Found".
  Response.StatusCode := 404;
  Response.Content :=
    '<HTML><TITLE>404 NOT FOUND</TITLE><H1><B>404 NOT FOUND</B></H1>' +
    'The file ' + Request.PathInfo + ' is not on this server.<br>' +
    'Please try again using another URL.' +
    '</HTML>';
end;

procedure TWebServer.TcpServerAccept(Sender: TObject;
  ClientSocket: TCustomIpClient);
var
  DataThread: TClientDataThread;
begin
  DataThread := TClientDataThread.Create(True);
  DataThread.Socket := ClientSocket;
  DataThread.WebServer := Self;
  DataThread.Resume;
  DataThread.WaitFor;
end;

procedure TWebServer.SetActive(Value: Boolean);
begin
  if Value then Start else Stop;
end;

{ TClientDataThread }

constructor TClientDataThread.Create(CreateSuspended: Boolean);
begin
  inherited Create(CreateSuspended);
  FreeOnTerminate := True;
end;

procedure TClientDataThread.Execute;
var
  Request: TWebRequest;
  Response: TWebResponse;
begin
  // Process incoming request
  Request := TWebRequest.Create(WebServer, Socket);
  Response := TWebResponse.Create(WebServer, Socket);

  try
    if Request.MethodType <> wmNone then
    begin
      WebServer.ProcessAction(Request, Response);
      if not Response.Sent then
        Response.SendResponse;
    end;
  finally
    Request.Free;
    Response.Free;
  end;
end;

end.

