
{*******************************************************}
{                                                       }
{ DWPL Delphi WDosX and Cross Platform Library          }
{ (automation)                                          }
{ DwpLptCtrls.pas                                       }
{                                                       }
{ for Delphi 5, 6, 7                                    }
{ Version 2.6.0                                         }
{ Copyright (c) 2000-2003 by Team-DWP                   }
{ e-mail: team-dwp@dwp42.org                            }
{ www: http://www.dwp42.org                             }
{                                                       }
{ The initial developer of the code is                  }
{   Immo Wache, immo.wache@dwp42.org                    }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ Licensees holding a valid Team-DWP No-Nonsense        }
{ License for this Software may use this file in        }
{ accordance with such license, which appears in        }
{ the file license.txt that came with this software.    }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/automation/DwpLptCtrls.pas,v 1.1 2003/10/08 18:44:27 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: DwpLptCtrls.pas,v $
  Revision 1.1  2003/10/08 18:44:27  iwache
  Unit with component TdwpLptCtrl for LPT port access added.

  ----------------------------------------------------------------------------
}

unit DwpLptCtrls;

interface

uses
  SysUtils, Classes;

type
  TLptDevice = (ldNull, ldLPT1, ldLPT2, ldLPT3);

const
  LptDeviceNames: array[TLptDevice] of string =
  ('NULL:', 'LPT1:', 'LPT2:', 'LPT3:');

type
  TdwpLptCtrl = class(TComponent)
  private
    FActive: Boolean;
    FCtrl: Byte;
    FData: Byte;
    FDevice: TLptDevice;
    FDeviceAddr: Word;
    FStreamedActive: Boolean;
    function GetCtrl: Byte;
    function GetData: Byte;
    function GetStatus: Byte;
    procedure SetActive(Value: Boolean);
    procedure SetCtrl(Value: Byte);
    procedure SetData(Value: Byte);
    procedure SetDevice(Value: TLptDevice);
  protected
    procedure Loaded; override;
  public
    destructor Destroy; override;
    procedure Close;
    procedure Open;
    property Ctrl: Byte read GetCtrl write SetCtrl;
    property Data: Byte read GetData write SetData;
    property Status: Byte read GetStatus;
  published
    property Active: Boolean read FActive write SetActive;
    property Device: TLptDevice read FDevice write SetDevice;
  end;

  ELptCtrlError = class(Exception)
  end;

implementation

{$IFDEF WDOSX}
uses
  WDosSysUtils;
{$ELSE}
uses
  WinPorts;
{$ENDIF}

resourcestring
  sLptCtrlNotOpen = 'Device %s is not open';
  sLptCtrlOnOpenError = 'Operation on opened device %s not allowed';
  sLptCtrlOpenError = 'Device %s can''t be opened';

type
  TLptOpenDevice = ldLPT1..ldLPT3;
  TLptOpenDevices = set of TLptOpenDevice;

var
  LptOpenDevices: TLptOpenDevices;

{
 ************************************* TwdxLptCtrl *****************************
}

destructor TdwpLptCtrl.Destroy;
begin
  { close device }
  Close;
  inherited Destroy;
end;

procedure TdwpLptCtrl.Close;
begin
  if FActive then
  begin
    Exclude(LptOpenDevices, FDevice);
    FActive := False;
  end;
end;

procedure TdwpLptCtrl.Loaded;
begin
  inherited Loaded;
  Active := FStreamedActive;
end;

procedure TdwpLptCtrl.Open;
begin
  if not FActive then
  begin
    { check for a valid device }
    if (FDevice = ldNull) or (FDevice in LptOpenDevices) then
      raise ELptCtrlError.CreateFmt(sLptCtrlOpenError,
        [LptDeviceNames[FDevice]]);
    Include(LptOpenDevices, FDevice);
    { syncronize Ctrl and Data }
    FCtrl := Port[FDeviceAddr + 2];
    Port[FDeviceAddr] := FData;
    { set Active }
    FActive := True;
  end;
end;

procedure TdwpLptCtrl.SetActive(Value: Boolean);
begin
  if csLoading in ComponentState then
    FStreamedActive := Value
  else if Value then
    Open
  else
    Close;
end;

function TdwpLptCtrl.GetCtrl: Byte;
begin
  if FActive then
    Result := Port[FDeviceAddr + 2]
  else
    raise ELptCtrlError.CreateFmt(sLptCtrlNotOpen, [LptDeviceNames[FDevice]]);
end;

procedure TdwpLptCtrl.SetCtrl(Value: Byte);
begin
  if FActive then
  begin
    if FCtrl <> Value then
    begin
      Port[FDeviceAddr + 2] := Value;
      FCtrl := Value;
    end;
  end
  else
    raise ELptCtrlError.CreateFmt(sLptCtrlNotOpen, [LptDeviceNames[FDevice]]);
end;

function TdwpLptCtrl.GetData: Byte;
begin
  if FActive then
    Result := FData
  else
    raise ELptCtrlError.CreateFmt(sLptCtrlNotOpen, [LptDeviceNames[FDevice]]);
end;

procedure TdwpLptCtrl.SetData(Value: Byte);
begin
  if FActive then
  begin
    if FData <> Value then
    begin
      Port[FDeviceAddr] := Value;
      FData := Value;
    end;
  end
  else
    raise ELptCtrlError.CreateFmt(sLptCtrlNotOpen, [LptDeviceNames[FDevice]]);
end;

procedure TdwpLptCtrl.SetDevice(Value: TLptDevice);

const
{$IFDEF WDOSX}
  Offset: array[TLptDevice] of Byte = (0, 8, 10, 12);
{$ELSE}
  DeviceAddr: array[TLptDevice] of Word = (0, $378, $278, $3BC);
{$ENDIF}

begin
  if Value <> FDevice then
  begin
    if not FActive then
    begin
      FDevice := Value;
      { get port address of device }
{$IFDEF WDOSX}
      FDeviceAddr := MemW[$400 + Offset[FDevice]];
{$ELSE}
      FDeviceAddr := DeviceAddr[Value];
{$ENDIF}
    end
    else
      raise ELptCtrlError.CreateFmt(sLptCtrlOnOpenError,
        [LptDeviceNames[FDevice]]);
  end;
end;

function TdwpLptCtrl.GetStatus: Byte;
begin
  if FActive then
    Result := Port[FDeviceAddr + 1]
  else
    raise ELptCtrlError.CreateFmt(sLptCtrlNotOpen, [LptDeviceNames[FDevice]]);
end;

end.

