
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Visual Library (clx)       }
{ QDwplCtrls.pas                                        }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Kylix and Delphi Cross-Platform Runtime Library   }
{     Copyright (C) 1996, 2001                          }
{     Borland Software Corporation                      }
{     www: http://www.borland.com                       }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}

{ *************************************************************************** }
{                                                                             }
{ Kylix and Delphi Cross-Platform Visual Component Library                    }
{                                                                             }
{ Copyright (c) 2000, 2001 Borland Software Corporation                       }
{                                                                             }
{ This file may be distributed and/or modified under the terms of the GNU     }
{ General Public License (GPL) version 2 as published by the Free Software    }
{ Foundation and appearing at http://www.borland.com/kylix/gpl.html.          }
{                                                                             }
{ Licensees holding a valid Borland No-Nonsense License for this Software may }
{ use this file in accordance with such license, which appears in the file    }
{ license.txt that came with this Software.                                   }
{                                                                             }
{ *************************************************************************** }
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/QDwplCtrls.pas,v 1.1 2003/01/30 18:41:46 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: QDwplCtrls.pas,v $
  Revision 1.1  2003/01/30 18:41:46  iwache
  Unit for DWPL specific components added. Initial component
  is TRtcTimer, a timer component which uses the RTC timer chip
  of the PC. In Windows emulation mode it is mapped to the normal
  PIC-Timer.

  ----------------------------------------------------------------------------
}
unit QDwplCtrls;

{.$DEFINE DESIGNTIME}

interface

uses
{$IFNDEF DESIGNTIME}
  WQ, QTypes, QForms,
{$ENDIF}
  SysUtils, Types, Classes;

type
  TIntFreq =(ifNone, if32768, if16384, if8192,
             if4096, if2048,  if1024,  if512,
             if256,  if128,   if64,    if32,
             if16,   if8,     if4,     if2);

const
  DIntFreq =if64;

type
{$IFDEF DESIGNTIME}
  TRtcTimer = class(TComponent)
{$ELSE}
  TRtcTimer = class(THandleComponent)
{$ENDIF}
  private
    FInterval: Cardinal;
    FOnTimer: TNotifyEvent;
    FEnabled: Boolean;
    FIntFreq: TIntFreq;
    procedure SetEnabled(Value: Boolean);
    procedure SetInterval(Value: Cardinal);
    function GetIntFreq: TIntFreq;
    procedure SetIntFreq(Value: TIntFreq);
  protected
{$IFNDEF DESIGNTIME}
    procedure CreateWidget; override;
    function GetHandle: TQRtcTimer;
    procedure HookEvents; override;
    procedure Timer(Sender: TObject); virtual;
    function QRtcTimer: TQRtcTimer;
{$ENDIF}
  public
    constructor Create(AOwner: TComponent); override;
{$IFNDEF DESIGNTIME}
    property Handle: TQRtcTimer read GetHandle;
{$ENDIF}
  published
    property Enabled: Boolean read FEnabled write SetEnabled default True;
    property Interval: Cardinal read FInterval write SetInterval default 1000;
    property OnTimer: TNotifyEvent read FOnTimer write FOnTimer;
    property IntFreq: TIntFreq read GetIntFreq write SetIntFreq default DIntFreq;
  end;

implementation

constructor TRtcTimer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  IntFreq := DIntFreq;
  Interval := 1000;
  Enabled := True;
end;

{$IFNDEF DESIGNTIME}
procedure TRtcTimer.CreateWidget;
begin
  if not (csDesigning in ComponentState) then
    FHandle := TQRtcTimer.Create;
end;

function TRtcTimer.GetHandle: TQRtcTimer;
begin
  HandleNeeded;
  Result := FHandle as TQRtcTimer;
end;

procedure TRtcTimer.HookEvents;
begin
  QRtcTimer.OnTimer := Timer;
  inherited HookEvents;
end;

function TRtcTimer.QRtcTimer: TQRtcTimer;
begin
  Result := Handle;
end;
{$ENDIF}

procedure TRtcTimer.SetEnabled(Value: Boolean);
begin
  if Value <> FEnabled then
  begin
    FEnabled := Value;
{$IFNDEF DESIGNTIME}
    if not (csDesigning in ComponentState) then
    begin
      if Value then
        QRtcTimer.Interval := FInterval;
      QRtcTimer.Enabled := Value;
    end;
{$ENDIF}
  end;
end;

procedure TRtcTimer.SetInterval(Value: Cardinal);
begin
  if Value <> FInterval then
  begin
    FInterval := Value;
{$IFNDEF DESIGNTIME}
    if Enabled and not (csDesigning in ComponentState) then
      QRtcTimer.Interval := FInterval;
{$ENDIF}
  end;
end;

{$IFNDEF DESIGNTIME}
procedure TRtcTimer.Timer(Sender: TObject);
begin
  if Assigned(FOnTimer) then
  try
    FOnTimer(Self);
  except
    Application.HandleException(Self);
  end;
end;
{$ENDIF}

function TRtcTimer.GetIntFreq: TIntFreq;
begin
  Result := FIntFreq;
end;

procedure TRtcTimer.SetIntFreq(Value: TIntFreq);
begin
{$IFNDEF DESIGNTIME}
  QRtcTimer.IntFreq :=Integer(Value);
{$ENDIF}
  FIntFreq := Value;
end;

end.
