
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Visual Library (clx/vi)    }
{ WQ.pas                                                }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/WQ.pas,v 1.21 2004/03/30 20:23:00 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WQ.pas,v $
  Revision 1.21  2004/03/30 20:23:00  iwache
  - TQDialog, TQClxFileDialog, TQDirSelDlg,
  and TQMultiLineEdit class added.
  - some WQ key codes added e.g. Key_Home.

  Revision 1.20  2003/12/01 21:04:54  iwache
  Function QMessageBox now converts message box result
  consts from Vi layer to VisualCLX layer.

  Revision 1.19  2003/10/16 19:08:10  iwache
  Wrong behaviour of TQOpenComboBox events OnChanged
  and OnSelected fixed.

  Revision 1.18  2003/10/07 19:21:24  iwache
  Application.OnIdle event support added.

  Revision 1.17  2003/03/04 10:05:38  iwache
  - TQPopupMenu implemented
  - TQWidget.MapToGlobal added

  Revision 1.16  2003/03/03 15:45:53  iwache
  TQOpenComboBox added

  Revision 1.15  2003/02/04 17:25:57  iwache
  Handling message loop for thread synchronisation added.

  Revision 1.14  2003/01/25 11:21:05  iwache
  DWPL-Bugs-673813:
  Description:
  The dblclick event of TEdit and TListBox is invalid.
  The ReadOnly property of TEdit is invalid.

  Solution for Doubleclick:
  DblClick Event handler added in TWidgetControl.EventFilter.
  TCustomListBox.DblClick modified to fit behaviour with
  the TV-ListBox on doubleclick. An other solution would
  be fit behaiour of TV-ListBox to the VisualCLX one. But this
  would need much more work and tests.

  Solution for ReadOnly property of TEdit:
  TViInputLine added two new properties 'ReadOnly' and 'Text'.
  TViInputLine.HandleEvent and TViInputLine.SetData modified
  to handle ReadOnly property right.

  WQ.pas: modified for both bugs.
  ViTypes.pas: event code evMouseDbl  = $0800; added.
  This is a dummy TV event code.

  Revision 1.13  2003/01/23 19:19:55  iwache
  - TQPushButton now creates a TViButton without
  Options ofPreProcess and ofPostProcess to avoid
  double KeyDown and KeyPress events with TButton.
  - QInputDialog now handles accelerator keys
  for ALabel correct (e.g. &text). This is useful
  for InputQuery and InputBox functions.

  Revision 1.12  2003/01/22 20:52:18  iwache
  2'nd Fix for dwpl-Bug #667896 :
  Description:
  Buttons do not handle Enter...
  Solution:
  function TButton.NeedKey implemented and some modifications
  done to VisualCLX, WQ-Layer and TV-Layer.

  Revision 1.11  2003/01/20 22:08:56  iwache
  New Vi const kbSpace and WQ const Key_Space added.
  Both for handling space bar with focused buttons.

  Revision 1.10  2002/10/27 22:25:15  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
unit WQ;

interface

uses
  SysUtils, Classes, ViDrivers, ViTypes, ViUtils, ViClasses,
  ViScreens, ViViews, ViMenus, ViApp, ViDialogs, ViMsgBox,
  ViStdDlg, ViColorSel, ViMessages, ViThreads, ViEditors,
{$IFDEF WIN}
  WinTimers;
{$ENDIF}
{$IFDEF WDOSX}
  WDosTimers;
{$ENDIF}

const

{QT key codes }

    QKey_Escape = 4096 { $1000 };
    QKey_Tab = 4097 { $1001 };
    QKey_Backtab = 4098 { $1002 };
    QKey_Backspace = 4099 { $1003 };
    QKey_Return = 4100 { $1004 };
    QKey_Enter = 4101 { $1005 };
    QKey_Insert = 4102 { $1006 };
    QKey_Delete = 4103 { $1007 };

//    QKey_Pause = 4104 { $1008 };
//    QKey_Print = 4105 { $1009 };
//    QKey_SysReq = 4106 { $100a };

    QKey_Home = 4112 { $1010 };
    QKey_End = 4113 { $1011 };
    QKey_Left = 4114 { $1012 };
    QKey_Up = 4115 { $1013 };
    QKey_Right = 4116 { $1014 };
    QKey_Down = 4117 { $1015 };
    QKey_Prior = 4118 { $1016 };
    QKey_PageUp = 4118 { $1016 };
    QKey_Next = 4119 { $1017 };

//    QKey_PageDown = 4119 { $1017 };
//    QKey_Shift = 4128 { $1020 };
//    QKey_Control = 4129 { $1021 };
//    QKey_Meta = 4130 { $1022 };
//    QKey_Alt = 4131 { $1023 };
//    QKey_CapsLock = 4132 { $1024 };
//    QKey_NumLock = 4133 { $1025 };
//    QKey_ScrollLock = 4134 { $1026 };

    QKey_F1 = 4144 { $1030 };
    QKey_F2 = 4145 { $1031 };
    QKey_F3 = 4146 { $1032 };
    QKey_F4 = 4147 { $1033 };
    QKey_F5 = 4148 { $1034 };
    QKey_F6 = 4149 { $1035 };
    QKey_F7 = 4150 { $1036 };
    QKey_F8 = 4151 { $1037 };
    QKey_F9 = 4152 { $1038 };
    QKey_F10 = 4153 { $1039 };
    QKey_F11 = 4154 { $103a };
    QKey_F12 = 4155 { $103b };
    QKey_F13 = 4156 { $103c };
    QKey_F14 = 4157 { $103d };
    QKey_F15 = 4158 { $103e };
    QKey_F16 = 4159 { $103f };
    QKey_F17 = 4160 { $1040 };
    QKey_F18 = 4161 { $1041 };
    QKey_F19 = 4162 { $1042 };
    QKey_F20 = 4163 { $1043 };
    QKey_F21 = 4164 { $1044 };
    QKey_F22 = 4165 { $1045 };
    QKey_F23 = 4166 { $1046 };
    QKey_F24 = 4167 { $1047 };
    QKey_F25 = 4168 { $1048 };
    QKey_F26 = 4169 { $1049 };
    QKey_F27 = 4170 { $104a };
    QKey_F28 = 4171 { $104b };
    QKey_F29 = 4172 { $104c };
    QKey_F30 = 4173 { $104d };
    QKey_F31 = 4174 { $104e };
    QKey_F32 = 4175 { $104f };
    QKey_F33 = 4176 { $1050 };
    QKey_F34 = 4177 { $1051 };
    QKey_F35 = 4178 { $1052 };

//    QKey_Super_L = 4179 { $1053 };
//    QKey_Super_R = 4180 { $1054 };
//    QKey_Menu = 4181 { $1055 };
//    QKey_Hyper_L = 4182 { $1056 };
//    QKey_Hyper_R = 4183 { $1057 };
//    QKey_Help = 4184 { $1058 };

    QKey_Space = 32 { $20 };

//    QKey_Any = 32 { $20 };

    QKey_Exclam = 33 { $21 };
    QKey_QuoteDbl = 34 { $22 };
    QKey_NumberSign = 35 { $23 };
    QKey_Dollar = 36 { $24 };
    QKey_Percent = 37 { $25 };
    QKey_Ampersand = 38 { $26 };
    QKey_Apostrophe = 39 { $27 };
    QKey_ParenLeft = 40 { $28 };
    QKey_ParenRight = 41 { $29 };
    QKey_Asterisk = 42 { $2a };
    QKey_Plus = 43 { $2b };
    QKey_Comma = 44 { $2c };
    QKey_Minus = 45 { $2d };
    QKey_Period = 46 { $2e };
    QKey_Slash = 47 { $2f };
    QKey_0 = 48 { $30 };
    QKey_1 = 49 { $31 };
    QKey_2 = 50 { $32 };
    QKey_3 = 51 { $33 };
    QKey_4 = 52 { $34 };
    QKey_5 = 53 { $35 };
    QKey_6 = 54 { $36 };
    QKey_7 = 55 { $37 };
    QKey_8 = 56 { $38 };
    QKey_9 = 57 { $39 };
    QKey_Colon = 58 { $3a };
    QKey_Semicolon = 59 { $3b };
    QKey_Less = 60 { $3c };
    QKey_Equal = 61 { $3d };
    QKey_Greater = 62 { $3e };
    QKey_Question = 63 { $3f };
    QKey_At = 64 { $40 };
    QKey_A = 65 { $41 };
    QKey_B = 66 { $42 };
    QKey_C = 67 { $43 };
    QKey_D = 68 { $44 };
    QKey_E = 69 { $45 };
    QKey_F = 70 { $46 };
    QKey_G = 71 { $47 };
    QKey_H = 72 { $48 };
    QKey_I = 73 { $49 };
    QKey_J = 74 { $4a };
    QKey_K = 75 { $4b };
    QKey_L = 76 { $4c };
    QKey_M = 77 { $4d };
    QKey_N = 78 { $4e };
    QKey_O = 79 { $4f };
    QKey_P = 80 { $50 };
    QKey_Q = 81 { $51 };
    QKey_R = 82 { $52 };
    QKey_S = 83 { $53 };
    QKey_T = 84 { $54 };
    QKey_U = 85 { $55 };
    QKey_V = 86 { $56 };
    QKey_W = 87 { $57 };
    QKey_X = 88 { $58 };
    QKey_Y = 89 { $59 };
    QKey_Z = 90 { $5a };
    QKey_BracketLeft = 91 { $5b };
    QKey_Backslash = 92 { $5c };
    QKey_BracketRight = 93 { $5d };
    QKey_AsciiCircum = 94 { $5e };
    QKey_Underscore = 95 { $5f };
    QKey_QuoteLeft = 96 { $60 };

    QKey_BraceLeft = 123 { $7b };
    QKey_Bar = 124 { $7c };
    QKey_BraceRight = 125 { $7d };
    QKey_AsciiTilde = 126 { $7e };

//    QKey_nobreakspace = 160 { $a0 };
//    QKey_exclamdown = 161 { $a1 };
//    QKey_cent = 162 { $a2 };
//    QKey_sterling = 163 { $a3 };
//    QKey_currency = 164 { $a4 };
//    QKey_yen = 165 { $a5 };
//    QKey_brokenbar = 166 { $a6 };
//    QKey_section = 167 { $a7 };
//    QKey_diaeresis = 168 { $a8 };
//    QKey_copyright = 169 { $a9 };
//    QKey_ordfeminine = 170 { $aa };
//    QKey_guillemotleft = 171 { $ab };
//    QKey_notsign = 172 { $ac };
//    QKey_hyphen = 173 { $ad };
//    QKey_registered = 174 { $ae };
//    QKey_macron = 175 { $af };
//    QKey_degree = 176 { $b0 };
//    QKey_plusminus = 177 { $b1 };
//    QKey_twosuperior = 178 { $b2 };
//    QKey_threesuperior = 179 { $b3 };
//    QKey_acute = 180 { $b4 };
//    QKey_mu = 181 { $b5 };
//    QKey_paragraph = 182 { $b6 };
//    QKey_periodcentered = 183 { $b7 };
//    QKey_cedilla = 184 { $b8 };
//    QKey_onesuperior = 185 { $b9 };
//    QKey_masculine = 186 { $ba };
//    QKey_guillemotright = 187 { $bb };
//    QKey_onequarter = 188 { $bc };
//    QKey_onehalf = 189 { $bd };
//    QKey_threequarters = 190 { $be };
//    QKey_questiondown = 191 { $bf };
//    QKey_Agrave = 192 { $c0 };
//    QKey_Aacute = 193 { $c1 };
//    QKey_Acircumflex = 194 { $c2 };
//    QKey_Atilde = 195 { $c3 };
//    QKey_Adiaeresis = 196 { $c4 };
//    QKey_Aring = 197 { $c5 };
//    QKey_AE = 198 { $c6 };
//    QKey_Ccedilla = 199 { $c7 };
//    QKey_Egrave = 200 { $c8 };
//    QKey_Eacute = 201 { $c9 };
//    QKey_Ecircumflex = 202 { $ca };
//    QKey_Ediaeresis = 203 { $cb };
//    QKey_Igrave = 204 { $cc };
//    QKey_Iacute = 205 { $cd };
//    QKey_Icircumflex = 206 { $ce };
//    QKey_Idiaeresis = 207 { $cf };
//    QKey_ETH = 208 { $d0 };
//    QKey_Ntilde = 209 { $d1 };
//    QKey_Ograve = 210 { $d2 };
//    QKey_Oacute = 211 { $d3 };
//    QKey_Ocircumflex = 212 { $d4 };
//    QKey_Otilde = 213 { $d5 };
//    QKey_Odiaeresis = 214 { $d6 };
//    QKey_multiply = 215 { $d7 };
//    QKey_Ooblique = 216 { $d8 };
//    QKey_Ugrave = 217 { $d9 };
//    QKey_Uacute = 218 { $da };
//    QKey_Ucircumflex = 219 { $db };
//    QKey_Udiaeresis = 220 { $dc };
//    QKey_Yacute = 221 { $dd };
//    QKey_THORN = 222 { $de };
//    QKey_ssharp = 223 { $df };
//    QKey_agrave_lower = 224 { $e0 };
//    QKey_aacute_lower = 225 { $e1 };
//    QKey_acircumflex_lower = 226 { $e2 };
//    QKey_atilde_lower = 227 { $e3 };
//    QKey_adiaeresis_lower = 228 { $e4 };
//    QKey_aring_lower = 229 { $e5 };
//    QKey_ae_lower = 230 { $e6 };
//    QKey_ccedilla_lower = 231 { $e7 };
//    QKey_egrave_lower = 232 { $e8 };
//    QKey_eacute_lower = 233 { $e9 };
//    QKey_ecircumflex_lower = 234 { $ea };
//    QKey_ediaeresis_lower = 235 { $eb };
//    QKey_igrave_lower = 236 { $ec };
//    QKey_iacute_lower = 237 { $ed };
//    QKey_icircumflex_lower = 238 { $ee };
//    QKey_idiaeresis_lower = 239 { $ef };
//    QKey_eth_lower = 240 { $f0 };
//    QKey_ntilde_lower = 241 { $f1 };
//    QKey_ograve_lower = 242 { $f2 };
//    QKey_oacute_lower = 243 { $f3 };
//    QKey_ocircumflex_lower = 244 { $f4 };
//    QKey_otilde_lower = 245 { $f5 };
//    QKey_odiaeresis_lower = 246 { $f6 };
//    QKey_division = 247 { $f7 };
//    QKey_oslash = 248 { $f8 };
//    QKey_ugrave_lower = 249 { $f9 };
//    QKey_uacute_lower = 250 { $fa };
//    QKey_ucircumflex_lower = 251 { $fb };
//    QKey_udiaeresis_lower = 252 { $fc };
//    QKey_yacute_lower = 253 { $fd };
//    QKey_thorn_lower = 254 { $fe };
//    QKey_ydiaeresis = 255 { $ff };
//    QKey_unknown = 65535 { $ffff };

{ Event codes }

  Q_EVMouseDown = ViTypes.evMouseDown;
  Q_EVMouseUp = ViTypes.evMouseUp;
  Q_EVMouseMove = ViTypes.evMouseMove;
  Q_EVMouseAuto = ViTypes.evMouseAuto;
  Q_EVMouseButtonDblClick = ViTypes.evMouseDbl;
  Q_EVKeyDown = ViTypes.evKeyDown;
  Q_EVKeyPress = ViTypes.evKeyPress;
  Q_EVKeyUp = ViTypes.evKeyUp;
  Q_EVCommand = ViTypes.evCommand;
  Q_EVBroadcast = ViTypes.evBroadcast;

{ Event masks }

  Q_EVNothing = ViTypes.evNothing;
  Q_EVMouse = ViTypes.evMouse;
  Q_EVKeyboard = ViTypes.evKeyboard;
  Q_EVMessage = ViTypes.evMessage;

{ Commands }

  Q_CMQuit = ViViews.cmQuit;
  QEventType_Close = ViViews.cmClose;
//  QEventType_Resize = ViViews.cmResize;
  QEventType_Resize = ViViews.cmxBoundsChanged;
  QEventType_FocusIn = ViViews.cmFocusIn;

  cmUser = $100;

  QEventType_ShowMaximized = cmUser;
  QEventType_ShowNormal = cmUser + 1;
  QEventType_ShowMinimized = cmUser + 2;

  QEventType_IsMaximized = cmUser + 3;
  QEventType_IsNormal = cmUser + 4;
  QEventType_IsMinimized = cmUser + 5;

{ Key codes }
  Key_Shift = $FFFF;
  Key_Control = $FFFF;
  Key_Alt = $FFFF;

  Key_Tab = kbTab;
  Key_Backtab = kbShiftTab;

  Key_Return = kbEnter;
  Key_Enter = kbEnter;
  Key_Escape = kbEsc;
  Key_Space = kbSpace;

  Key_Left = kbLeft;
  Key_Right = kbRight;
  Key_Up = kbUp;
  Key_Down = kbDown;

  Key_ShiftInsert = kbShiftIns;
  Key_ShiftDelete = kbShiftDel;
  Key_CtrlV = kbCtrlV;
  Key_CtrlX = kbCtrlX;
  Key_CtrlC = kbCtrlC;

  Key_Home = kbHome;
  Key_End = kbEnd;
  Key_PageUp = kbPgUp;
  Key_PageDown = kbPgDn;
  Key_Backspace = kbBack;
  Key_Insert = kbIns;
  Key_Delete = kbDel;

{ Message box flags }

  mfWarning = ViMsgBox.mfWarning;
  mfError = ViMsgBox.mfError;
  mfInformation = ViMsgBox.mfInformation;
  mfConfirmation = ViMsgBox.mfConfirmation;
  mfInsertInApp = ViMsgBox.mfInsertInApp;

  mfYesButton = ViMsgBox.mfYesButton;
  mfNoButton = ViMsgBox.mfNoButton;
  mfOKButton = ViMsgBox.mfOKButton;
  mfCancelButton = ViMsgBox.mfCancelButton;
  mfAbort = ViMsgBox.mfAbort;
  mfRetry = ViMsgBox.mfRetry;
  mfIgnore = ViMsgBox.mfIgnore;

{ Mouse button masks }

  mbLeftButton  = ViTypes.mbLeftButton;
  mbRightButton = ViTypes.mbRightButton;
  mbMiddleButton = 0;

{ TQLineEdit }

  leMaxLength = 32767;

  cHotkeyPrefix = '&';

const
  WidgetFlags_WType_TopLevel = 1 { $1 };
  WidgetFlags_WType_Modal = 2 { $2 };
  WidgetFlags_WType_Popup = 4 { $4 };
  WidgetFlags_WType_Desktop = 8 { $8 };
  WidgetFlags_WType_Mask = 15 { $f };
  WidgetFlags_WStyle_Customize = 16 { $10 };
  WidgetFlags_WStyle_NormalBorder = 32 { $20 };
  WidgetFlags_WStyle_DialogBorder = 64 { $40 };
  WidgetFlags_WStyle_NoBorder = 0 { $0 };
  WidgetFlags_WStyle_Title = 128 { $80 };
  WidgetFlags_WStyle_SysMenu = 256 { $100 };
  WidgetFlags_WStyle_Minimize = 512 { $200 };
  WidgetFlags_WStyle_Maximize = 1024 { $400 };
  WidgetFlags_WStyle_MinMax = 1536 { $600 };
  WidgetFlags_WStyle_Tool = 2048 { $800 };
  WidgetFlags_WStyle_StaysOnTop = 4096 { $1000 };
  WidgetFlags_WStyle_Dialog = 8192 { $2000 };
  WidgetFlags_WStyle_ContextHelp = 16384 { $4000 };
  WidgetFlags_WStyle_NoBorderEx = 32768 { $8000 };
  WidgetFlags_WStyle_Mask = 65520 { $fff0 };
  WidgetFlags_WDestructiveClose = 65536 { $10000 };
  WidgetFlags_WPaintDesktop = 131072 { $20000 };
  WidgetFlags_WPaintUnclipped = 262144 { $40000 };
  WidgetFlags_WPaintClever = 524288 { $80000 };
  WidgetFlags_WResizeNoErase = 1048576 { $100000 };
  WidgetFlags_WMouseNoMask = 2097152 { $200000 };
  WidgetFlags_WNorthWestGravity = 4194304 { $400000 };
  WidgetFlags_WRepaintNoErase = 8388608 { $800000 };
  WidgetFlags_WX11BypassWM = 16777216 { $1000000 };
  WidgetFlags_WGroupLeader = 33554432 { $2000000 };

type

  { Widget flags }
(*
  WidgetFlags = (
    WidgetFlags_WType_TopLevel = 1 { $1 },
    WidgetFlags_WType_Modal = 2 { $2 },
    WidgetFlags_WType_Popup = 4 { $4 },
    WidgetFlags_WType_Desktop = 8 { $8 },
    WidgetFlags_WType_Mask = 15 { $f },
    WidgetFlags_WStyle_Customize = 16 { $10 },
    WidgetFlags_WStyle_NormalBorder = 32 { $20 },
    WidgetFlags_WStyle_DialogBorder = 64 { $40 },
    WidgetFlags_WStyle_NoBorder = 0 { $0 },
    WidgetFlags_WStyle_Title = 128 { $80 },
    WidgetFlags_WStyle_SysMenu = 256 { $100 },
    WidgetFlags_WStyle_Minimize = 512 { $200 },
    WidgetFlags_WStyle_Maximize = 1024 { $400 },
    WidgetFlags_WStyle_MinMax = 1536 { $600 },
    WidgetFlags_WStyle_Tool = 2048 { $800 },
    WidgetFlags_WStyle_StaysOnTop = 4096 { $1000 },
    WidgetFlags_WStyle_Dialog = 8192 { $2000 },
    WidgetFlags_WStyle_ContextHelp = 16384 { $4000 },
    WidgetFlags_WStyle_NoBorderEx = 32768 { $8000 },
    WidgetFlags_WStyle_Mask = 65520 { $fff0 },
    WidgetFlags_WDestructiveClose = 65536 { $10000 },
    WidgetFlags_WPaintDesktop = 131072 { $20000 },
    WidgetFlags_WPaintUnclipped = 262144 { $40000 },
    WidgetFlags_WPaintClever = 524288 { $80000 },
    WidgetFlags_WResizeNoErase = 1048576 { $100000 },
    WidgetFlags_WMouseNoMask = 2097152 { $200000 },
    WidgetFlags_WNorthWestGravity = 4194304 { $400000 },
    WidgetFlags_WRepaintNoErase = 8388608 { $800000 },
    WidgetFlags_WX11BypassWM = 16777216 { $1000000 },
    WidgetFlags_WGroupLeader = 33554432 { $2000000 });
*)

  TEvent = ViTypes.TEvent;

  TPoint = ViTypes.TPoint;
  TRect = ViTypes.TRect;

  TOrientation = ViTypes.TOrientation;

  TQWindow = class;

  TQApplication = class(TObject)
  private
    FViApplication: TViApplication;
    FOnEvent: TEventFilterEvent;
    FOnIdle: TNotifyEvent;
    function DoEvent(Sender: TObject; var Event: TEvent): Boolean;
    procedure DoIdle(Sender: TObject);
  public
    constructor Create;
    destructor Destroy; override;
    procedure ProcessOneEvent;
    procedure ProcessEvents;
    function Height: Integer;
    function Width: Integer;
    property OnEvent: TEventFilterEvent read FOnEvent write FOnEvent;
    property OnIdle: TNotifyEvent read FOnIdle write FOnIdle;
  end;

  TQWidget = class(TObject)
  private
    FOnEvent: TEventFilterEvent;
    FWFlags: Cardinal;
    FCurrent: TViView;
    function GetEnabled: Boolean;
  protected
    function GetCaption: string; virtual;
    procedure SetCaption(const Value: string); virtual;
    function GetWFlags: Cardinal;
    procedure SetWFlags(Value: Cardinal); virtual;
    function ViView: TViView; virtual;
    function DoEvent(Sender: TObject; var Event: TEvent): Boolean; virtual;
    procedure SetEnabled(Value: Boolean); virtual;
    procedure RestoreCurrent;
    procedure SaveCurrent(AGroup: TViGroup);
    procedure BoundsChanged(const R: TRect); virtual;
    function ViClientView: TViGroup; virtual;
  public
    procedure Show; virtual;
    procedure Hide; virtual;
    procedure SetFocus;
    procedure BringToFront;
    procedure SendToBack;
    procedure SetGeometry(ALeft, ATop, AWidth, AHeight: Integer); virtual;
    procedure GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer); virtual;
    procedure MapFromGlobal(Source: TPoint; var Dest: TPoint);
    procedure ShowNormal; virtual;
    procedure ShowMinimized; virtual;
    procedure ShowMaximized; virtual;
    procedure Reparent(ParentWidget: TQWidget);
    procedure MapToGlobal(Source: TPoint; var Dest: TPoint);
    property Caption: string read GetCaption write SetCaption;
    property OnEvent: TEventFilterEvent read FOnEvent write FOnEvent;
    property WFlags: Cardinal read GetWFlags write SetWFlags;
    property Enabled: Boolean read GetEnabled write SetEnabled;
  end;

  TQWindow = class(TQWidget)
  private
    FViWindow: TViWindow;
    FViClientGroup: TViGroup;
    FViBackground: TViClientFrame;
  protected
    function GetCaption: string; override;
    procedure SetCaption(const Value: string); override;
    function ViView: TViView; override;
    procedure SetWFlags(Value: Cardinal); override;
    function DoEvent(Sender: TObject; var Event: TEvent): Boolean; override;
    function DoFrameEvent(Sender: TObject; var Event: TEvent): Boolean;
    function ViClientView: TViGroup; override;
    function DoClientEvent(Sender: TObject; var Event: TEvent): Boolean; virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure ShowMaximized; override;
    procedure ShowMinimized; override;
    procedure ShowNormal; override;
    procedure SetSizeLimits(MinWidth, MinHeight, MaxWidth, MaxHeight: Integer);
    procedure ExitLoop;
    procedure ShowModal;
    procedure InitClientView; virtual;
    procedure GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure SetGeometry(ALeft, ATop, AWidth, AHeight: Integer); override;
  end;

  TQLabel = class(TQWidget)
  private
    FViLabel: TViLabel;
    FCaption: string;
  protected
    function ViView: TViView; override;
    function GetCaption: string; override;
    procedure SetCaption(const Value: string); override;
    function DoEvent(Sender: TObject; var Event: TEvent): Boolean; override;
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
  end;

  TQButton = class(TQWidget)
  private
    FOnClicked: TNotifyEvent;
  protected
    procedure DoClicked(Sender: TObject); virtual;
    function DoEvent(Sender: TObject; var Event: TEvent): Boolean; override;
  public
    property OnClicked: TNotifyEvent read FOnClicked write FOnClicked;
  end;

  TQPushButton = class(TQButton)
  private
    FViButton: TViButton;
    FCaption: string;
  protected
    function ViView: TViView; override;
    function GetCaption: string; override;
    procedure SetCaption(const Value: string); override;
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    function GetClientRect: TRect;
    procedure SetDefault(Value: Boolean);
  end;

  TQLineEdit = class(TQWidget)
  private
    FViInputLine: TViInputLine;
    FOnTextChanged: TChangedEvent;
    FEdited: Boolean;
    function GetMaxLength: Integer;
    procedure SetMaxLength(Value: Integer);
    function GetAutoSelect: Boolean;
    procedure SetAutoSelect(Value: Boolean);
    function GetHideSelection: Boolean;
    procedure SetHideSelection(Value: Boolean);
    function GetReadOnly: Boolean;
    procedure SetReadOnly(Value: Boolean);
    procedure SetEdited(Value: Boolean);
    function GetMarkedText: string;
    procedure SetMarkedText(const Value: string);
    procedure SetEchoMode(Value: Integer);
    function GetEchoMode: Integer;
  protected
    function ViView: TViView; override;
    function GetCaption: string; override;
    procedure SetCaption(const Value: string); override;
    procedure DoTextChanged(Sender: TObject); virtual;
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    procedure Clear;
    procedure ClearSelection;
    procedure Copy;
    procedure Cut;
    procedure SelectAll;
    function SelStart: Integer;
    procedure SetSelection(SelStart, SelLength: Integer);
    procedure SetSelStart(Value: Integer);
    property MaxLength: Integer read GetMaxLength write SetMaxLength;
    property AutoSelect: Boolean read GetAutoSelect write SetAutoSelect;
    property HideSelection: Boolean read GetHideSelection write
        SetHideSelection;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly;
    property OnTextChanged: TChangedEvent read FOnTextChanged write
        FOnTextChanged;
    property Edited: Boolean read FEdited write SetEdited;
    property MarkedText: string read GetMarkedText write SetMarkedText;
    property EchoMode: Integer read GetEchoMode write SetEchoMode;
  end;

  TClipboard = class(TObject)
  private
    FAsText: string;
    procedure SetAsText(const Value: string);
  protected
  public
    property AsText: string read FAsText write SetAsText;
  end;

  TQListBox = class(TQWidget)
  private
    FViListBox: TViListBox;
    FViScrollBar: TViScrollBar;
    FOnChanged: TChangedEvent;
    function GetData(Index: Integer): Integer;
    procedure SetData(Index: Integer; Value: Integer);
  protected
    function GetCurrentItem: Integer; virtual;
    procedure SetCurrentItem(Value: Integer); virtual;
    function DoEvent(Sender: TObject; var Event: TEvent): Boolean; override;
    function ViView: TViView; override;
    procedure DoChanged(Sender: TObject);
    procedure SetEnabled(Value: Boolean); override;
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    procedure GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure SetGeometry(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure Clear;
    function Count: Integer;
    procedure RemoveItem(Index: Integer);
    procedure InsertItem(const S: string; Index: Integer);
    function GetItem(Index: Integer): string;
    procedure Sort;
    procedure Update;
    property CurrentItem: Integer read GetCurrentItem write SetCurrentItem;
    property Data[Index: Integer]: Integer read GetData write SetData;
    property OnChanged: TChangedEvent read FOnChanged write FOnChanged;
  end;

  TQScrollBar = class(TQWidget)
  private
    FViScrollBar: TViScrollBar;
    FOnChange: TChangedEvent;
    FOnScroll: TScrollEvent;
  protected
    function DoEvent(Sender: TObject; var Event: TEvent): Boolean; override;
    function ViView: TViView; override;
    procedure BoundsChanged(const R: TRect); override;
    procedure DoChange(Sender: TObject); virtual;
    procedure DoScroll(Sender: TObject; ScrollPart: Integer); virtual;
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    procedure SetSteps(SmallChange, LargeChange: Integer);
    procedure SetOrientation(Value: TOrientation);
    procedure SetRange(Min, Max: Integer);
    procedure SetValue(Position: Integer);
    function PageStep: Integer;
    function LineStep: Integer;
    function MaxValue: Integer;
    function MinValue: Integer;
    function Value: Integer;
    property OnChange: TChangedEvent read FOnChange write FOnChange;
    property OnScroll: TScrollEvent read FOnScroll write FOnScroll;
  end;

  TQRadioButton = class(TQButton)
  private
    FViRadioButton: TViRadioButtons;
    FCaption: string;
  protected
    function ViView: TViView; override;
    function GetCaption: string; override;
    procedure SetCaption(const Value: string); override;
    procedure SetEnabled(Value: Boolean); override;
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    function IsChecked: Boolean;
    procedure SetChecked(Value: Boolean);
  end;

  TQCheckBox = class(TQButton)
  private
    TViMultiCheckBox: TViMultiCheckBoxes;
    FCaption: string;
  protected
    function ViView: TViView; override;
    function GetCaption: string; override;
    procedure SetCaption(const Value: string); override;
    procedure SetEnabled(Value: Boolean); override;
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    function IsChecked: Boolean;
    procedure SetChecked(Value: Boolean);
    procedure SetNoChange;
    procedure SetTriState(Value: Boolean);
    function State: Integer;
  end;

  TQFrame = class(TQWidget)
  private
    FViClientGroup: TViGroup;
    FViBackground: TViClientFrame;
    FViLabel: TViLabel;
    FCaption: string;
  protected
    function ViView: TViView; override;
    function GetCaption: string; override;
    procedure SetCaption(const Value: string); override;
    function DoEvent(Sender: TObject; var Event: TEvent): Boolean; override;
    function ViClientView: TViGroup; override;
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    procedure InitBackground; virtual;
    procedure SetFrame(Value: Boolean);
    procedure SetGeometry(ALeft, ATop, AWidth, AHeight: Integer); override;
  end;

  TQGroupBox = class(TQWidget)
  private
    FViBaseGroup: TViGroup;
    FViClientGroup: TViGroup;
    FViBackground: TViClientFrame;
    FViLabel: TViLabel;
    FCaption: string;
  protected
    function ViView: TViView; override;
    function GetCaption: string; override;
    procedure SetCaption(const Value: string); override;
    function DoEvent(Sender: TObject; var Event: TEvent): Boolean; override;
    function ViClientView: TViGroup; override;
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    procedure InitBackground; virtual;
    procedure SetFrame(Value: Boolean);
    procedure SetGeometry(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer); override;
  end;

  TQButtonGroup = class(TQGroupBox)
  private
    FColumns: Integer;
    FOrientation: Integer;
  protected
    procedure UpdateLayout;
  public
    constructor Create(ParentWidget: TQWidget; Columns: Integer; Orientation:
        Integer);
    procedure SetColumnLayout(Columns, Orientation: Integer);
    procedure SetGeometry(ALeft, ATop, AWidth, AHeight: Integer); override;
  end;

  TQMenuData = class;

  TQMenuBar = class(TQWidget)
  private
    FMenuData: TQMenuData;
    FOnActivated: TActivatedEvent;
  protected
    procedure DoActivated(Sender: TObject; ID: Integer); virtual;
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    procedure Hide; override;
    procedure Show; override;
    property MenuData: TQMenuData read FMenuData;
    property OnActivated: TActivatedEvent read FOnActivated write FOnActivated;
  end;

  TQPopupMenu = class(TQWidget)
  private
    FViMenuPopup: TViMenuPopup;
    FMenuData: TQMenuData;
    FOnActivated: TActivatedEvent;
    function GetMenu: PMenu;
  protected
    procedure DoActivated(Sender: TObject; ID: Integer); virtual;
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    procedure Execute(Point: TPoint);
    property MenuData: TQMenuData read FMenuData;
    property Menu: PMenu read GetMenu;
    property OnActivated: TActivatedEvent read FOnActivated write FOnActivated;
  end;

  TQMenuData = class(TObject)
  private
    FMenu: PMenu;
  public
    function InsertItem(const ACaption: WideString; Index: Integer): Integer;
    function InsertSubMenu(const ACaption: WideString; AWidget: TQWidget; Index:
        Integer): Integer;
    constructor Create;
    procedure ChangeItem(ID: Integer; const ACaption: WideString);
    procedure RemoveItem(ID: Integer);
    function InsertSeparator(Index: Integer): Integer;
    destructor Destroy; override;
    procedure DisposeMenu(Menu: PMenu);
    procedure SetItemEnabled(ID: Integer; Value: Boolean);
    procedure SetItemChecked(ID: Integer; Value: Boolean);
    procedure SetAccel(const Param: string; ShortCut: Word; ID: Integer);
    property Menu: PMenu read FMenu;
  end;

  TQTimer = class(TQWidget)
  private
    FViTimer: TViTimer;
    FOnTimer: TNotifyEvent;
    procedure SetEnabled(Value: Boolean); reintroduce;
    procedure SetInterval(Value: Cardinal);
    function GetEnabled: Boolean;
    function GetInterval: Cardinal;
  protected
    procedure Timer(Sender: TObject); dynamic;
  public
    constructor Create; 
    destructor Destroy; override;
    property Enabled: Boolean read GetEnabled write SetEnabled;
    property Interval: Cardinal read GetInterval write SetInterval;
    property OnTimer: TNotifyEvent read FOnTimer write FOnTimer;
  end;

  TQRtcTimer = class(TQWidget)
  private
    FViRtcTimer: TViRtcTimer;
    FOnTimer: TNotifyEvent;
    procedure SetEnabled(Value: Boolean); reintroduce;
    procedure SetInterval(Value: Cardinal);
    function GetEnabled: Boolean;
    function GetInterval: Cardinal;
    function GetIntFreq: Integer;
    procedure SetIntFreq(Value: Integer);
  protected
    procedure Timer(Sender: TObject); dynamic;
  public
    constructor Create;
    destructor Destroy; override;
    property Enabled: Boolean read GetEnabled write SetEnabled;
    property Interval: Cardinal read GetInterval write SetInterval;
    property OnTimer: TNotifyEvent read FOnTimer write FOnTimer;
    property IntFreq: Integer read GetIntFreq write SetIntFreq;
  end;

  TQOpenComboBoxStyle = (csDropDown, csDropDownList, csOwnerDrawFixed,
    csOwnerDrawVariable);

  TQOpenComboBox = class(TQListBox)
  private
    FViInputLine: TViInputLine;
    FViHistory: TViHistory;
    FOnTextChanged: TChangedEvent;
    FEdited: Boolean;
    FStyle: Integer;
    function GetMaxLength: Integer;
    procedure SetMaxLength(Value: Integer);
    function GetAutoSelect: Boolean;
    procedure SetAutoSelect(Value: Boolean);
    function GetHideSelection: Boolean;
    procedure SetHideSelection(Value: Boolean);
    function GetReadOnly: Boolean;
    procedure SetReadOnly(Value: Boolean);
    procedure SetEdited(Value: Boolean);
    function GetMarkedText: string;
    procedure SetMarkedText(const Value: string);
    procedure SetEchoMode(Value: Integer);
    function GetEchoMode: Integer;
    procedure SetStyle(Value: Integer);
  protected
    function ViView: TViView; override;
    function GetCurrentItem: Integer; override;
    procedure SetCurrentItem(Index: Integer); override;
    function GetCaption: string; override;
    procedure SetCaption(const Value: string); override;
    procedure DoTextChanged(Sender: TObject); virtual;
    procedure SetEnabled(Value: Boolean); override;
    procedure DoListChanged(Sender: TObject);
    procedure DoSelected(Sender: TObject);
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    procedure Clear;
    procedure ClearSelection;
    procedure Copy;
    procedure Cut;
    procedure SelectAll;
    function SelStart: Integer;
    procedure SetSelection(SelStart, SelLength: Integer);
    procedure SetSelStart(Value: Integer);
    procedure GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure SetGeometry(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure Hide; override;
    procedure Show; override;
    property MaxLength: Integer read GetMaxLength write SetMaxLength;
    property AutoSelect: Boolean read GetAutoSelect write SetAutoSelect;
    property HideSelection: Boolean read GetHideSelection write
        SetHideSelection;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly;
    property OnTextChanged: TChangedEvent read FOnTextChanged write
        FOnTextChanged;
    property Edited: Boolean read FEdited write SetEdited;
    property MarkedText: string read GetMarkedText write SetMarkedText;
    property EchoMode: Integer read GetEchoMode write SetEchoMode;
    property Style: Integer read FStyle write SetStyle;
  end;

  TQDialog = class(TObject)
  private
    FModal: Boolean;
    FWidgetFlags: Integer;
    FOnEvent: TEventFilterEvent;
  protected
    function ViDialog: TViDialog; virtual; abstract;
    function GetCaption: string; virtual;
    procedure SetCaption(const Value: string); virtual;
    procedure CreateWidget; virtual;
    procedure DestroyWidget; virtual;
  public
    constructor Create(Modal: Boolean; WidgetFlags: Integer);
    procedure Show; virtual;
    procedure Hide; virtual;
    function DialogResult: Boolean; virtual;
    procedure GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer); virtual;
    procedure SetGeometry(ALeft, ATop, AWidth, AHeight: Integer); virtual;
    procedure BringToFront;
    destructor Destroy; override;
    property Caption: string read GetCaption write SetCaption;
    property OnEvent: TEventFilterEvent read FOnEvent write FOnEvent;
  end;

  TQClxFileDialog = class(TQDialog)
  private
    FViFileDialog: TViFileDialog;
    FHistoryList: TStrings;
    FOnCloseRequest: TCloseRequestEvent;
  protected
    function ViDialog: TViDialog; override;
    procedure CreateWidget; override;
    procedure DestroyWidget; override;
  public
    function Execute: Boolean; virtual;
    procedure RereadDir; virtual;
    function SelectedFile: string;
    procedure DoCloseRequest(Sender: TObject; var CanClose: Boolean);
    procedure SetSaveMode;
    procedure SetEditText(const Value: string);
    procedure SetDefaultExt(const Value: string);
    procedure SetDir(const Value: string);
    procedure SetHistoryList(Value: TStrings);
    property OnCloseRequest: TCloseRequestEvent read FOnCloseRequest write
        FOnCloseRequest;
  end;

  TQDirSelDlg = class(TObject)
  private
    FCaption: string;
    FRootDirectory: string;
    FSelectedDir: string;
    procedure SetCaption(const Value: string);
    procedure SetRootDirectory(const Value: string);
    function GetSelectedDir: string;
  public
    function Execute: Boolean; virtual;
    property Caption: string read FCaption write SetCaption;
    property RootDirectory: string read FRootDirectory write SetRootDirectory;
    property SelectedDir: string read GetSelectedDir;
  end;

  TQMultiLineEdit = class(TQWidget)
  private
    FViMemo: TViMemo;
    FViHScrollBar: TViScrollBar;
    FViVScrollBar: TViScrollBar;
    FOnChanged: TChangedEvent;
    FOnReturnPressed: TNotifyEvent;
  protected
    function ViView: TViView; override;
    function DoEvent(Sender: TObject; var Event: TEvent): Boolean; override;
    procedure DoChanged(Sender: TObject);
    procedure DoReturnPressed(Sender: TObject);
  public
    constructor Create(ParentWidget: TQWidget);
    destructor Destroy; override;
    procedure GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure SetGeometry(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure Append(const Value: string);
    function AtEnd: Boolean;
    procedure Clear;
    function NumLines: Integer;
    procedure RemoveLine(Index: Integer);
    function TextLine(Index: Integer): string;
    procedure InsertLine(Index: Integer; const Line: string);
    procedure SetText(const Value: string);
    function GetText: string;
    function TextWidth(LineNum: Integer): Integer;
    procedure Deselect;
    procedure Insert(const Value: string; Mark: Boolean = False);
    function HasMarkedText: Boolean;
    procedure Copy;
    procedure Cut;
    procedure Paste;
    procedure CursorPosition(var Line, Column: Integer);
    function Edited: Boolean;
    function IsReadOnly: Boolean;
    procedure GetMarkedRegion(var Line1, Column1, Line2, Column2: Integer);
    function MarkedText: string;
    procedure InsertAt(const Text: string; Column, Row: Integer; Mark: Boolean = 
        False); overload;
    procedure SelectAll;
    procedure SetEdited(const Value: Boolean); virtual;
    procedure SetReadOnly(const Value: Boolean); virtual;
    function LineLength(Index: Integer): Integer;
    procedure SetCursorPosition(Line, Column: Integer);
    property OnChanged: TChangedEvent read FOnChanged write FOnChanged;
    property OnReturnPressed: TNotifyEvent read FOnReturnPressed write
        FOnReturnPressed;
  end;

var
  Clipboard: TClipboard = nil;
  ChDirHistory: TStrings = nil;

procedure RectToViRect(var R: TRect);

procedure ViRectToRect(var R: TRect);

procedure PointToViPoint(var P: TPoint);

procedure ViPointToPoint(var P: TPoint);

function QMessageBox(const Msg: string; AOptions, X, Y: Integer): Integer;

function QInputDialog(const Title, ALabel: string; var S: string): Boolean;

function TranslateAltKeyToKey(KeyCode: Word): Integer;

function TranslateAccelerator(const Caption: string): string;


implementation

{
********************************* TQApplication ********************************
}

procedure RectToViRect(var R: TRect);
var
  X, Y: Integer;
begin
  R.A.X := R.A.X div 8;
  R.A.Y := R.A.Y div 16;
  X := R.B.X div 8;
  if (R.B.X) mod 8 > 1 then Inc(X);
  R.B.X := X;
  Y := R.B.Y div 16;
  if (R.B.Y) mod 16 > 1 then Inc(Y);
  R.B.Y := Y;
end;

procedure ViRectToRect(var R: TRect);
begin
  R.A.X := R.A.X * 8;
  R.A.Y := R.A.Y * 16;
  R.B.X := R.B.X * 8;
  R.B.Y := R.B.Y * 16;
end;

procedure PointToViPoint(var P: TPoint);
begin
  P.X := P.X div 8;
  P.Y := P.Y div 16;
end;

procedure ViPointToPoint(var P: TPoint);
begin
  P.X := P.X * 8;
  P.Y := P.Y * 16;
end;

function QMessageBox(const Msg: string; AOptions, X, Y: Integer): Integer;
const
  MessageBoxResult: array[0..8] of Integer = (
      mfOKButton, mfCancelButton, mfYesButton, mfNoButton,
      0, 0, mfAbort, mfRetry, mfIgnore);
var
  R: TRect;
  P: TPoint;
begin
  RectAssign(R, 0, 0, 40, 9);
  if (X = -1) and (Y = -1) then
  begin
    if AOptions and mfInsertInApp = 0 then
      RectMove(R, (ViDesktop.Size.X - R.B.X) div 2,
          (ViDesktop.Size.Y - R.B.Y) div 2)
    else RectMove(R, (ViApplication.Size.X - R.B.X) div 2,
        (ViApplication.Size.Y - R.B.Y) div 2)
  end
  else
  begin
    P.X := X;
    P.Y := Y;
    PointToViPoint(P);
    RectMove(R, P.X, P.Y);
  end;
  Result := MessageBoxRect(R, Msg, AOptions);
  // convert message box result consts
  Result := MessageBoxResult[Result - 10];
end;

function QInputDialog(const Title, ALabel: string; var S: string): Boolean;
begin
  Result := InputBox(Title, TranslateAccelerator(ALabel), S, MaxInt)
      <> cmCancel;
end;

function TranslateAltKeyToKey(KeyCode: Word): Integer;
begin
  Result := Ord(GetAltChar(KeyCode));
end;

function TranslateAccelerator(const Caption: string): string;
var
  I, L: Integer;
begin
  Result := '';
  I := 1;
  L := Length(Caption) - 1;
  while I <= L do
  begin
    if Caption[I] = cHotkeyPrefix then
    begin
      Inc(I);
      if Caption[I] = cHotkeyPrefix then
        Result := Result + cHotkeyPrefix
      else
        Result := Result + '~' + Caption[I] + '~';
    end
    else Result := Result + Caption[I];
    Inc(I);
  end;
  if I = (L + 1) then
    Result := Result + Caption[L + 1];
end;

constructor TQApplication.Create;
begin
  inherited Create;
  FViApplication := TViApplication.Create;
  FViApplication.OnEvent := DoEvent;
  FViApplication.OnIdle := DoIdle;
end;

destructor TQApplication.Destroy;
begin
  FViApplication.Free;
  inherited Destroy;
end;

procedure TQApplication.ProcessOneEvent;
var
  E: TEvent;
begin
  FViApplication.GetEvent(E);
  FViApplication.HandleEvent(E);
  if E.What <> evNothing then FViApplication.EventError(E);
  ProcessMessages;
end;

procedure TQApplication.ProcessEvents;
var
  E: TEvent;
begin
  repeat
    FViApplication.GetEvent(E);
    FViApplication.HandleEvent(E);
    if E.What <> evNothing then FViApplication.EventError(E);
    ProcessMessages;
  until E.What = evNothing;
end;

function TQApplication.DoEvent(Sender: TObject; var Event: TEvent):
    Boolean;
begin
  if Event.What = evKeyDown then
    Result := False
  else
    if Assigned(OnEvent) then Result := OnEvent(Self, Event)
    else Result := False;
end;

function TQApplication.Height: Integer;
begin
  Result := FViApplication.Size.Y * 16;
end;

function TQApplication.Width: Integer;
begin
  Result := FViApplication.Size.X * 8;
end;

procedure TQApplication.DoIdle(Sender: TObject);
begin
  if Assigned(FOnIdle) then OnIdle(Self);
end;


{
********************************* TQWidget *************************************
}
function TQWidget.GetCaption: string;
begin
  Result := '';
end;

procedure TQWidget.SetCaption(const Value: string);
begin
end;

procedure TQWidget.Show;
begin
  SaveCurrent(ViView.Owner);
  ViView.Show;
  ViView.DrawView;
  RestoreCurrent;
end;

procedure TQWidget.Hide;
begin
  SaveCurrent(ViView.Owner);
  if FCurrent = ViView then FCurrent := nil;
  ViView.Hide;
  RestoreCurrent;
end;

procedure TQWidget.SetFocus;
begin
//  ViView.Select;
  ViView.Focus;
end;

function TQWidget.ViView: TViView;
begin
  Result := nil;
end;

function TQWidget.DoEvent(Sender: TObject; var Event: TEvent): Boolean;
begin
  Result := False;
  if Event.What = evNothing then Exit;
  if (Event.What = evKeyDown) and ((ViView.State and sfFocused = 0)
      or (ViView.Owner.Phase <> phFocused)) then
    Exit
  else
    if Assigned(OnEvent) then
    begin
      if (Event.What = evMouseDown) and Event.Double then
        Event.What := evMouseDbl;
      Result := OnEvent(Self, Event);
      if (Event.What = evMouseDbl) and Event.Double then
        Event.What := evMouseDown;
    end;
end;

procedure TQWidget.BringToFront;
begin
  ViView.MakeFirst;
end;

procedure TQWidget.SendToBack;
var
  Last: TViView;
begin
  Last := ViView.Owner.Last;
  ViView.PutInFrontOf(Last);
  Last.PutInFrontOf(ViView);
end;

procedure TQWidget.SetGeometry(ALeft, ATop, AWidth, AHeight: Integer);
var
  R: TRect;
begin
  RectAssign(R, ALeft, ATop, ALeft + AWidth, ATop + AHeight);
  RectToViRect(R);
  ViView.ChangeBounds(R);
  BoundsChanged(R);
  ViView.Owner.Redraw;
end;

procedure TQWidget.GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer);
var
  R: TRect;
begin
  ViView.GetBounds(R);
  ViRectToRect(R);
  ALeft := R.A.X;
  ATop := R.A.Y;
  AWidth := R.B.X - ALeft;
  AHeight := R.B.Y - ATop;
end;

procedure TQWidget.MapFromGlobal(Source: TPoint; var Dest: TPoint);
begin
  ViView.MakeLocal(Source, Dest);
end;

function TQWidget.GetWFlags: Cardinal;
begin
  Result := FWFlags;
end;

procedure TQWidget.SetWFlags(Value: Cardinal);
begin
  FWFlags := Value;
end;

procedure TQWidget.ShowNormal;
begin
  Show;
end;

procedure TQWidget.ShowMinimized;
begin
  Show;
end;

procedure TQWidget.ShowMaximized;
begin
  Show;
end;

function TQWidget.GetEnabled: Boolean;
begin
  Result := (ViView.State and sfDisabled) = 0;
end;

procedure TQWidget.SetEnabled(Value: Boolean);
begin
  if Value <> Enabled then
  begin
    if Value then
      ViView.State := ViView.State and not sfDisabled
    else
      ViView.State := ViView.State or sfDisabled;
    ViView.Draw;
  end;
end;

procedure TQWidget.Reparent(ParentWidget: TQWidget);
var
  Owner: TViGroup;
begin
  if Self is TQWindow then Exit;

  Owner := ViView.Owner;
//  if not Assigned(ParentWidget) or (ParentWidget.ViView <> Owner) then
  if not Assigned(ParentWidget) or (ParentWidget.ViClientView <> Owner) then
  begin
    if Assigned(Owner) then
    begin
      SaveCurrent(Owner);
      ViView.Hide;
      Owner.Delete(ViView);
      if FCurrent = ViView then
      begin
        FCurrent := nil;
        Abort;
      end;
      RestoreCurrent;
    end;
    if ParentWidget <> nil then
    begin
//      Owner :=ParentWidget.ViView as TViGroup;
      Owner :=ParentWidget.ViClientView;
      SaveCurrent(Owner);
      Owner.Insert(ViView);
      RestoreCurrent;
    end;
  end;
end;

procedure TQWidget.RestoreCurrent;
begin
  if Assigned(FCurrent) then FCurrent.Select;
end;

procedure TQWidget.SaveCurrent(AGroup: TViGroup);
begin
  if Assigned(AGroup) then
    FCurrent := AGroup.Current
  else
    FCurrent := nil;
end;

procedure TQWidget.BoundsChanged(const R: TRect);
begin
  // do nothing
end;

function TQWidget.ViClientView: TViGroup;
begin
  Result := nil;
end;

procedure TQWidget.MapToGlobal(Source: TPoint; var Dest: TPoint);
begin
  PointToViPoint(Source);
  ViView.MakeGlobal(Source, Dest);
  ViPointToPoint(Dest);
end;



{
********************************* TQWindow *************************************
}
constructor TQWindow.Create;
var
  R: TRect;
begin
  inherited Create;
  RectAssign(R, 0, 0, 45, 7);
  FViWindow := TViDialog.Create(R, 'Window');
  FViWindow.Options := (ofSelectable + ofBuffered + ofTopSelect);
  FViWindow.GrowMode := 0;
  FViWindow.Flags := wfMove;
  FViWindow.WindowPalette := dpBlueDialog;
  FViWindow.State := FViWindow.State and not sfVisible;
  FViWindow.Options := FViWindow.Options or ofTileable;


  InitClientView;
  if FViClientGroup <> nil then FViWindow.Insert(FViClientGroup);
  FViClientGroup.OnEvent := DoClientEvent;


  ViDesktop.Insert(FViWindow);
  FViWindow.OnEvent := DoEvent;
  with FViWindow.Frame do
  begin
    OnEvent := DoFrameEvent;
    EventMask := EventMask + evMouseUp + evMouseMove;
  end;
end;

destructor TQWindow.Destroy;
begin
  FreeAndNil(FViBackground);
  FreeAndNil(FViClientGroup);
  FreeAndNil(FViWindow);
  inherited Destroy;
end;

function TQWindow.GetCaption: string;
begin
  Result := FViWindow.GetTitle;
end;

procedure TQWindow.SetCaption(const Value: string);
begin
  FViWindow.Title := Value;
  FViWindow.DrawView;
end;

function TQWindow.ViView: TViView;
begin
  Result := FViWindow;
end;

procedure TQWindow.SetWFlags(Value: Cardinal);

  procedure SetFlags(AState: Word; Enable: Boolean);
  begin
    if Enable then
      FViWindow.Flags := FViWindow.Flags or AState
    else
      FViWindow.Flags := FViWindow.Flags and not AState;
  end;

begin
  inherited SetWFlags(Value);
  SetFlags(wfZoom, Value and Integer(WidgetFlags_WStyle_Maximize) <> 0);
  SetFlags(wfClose, Value and Integer(WidgetFlags_WStyle_SysMenu) <> 0);
  SetFlags(wfGrow, Value and Integer(WidgetFlags_WStyle_NormalBorder) <> 0);

  FViWindow.WindowPalette := dpBlueDialog;
  if Value and Integer(WidgetFlags_WStyle_Dialog) <> 0 then
    FViWindow.WindowPalette := dpGrayDialog;
  if Value and Integer(WidgetFlags_WStyle_Tool) <> 0 then
    FViWindow.WindowPalette := dpCyanDialog;
end;

procedure TQWindow.ShowMaximized;
begin
  inherited;
  FViWindow.ShowMaximized;
end;

procedure TQWindow.ShowMinimized;
begin
  ShowNormal;
end;

procedure TQWindow.ShowNormal;
begin
  inherited;
  FViWindow.ShowNormal;
end;

function TQWindow.DoEvent(Sender: TObject; var Event: TEvent): Boolean;
var
  Max, Min: TPoint;
begin
  if (Event.What = evCommand) then
  begin
    case Event.Command of
      cmZoom:
        begin
          with FViWindow do
          begin
            SizeLimits(Min, Max);
            if (Size.X <> Max.X) or (Size.Y <> Max.Y) then
              Event.Command := QEventType_ShowMaximized
            else
              Event.Command := QEventType_ShowNormal;
          end;
        end;
      cmZoomed:
        begin
          with FViWindow do
          begin
            SizeLimits(Min, Max);
            if (Size.X = Max.X) and (Size.Y = Max.Y) then
              Event.Command := QEventType_IsMaximized
            else
              Event.Command := QEventType_IsNormal;
          end;
        end;
    end;
  end;
  if (Sender = FViWindow) and (Event.What and PositionalEvents <> 0) then
  begin
    Result := False;
  end
  else
    if (Event.What = evKeyDown) then
      Result := False
    else
      Result := inherited DoEvent(Sender, Event);
end;

function TQWindow.DoFrameEvent(Sender: TObject; var Event: TEvent):
    Boolean;
begin
  if (Event.What and PositionalEvents <> 0) then
    Result := inherited DoEvent(Sender, Event)
  else
    Result := False;
end;

procedure TQWindow.SetSizeLimits(MinWidth, MinHeight, MaxWidth, MaxHeight: 
    Integer);
var
  R: TRect;
begin
  R.A.X := MinWidth;
  R.A.Y := MinHeight;
  R.B.X := MaxWidth;
  R.B.Y := MaxHeight;
  RectToViRect(R);
  FViWindow.Constraints := R;
end;

procedure TQWindow.ExitLoop;
begin
  FViWindow.EndModal(cmCancel);
end;

procedure TQWindow.ShowModal;
begin
  FViWindow.SetState(sfModal, True);
  FViWindow.Execute;
  FViWindow.SetState(sfModal, False);
end;

procedure TQWindow.InitClientView;
var
  R: TRect;
begin
  FViWindow.GetExtent(R);
  RectGrow(R, -1, -1);
  FViClientGroup := TViGroup.Create(R);
  with FViClientGroup do
  begin
    Options := Options or (ofFirstClick + ofPreProcess);
    GrowMode := gfGrowHiX + gfGrowHiY;
  end;
  FViClientGroup.GetExtent(R);
  FViBackground := TViClientFrame.Create(R);
  with FViBackground do
  begin
    GrowMode := gfGrowHiX + gfGrowHiY;
    EventMask := EventMask or evBroadcast;
  end;
  FViClientGroup.Insert(FViBackground);
end;

function TQWindow.ViClientView: TViGroup;
begin
  Result := FViClientGroup;
end;

function TQWindow.DoClientEvent(Sender: TObject; var Event: TEvent): Boolean;
begin
  if (Event.What = evKeyDown) then
    Result := False
  else
    Result := inherited DoEvent(Sender, Event);
end;

procedure TQWindow.GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer);
var
  R: TRect;
begin
  ViView.GetBounds(R);
  RectGrow(R, -1, -1);
  ViRectToRect(R);
  ALeft := R.A.X;
  ATop := R.A.Y;
  AWidth := R.B.X - ALeft;
  AHeight := R.B.Y - ATop;
end;

procedure TQWindow.SetGeometry(ALeft, ATop, AWidth, AHeight: Integer);
var
  R: TRect;
begin
  RectAssign(R, ALeft, ATop, ALeft + AWidth, ATop + AHeight);
  RectToViRect(R);
  RectGrow(R, 1, 1);
  ViView.ChangeBounds(R);
  BoundsChanged(R);
  ViView.Owner.Redraw;
end;


{
********************************* TQLabel **************************************
}
constructor TQLabel.Create(ParentWidget: TQWidget);
var
  R: TRect;
begin
  inherited Create;
  RectAssign(R, 2, 3, 12, 4);
  FViLabel := TViLabel.Create(R, '', nil);
  Assert(Assigned(ParentWidget));
  ParentWidget.ViClientView.Insert(FViLabel);
  FViLabel.EventMask := FViLabel.EventMask + evMouseUp + evMouseMove;
  FViLabel.OnEvent := DoEvent;
end;

destructor TQLabel.Destroy;
begin
  FViLabel.Free;
  inherited Destroy;
end;

function TQLabel.ViView: TViView;
begin
  Result := FViLabel;
end;

function TQLabel.GetCaption: string;
begin
  Result := FCaption;
end;

procedure TQLabel.SetCaption(const Value: string);
begin
  FCaption := Value;
  FViLabel.Text := TranslateAccelerator(Value);
  FViLabel.DrawView;
end;

function TQLabel.DoEvent(Sender: TObject; var Event: TEvent): Boolean;
begin
  if Event.What = evCommand then
  begin
    Result := False;
    Exit;
  end;
  Result := inherited DoEvent(Sender, Event);
end;


{
********************************* TQButton *************************************
}
procedure TQButton.DoClicked(Sender: TObject);
begin
  if Assigned(OnClicked) then OnClicked(Self);
end;

function TQButton.DoEvent(Sender: TObject; var Event: TEvent): Boolean;
begin
  if (Event.What = evCommand) then
  begin
    if Event.Command = cmFocusIn then
    begin
      inherited DoEvent(Sender, Event);
      Result := True;
    end
    else
    begin
      Result := False;
      Exit;
    end;
  end
  else Result := inherited DoEvent(Sender, Event);
end;

 
{
********************************* TQPushButton *********************************
}
constructor TQPushButton.Create(ParentWidget: TQWidget);
var
  R: TRect;
begin
  inherited Create;
  RectAssign(R, 2, 3, 12, 4);
  FViButton := TViButton.Create(R, '', 0, 0);
  FViButton.Options := FViButton.Options and not(ofPreProcess + ofPostProcess);
  Assert(Assigned(ParentWidget));
  ParentWidget.ViClientView.Insert(FViButton);
  FViButton.EventMask := FViButton.EventMask + evMouseUp + evMouseMove;
  FViButton.OnEvent := DoEvent;
end;

destructor TQPushButton.Destroy;
begin
  FViButton.Free;
  inherited Destroy;
end;

function TQPushButton.GetCaption: string;
begin
  Result := FCaption;
end;

procedure TQPushButton.SetCaption(const Value: string);
begin
  FCaption := Value;
  FViButton.Title := TranslateAccelerator(Value);
  FViButton.DrawView;
end;

function TQPushButton.ViView: TViView;
begin
  Result := FViButton;
end;

function TQPushButton.GetClientRect: TRect;
begin
  FViButton.GetExtent(Result);
  Inc(Result.A.X);
//  Dec(Result.B.X);
  Dec(Result.B.Y);
  ViRectToRect(Result);
end;

procedure TQPushButton.SetDefault(Value: Boolean);
begin
  FViButton.MakeDefault(Value);
end;


constructor TQLineEdit.Create(ParentWidget: TQWidget);
var
  R: TRect;
begin
  inherited Create;
  RectAssign(R, 2, 3, 12, 4);
  FViInputLine := TViInputLine.Create(R, leMaxLength);
  Assert(Assigned(ParentWidget));
  ParentWidget.ViClientView.Insert(FViInputLine);
  FViInputLine.EventMask := FViInputLine.EventMask + evMouseUp + evMouseMove;
  FViInputLine.OnEvent := DoEvent;
  FViInputLine.OnTextChanged := DoTextChanged;
  FViInputLine.EchoMode := emNormal;

//  FViInputLine.Options := FViInputLine.Options and ofPreProcess;
end;

destructor TQLineEdit.Destroy;
begin
  FViInputLine.Free;
  inherited Destroy;
end;

function TQLineEdit.ViView: TViView;
begin
  Result := FViInputLine;
end;

function TQLineEdit.GetCaption: string;
begin
  Result := FViInputLine.Data;
end;

procedure TQLineEdit.SetCaption(const Value: string);
begin
  FViInputLine.Text := Value;
  FViInputLine.DrawView;
end;

function TQLineEdit.GetMaxLength: Integer;
begin
  Result := FViInputLine.MaxLen;
end;

procedure TQLineEdit.SetMaxLength(Value: Integer);
begin
  FViInputLine.MaxLen := Value;
end;

procedure TQLineEdit.Clear;
begin
  FViInputLine.Data := '';
  ClearSelection;
end;

procedure TQLineEdit.ClearSelection;
begin
  FViInputLine.SelectAll(False);
end;

procedure TQLineEdit.Copy;
var
  I: Integer;
  TextLen: Integer;
  Text: string;
begin
  if MarkedText <> '' then
    if EchoMode = 0 then
      Clipboard.AsText := MarkedText
    else
    begin
      TextLen := Length(MarkedText);
      for I := 1 to TextLen do
        Text := Text + '*';
      Clipboard.AsText := Text;
    end;
end;

procedure TQLineEdit.Cut;
begin
  Copy;
  MarkedText := '';
end;

procedure TQLineEdit.SelectAll;
begin
  FViInputLine.SelectAll(True);
end;

function TQLineEdit.SelStart: Integer;
begin
  Result := FViInputLine.SelStart;
end;

function TQLineEdit.GetAutoSelect: Boolean;
begin
  Result := FViInputLine.AutoSelect;
end;

procedure TQLineEdit.SetAutoSelect(Value: Boolean);
begin
  FViInputLine.AutoSelect := Value;
end;

function TQLineEdit.GetHideSelection: Boolean;
begin
  Result := FViInputLine.HideSelection;
end;

procedure TQLineEdit.SetHideSelection(Value: Boolean);
begin
  FViInputLine.HideSelection := Value;
end;

function TQLineEdit.GetReadOnly: Boolean;
begin
  Result := FViInputLine.ReadOnly;
end;

procedure TQLineEdit.SetReadOnly(Value: Boolean);
begin
  FViInputLine.ReadOnly := Value;
end;

procedure TQLineEdit.DoTextChanged(Sender: TObject);
begin
  Edited := True;
  if Assigned(OnTextChanged) then OnTextChanged(Self);
end;

procedure TQLineEdit.SetEdited(Value: Boolean);
begin
  if FEdited <> Value then
  begin
    FEdited := Value;
  end;
end;

function TQLineEdit.GetMarkedText: string;
begin
  with FViInputLine do
    Result := System.Copy(Data, SelStart + 1, SelEnd - SelStart);
end;

procedure TQLineEdit.SetMarkedText(const Value: string);
var
  S: string;
  SelLength: Integer;
begin
  S := FViInputLine.Data;
  with FViInputLine do
  begin
    SelLength := SelEnd - SelStart;
    Delete(S, SelStart + 1, SelLength);

    if CurPos > SelStart then
      CurPos := CurPos - (SelLength);

    Insert(Value, S, SelStart + 1);
    SelEnd := SelStart + Length(Value);
  end;
  FViInputLine.Data := S;
  FViInputLine.DrawView;
end;

procedure TQLineEdit.SetSelection(SelStart, SelLength: Integer);
begin
  FViInputLine.SelStart := SelStart;
  FViInputLine.SelEnd := SelStart + SelLength;
  FViInputLine.DrawView;
end;

procedure TQLineEdit.SetSelStart(Value: Integer);
begin
  FViInputLine.SelStart := Value;
  FViInputLine.DrawView;
end;

procedure TQLineEdit.SetEchoMode(Value: Integer);
begin
  FViInputLine.EchoMode := Value;
end;

function TQLineEdit.GetEchoMode: Integer;
begin
  Result := FViInputLine.EchoMode;
end;

procedure TClipboard.SetAsText(const Value: string);
begin
  FAsText := Value;
end;


constructor TQListBox.Create(ParentWidget: TQWidget);
var
  Current: TViView;
  Owner: TViGroup;
  R: TRect;
begin
  inherited Create;
  RectAssign(R, 14, 2, 15, 7);
  FViScrollBar := TViScrollBar.Create(R);
  FViScrollBar.GrowMode := 0;
  Owner := ParentWidget.ViClientView as TViGroup;
  Current := Owner.Current;
  Owner.Insert(FViScrollBar);
  RectAssign(R, 2, 2, 14, 7);
  FViListBox := TViListBox.Create(R, 1, FViScrollBar);
  Owner.Insert(FViListBox);
  FViListBox.EventMask := FViListBox.EventMask + evMouseUp + evMouseMove;
  FViListBox.OnEvent := DoEvent;
  FViListBox.OnChanged := DoChanged;
  if Assigned(Current) then
    Current.Select;
end;

destructor TQListBox.Destroy;
begin
  FViListBox.Free;
  FViScrollBar.Free;
  inherited Destroy;
end;

function TQListBox.DoEvent(Sender: TObject; var Event: TEvent): Boolean;
begin
(*
  if Event.What = evCommand then
  begin
    Result := False;
    Exit;
  end;
*)
  Result := inherited DoEvent(Sender, Event);
end;

function TQListBox.ViView: TViView;
begin
  Result := FViListBox;
end;

procedure TQListBox.GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer);
var
  R: TRect;
begin
  ViView.GetBounds(R);
  R.B.X := R.B.X + 1;
  ViRectToRect(R);
  ALeft := R.A.X;
  ATop := R.A.Y;
  AWidth := R.B.X - ALeft;
  AHeight := R.B.Y - ATop;
end;

procedure TQListBox.SetGeometry(ALeft, ATop, AWidth, AHeight: Integer);
var
  R: TRect;
begin
  RectAssign(R, ALeft, ATop, ALeft + AWidth, ATop + AHeight);
  RectToViRect(R);
  R.B.X := R.B.X - 1;
  FViListBox.ChangeBounds(R);
  R.A.X := R.B.X;
  R.B.X := R.B.X + 1;
  FViScrollBar.ChangeBounds(R);
  ViView.Owner.Redraw;
end;

procedure TQListBox.Clear;
begin
  FViListBox.List.Clear;
  Update;
end;

function TQListBox.Count: Integer;
begin
  Result := FViListBox.List.Count;
end;

procedure TQListBox.RemoveItem(Index: Integer);
begin
  with FViListBox do
    if Index < Focused then
      FocusItem(Focused - 1);
  FViListBox.List.Delete(Index);
  Update;
end;

procedure TQListBox.InsertItem(const S: string; Index: Integer);
begin
  with FViListBox do
    if Index <= Focused then
      FocusItem(Focused + 1);
  FViListBox.List.Insert(Index, S);
  Update;
end;

procedure TQListBox.Update;
begin
  with FViListBox do
  begin
    SetRange(List.Count);
    if Range > 0 then
    begin
      if Focused >= List.Count - 1 then
        FocusItem(List.Count - 1);
    end;
    DrawView;
  end;
end;

function TQListBox.GetCurrentItem: Integer;
begin
  if FViListBox.Range > 0 then
    Result := FViListBox.Focused
  else
    Result := -1;
end;

procedure TQListBox.SetCurrentItem(Value: Integer);
begin
(*
  if FViListBox.Range > 0 then
  begin
    if Value > FViListBox.Range then
      Value := FViListBox.Range - 1;
    FViListBox.FocusItem(Value);
  end;
*)
  FViListBox.FocusItemNum(Value);
end;

function TQListBox.GetItem(Index: Integer): string;
begin
  Result := FViListBox.List[Index];
end;

procedure TQListBox.Sort;
begin
  (FViListBox.List as TStringList).Sort;
  with FViListBox do
  begin
    if Range > 0 then FocusItem(0);
    DrawView;
  end;
end;

function TQListBox.GetData(Index: Integer): Integer;
begin
  Result := Integer(FViListBox.List.Objects[Index]);
end;

procedure TQListBox.SetData(Index: Integer; Value: Integer);
begin
  FViListBox.List.Objects[Index] := Pointer(Value);
end;

procedure TQListBox.DoChanged(Sender: TObject);
begin
  if Assigned(OnChanged) then OnChanged(Self);
end;

procedure TQListBox.SetEnabled(Value: Boolean);
begin
  if Value <> Enabled then
  begin
    if Value then
    begin
      FViListBox.State := FViListBox.State and not sfDisabled;
      FViScrollBar.State := FViScrollBar.State and not sfDisabled;
    end
    else
    begin
      FViListBox.State := FViListBox.State or sfDisabled;
      FViScrollBar.State := FViScrollBar.State or sfDisabled;
    end;
    ViView.DrawView;
  end;
end;

constructor TQScrollBar.Create(ParentWidget: TQWidget);
var
  R: TRect;
begin
  inherited Create;
  RectAssign(R, 2, 3, 12, 4);
  FViScrollBar := TViScrollBar.Create(R);
  FViScrollBar.Min := 0;
  FViScrollBar.Max := 100;
  Assert(Assigned(ParentWidget));
  ParentWidget.ViClientView.Insert(FViScrollBar);
  FViScrollBar.EventMask := FViScrollBar.EventMask + evMouseUp + evMouseMove;
  FViScrollBar.Options := FViScrollBar.Options or ofSelectable or ofFirstClick;
  FViScrollBar.OnEvent := DoEvent;
  FViScrollBar.OnChange := DoChange;
  FViScrollBar.OnScroll := DoScroll;
end;

destructor TQScrollBar.Destroy;
begin
  FViScrollBar.Free;
  inherited Destroy;
end;

function TQScrollBar.DoEvent(Sender: TObject; var Event: TEvent): Boolean;
begin
  if (Event.What = evCommand) then
  begin
    if Event.Command = cmFocusIn then
    begin
      inherited DoEvent(Sender, Event);
      Result := True;
    end
    else
    begin
      Result := False;
      Exit;
    end;
  end
  else Result := inherited DoEvent(Sender, Event);
end;

function TQScrollBar.ViView: TViView;
begin
  Result := FViScrollBar;
end;

procedure TQScrollBar.BoundsChanged(const R: TRect);
begin
  FViScrollBar.BoundsChanged;
end;

procedure TQScrollBar.SetSteps(SmallChange, LargeChange: Integer);
begin
  FViScrollBar.SetStep(LargeChange, SmallChange);
end;

procedure TQScrollBar.SetOrientation(Value: TOrientation);
begin
  FViScrollBar.SetOrientation(Value);
end;

procedure TQScrollBar.SetRange(Min, Max: Integer);
begin
  FViScrollBar.SetRange(Min, Max);
end;

procedure TQScrollBar.SetValue(Position: Integer);
begin
  FViScrollBar.SetValue(Position);
end;

function TQScrollBar.PageStep: Integer;
begin
  Result := FViScrollBar.PgStep;
end;

function TQScrollBar.LineStep: Integer;
begin
  Result := FViScrollBar.ArStep;
end;

function TQScrollBar.MaxValue: Integer;
begin
  Result := FViScrollBar.Max;
end;

function TQScrollBar.MinValue: Integer;
begin
  Result := FViScrollBar.Min;
end;

function TQScrollBar.Value: Integer;
begin
  Result := FViScrollBar.Value;
end;

procedure TQScrollBar.DoChange(Sender: TObject);
begin
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TQScrollBar.DoScroll(Sender: TObject; ScrollPart: Integer);
begin
  if Assigned(OnScroll) then OnScroll(Self, ScrollPart);
end;

constructor TQRadioButton.Create(ParentWidget: TQWidget);
var
  R: TRect;
  Strings: TStrings;
begin
  inherited Create;
  RectAssign(R, 2, 3, 12, 4);
  Strings := TStringList.Create;
  try
    Strings.Add('AItem');
    FViRadioButton := TViRadioButtons.Create(R, Strings);
  finally
    Strings.Free;
  end;
  Assert(Assigned(ParentWidget));
  ParentWidget.ViClientView.Insert(FViRadioButton);
  FViRadioButton.EventMask := FViRadioButton.EventMask
      + evMouseUp + evMouseMove;
  FViRadioButton.OnEvent := DoEvent;
  FViRadioButton.OnClicked := DoClicked;
  SetChecked(False);
end;

destructor TQRadioButton.Destroy;
begin
  FViRadioButton.Free;
  inherited Destroy;
end;

function TQRadioButton.GetCaption: string;
begin
  Result := FCaption;
end;

procedure TQRadioButton.SetCaption(const Value: string);
begin
  FCaption := Value;
  FViRadioButton.Strings[0] := TranslateAccelerator(Value);
  FViRadioButton.DrawView;
end;

function TQRadioButton.ViView: TViView;
begin
  Result := FViRadioButton;
end;

function TQRadioButton.IsChecked: Boolean;
begin
  Result := FViRadioButton.Mark(0);
end;

procedure TQRadioButton.SetChecked(Value: Boolean);
begin
  if Value then
    FViRadioButton.Press(0)
  else
    FViRadioButton.Press(1);
  FViRadioButton.DrawView;
end;

procedure TQRadioButton.SetEnabled(Value: Boolean);
begin
  inherited SetEnabled(Value);
  if Value then
    FViRadioButton.EnableMask := $FFFFFFFF
  else
    FViRadioButton.EnableMask := $FFFFFFFE;
  FViRadioButton.DrawView;
end;


constructor TQCheckBox.Create(ParentWidget: TQWidget);
var
  R: TRect;
  Strings: TStrings;
begin
  inherited Create;
  RectAssign(R, 2, 3, 12, 4);
  Strings := TStringList.Create;
  try
    Strings.Add('AItem');
    TViMultiCheckBox :=
        TViMultiCheckBoxes.Create(R, Strings, 2, cfTwoBits, ' X'#63#254);
  finally
    Strings.Free;
  end;
  Assert(Assigned(ParentWidget));
  ParentWidget.ViClientView.Insert(TViMultiCheckBox);
  TViMultiCheckBox.EventMask := TViMultiCheckBox.EventMask
      + evMouseUp + evMouseMove;
  TViMultiCheckBox.OnEvent := DoEvent;
  TViMultiCheckBox.OnClicked := DoClicked;
  SetChecked(False);
end;

destructor TQCheckBox.Destroy;
begin
  TViMultiCheckBox.Free;
  inherited Destroy;
end;

function TQCheckBox.GetCaption: string;
begin
  Result := FCaption;
end;

function TQCheckBox.IsChecked: Boolean;
begin
  Result := TViMultiCheckBox.MultiMark(0) = 1;
end;

procedure TQCheckBox.SetCaption(const Value: string);
begin
  FCaption := Value;
  TViMultiCheckBox.Strings[0] := TranslateAccelerator(Value);
  TViMultiCheckBox.DrawView;
end;

procedure TQCheckBox.SetChecked(Value: Boolean);
begin
  if Value then
    TViMultiCheckBox.Value := 1
  else
    TViMultiCheckBox.Value := 0;
  TViMultiCheckBox.DrawView;
  DoClicked(nil);
end;

function TQCheckBox.ViView: TViView;
begin
  Result := TViMultiCheckBox;
end;

procedure TQCheckBox.SetNoChange;
begin
  if TViMultiCheckBox.SelRange = 3 then
    TViMultiCheckBox.Value := 2
  else
    TViMultiCheckBox.Value := 1;
  TViMultiCheckBox.DrawView;
  DoClicked(nil);
end;

procedure TQCheckBox.SetTriState(Value: Boolean);
begin
  if Value then
    TViMultiCheckBox.SelRange := 3
  else
    TViMultiCheckBox.SelRange := 2;
  TViMultiCheckBox.DrawView;
end;

function TQCheckBox.State: Integer;
begin
  Result := TViMultiCheckBox.MultiMark(0);
  case Result of
    1: Result := 2;
    2: Result := 1;
  end;
end;

procedure TQCheckBox.SetEnabled(Value: Boolean);
begin
  inherited SetEnabled(Value);
  if Value then
    TViMultiCheckBox.EnableMask := $FFFFFFFF
  else
    TViMultiCheckBox.EnableMask := $FFFFFFFE;
  TViMultiCheckBox.DrawView;
end;

constructor TQFrame.Create(ParentWidget: TQWidget);
var
  R: TRect;
begin
  inherited Create;
  RectAssign(R, 2, 3, 12, 20);
  FViClientGroup := TViClientGroup.Create(R);
  InitBackground;
  if FViBackground <> nil then FViClientGroup.Insert(FViBackground);
  RectAssign(R, 2, 2, 2, 3);
  FViLabel := TViLabel.Create(R, '', nil);
  FViClientGroup.Insert(FViLabel);
  Assert(Assigned(ParentWidget));
  ParentWidget.ViClientView.Insert(FViClientGroup);
  FViClientGroup.Options :=
      FViClientGroup.Options or (ofFramed + ofFirstClick);
//  FViClientGroup.OnEvent := DoEvent;
  FViBackground.OnEvent := DoEvent;
  FViLabel.OnEvent := DoEvent;
end;

destructor TQFrame.Destroy;
begin
  FViClientGroup.OnEvent := nil;
  FreeAndNil(FViLabel);
  FreeAndNil(FViBackground);
  FreeAndNil(FViClientGroup);
  inherited Destroy;
end;

function TQFrame.DoEvent(Sender: TObject; var Event: TEvent): Boolean;
begin
  if Event.What = evCommand then
    Result := False
  else if (Event.What = evKeyDown) then
    Result := False
  else
    Result := inherited DoEvent(Sender, Event);
end;

function TQFrame.GetCaption: string;
begin
  Result := FCaption;
end;

procedure TQFrame.SetCaption(const Value: string);
var
  ALeft, ATop, AWidth, AHeight: Integer;
begin
  FCaption := Value;
  FViLabel.Text := TranslateAccelerator(Value);
  FViLabel.Size.X := CStrLen(FViLabel.Text);
  GetGeometry(ALeft, ATop, AWidth, AHeight);
  SetGeometry(ALeft, ATop, AWidth, AHeight);
end;

function TQFrame.ViView: TViView;
begin
  Result := FViClientGroup;
end;

procedure TQFrame.InitBackground;
var
  R: TRect;
begin
  FViClientGroup.GetExtent(R);
  FViBackground := TViClientFrame.Create(R);
  with FViBackground do
  begin
    GrowMode := gfGrowHiX + gfGrowHiY;
    EventMask := EventMask or evBroadcast;
  end;
end;

function TQFrame.ViClientView: TViGroup;
begin
  Result := FViClientGroup;
end;

procedure TQFrame.SetFrame(Value: Boolean);
begin
  with FViClientGroup do
  begin
    if Value then
      Options := Options or ofFramed
    else
      Options := Options and not ofFramed;
    Redraw;
  end;
end;

procedure TQFrame.SetGeometry(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited SetGeometry(ALeft, ATop, AWidth, AHeight);
  FViLabel.Origin.X := (FViClientGroup.Size.X - FViLabel.Size.X) div 2;
  FViLabel.Origin.Y := (FViClientGroup.Size.Y - FViLabel.Size.Y) div 2;
  FViClientGroup.Redraw;
end;

constructor TQGroupBox.Create(ParentWidget: TQWidget);
var
  R: TRect;
begin
  inherited Create;
  RectAssign(R, 2, 3, 12, 20);

  FViBaseGroup := TViClientGroup.Create(R);
  RectMove(R, - R.A.X, - R.A.Y);
  RectGrow(R, 1, 1);
  FViClientGroup := TViClientGroup.Create(R);
  FViBaseGroup.Insert(FViClientGroup);

  InitBackground;
  if FViBackground <> nil then FViClientGroup.Insert(FViBackground);
  RectAssign(R, 2, 2, 2, 3);
  FViLabel := TViLabel.Create(R, '', nil);
  Assert(Assigned(ParentWidget));

  ParentWidget.ViClientView.Insert(FViBaseGroup);
  ParentWidget.ViClientView.Insert(FViLabel);
  FViBaseGroup.Options :=
      FViBaseGroup.Options or (ofFramed + ofFirstClick);
  FViClientGroup.Options :=
      FViClientGroup.Options or (ofFirstClick);
  FViClientGroup.GrowMode := gfGrowHiX + gfGrowHiY;

  FViBackground.OnEvent := DoEvent;
  FViLabel.OnEvent := DoEvent;
end;

destructor TQGroupBox.Destroy;
begin
  FViClientGroup.OnEvent := nil;
  FreeAndNil(FViLabel);
  FreeAndNil(FViBackground);
  FreeAndNil(FViClientGroup);
  FreeAndNil(FViBaseGroup);
  inherited Destroy;
end;

function TQGroupBox.DoEvent(Sender: TObject; var Event: TEvent): Boolean;
begin
  if Event.What = evCommand then
    Result := False
  else if (Event.What = evKeyDown) then
    Result := False
  else
    Result := inherited DoEvent(Sender, Event);
end;

function TQGroupBox.GetCaption: string;
begin
  Result := FCaption;
end;

procedure TQGroupBox.InitBackground;
var
  R: TRect;
begin
  FViClientGroup.GetExtent(R);
  FViBackground := TViClientFrame.Create(R);
  with FViBackground do
  begin
    GrowMode := gfGrowHiX + gfGrowHiY;
    EventMask := EventMask or evBroadcast;
  end;
end;

procedure TQGroupBox.SetCaption(const Value: string);
var
  ALeft, ATop, AWidth, AHeight: Integer;
begin
  FCaption := Value;
  FViLabel.Text := TranslateAccelerator(Value);
  FViLabel.Size.X := CStrLen(FViLabel.Text);
  GetGeometry(ALeft, ATop, AWidth, AHeight);
  SetGeometry(ALeft, ATop, AWidth, AHeight);
end;

procedure TQGroupBox.SetFrame(Value: Boolean);
begin
  with FViClientGroup do
  begin
    if Value then
      Options := Options or ofFramed
    else
      Options := Options and not ofFramed;
    Redraw;
  end;
end;

procedure TQGroupBox.SetGeometry(ALeft, ATop, AWidth, AHeight: Integer);
begin
  ALeft := ALeft + 8;
  ATop := ATop + 16;
  AWidth := AWidth - 16;
  AHeight := AHeight - 16;
  inherited SetGeometry(ALeft, ATop, AWidth, AHeight);
  FViLabel.Origin.X := FViBaseGroup.Origin.X + 1;
  FViLabel.Origin.Y := FViBaseGroup.Origin.Y - 1;
  FViBaseGroup.Redraw;
end;

function TQGroupBox.ViClientView: TViGroup;
begin
  Result := FViClientGroup;
end;

function TQGroupBox.ViView: TViView;
begin
  Result := FViBaseGroup;
end;

procedure TQGroupBox.GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited GetGeometry(ALeft, ATop, AWidth, AHeight);
  ALeft := ALeft - 8;
  ATop := ATop - 16;
  AWidth := AWidth + 16;
  AHeight := AHeight + 16;
end;


{
********************************* TQButtonGroup *************************************
}
constructor TQButtonGroup.Create(ParentWidget: TQWidget; Columns: Integer;
    Orientation: Integer);
begin
  inherited Create(ParentWidget);
  FColumns := Columns;
  FOrientation := Orientation;
end;

procedure TQButtonGroup.SetColumnLayout(Columns, Orientation: Integer);
begin
  FColumns := Columns;
  FOrientation := Orientation;
  UpdateLayout;
end;

procedure TQButtonGroup.UpdateLayout;
var
  ControlsCount: Integer;
  Curr: TViView;
  StepX: Integer;
  Rows: Integer;
  Index: Integer;

  procedure RealignView(Index, Rows, StepX: Integer);
  var
    R: TRect;
    Left: Integer;
    Top: Integer;
    Width: Integer;
    Row, Column: Integer;
    ClientHeight: Integer;

  begin
    Row := Index mod Rows;
    Column := Index div Rows;

    Left := Column * StepX + 1;
    if Column < FColumns - 1 then
      Width := StepX - 1
    else
      Width := FViClientGroup.Size.X - Pred(FColumns) * StepX;

    ClientHeight := FViClientGroup.Size.Y - 4;
    if Row = 0 then
      Top := 1
    else if Row = Rows - 1 then
      Top := FViClientGroup.Size.Y - 2
    else
      Top := (ClientHeight * Row div (Rows - 1)) + 2;

    RectAssign(R, Left, Top, Left + Width, Top + 1);
    Curr.ChangeBounds(R);
  end;

begin
  ControlsCount := -1;
  Curr := FViClientGroup.Last;
  if Assigned(Curr) then
    repeat
      Curr := Curr.Next;
      if Curr.State and sfVisible <> 0 then
        Inc(ControlsCount);
    until Curr = FViClientGroup.Last;
  if ControlsCount <= 0 then Exit;
  Rows := ControlsCount div FColumns;
  if (ControlsCount mod FColumns) > 0 then
    Inc(Rows);
  StepX := FViClientGroup.Size.X div (FColumns);
//  StepY := FViClientGroup.Size.Y div (Rows * 2 + 1);
  Index := ControlsCount - 1;
  Curr := FViClientGroup.Last;
  if Assigned(Curr) then
    repeat
      Curr := Curr.Next;
      if Curr is TViRadioButtons then
      begin
        RealignView(Index, Rows, StepX);
        Dec(Index);
      end;
    until Curr = FViClientGroup.Last;
  (ViView as TViGroup).Redraw;
end;

procedure TQButtonGroup.SetGeometry(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited SetGeometry(ALeft, ATop, AWidth, AHeight);
  UpdateLayout;
end;


{
********************************* TQMenuBar *************************************
}
constructor TQMenuBar.Create(ParentWidget: TQWidget);
begin
  inherited Create;
  FMenuData := TQMenuData.Create;
end;

destructor TQMenuBar.Destroy;
begin
  if ViMenuBar.Menu = FMenuData.Menu then
    Hide;
  FMenuData.Free;
  inherited Destroy;
end;

procedure TQMenuBar.Hide;
begin
  ViMenuBar.Current := nil;
  ViMenuBar.Menu := nil;
  ViMenuBar.OnActivated := nil;
  ViMenuBar.DrawView;
end;

procedure TQMenuBar.Show;
begin
  ViMenuBar.Current := nil;
  ViMenuBar.Menu := FMenuData.Menu;
  ViMenuBar.OnActivated := DoActivated;
  ViMenuBar.DrawView;
end;

procedure TQMenuBar.DoActivated(Sender: TObject; ID: Integer);
begin
  if Assigned(OnActivated) then OnActivated(Self, ID);
end;


{
********************************* TQPopupMenu *************************************
}
constructor TQPopupMenu.Create(ParentWidget: TQWidget);
begin
  inherited Create;
  FMenuData := TQMenuData.Create;
end;

destructor TQPopupMenu.Destroy;
begin
  FMenuData.Free;
  inherited Destroy;
end;

function TQPopupMenu.GetMenu: PMenu;
begin
  Result := FMenuData.Menu;
end;

procedure TQPopupMenu.Execute(Point: TPoint);
var
  R: TRect;
begin
  PointToViPoint(Point);
  RectAssign(R, 0, 0, 80, 40);
  RectMove(R, Point.X + 1, Point.Y);
  FViMenuPopup := TViMenuPopup.Create(R, FMenuData.Menu);
  try
    FViMenuPopup.OnActivated := DoActivated;
    FViMenuPopup.DoSelect;
  finally
    FViMenuPopup.Menu := nil;
    FViMenuPopup.Free;
  end;
end;

procedure TQPopupMenu.DoActivated(Sender: TObject; ID: Integer);
begin
  if Assigned(OnActivated) then OnActivated(Self, ID);
end;


{
********************************* TQMenuData *************************************
}
function TQMenuData.InsertItem(const ACaption: WideString; Index: Integer):
    Integer;
var
  I: Integer;
  OldNext, AItem: PMenuItem;
  Caption: string;
begin
  OldNext := FMenu.Items;
  if Index > 0 then
  begin
    for I := 1 to Index do
      OldNext := OldNext.Next;
  end;
  Caption := TranslateAccelerator(ACaption);
  if Caption = '' then Caption := Format('Item [%d]', [Index]);
  AItem := NewItem(Caption, '', 0, $FFFF, 0, OldNext);
  if Index = 0 then
  begin
    FMenu.Items := AItem;
    FMenu.Default := AItem;
  end
  else
  begin
    OldNext := FMenu.Items;
    for I := 1 to Index - 1 do
      OldNext := OldNext.Next;
    OldNext.Next := AItem;
  end;
  Result := Integer(AItem);
end;

function TQMenuData.InsertSubMenu(const ACaption: WideString; AWidget: TQWidget;
    Index: Integer): Integer;
var
  AItem: PMenuItem;
  AMenu: PMenu;
begin
  Result := InsertItem(ACaption, Index);
  AMenu := (AWidget as TQPopupMenu).Menu;
  AItem := PMenuItem(Result);
  AItem.SubMenu := AMenu;
  AItem.Command := 0;
end;

constructor TQMenuData.Create;
begin
  inherited Create;
  FMenu := NewMenu(nil);
end;

destructor TQMenuData.Destroy;
begin
  DisposeMenu(FMenu);
  inherited Destroy;
end;

procedure TQMenuData.ChangeItem(ID: Integer; const ACaption: WideString);
var
  MenuItem: PMenuItem;
begin
  MenuItem := PMenuItem(ID);
  MenuItem.Name := TranslateAccelerator(ACaption);
end;

procedure TQMenuData.RemoveItem(ID: Integer);
var
  MenuItem: PMenuItem;
  NextItem: PMenuItem;
  PrevItem: PMenuItem;
begin
  MenuItem := PMenuItem(ID);
  NextItem := MenuItem.Next;
  PrevItem := FMenu.Items;
  if PrevItem = MenuItem then
    PrevItem := nil;
  while Assigned(PrevItem) and (PrevItem.Next <> MenuItem) do
    PrevItem := PrevItem.Next;
  if Assigned(PrevItem) then
    PrevItem.Next := NextItem
  else
    Menu.Items := NextItem;
  if FMenu.Default = MenuItem then
    if Assigned(NextItem) then
      FMenu.Default := NextItem
    else
      FMenu.Default := PrevItem;
  Dispose(MenuItem);
end;

function TQMenuData.InsertSeparator(Index: Integer): Integer;
var
  I: Integer;
  OldNext, AItem: PMenuItem;
begin
  OldNext := FMenu.Items;
  if Index > 0 then
  begin
    for I := 1 to Index do
      OldNext := OldNext.Next;
  end;
  AItem := NewLine(OldNext);
  if Index = 0 then
    FMenu.Items := AItem
  else
  begin
    OldNext := FMenu.Items;
    for I := 1 to Index - 1 do
      OldNext := OldNext.Next;
    OldNext.Next := AItem;
  end;
  Result := Integer(AItem);
end;

procedure TQMenuData.DisposeMenu(Menu: PMenu);
var
  P, Q: PMenuItem;
begin
  if Menu <> nil then
  begin
    P := Menu^.Items;
    while P <> nil do
    begin
      Q := P;
      P := P^.Next;
      Dispose(Q);
    end;
    Dispose(Menu);
  end;
end;

procedure TQMenuData.SetItemEnabled(ID: Integer; Value: Boolean);
var
  MenuItem: PMenuItem;
begin
  MenuItem := PMenuItem(ID);
  MenuItem.Disabled := not Value;
end;

procedure TQMenuData.SetItemChecked(ID: Integer; Value: Boolean);
var
  MenuItem: PMenuItem;
begin
  MenuItem := PMenuItem(ID);
  MenuItem.Checked := Value;
end;

procedure TQMenuData.SetAccel(const Param: string; ShortCut: Word; ID: Integer);
var
  MenuItem: PMenuItem;
begin
  MenuItem := PMenuItem(ID);
  MenuItem.KeyCode := ShortCut;
  MenuItem.Param := Param;
end;


{
********************************* TQTimer *************************************
}
constructor TQTimer.Create;
begin
  inherited Create;
  FViTimer := TViTimer.Create;
  FViTimer.OnTimer := Timer;
end;

destructor TQTimer.Destroy;
begin
  FViTimer.Free;
  inherited Destroy;
end;

procedure TQTimer.SetEnabled(Value: Boolean);
begin
  FViTimer.Enabled := Value;
end;

procedure TQTimer.SetInterval(Value: Cardinal);
begin
  FViTimer.Interval := Value;
end;

procedure TQTimer.Timer(Sender: TObject);
begin
  if Assigned(FOnTimer) then FOnTimer(Self);
end;

function TQTimer.GetEnabled: Boolean;
begin
  Result := FViTimer.Enabled;
end;

function TQTimer.GetInterval: Cardinal;
begin
  Result := FViTimer.Interval;
end;


{
********************************* TQRtcTimer *************************************
}
constructor TQRtcTimer.Create;
begin
  inherited Create;
  FViRtcTimer := TViRtcTimer.Create;
  FViRtcTimer.OnTimer := Timer;
end;

destructor TQRtcTimer.Destroy;
begin
  FViRtcTimer.Free;
  inherited Destroy;
end;

function TQRtcTimer.GetEnabled: Boolean;
begin
  Result := FViRtcTimer.Enabled;
end;

function TQRtcTimer.GetInterval: Cardinal;
begin
  Result := FViRtcTimer.Interval;
end;

procedure TQRtcTimer.SetEnabled(Value: Boolean);
begin
  FViRtcTimer.Enabled := Value;
end;

procedure TQRtcTimer.SetInterval(Value: Cardinal);
begin
  FViRtcTimer.Interval := Value;
end;

procedure TQRtcTimer.Timer(Sender: TObject);
begin
  if Assigned(FOnTimer) then FOnTimer(Self);
end;

function TQRtcTimer.GetIntFreq: Integer;
begin
  Result :=Integer(FViRtcTimer.IntFreq);
end;

procedure TQRtcTimer.SetIntFreq(Value: Integer);
begin
  FViRtcTimer.IntFreq :=TIntFreq(Value);
end;


{
********************************* TQOpenComboBox *************************************
}
constructor TQOpenComboBox.Create(ParentWidget: TQWidget);
var
  R: TRect;
begin
  inherited Create(ParentWidget);

  FViScrollBar.Hide;
  FViListBox.Hide;

  RectAssign(R, 2, 3, 12, 4);
  FViInputLine := TViInputLine.Create(R, leMaxLength);
  Assert(Assigned(ParentWidget));
  ParentWidget.ViClientView.Insert(FViInputLine);
  FViInputLine.EventMask := FViInputLine.EventMask + evMouseUp + evMouseMove;
  FViInputLine.OnEvent := DoEvent;
  FViInputLine.OnTextChanged := DoTextChanged;
  FViInputLine.EchoMode := emNormal;

  RectAssign(R, 12, 3, 15, 4);
  FViHistory  := TViHistory.Create(R, FViInputLine, FViListBox.List);
  FViHistory.OnChanged := DoListChanged;
  FViHistory.OnSelected := DoSelected;
  Assert(Assigned(ParentWidget));
  ParentWidget.ViClientView.Insert(FViHistory);
end;

destructor TQOpenComboBox.Destroy;
begin
  FViHistory.Free;
  FViInputLine.Free;
  inherited Destroy;
end;

procedure TQOpenComboBox.Clear;
begin
  FViInputLine.Data := '';
  ClearSelection;
end;

procedure TQOpenComboBox.ClearSelection;
begin
  FViInputLine.SelectAll(False);
end;

procedure TQOpenComboBox.Copy;
var
  I: Integer;
  TextLen: Integer;
  Text: string;
begin
  if MarkedText <> '' then
    if EchoMode = 0 then
      Clipboard.AsText := MarkedText
    else
    begin
      TextLen := Length(MarkedText);
      for I := 1 to TextLen do
        Text := Text + '*';
      Clipboard.AsText := Text;
    end;
end;

procedure TQOpenComboBox.Cut;
begin
  Copy;
  MarkedText := '';
end;

procedure TQOpenComboBox.DoTextChanged(Sender: TObject);
begin
  Edited := True;
  if Assigned(OnTextChanged) then OnTextChanged(Self);
end;

function TQOpenComboBox.GetAutoSelect: Boolean;
begin
  Result := FViInputLine.AutoSelect;
end;

function TQOpenComboBox.GetCaption: string;
begin
  Result := FViInputLine.Data;
end;

function TQOpenComboBox.GetEchoMode: Integer;
begin
  Result := FViInputLine.EchoMode;
end;

function TQOpenComboBox.GetHideSelection: Boolean;
begin
  Result := FViInputLine.HideSelection;
end;

function TQOpenComboBox.GetMarkedText: string;
begin
  with FViInputLine do
    Result := System.Copy(Data, SelStart + 1, SelEnd - SelStart);
end;

function TQOpenComboBox.GetMaxLength: Integer;
begin
  Result := FViInputLine.MaxLen;
end;

function TQOpenComboBox.GetReadOnly: Boolean;
begin
  Result := FViInputLine.ReadOnly;
end;

procedure TQOpenComboBox.SelectAll;
begin
  FViInputLine.SelectAll(True);
end;

function TQOpenComboBox.SelStart: Integer;
begin
  Result := FViInputLine.SelStart;
end;

procedure TQOpenComboBox.SetAutoSelect(Value: Boolean);
begin
  FViInputLine.AutoSelect := Value;
end;

procedure TQOpenComboBox.SetCaption(const Value: string);
begin
  FViInputLine.Text := Value;
  FViInputLine.DrawView;
end;

procedure TQOpenComboBox.SetEchoMode(Value: Integer);
begin
  FViInputLine.EchoMode := Value;
end;

procedure TQOpenComboBox.SetEdited(Value: Boolean);
begin
  if FEdited <> Value then
  begin
    FEdited := Value;
  end;
end;

procedure TQOpenComboBox.SetHideSelection(Value: Boolean);
begin
  FViInputLine.HideSelection := Value;
end;

procedure TQOpenComboBox.SetMarkedText(const Value: string);
var
  S: string;
  SelLength: Integer;
begin
  S := FViInputLine.Data;
  with FViInputLine do
  begin
    SelLength := SelEnd - SelStart;
    Delete(S, SelStart + 1, SelLength);

    if CurPos > SelStart then
      CurPos := CurPos - (SelLength);

    Insert(Value, S, SelStart + 1);
    SelEnd := SelStart + Length(Value);
  end;
  FViInputLine.Data := S;
  FViInputLine.DrawView;
end;

procedure TQOpenComboBox.SetMaxLength(Value: Integer);
begin
  FViInputLine.MaxLen := Value;
end;

procedure TQOpenComboBox.SetReadOnly(Value: Boolean);
begin
  FViInputLine.ReadOnly := Value;
end;

procedure TQOpenComboBox.SetSelection(SelStart, SelLength: Integer);
begin
  FViInputLine.SelStart := SelStart;
  FViInputLine.SelEnd := SelStart + SelLength;
  FViInputLine.DrawView;
end;

procedure TQOpenComboBox.SetSelStart(Value: Integer);
begin
  FViInputLine.SelStart := Value;
  FViInputLine.DrawView;
end;

function TQOpenComboBox.ViView: TViView;
begin
  Result := FViInputLine;
end;

procedure TQOpenComboBox.GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer);
var
  R: TRect;
begin
  ViView.GetBounds(R);
  R.B.X := R.B.X + 3;
  ViRectToRect(R);
  ALeft := R.A.X;
  ATop := R.A.Y;
  AWidth := R.B.X - ALeft;
  AHeight := R.B.Y - ATop;
end;

procedure TQOpenComboBox.SetGeometry(ALeft, ATop, AWidth, AHeight: Integer);
var
  R: TRect;
begin
  RectAssign(R, ALeft, ATop, ALeft + AWidth, ATop + AHeight);
  RectToViRect(R);
  R.B.X := R.B.X - 3;
  FViInputLine.ChangeBounds(R);
  R.A.X := R.B.X;
  R.B.X := R.B.X + 3;
  FViHistory.ChangeBounds(R);
  ViView.Owner.Redraw;
end;

procedure TQOpenComboBox.SetCurrentItem(Index: Integer);
begin
  inherited SetCurrentItem(Index);
  if FViListBox.List.Count > 0 then
    Caption := FViListBox.List[Index];
end;

procedure TQOpenComboBox.Hide;
begin
  FViHistory.Hide;
  inherited;
end;

procedure TQOpenComboBox.Show;
begin
  inherited;
  FViHistory.Show;
  FViHistory.DrawView;
end;

procedure TQOpenComboBox.SetEnabled(Value: Boolean);
begin
  if Value <> Enabled then
  begin
    if Value then
    begin
      FViInputLine.State := FViInputLine.State and not sfDisabled;
      FViHistory.State := FViHistory.State and not sfDisabled;
    end
    else
    begin
      FViInputLine.State := FViInputLine.State or sfDisabled;
      FViHistory.State := FViHistory.State or sfDisabled;
    end;
    ViView.DrawView;
  end;
end;

function TQOpenComboBox.GetCurrentItem: Integer;
begin
  Result := FViListBox.List.IndexOf(Caption);
end;

procedure TQOpenComboBox.DoListChanged(Sender: TObject);
begin
  with Sender as TViListViewer do
    Caption := GetText(Focused, 0);
  if Assigned(OnTextChanged) then OnTextChanged(Self);
end;

procedure TQOpenComboBox.DoSelected(Sender: TObject);
begin
  if Assigned(OnChanged) then OnChanged(Self);
end;

procedure TQOpenComboBox.SetStyle(Value: Integer);
var
  QStyle: TQOpenComboBoxStyle;
begin
  if Value <> FStyle then
  begin
    if not (Value in [0..3]) then
      raise EComponentError.Create('Invalid ComboBox style');
    FStyle := Value;
    QStyle := TQOpenComboBoxStyle(FStyle);
    case QStyle of
      csDropDown:
        ReadOnly := False;
      csDropDownList:
        ReadOnly := True;
      csOwnerDrawFixed, csOwnerDrawVariable:
      begin
        raise EComponentError.Create('Invalid ComboBox style, '
            + 'csOwnerDraw not supported');
      end;
    end;
  end;
end;


{
********************************* TQDialog *************************************
}
constructor TQDialog.Create(Modal: Boolean; WidgetFlags: Integer);
begin
  inherited Create;
  FModal := Modal;
  FWidgetFlags := WidgetFlags;
  CreateWidget;
end;

destructor TQDialog.Destroy;
begin
  DestroyWidget;
  inherited Destroy;
end;

procedure TQDialog.Show;
begin
  // TODO -cMM: TQDialog.Show default body inserted
end;

procedure TQDialog.Hide;
begin
  // TODO -cMM: TQDialog.Hide default body inserted
end;

function TQDialog.DialogResult: Boolean;
begin
  Result := False;
end;

procedure TQDialog.GetGeometry(var ALeft, ATop, AWidth, AHeight: Integer);
var
  R: TRect;
begin
  ViDialog.GetBounds(R);
  ViRectToRect(R);
  ALeft := R.A.X;
  ATop := R.A.Y;
  AWidth := R.B.X - ALeft;
  AHeight := R.B.Y - ATop;
end;

procedure TQDialog.SetGeometry(ALeft, ATop, AWidth, AHeight: Integer);
var
  R: TRect;
begin
  RectAssign(R, ALeft, ATop, ALeft + AWidth, ATop + AHeight);
  RectToViRect(R);
//  ViDialog.ChangeBounds(R);

//  BoundsChanged(R);
//  ViDialog.Owner.Redraw;
end;

function TQDialog.GetCaption: string;
begin
  Result := ViDialog.GetTitle;
end;

procedure TQDialog.SetCaption(const Value: string);
begin
  ViDialog.Title := Value;
  ViDialog.DrawView;
end;

procedure TQDialog.BringToFront;
begin
  ViDialog.MakeFirst;
end;

procedure TQDialog.CreateWidget;
begin
end;

procedure TQDialog.DestroyWidget;
begin
end;


{
********************************* TQClxFileDialog ******************************
}
procedure TQClxFileDialog.CreateWidget;
const
  FDOptions: Word = fdOpenButton or fdNoLoadDir;
begin
  FHistoryList := TStringList.Create;
  FViFileDialog := TViFileDialog.Create('*.*', 'ATitle',
      'File ~n~ame', FDOptions, FHistoryList);
  FViFileDialog.OnCloseRequest := DoCloseRequest;
end;

procedure TQClxFileDialog.DestroyWidget;
begin
  FViFileDialog.Free;
  FHistoryList.Free;
end;

function TQClxFileDialog.ViDialog: TViDialog;
begin
  Result := FViFileDialog as TViDialog;
end;

function TQClxFileDialog.Execute: Boolean;
begin
  Result := ViApplication.ExecuteDialog(FViFileDialog, nil) <> cmCancel;
end;

procedure TQClxFileDialog.RereadDir;
begin
  FViFileDialog.ReadDirectory;
end;

function TQClxFileDialog.SelectedFile: string;
begin
  Result := FViFileDialog.GetFileName;
end;

procedure TQClxFileDialog.DoCloseRequest(Sender: TObject; var CanClose: 
    Boolean);
begin
  if Assigned(OnCloseRequest) then
    OnCloseRequest(Self, CanClose);
end;

procedure TQClxFileDialog.SetSaveMode;
begin
  with FViFileDialog.OpenSaveButton do
  begin
    Title := '~S~ave';
    DrawView;
  end;
end;

procedure TQClxFileDialog.SetEditText(const Value: string);
begin
  FViFileDialog.FileName.Data := Value;
end;

procedure TQClxFileDialog.SetDefaultExt(const Value: string);
var
  aValue: string;
begin
  if Value = '' then
    aValue := '*'
  else
    aValue := Value;
  FViFileDialog.WildCard := '*.' + aValue;
end;

procedure TQClxFileDialog.SetDir(const Value: string);
begin
  FViFileDialog.Directory := Value;
end;

procedure TQClxFileDialog.SetHistoryList(Value: TStrings);
begin
  FViFileDialog.ViHistory.List.Assign(Value);
end;

{
********************************* TQDirSelDlg ******************************
}
procedure TQDirSelDlg.SetCaption(const Value: string);
begin
  if FCaption <> Value then
  begin
    FCaption := Value;
  end;
end;

procedure TQDirSelDlg.SetRootDirectory(const Value: string);
begin
  if FRootDirectory <> Value then
  begin
    FRootDirectory := Value;
  end;
end;

function TQDirSelDlg.GetSelectedDir: string;
begin
  Result := FSelectedDir;
end;

function TQDirSelDlg.Execute: Boolean;
var
  ChDirDialog: TViChDirDialog;
begin
  ChDirDialog := TViChDirDialog.Create(Caption, cdNormal,
      ChDirHistory);
  try
    ChDirDialog.DirName := RootDirectory;
    Result := (ViApplication.ExecuteDialog(ChDirDialog, nil) <> cmCancel);
    if Result then
    begin
      FSelectedDir := ChDirDialog.DirName;
      if ChDirHistory.IndexOf(FSelectedDir) < 0 then
        ChDirHistory.Add(FSelectedDir);
    end;
  finally
    ChDirDialog.Free;
  end;
end;

{
********************************* TQMultiLineEdit ******************************
}
constructor TQMultiLineEdit.Create(ParentWidget: TQWidget);
var
  Current: TViView;
  Owner: TViGroup;
  R: TRect;
begin
  inherited Create;
  Owner := ParentWidget.ViClientView as TViGroup;
  Current := Owner.Current;

  RectAssign(R, 14, 2, 15, 7);
  FViVScrollBar := TViScrollBar.Create(R);
  FViVScrollBar.GrowMode := 0;
  Owner.Insert(FViVScrollBar);

  RectAssign(R, 2, 7, 14, 8);
  FViHScrollBar := TViScrollBar.Create(R);
  FViHScrollBar.GrowMode := 0;
  Owner.Insert(FViHScrollBar);

  RectAssign(R, 2, 2, 14, 7);
  FViMemo := TViMemo.Create(R, FViHScrollBar, FViVScrollBar, True);
  Owner.Insert(FViMemo);

  FViMemo.EventMask := FViMemo.EventMask + evMouseUp + evMouseMove;
  FViMemo.OnEvent := DoEvent;
  FViMemo.OnChanged := DoChanged;
  FViMemo.OnReturnPressed := DoReturnPressed;

  if Assigned(Current) then
    Current.Select;
end;

destructor TQMultiLineEdit.Destroy;
begin
  FViMemo.Free;
  FViHScrollBar.Free;
  FViVScrollBar.Free;
  inherited Destroy;
end;

function TQMultiLineEdit.ViView: TViView;
begin
  Result := FViMemo;
end;

procedure TQMultiLineEdit.GetGeometry(var ALeft, ATop, AWidth, AHeight: 
    Integer);
var
  R: TRect;
begin
  FViMemo.GetBounds(R);
  R.B.X := R.B.X + 1;
  R.B.Y := R.B.Y + 1;
  ViRectToRect(R);
  ALeft := R.A.X;
  ATop := R.A.Y;
  AWidth := R.B.X - ALeft;
  AHeight := R.B.Y - ATop;
end;

procedure TQMultiLineEdit.SetGeometry(ALeft, ATop, AWidth, AHeight: Integer);
var
  R: TRect;
  AX: Integer;
begin
  RectAssign(R, ALeft, ATop, ALeft + AWidth, ATop + AHeight);
  RectToViRect(R);
  R.B.X := R.B.X - 1;
  R.B.Y := R.B.Y - 1;
  FViMemo.ChangeBounds(R);
  AX := R.A.X;
  R.A.X := R.B.X;
  R.B.X := R.B.X + 1;
  FViVScrollBar.ChangeBounds(R);
  R.A.X := AX;
  R.A.Y := R.B.Y;
  R.B.X := R.B.X - 1;
  R.B.Y := R.B.Y + 1;
  FViHScrollBar.ChangeBounds(R);
  ViView.Owner.Redraw;
end;

procedure TQMultiLineEdit.Append(const Value: string);
begin
  FViMemo.Append(Value);
end;

function TQMultiLineEdit.AtEnd: Boolean;
begin
  Result := (FViMemo.CurPtr = FViMemo.BufLen);
end;

function TQMultiLineEdit.DoEvent(Sender: TObject; var Event: TEvent): Boolean;
begin
(*
  if Event.What = evCommand then
  begin
    Result := False;
    Exit;
  end;
*)
  Result := inherited DoEvent(Sender, Event);
end;

procedure TQMultiLineEdit.Clear;
begin
  FViMemo.Clear;
end;

function TQMultiLineEdit.NumLines: Integer;
begin
  Result := FViMemo.NumLines;
end;

procedure TQMultiLineEdit.RemoveLine(Index: Integer);
begin
  FViMemo.RemoveLine(Index);
end;

function TQMultiLineEdit.TextLine(Index: Integer): string;
begin
  Result := FViMemo.TextLine(Index);
end;

procedure TQMultiLineEdit.InsertLine(Index: Integer; const Line: string);
begin
  FViMemo.InsertLine(Index, Line);
end;

procedure TQMultiLineEdit.SetText(const Value: string);
begin
  FViMemo.SetText(Value);
end;

function TQMultiLineEdit.GetText: string;
begin
  Result := FViMemo.GetText;
end;

function TQMultiLineEdit.TextWidth(LineNum: Integer): Integer;
begin
  Result := LineLength(LineNum) * 8;
end;

procedure TQMultiLineEdit.Deselect;
begin
  FViMemo.Deselect;
end;

procedure TQMultiLineEdit.Insert(const Value: string; Mark: Boolean = False);
begin
  FViMemo.Insert(Value, Mark);
end;

function TQMultiLineEdit.HasMarkedText: Boolean;
begin
  Result := FViMemo.HasMarkedText;
end;

procedure TQMultiLineEdit.Copy;
begin
  FViMemo.Copy;
end;

procedure TQMultiLineEdit.Cut;
begin
  FViMemo.Cut;
end;

procedure TQMultiLineEdit.Paste;
begin
  FViMemo.Paste;
end;

procedure TQMultiLineEdit.CursorPosition(var Line, Column: Integer);
begin
  Line := FViMemo.CurPos.Y;
  Column := FViMemo.CurPos.X;
end;

function TQMultiLineEdit.Edited: Boolean;
begin
  Result := FViMemo.Modified;
end;

function TQMultiLineEdit.IsReadOnly: Boolean;
begin
  Result := FViMemo.ReadOnly;
end;

procedure TQMultiLineEdit.GetMarkedRegion(var Line1, Column1, Line2, Column2:
    Integer);
begin
  FViMemo.GetMarkedRegion(Line1, Column1, Line2, Column2);
end;

procedure TQMultiLineEdit.SetEdited(const Value: Boolean);
begin
  FViMemo.SetEdited(Value);
end;

procedure TQMultiLineEdit.SetReadOnly(const Value: Boolean);
begin
  FViMemo.ReadOnly := Value;
end;

function TQMultiLineEdit.LineLength(Index: Integer): Integer;
begin
  Result := Length(TextLine(Index));
end;

function TQMultiLineEdit.MarkedText: string;
begin
  Result := FViMemo.MarkedText;
end;

procedure TQMultiLineEdit.InsertAt(const Text: string; Column, Row: Integer;
    Mark: Boolean = False);
begin
  FViMemo.InsertAt(Text, Column, Row, Mark);
end;

procedure TQMultiLineEdit.SelectAll;
begin
  FViMemo.SetSelect(0, FViMemo.BufLen, True);
end;

procedure TQMultiLineEdit.SetCursorPosition(Line, Column: Integer);
begin
  FViMemo.SetCursorPosition(Line, Column);
end;

procedure TQMultiLineEdit.DoChanged(Sender: TObject);
begin
  if Assigned(OnChanged) then OnChanged(Self);
end;

procedure TQMultiLineEdit.DoReturnPressed(Sender: TObject);
begin
  if Assigned(OnReturnPressed) then OnReturnPressed(Self);
end;


initialization
  Clipboard := TClipboard.Create;
  ChDirHistory := TStringList.Create;

finalization
  FreeAndNil(Clipboard);
  FreeAndNil(ChDirHistory);
end.
