
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViColorSel.pas                                        }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViColorsel.pas,v 1.4 2003/12/01 21:10:38 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViColorsel.pas,v $
  Revision 1.4  2003/12/01 21:10:38  iwache
  Constructor TViColorDialog.Create now takes care about
  ViMode objects (the dialog himself and the buttons).
  This fix the bug which ignores any keys in the dialog.

  Revision 1.3  2002/10/27 22:25:16  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
unit ViColorSel;

{$X+,I-}

interface

uses
  Classes, ViTypes, ViUtils, ViClasses, ViViews, ViDialogs;

const
  cmColorForegroundChanged = 71;
  cmColorBackgroundChanged = 72;
  cmColorSet = 73;
  cmNewColorItem = 74;
  cmNewColorIndex = 75;
  cmSaveColorIndex = 76;

  cBackgroundBlink: Boolean = False;

{ True: 8 available/selectable background colors}
{ False:16 available/selectable colors}

type

  { TColorItem }

  PColorItem = ^TColorItem;
  TColorItem = record
    Name: string;
    Index: Byte;
    Next: PColorItem;
  end;

  { TColorGroup }

  PColorGroup = ^TColorGroup;
  TColorGroup = record
    Name: string;
    Index: Byte;
    Items: PColorItem;
    Next: PColorGroup;
  end;

  { TColorIndexes }

  PColorIndex = ^TColorIndex;
  TColorIndex = record
    GroupIndex: Byte;
    ColorSize: Byte;
    ColorIndex: array[0..255] of Byte;
  end;

  { TColorSelector }

  TColorSel = (csBackground, csForeground);

  TViColorSelector = class(TViView)
    Color: Byte;
    SelType: TColorSel;
    constructor Create(var Bounds: TRect; ASelType: TColorSel);
    procedure Draw; override;
    procedure HandleEvent(var Event: TEvent); override;
  end;

  { TMonoSelector }

  TViMonoSelector = class(TViCluster)
    constructor Create(var Bounds: TRect);
    procedure Draw; override;
    procedure HandleEvent(var Event: TEvent); override;
    function Mark(Item: Integer): Boolean; override;
    procedure NewColor;
    procedure Press(Item: Integer); override;
    procedure MovedTo(Item: Integer); override;
  end;

  { TColorDisplay }

  TViColorDisplay = class(TViView)
    Color: ^Byte;
    Text: string;
    constructor Create(var Bounds: TRect; AText: string);
    procedure Draw; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure SetColor(var AColor: Byte); virtual;
  end;

  { TColorGroupList }

  TViColorGroupList = class(TViListViewer)
    Groups: PColorGroup;
    constructor Create(var Bounds: TRect; AScrollBar: TViScrollBar; AGroups:
      PColorGroup);
    destructor Destroy; override;
    procedure FocusItem(Item: Integer); override;
    function GetText(Item: Integer; MaxLen: Integer): string; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure SetGroupIndex(GroupNum, ItemNum: Byte);
    function GetGroup(GroupNum: Byte): PColorGroup;
    function GetGroupIndex(GroupNum: Byte): Byte;
    function GetNumGroups: Byte;
  end;

  { TColorItemList }

  TViColorItemList = class(TViListViewer)
    Items: PColorItem;
    constructor Create(var Bounds: TRect; AScrollBar: TViScrollBar;
      AItems: PColorItem);
    procedure FocusItem(Item: Integer); override;
    function GetText(Item: Integer; MaxLen: Integer): string; override;
    procedure HandleEvent(var Event: TEvent); override;
  end;

  { TColorDialog }

  TViColorDialog = class(TViDialog)
    GroupIndex: Byte;
    Display: TViColorDisplay;
    Groups: TViColorGroupList;
    ForLabel: TViLabel;
    ForSel: TViColorSelector;
    BakLabel: TViLabel;
    BakSel: TViColorSelector;
    MonoLabel: TViLabel;
    MonoSel: TViMonoSelector;
    Pal: TPalette;
    constructor Create(APalette: TPalette; AGroups: PColorGroup);
    procedure HandleEvent(var Event: TEvent); override;
    procedure GetIndexes(var Colors: PColorIndex);
    procedure SetIndexes(var Colors: PColorIndex);
  end;

{ Pointer to saved color list item indexes }
var
  ColorIndexes: PColorIndex = nil;

{ Load and Store Palette routines }

procedure StoreIndexes(var S: TStream);
procedure LoadIndexes(var S: TStream);

{ Color list building routines }

function ColorItem(const Name: string; Index: Byte;
  Next: PColorItem): PColorItem;
function ColorGroup(const Name: string; Items: PColorItem;
  Next: PColorGroup): PColorGroup;

{ Standard color items functions }

function DesktopColorItems(const Next: PColorItem): PColorItem;
function MenuColorItems(const Next: PColorItem): PColorItem;
function DialogColorItems(Palette: Word; const Next: PColorItem): PColorItem;
function WindowColorItems(Palette: Word; const Next: PColorItem): PColorItem;

implementation

{ TColorSelector }

constructor TViColorSelector.Create(var Bounds: TRect; ASelType: TColorSel);
begin
  inherited Create(Bounds);
  Options := Options or (ofSelectable + ofFirstClick + ofFramed);
  EventMask := EventMask or evBroadcast;
  SelType := ASelType;
  Color := 0;
end;

procedure TViColorSelector.Draw;
var
  B: TDrawBuffer;
  C, I, J: Integer;
begin
  MoveChar(B, ' ', $70, Size.X);
  for I := 0 to Size.Y do
  begin
    if I < 4 then
      for J := 0 to 3 do
      begin
        C := I * 4 + J;
        MoveChar(B[J * 3], ldBlockFull, C, 3);
        if C = Byte(Color) then
        begin
          WordRec(B[J * 3 + 1]).Lo := 8;
          if C = 0 then WordRec(B[J * 3 + 1]).Hi := $70;
        end;
      end;
    WriteLine(0, I, Size.X, 1, B);
  end;
end;

procedure TViColorSelector.HandleEvent(var Event: TEvent);
const
  Width = 4;
var
  MaxCol: Byte;
  Mouse: TPoint;
  OldColor: Byte;

  procedure ColorChanged;
  var
    Msg: Integer;
  begin
    if SelType = csForeground then
      Msg := cmColorForegroundChanged
    else
      Msg := cmColorBackgroundChanged;
    SendMessage(Owner, evBroadcast, Msg, Pointer(Color));
  end;

begin
  inherited HandleEvent(Event);
  case Event.What of
    evMouseDown:
      begin
        OldColor := Color;
        repeat
          if MouseInView(Event.Where) then
          begin
            MakeLocal(Event.Where, Mouse);
            Color := Mouse.Y * 4 + Mouse.X div 3;
          end
          else
            Color := OldColor;
          ColorChanged;
          DrawView;
        until not MouseEvent(Event, evMouseMove);
      end;
    evKeyDown:
      begin
        if (SelType = csBackground) and cBackgroundBlink then
          MaxCol := 7
        else
          MaxCol := 15;
        case CtrlToArrow(Event.KeyCode) of
          kbLeft:
            if Color > 0 then
              Dec(Color)
            else
              Color := MaxCol;
          kbRight:
            if Color < MaxCol then
              Inc(Color)
            else
              Color := 0;
          kbUp:
            if Color > Width - 1 then
              Dec(Color, Width)
            else if Color = 0 then
              Color := MaxCol
            else
              Inc(Color, MaxCol - Width);
          kbDown:
            if Color < MaxCol - (Width - 1) then
              Inc(Color, Width)
            else if Color = MaxCol then
              Color := 0
            else
              Dec(Color, MaxCol - Width);
        else
          Exit;
        end;
      end;
    evBroadcast:
      if Event.Command = cmColorSet then
      begin
        if SelType = csBackground then
          Color := Event.InfoByte shr 4
        else
          Color := Event.InfoByte and $0F;
        DrawView;
        Exit;
      end
      else
        Exit;
  else
    Exit;
  end;
  DrawView;
  ColorChanged;
  ClearEvent(Event);
end;

{ TMonoSelector }

const
  MonoColors: array[0..4] of Byte = ($07, $0F, $01, $70, $09);

constructor TViMonoSelector.Create(var Bounds: TRect);
var
  Items: TStrings;
begin
  Items := TStringList.Create;
  try
    with Items do
    begin
      Add('Normal');
      Add('Highlight');
      Add('Underline');
      Add('Inverse');
    end;

    inherited Create(Bounds, Items);
    EventMask := EventMask or evBroadcast;
  finally
    Items.Free;
  end;
end;

procedure TViMonoSelector.Draw;
const
  Button = ' ( ) ';
begin
  DrawBox(Button, #7);
end;

procedure TViMonoSelector.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if (Event.What = evBroadcast) and (Event.Command = cmColorSet) then
  begin
    Value := Event.InfoByte;
    DrawView;
  end;
end;

function TViMonoSelector.Mark(Item: Integer): Boolean;
begin
  Mark := MonoColors[Item] = Value;
end;

procedure TViMonoSelector.NewColor;
begin
  SendMessage(Owner, evBroadcast, cmColorForegroundChanged,
    Pointer(Value and $0F));
  SendMessage(Owner, evBroadcast, cmColorBackgroundChanged,
    Pointer((Value shr 4) and $0F));
end;

procedure TViMonoSelector.Press(Item: Integer);
begin
  Value := MonoColors[Item];
  NewColor;
end;

procedure TViMonoSelector.MovedTo(Item: Integer);
begin
  Value := MonoColors[Item];
  NewColor;
end;

{ TColorDisplay }

constructor TViColorDisplay.Create(var Bounds: TRect; AText: string);
begin
  inherited Create(Bounds);
  EventMask := EventMask or evBroadcast;
  Text := AText;
  Color := nil;
end;

procedure TViColorDisplay.Draw;
var
  B: TDrawBuffer;
  I: Integer;
  C: Byte;
begin
  C := Color^;
  if C = 0 then C := ErrorAttr;
  for I := 0 to Size.X div Length(Text) do
    MoveStr(B[I * Length(Text)], Text, C);
  WriteLine(0, 0, Size.X, Size.Y, B);
end;

procedure TViColorDisplay.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  case Event.What of
    evBroadcast:
      case Event.Command of
        cmColorBackgroundChanged:
          begin
            Color^ := (Color^ and $0F) or (Event.InfoByte shl 4 and $F0);
            DrawView;
          end;
        cmColorForegroundChanged:
          begin
            Color^ := (Color^ and $F0) or (Event.InfoByte and $0F);
            DrawView;
          end;
      end;
  end;
end;

procedure TViColorDisplay.SetColor(var AColor: Byte);
begin
  Color := @AColor;
  SendMessage(Owner, evBroadcast, cmColorSet, Pointer(Color^));
  DrawView;
end;

{ TColorGroupList }

constructor TViColorGroupList.Create(var Bounds: TRect; AScrollBar:
  TViScrollBar; AGroups: PColorGroup);
var
  I: Integer;
begin
  inherited Create(Bounds, 1, nil, AScrollBar);
  Groups := AGroups;
  I := 0;
  while AGroups <> nil do
  begin
    AGroups := AGroups^.Next;
    Inc(I);
  end;
  SetRange(I);
end;

destructor TViColorGroupList.Destroy;

  procedure FreeItems(CurItem: PColorItem);
  var
    P: PColorItem;
  begin
    while CurItem <> nil do
    begin
      P := CurItem;
      CurItem := CurItem^.Next;
      Dispose(P);
    end;
  end;

  procedure FreeGroups(CurGroup: PColorGroup);
  var
    P: PColorGroup;
  begin
    while CurGroup <> nil do
    begin
      P := CurGroup;
      FreeItems(CurGroup^.Items);
      CurGroup := CurGroup^.Next;
      Dispose(P);
    end
  end;

begin
  FreeGroups(Groups);
  inherited Destroy;
end;

procedure TViColorGroupList.FocusItem(Item: Integer);
var
  CurGroup: PColorGroup;
begin
  inherited FocusItem(Item);
  CurGroup := Groups;
  while Item > 0 do
  begin
    CurGroup := CurGroup^.Next;
    Dec(Item);
  end;
  SendMessage(Owner, evBroadcast, cmNewColorItem, CurGroup);
end;

function TViColorGroupList.GetText(Item: Integer; MaxLen: Integer): string;
var
  CurGroup: PColorGroup;
begin
  CurGroup := Groups;
  while Item > 0 do
  begin
    CurGroup := CurGroup^.Next;
    Dec(Item);
  end;
  GetText := CurGroup^.Name;
end;

procedure TViColorGroupList.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if Event.What = evBroadcast then
    if Event.Command = cmSaveColorIndex then
      SetGroupIndex(Focused, Event.InfoByte);
end;

procedure TViColorGroupList.SetGroupIndex(GroupNum, ItemNum: Byte);
var
  Group: PColorGroup;
begin
  Group := GetGroup(GroupNum);
  if Group <> nil then
    Group^.Index := ItemNum;
end;

function TViColorGroupList.GetGroupIndex(GroupNum: Byte): Byte;
var
  Group: PColorGroup;
begin
  Group := GetGroup(GroupNum);
  if Group <> nil then
    GetGroupIndex := Group^.Index
  else
    GetGroupIndex := 0;
end;

function TViColorGroupList.GetGroup(GroupNum: Byte): PColorGroup;
var
  Group: PColorGroup;
begin
  Group := Groups;
  while GroupNum > 0 do
  begin
    Group := Group^.Next;
    Dec(GroupNum);
  end;
  GetGroup := Group;
end;

function TViColorGroupList.GetNumGroups: Byte;
var
  Index: Byte;
  Group: PColorGroup;
begin
  Index := 0;
  Group := Groups;
  while Group <> nil do
  begin
    Inc(Index);
    Group := Group^.Next;
  end;
  GetNumGroups := Index;
end;

{ TColorItemList }

constructor TViColorItemList.Create(var Bounds: TRect; AScrollBar:
  TViScrollBar; AItems: PColorItem);
var
  I: Integer;
begin
  inherited Create(Bounds, 1, nil, AScrollBar);
  EventMask := EventMask or evBroadcast;
  Items := AItems;
  I := 0;
  while AItems <> nil do
  begin
    AItems := AItems^.Next;
    Inc(I);
  end;
  SetRange(I);
end;

procedure TViColorItemList.FocusItem(Item: Integer);
var
  CurItem: PColorItem;
begin
  inherited FocusItem(Item);
  SendMessage(Owner, evBroadcast, cmSaveColorIndex, Pointer(Item));
  CurItem := Items;
  while Item > 0 do
  begin
    CurItem := CurItem^.Next;
    Dec(Item);
  end;
  SendMessage(Owner, evBroadcast, cmNewColorIndex, Pointer(CurItem^.Index));
end;

function TViColorItemList.GetText(Item: Integer; MaxLen: Integer): string;
var
  CurItem: PColorItem;
begin
  CurItem := Items;
  while Item > 0 do
  begin
    CurItem := CurItem^.Next;
    Dec(Item);
  end;
  GetText := CurItem^.Name;
end;

procedure TViColorItemList.HandleEvent(var Event: TEvent);
var
  CurItem: PColorItem;
  Group: PColorGroup;
  I: Integer;
begin
  inherited HandleEvent(Event);
  if Event.What = evBroadcast then
    case Event.Command of
      cmNewColorItem:
        begin
          Group := Event.InfoPtr;
          Items := Group^.Items;
          CurItem := Items;
          I := 0;
          while CurItem <> nil do
          begin
            CurItem := CurItem^.Next;
            Inc(I);
          end;
          SetRange(I);
          FocusItem(Group^.Index);
          DrawView;
        end;
    end;
end;

{ TColorDialog }

constructor TViColorDialog.Create(APalette: TPalette; AGroups: PColorGroup);
var
  R: TRect;
  P: TViView;
begin
  RectAssign(R, 0, 0, 61, 18);
  inherited Create(R, 'Colors', True);
  Options := Options or ofCentered;
  Pal := APalette;

  RectAssign(R, 18, 3, 19, 14);
  P := TViScrollBar.Create(R);
  Insert(P);
  RectAssign(R, 3, 3, 18, 14);
  Groups := TViColorGroupList.Create(R, TViScrollBar(P), AGroups);
  Insert(Groups);
  RectAssign(R, 2, 2, 8, 3);
  Insert(TViLabel.Create(R, '~G~roup', Groups));

  RectAssign(R, 41, 3, 42, 14);
  P := TViScrollBar.Create(R);
  Insert(P);
  RectAssign(R, 21, 3, 41, 14);
  P := TViColorItemList.Create(R, TViScrollBar(P), AGroups^.Items);
  Insert(P);
  RectAssign(R, 20, 2, 25, 3);
  Insert(TViLabel.Create(R, '~I~tem', P));

  RectAssign(R, 45, 3, 57, 7);
  ForSel := TViColorSelector.Create(R, csForeground);
  Insert(ForSel);
  Dec(R.A.Y);
  R.B.Y := R.A.Y + 1;
  ForLabel := TViLabel.Create(R, '~F~oreground', ForSel);
  Insert(ForLabel);

  Inc(R.A.Y, 7);
  Inc(R.B.Y, 10);
  if cBackgroundBlink then Dec(R.B.Y, 2);
  BakSel := TViColorSelector.Create(R, csBackground);
  Insert(BakSel);
  Dec(R.A.Y);
  R.B.Y := R.A.Y + 1;
  BakLabel := TViLabel.Create(R, '~B~ackground', BakSel);
  Insert(BakLabel);

  Dec(R.A.X);
  Inc(R.B.X);
  Inc(R.A.Y, 6);
  Inc(R.B.Y, 7);
  if cBackgroundBlink then
  begin
    Dec(R.A.Y, 2);
    Dec(R.B.Y, 2);
  end;
  Display := TViColorDisplay.Create(R, 'Text ');
  Insert(Display);

  RectAssign(R, 44, 3, 59, 8);
  MonoSel := TViMonoSelector.Create(R);
  MonoSel.Hide;
  Insert(MonoSel);
  RectAssign(R, 43, 2, 49, 3);
  MonoLabel := TViLabel.Create(R, '~C~olor', MonoSel);
  MonoLabel.Hide;
  Insert(MonoLabel);

  if (AGroups <> nil) and (AGroups^.Items <> nil) then
    Display.SetColor(Byte(Pal[AGroups^.Items^.Index]));

  RectAssign(R, 16, 15, 26, 17);
  P := TViButton.Create(R, 'O~K~', cmOk, bfDefault, True);
  Insert(P);
  RectAssign(R, 28, 15, 38, 17);
  P := TViButton.Create(R, 'Cancel', cmCancel, bfNormal, True);
  Insert(P);
  SelectNext(False);
end;

procedure TViColorDialog.HandleEvent(var Event: TEvent);
begin
  if Event.What = evBroadcast then
    if Event.Command = cmNewColorItem then
      GroupIndex := Groups.Focused;
  inherited HandleEvent(Event);
  if Event.What = evBroadcast then
    if Event.Command = cmNewColorIndex then
      Display.SetColor(Byte(Pal[Event.InfoByte]));
end;

procedure TViColorDialog.SetIndexes(var Colors: PColorIndex);
var
  NumGroups, Index: Byte;
begin
  NumGroups := Groups.GetNumGroups;
  if (Colors <> nil) and (Colors^.ColorSize <> NumGroups) then
  begin
    FreeMem(Colors, 2 + Colors^.ColorSize);
    Colors := nil;
  end;
  if Colors = nil then
  begin
    GetMem(Colors, 2 + NumGroups);
    FillChar(Colors^, 2 + NumGroups, 0);
    Colors^.ColorSize := NumGroups;
  end;
  for Index := 0 to NumGroups - 1 do
    Groups.SetGroupIndex(Index, Colors^.ColorIndex[Index]);
  GroupIndex := Colors^.GroupIndex;
end;

procedure TViColorDialog.GetIndexes(var Colors: PColorIndex);
var
  NumGroups, Index: Byte;
begin
  NumGroups := Groups.GetNumGroups;
  if Colors = nil then
  begin
    GetMem(Colors, 2 + NumGroups);
    FillChar(Colors^, 2 + NumGroups, 0);
    Colors^.ColorSize := NumGroups;
  end;
  Colors^.GroupIndex := GroupIndex;
  for Index := 0 to NumGroups - 1 do
    Colors^.ColorIndex[Index] := Groups.GetGroupIndex(Index);
end;

{ Load and Store Palette routines }

procedure LoadIndexes(var S: TStream);
var
  ColorSize: Byte;
begin
  S.Read(ColorSize, Sizeof(ColorSize));
  if ColorSize > 0 then
  begin
    if ColorIndexes <> nil then
      FreeMem(ColorIndexes, 2 + ColorIndexes^.ColorSize);
    getmem(ColorIndexes, ColorSize);
    S.Read(ColorIndexes^, ColorSize);
  end;
end;

procedure StoreIndexes(var S: TStream);
var
  ColorSize: Byte;
begin
  if ColorIndexes <> nil then
    ColorSize := 2 + ColorIndexes^.ColorSize
  else
    ColorSize := 0;
  S.Write(ColorSize, Sizeof(ColorSize));
  if ColorSize > 0 then
    S.Write(ColorIndexes^, ColorSize);
end;

{ -- Color list building routines -- }

function ColorItem(const Name: string; Index: Byte;
  Next: PColorItem): PColorItem;
var
  Item: PColorItem;
begin
  New(Item);
  Item^.Name := Name;
  Item^.Index := Index;
  Item^.Next := Next;
  ColorItem := Item;
end;

function ColorGroup(const Name: string; Items: PColorItem;
  Next: PColorGroup): PColorGroup;
var
  Group: PColorGroup;
begin
  New(Group);
  Group^.Name := Name;
  Group^.Index := 0;
  Group^.Items := Items;
  Group^.Next := Next;
  ColorGroup := Group;
end;

{ Standard color items functions }

function DesktopColorItems(const Next: PColorItem): PColorItem;
begin
  DesktopColorItems :=
    ColorItem('Color', 1,
    Next);
end;

function MenuColorItems(const Next: PColorItem): PColorItem;
begin
  MenuColorItems :=
    ColorItem('Normal', 2,
    ColorItem('Disabled', 3,
    ColorItem('Shortcut', 4,
    ColorItem('Selected', 5,
    ColorItem('Selected disabled', 6,
    ColorItem('Shortcut selected', 7,
    Next))))));
end;

function DialogColorItems(Palette: Word; const Next: PColorItem): PColorItem;
const
  COffset: array[dpBlueDialog..dpGrayDialog] of Byte =
  (64, 96, 32);
var
  Offset: Byte;
begin
  Offset := COffset[Palette];
  DialogColorItems :=
    ColorItem('Frame/background', Offset + 1,
    ColorItem('Frame icons', Offset + 2,
    ColorItem('Scroll bar page', Offset + 3,
    ColorItem('Scroll bar icons', Offset + 4,
    ColorItem('Static text', Offset + 5,

    ColorItem('Label normal', Offset + 6,
    ColorItem('Label selected', Offset + 7,
    ColorItem('Label shortcut', Offset + 8,

    ColorItem('Button normal', Offset + 9,
    ColorItem('Button default', Offset + 10,
    ColorItem('Button selected', Offset + 11,
    ColorItem('Button disabled', Offset + 12,
    ColorItem('Button shortcut', Offset + 13,
    ColorItem('Button shadow', Offset + 14,

    ColorItem('Cluster normal', Offset + 15,
    ColorItem('Cluster selected', Offset + 16,
    ColorItem('Cluster shortcut', Offset + 17,

    ColorItem('Input normal', Offset + 18,
    ColorItem('Input selected', Offset + 19,
    ColorItem('Input arrow', Offset + 20,

    ColorItem('History button', Offset + 21,
    ColorItem('History sides', Offset + 22,
    ColorItem('History bar page', Offset + 23,
    ColorItem('History bar icons', Offset + 24,

    ColorItem('List normal', Offset + 25,
    ColorItem('List focused', Offset + 26,
    ColorItem('List selected', Offset + 27,
    ColorItem('List divider', Offset + 28,

    ColorItem('Information pane', Offset + 29,
    Next)))))))))))))))))))))))))))));
end;

function WindowColorItems(Palette: Word;
  const Next: PColorItem): PColorItem;
const
  COffset: array[wpBlueWindow..wpGrayWindow] of Byte =
  (8, 16, 24);
var
  Offset: Word;
begin
  Offset := COffset[Palette];
  WindowColorItems :=
    ColorItem('Frame passive', Offset + 0,
    ColorItem('Frame active', Offset + 1,
    ColorItem('Frame icons', Offset + 2,
    ColorItem('Scroll bar page', Offset + 3,
    ColorItem('Scroll bar icons', Offset + 4,
    ColorItem('Normal text', Offset + 5,
    Next))))));
end;

{ ColorSel registration procedure }

initialization

finalization
  if Assigned(ColorIndexes) then
    FreeMem(ColorIndexes);
end.

