
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViDialogs.pas                                         }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViDialogs.pas,v 1.15 2004/03/30 20:25:54 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViDialogs.pas,v $
  Revision 1.15  2004/03/30 20:25:54  iwache
  Method TViHistory.RecordHistory adds only new strings not
  already exist in the list.

  Revision 1.14  2003/12/01 21:00:43  iwache
  New commands cmAbort, cmRetry, cmIgnore now also terminate
  the dialog in procedure TViDialog.HandleEvent.

  Revision 1.13  2003/11/19 18:49:54  iwache
  For Vi only mode TViDialog.Create has a
  new default parameter ViMode: Boolean = False.
  Set it to true if you like to use the class
  unconnected to WQ and CLX.

  Revision 1.12  2003/10/16 19:09:27  iwache
  Wrong behaviour of TQOpenComboBox events OnChanged
  and OnSelected fixed.

  Revision 1.11  2003/03/03 16:00:33  iwache
  TViHistoryViewer:
  - constructor Create now use a Strings list
    instead of the old HistoryId
  TViHistoryWindow
  - constructor Create now use a Strings list
    instead of the old HistoryId
  - procedure InitViewer now use a Strings list
    instead of the old HistoryId
  TViHistory
  - constructor Create now use a Strings list
    instead of the old HistoryId
  - property OnChanged added
  - old RecordHistory behaviour from TV removed

  Revision 1.10  2003/01/25 11:21:16  iwache
  DWPL-Bugs-673813:
  Description:
  The dblclick event of TEdit and TListBox is invalid.
  The ReadOnly property of TEdit is invalid.

  Solution for Doubleclick:
  DblClick Event handler added in TWidgetControl.EventFilter.
  TCustomListBox.DblClick modified to fit behaviour with
  the TV-ListBox on doubleclick. An other solution would
  be fit behaiour of TV-ListBox to the VisualCLX one. But this
  would need much more work and tests.

  Solution for ReadOnly property of TEdit:
  TViInputLine added two new properties 'ReadOnly' and 'Text'.
  TViInputLine.HandleEvent and TViInputLine.SetData modified
  to handle ReadOnly property right.

  WQ.pas: modified for both bugs.
  ViTypes.pas: event code evMouseDbl  = $0800; added.
  This is a dummy TV event code.

  Revision 1.9  2003/01/23 19:21:32  iwache
  - TViDialog.HandleEvent and TViButton.HandleEvent
  have now slightly different behaviour when option ofViInternal
  is set for the component. This is to fix the enter key bug
  in Vi created dialogs like MessageBoxRect.

  Revision 1.8  2003/01/22 20:52:33  iwache
  2'nd Fix for dwpl-Bug #667896 :
  Description:
  Buttons do not handle Enter...
  Solution:
  function TButton.NeedKey implemented and some modifications
  done to VisualCLX, WQ-Layer and TV-Layer.

  Revision 1.7  2003/01/20 22:18:19  iwache
  Fix for dwpl-Bug #667896 :
  Description:
  Buttons do not handle Enter, it is simply ignored, even if
  the focus is on a button (not to mention `default' buttons)
  (this applies to all DWPL versions up to v2.2.0 b3)
  Function-Test:
  Enter function tested (property Default) - OK
  ESC function tested (property Escape) - OK
  Space bar don't works with focused buttons - Error
  Solution:
  Simulate a KeyUp event for space key on buttons.
  This behaviour is different to Windows CLX where
  the event fires only once when the space key fires
  a keyup event. Unfortunately we don't have any keyup
  events with DOS and WDosX :-(

  Revision 1.6  2002/10/27 22:25:17  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
unit ViDialogs;

{$X+,I-}

interface

uses
  Classes, ViTypes, ViUtils, ViClasses, ViViews;

const

{ Color palettes }

  CGrayDialog = #32#33#34#35#36#37#38#39#40#41#42#43#44#45#46#47 +
    #48#49#50#51#52#53#54#55#56#57#58#59#60#61#62#63;
  CBlueDialog = #64#65#66#67#68#69#70#71#72#73#74#75#76#77#78#79 +
    #80#81#82#83#84#85#86#87#88#89#90#91#92#92#94#95;
  CCyanDialog = #96#97#98#99#100#101#102#103#104#105#106#107#108 +
    #109#110#111#112#113#114#115#116#117#118#119#120 +
    #121#122#123#124#125#126#127;

  CDialog = CGrayDialog;

  CStaticText = #6;
  CLabel = #7#8#9#9;
  CButton = #10#11#12#13#14#14#14#15;
  CCluster = #16#17#18#18#31;
  CInputLine = #19#19#20#21;
  CHistory = #22#23;
  CHistoryWindow = #19#19#21#24#25#19#20;
  CHistoryViewer = #6#6#7#6#6;

{ TDialog palette entires }

  dpBlueDialog = 0;
  dpCyanDialog = 1;
  dpGrayDialog = 2;

{ TButton flags }

  bfNormal = $00;
  bfDefault = $01;
  bfLeftJust = $02;
  bfBroadcast = $04;
  bfGrabFocus = $08;

{ TMultiCheckboxes flags }
{ hibyte = number of bits }
{ lobyte = bit mask }

  cfOneBit = $0101;
  cfTwoBits = $0203;
  cfFourBits = $040F;
  cfEightBits = $08FF;

type

{ TDialog class }

  { Palette layout }
  {  1 = Frame passive }
  {  2 = Frame active }
  {  3 = Frame icon }
  {  4 = ScrollBar page area }
  {  5 = ScrollBar controls }
  {  6 = StaticText }
  {  7 = Label normal }
  {  8 = Label selected }
  {  9 = Label shortcut }
  { 10 = Button normal }
  { 11 = Button default }
  { 12 = Button selected }
  { 13 = Button disabled }
  { 14 = Button shortcut }
  { 15 = Button shadow }
  { 16 = Cluster normal }
  { 17 = Cluster selected }
  { 18 = Cluster shortcut }
  { 19 = InputLine normal text }
  { 20 = InputLine selected text }
  { 21 = InputLine arrows }
  { 22 = History arrow }
  { 23 = History sides }
  { 24 = HistoryWindow scrollbar page area }
  { 25 = HistoryWindow scrollbar controls }
  { 26 = ListViewer normal }
  { 27 = ListViewer focused }
  { 28 = ListViewer selected }
  { 29 = ListViewer divider }
  { 30 = InfoPane }
  { 31 = Cluster disabled }
  { 32 = Reserved }

  TViDialog = class(TViWindow)
  public
    constructor Create(var Bounds: TRect; ATitle: TTitleStr;
        ViMode: Boolean = False);
    procedure HandleEvent(var Event: TEvent); override;
    function Valid(Command: Word): Boolean; override;
    procedure SizeLimits(var Min, Max: TPoint); override; // Added for VPPM
  end;

{ TSItem }

  PSItem = ^TSItem;
  TSItem = record
    Value: string;
    Next: PSItem;
  end;

{ TInputLine class }

  { Palette layout }
  { 1 = Passive }
  { 2 = Active }
  { 3 = Selected }
  { 4 = Arrows }

  TViInputLine = class(TViView)
  private
    FOnTextChanged: TChangedEvent;
    FData: string;
    FAutoSelect: Boolean;
    FHideSelection: Boolean;
    FReadOnly: Boolean;
    function CanScroll(Delta: Integer): Boolean;
    procedure SetData(const Value: string);
    function GetText: string;
    procedure SetText(const Value: string);
  protected
    procedure DoTextChanged;
    procedure InsertData(const Source: string; Index: Integer);
    procedure DeleteData(Index, Count: Integer);
  public
    EchoMode: Integer;
    MaxLen: Integer;
    CurPos: Integer;
    FirstPos: Integer;
    SelStart: Integer;
    SelEnd: Integer;
    constructor Create(var Bounds: TRect; AMaxLen: Integer);
    procedure Draw; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure SelectAll(Enable: Boolean);
    procedure SetState(AState: Word; Enable: Boolean); override;
    function Valid(Command: Word): Boolean; override;
    property OnTextChanged: TChangedEvent read FOnTextChanged write
        FOnTextChanged;
    property Data: string read FData write SetData;
    property AutoSelect: Boolean read FAutoSelect write FAutoSelect;
    property HideSelection: Boolean read FHideSelection write FHideSelection;
    property ReadOnly: Boolean read FReadOnly write FReadOnly;
    property Text: string read GetText write SetText;
  end;

{ TButton class }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Default text }
  { 3 = Selected text }
  { 4 = Disabled text }
  { 5 = Normal shortcut }
  { 6 = Default shortcut }
  { 7 = Selected shortcut }
  { 8 = Shadow }

  TViButton = class(TViView)
  private
    Down: Boolean;
  public
    Title: string;
    Command: Word;
    Flags: Byte;
    AmDefault: Boolean;
    constructor Create(var Bounds: TRect; ATitle: TTitleStr; ACommand: Word;
      AFlags: Word; ViMode: Boolean = False);
    procedure Draw; override;
    procedure DrawState(Down: Boolean);
    procedure HandleEvent(var Event: TEvent); override;
    procedure MakeDefault(Enable: Boolean);
    procedure Press; virtual;
    procedure SetState(AState: Word; Enable: Boolean); override;
  end;

{ TCluster }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Selected text }
  { 3 = Normal shortcut }
  { 4 = Selected shortcut }
  { 5 = Disabled text }

  TViCluster = class(TViView)
  private
    FOnClicked: TNotifyEvent;
    function Column(Item: Integer): Integer;
    function FindSel(P: TPoint): Integer;
    function Row(Item: Integer): Integer;
  protected
    procedure DoClicked;
  public
    Value: Longint;
    Sel: Integer;
    EnableMask: LongWord;
    Strings: TStrings;
    constructor Create(var Bounds: TRect; AStrings: TStrings);
    destructor Destroy; override;
    function ButtonState(Item: Integer): Boolean;
    procedure DrawBox(const Icon: string; Marker: Char);
    procedure DrawMultiBox(const Icon, Marker: string);
    function GetHelpCtx: Word; override;
    procedure HandleEvent(var Event: TEvent); override;
    function Mark(Item: Integer): Boolean; virtual;
    function MultiMark(Item: Integer): Byte; virtual;
    procedure Press(Item: Integer); virtual;
    procedure MovedTo(Item: Integer); virtual;
    procedure SetButtonState(AMask: LongWord; Enable: Boolean);
    procedure SetState(AState: Word; Enable: Boolean); override;
    property OnClicked: TNotifyEvent read FOnClicked write FOnClicked;
  end;

{ TRadioButtons }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Selected text }
  { 3 = Normal shortcut }
  { 4 = Selected shortcut }

  TViRadioButtons = class(TViCluster)
  public
    procedure Draw; override;
    function Mark(Item: Integer): Boolean; override;
    procedure MovedTo(Item: Integer); override;
    procedure Press(Item: Integer); override;
  end;

{ TCheckBoxes }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Selected text }
  { 3 = Normal shortcut }
  { 4 = Selected shortcut }

  TViCheckBoxes = class(TViCluster)
  public
    procedure Draw; override;
    function Mark(Item: Integer): Boolean; override;
    procedure Press(Item: Integer); override;
  end;

{ TMultiCheckBoxes }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Selected text }
  { 3 = Normal shortcut }
  { 4 = Selected shortcut }

  TViMultiCheckBoxes = class(TViCluster)
  public
    SelRange: ShortInt;
    Flags: Word;
    States: string;
    constructor Create(var Bounds: TRect; AStrings: TStrings; ASelRange: Byte;
      AFlags: Word; const AStates: string);
    procedure Draw; override;
    function MultiMark(Item: Integer): Byte; override;
    procedure Press(Item: Integer); override;
  end;

{ TListBox }

  { Palette layout }
  { 1 = Active }
  { 2 = Inactive }
  { 3 = Focused }
  { 4 = Selected }
  { 5 = Divider }

  TViListBox = class(TViListViewer)
  private
  protected
  public
    List: TStrings;
    constructor Create(var Bounds: TRect; ANumCols: Word;
      AScrollBar: TViScrollBar);
    function GetText(Item: Integer; MaxLen: Integer): string; override;
    procedure AssignList(AList: TStrings); virtual;
    destructor Destroy; override;
  end;

{ TStaticText }

  { Palette layout }
  { 1 = Text }

  TViStaticText = class(TViView)
  public
    Text: string;
    constructor Create(var Bounds: TRect; const AText: string);
    procedure Draw; override;
    procedure GetText(var S: string); virtual;
  end;

{ TLabel }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Selected text }
  { 3 = Normal shortcut }
  { 4 = Selected shortcut }

  TViLabel = class(TViStaticText)
  public
    Link: TViView;
    Light: Boolean;
    constructor Create(var Bounds: TRect; const AText: string; ALink: TViView);
    procedure Draw; override;
    procedure HandleEvent(var Event: TEvent); override;
  end;

{ THistoryViewer }

  { Palette layout }
  { 1 = Active }
  { 2 = Inactive }
  { 3 = Focused }
  { 4 = Selected }
  { 5 = Divider }

  TViHistoryViewer = class(TViListViewer)
  private
    FStrings: TStrings;
  public
    HistoryId: Integer;
    constructor Create(var Bounds: TRect; AHScrollBar, AVScrollBar: TViScrollBar;
        AStrings: TStrings);
    function GetText(Item: Integer; MaxLen: Integer): string; override;
    procedure HandleEvent(var Event: TEvent); override;
    function HistoryWidth: Integer;
  end;

{ THistoryWindow }

  { Palette layout }
  { 1 = Frame passive }
  { 2 = Frame active }
  { 3 = Frame icon }
  { 4 = ScrollBar page area }
  { 5 = ScrollBar controls }
  { 6 = HistoryViewer normal text }
  { 7 = HistoryViewer selected text }

  TViHistoryWindow = class(TViWindow)
  private
    FOnChanged: TChangedEvent;
  protected
    procedure DoChanged(Sender: TObject);
  public
    Viewer: TViListViewer;
    constructor Create(var Bounds: TRect; Strings: TStrings);
    function GetSelection: string; virtual;
    procedure InitViewer(Strings: TStrings); virtual;
    property OnChanged: TChangedEvent read FOnChanged write FOnChanged;
  end;

{ THistory }

  { Palette layout }
  { 1 = Arrow }
  { 2 = Sides }

  TViHistory = class(TViView)
  private
    FOnChanged: TChangedEvent;
    FOnSelected: TChangedEvent;
  protected
    procedure DoChanged(Sender: TObject);
    procedure DoSelected(Sender: TObject);
  public
    Link: TViInputLine;
    List: TStrings;
    constructor Create(var Bounds: TRect; ALink: TViInputLine; AList: TStrings);
    procedure Draw; override;
    procedure HandleEvent(var Event: TEvent); override;
    function InitHistoryWindow(var Bounds: TRect): TViHistoryWindow; virtual;
    procedure RecordHistory(const S: string); virtual;
    property OnChanged: TChangedEvent read FOnChanged write FOnChanged;
    property OnSelected: TChangedEvent read FOnSelected write FOnSelected;
  end;

{ SItem routines }

function NewSItem(const Str: string; ANext: PSItem): PSItem;

const

{ Dialog broadcast commands }

  cmRecordHistory = 60;

implementation

uses
  ViEventsClass, ViHistList, ViApp;

const

{ TButton messages }

  cmGrabDefault = 61;
  cmReleaseDefault = 62;

{ Utility functions }

function IsBlank(Ch: Char): Boolean;
begin
  IsBlank := (Ch = ' ') or (Ch = #13) or (Ch = #10);
end;

{ TDialog }

constructor TViDialog.Create(var Bounds: TRect; ATitle: TTitleStr; ViMode: 
    Boolean = False);
begin
  inherited Create(Bounds, ATitle, wnNoNumber);
  FPalettes[dpBlueDialog] := CBlueDialog;
  FPalettes[dpCyanDialog] := CCyanDialog;
  FPalettes[dpGrayDialog] := CGrayDialog;
  Options := Options or ofVersion20;
  if ViMode then
    Options := Options or ofViInternal;
  GrowMode := 0;
  Flags := wfMove + wfClose;
  WindowPalette := dpGrayDialog;
end;

procedure TViDialog.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  case Event.What of
    evKeyDown:
      if Options and ofViInternal <> 0 then
        case Event.KeyCode of
          kbEsc:
            begin
              Event.What := evCommand;
              Event.Command := cmCancel;
              Event.InfoPtr := nil;
              PutEvent(Event);
              ClearEvent(Event);
            end;
          kbEnter:
            begin
              Event.What := evBroadcast;
              Event.Command := cmDefault;
              Event.InfoPtr := nil;
              PutEvent(Event);
              ClearEvent(Event);
            end;
        end;
    evCommand:
      case Event.Command of
        cmOk, cmCancel, cmYes, cmNo, cmAbort, cmRetry, cmIgnore:
          if State and sfModal <> 0 then
          begin
            EndModal(Event.Command);
            ClearEvent(Event);
          end;
      end;
  end;
end;

function TViDialog.Valid(Command: Word): Boolean;
begin
  if Command = cmCancel then
    Valid := True
  else
    Valid := inherited Valid(Command);
end;

procedure TViDialog.SizeLimits(var Min, Max: TPoint);
begin
  inherited SizeLimits(Min, Max);
(* removed because of problems with Zoom
  Max.X := MaxLongint;
  Max.Y := MaxLongint;
*)
end;

function NewSItem(const Str: string; ANext: PSItem): PSItem;
var
  Item: PSItem;
begin
  New(Item);
  Item^.Value := Str;
  Item^.Next := ANext;
  NewSItem := Item;
end;

function HotKey(const S: string): Char;
var
  P: Word;
begin
  HotKey := #0;
  if S = '' then
    Exit;
  P := Pos('~', S);
  if P <> 0 then HotKey := UpCase(S[P + 1]);
end;

{ TInputLine }

constructor TViInputLine.Create(var Bounds: TRect; AMaxLen: Integer);
begin
  inherited Create(Bounds);
  FAutoSelect := True;
  FHideSelection := True;
  Palette := CInputLine;
  State := State or sfCursorVis;
  Options := Options or (ofSelectable + ofFirstClick + ofVersion20);
  MaxLen := AMaxLen;
end;

function TViInputLine.CanScroll(Delta: Integer): Boolean;
begin
  if Delta < 0 then
    CanScroll := FirstPos > 0
  else if Delta > 0 then
    CanScroll := Length(Data) - FirstPos + 2 > Size.X
  else
    CanScroll := False;
end;

procedure TViInputLine.Draw;
var
  Color: Byte;
  L, R: Integer;
  B: TDrawBuffer;
  S: string;
begin
  if State and sfFocused = 0 then
    Color := GetColor(1)
  else
    Color := GetColor(2);
  MoveChar(B, ' ', Color, Size.X);

  S := Copy(Data, FirstPos + 1, Size.X - 2);
  if EchoMode = emNormal then
    MoveStr(B[1], S, Color)
  else if EchoMode = emPassword then
    MoveChar(B[1], '*', Color, Length(S));

  if EchoMode <> emNone then
  begin
    if CanScroll(1) then MoveChar(B[Size.X - 1], ldRightScroll, GetColor(4), 1);
    if (State and sfFocused <> 0) or not FHideSelection then
    begin
      if CanScroll(-1) then MoveChar(B[0], ldLeftScroll, GetColor(4), 1);
      L := SelStart - FirstPos;
      R := SelEnd - FirstPos;
      if L < 0 then L := 0;
      if R > Size.X - 2 then R := Size.X - 2;
      if L < R then MoveChar(B[L + 1], #0, GetColor(3), R - L);
    end;
  end;

  WriteLine(0, 0, Size.X, Size.Y, B);
  if EchoMode = emNone then
    SetCursor(1, 0)
  else
    SetCursor(CurPos - FirstPos + 1, 0);
end;

procedure TViInputLine.HandleEvent(var Event: TEvent);
const
  PadKeys = [$47, $4B, $4D, $4F, $73, $74];
var
  Delta, Anchor, I: Integer;
  ExtendBlock: Boolean;
  OldData: string;
  OldCurPos, OldFirstPos,
    OldSelStart, OldSelEnd: Integer;
  WasAppending: Boolean;

  function MouseDelta: Integer;
  var
    Mouse: TPoint;
  begin
    MakeLocal(Event.Where, Mouse);
    if Mouse.X <= 0 then
      MouseDelta := -1
    else if Mouse.X >= Size.X - 1 then
      MouseDelta := 1
    else
      MouseDelta := 0;
  end;

  function MousePos: Integer;
  var
    Pos: Integer;
    Mouse: TPoint;
  begin
    MakeLocal(Event.Where, Mouse);
    if Mouse.X < 1 then Mouse.X := 1;
    Pos := Mouse.X + FirstPos - 1;
    if Pos < 0 then Pos := 0;
    if Pos > Length(Data) then Pos := Length(Data);
    MousePos := Pos;
  end;

  procedure DeleteSelect;
  begin
    if SelStart <> SelEnd then
    begin
      DeleteData(SelStart + 1, SelEnd - SelStart);
      CurPos := SelStart;
    end;
  end;

  procedure AdjustSelectBlock;
  begin
    if CurPos < Anchor then
    begin
      SelStart := CurPos;
      SelEnd := Anchor;
    end
    else
    begin
      SelStart := Anchor;
      SelEnd := CurPos;
    end;
  end;

  procedure SaveState;
  begin
(*
  if Validator <> nil then
*)
    begin
      OldData := Data;
      OldCurPos := CurPos;
      OldFirstPos := FirstPos;
      OldSelStart := SelStart;
      OldSelEnd := SelEnd;
      WasAppending := Length(Data) = CurPos;
    end;
  end;

  procedure RestoreState;
  begin
(*
  if Validator <> nil then
*)
    begin
      Data := OldData;
      CurPos := OldCurPos;
      FirstPos := OldFirstPos;
      SelStart := OldSelStart;
      SelEnd := OldSelEnd;
    end;
  end;

  function CheckValid(NoAutoFill: Boolean): Boolean;
(*
  var
    OldLen: Integer;
    NewData: string;
*)
  begin
(*
  if Validator <> nil then
  begin
    CheckValid := False;
    OldLen := Length(Data^);
    if (Validator^.Options and voOnAppend = 0) or
      (WasAppending and (CurPos = OldLen)) then
    begin
      NewData := Data^;
      if not Validator^.IsValidInput(NewData, NoAutoFill) then
        RestoreState
      else
      begin
        if Length(NewData) > MaxLen then NewData[0] := Char(MaxLen);
        Data^ := NewData;
        if (CurPos >= OldLen) and (Length(Data^) > OldLen) then
          CurPos := Length(Data^);
        CheckValid := True;
      end;
    end
    else
    begin
      CheckValid := True;
      if CurPos = OldLen then
        if not Validator^.IsValidInput(Data^, False) then
        begin
          Validator^.Error;
          CheckValid := False;
        end;
    end;
  end
  else
*)
    CheckValid := True;
  end;

var
  S: string;
begin
  if State and sfSelected <> 0 then
  begin
    case Event.What of
      evCommand, evMouseUp, evMouseMove:
        inherited HandleEvent(Event);
      evMouseDown:
        begin
          Delta := MouseDelta;
          if CanScroll(Delta) then
          begin
            repeat
              inherited HandleEvent(Event);
              if CanScroll(Delta) then
              begin
                Inc(FirstPos, Delta);
                DrawView;
              end;
            until not MouseEvent(Event, evMouseAuto);
            inherited HandleEvent(Event);
          end
          else if Event.Double then
          begin
            SelectAll(True);
            inherited HandleEvent(Event);
          end
          else
          begin
            Anchor := MousePos;
            repeat
              inherited HandleEvent(Event);
              if Event.What = evMouseAuto then
              begin
                Delta := MouseDelta;
                if CanScroll(Delta) then Inc(FirstPos, Delta);
              end;
              CurPos := MousePos;
              AdjustSelectBlock;
              DrawView;
            until not MouseEvent(Event, evMouseMove + evMouseAuto);
            inherited HandleEvent(Event);
          end;
          ClearEvent(Event);
        end;
      evKeyDown:
        begin
          if (Event.KeyCode <> kbLeft) and (Event.KeyCode <> kbRight) then
            inherited HandleEvent(Event);
          if Event.What = evNothing then Exit;
          SaveState;
          Event.KeyCode := CtrlToArrow(Event.KeyCode);
          if (Event.ScanCode in PadKeys) and
            (GetShiftState and $03 <> 0) then
          begin
            Event.CharCode := #0;
            if CurPos = SelEnd then
              Anchor := SelStart
            else
              Anchor := SelEnd;
            ExtendBlock := True;
          end
          else
            ExtendBlock := False;
          case Event.KeyCode of
            kbLeft:
              if CurPos > 0 then Dec(CurPos);
            kbRight:
              if CurPos < Length(Data) then
              begin
                Inc(CurPos);
                CheckValid(True);
              end;
            kbHome:
              CurPos := 0;
            kbEnd:
              begin
                CurPos := Length(Data);
                CheckValid(True);
              end;
            kbBack:
              if CurPos > 0 then
              begin
                DeleteData(CurPos, 1);
                if not FReadOnly then Dec(CurPos);
                if FirstPos > 0 then Dec(FirstPos);
                CheckValid(True);
              end;
            kbDel:
              begin
                if SelStart = SelEnd then
                  if CurPos < Length(Data) then
                  begin
                    SelStart := CurPos;
                    SelEnd := CurPos + 1;
                  end;
                DeleteSelect;
                CheckValid(True);
              end;
            kbIns:
              SetState(sfCursorIns, State and sfCursorIns = 0);
          else
            case Event.CharCode of
              ' '..#255:
                begin
                  S := Data;
                  if State and sfCursorIns <> 0 then
                    Delete(S, CurPos + 1, 1)
                  else
                  begin
                    // DeleteSelect;
                    if SelStart <> SelEnd then
                    begin
                      Delete(S, SelStart + 1, SelEnd - SelStart);
                      if not FReadOnly then CurPos := SelStart;
                    end;
                  end;
                  if CheckValid(True) then
                  begin
                    if Length(S) < MaxLen then
                    begin
                      if FirstPos > CurPos then FirstPos := CurPos;
                      if not FReadOnly then Inc(CurPos);
                      Insert(Event.CharCode, S, CurPos);
                    end;
                    CheckValid(False);
                  end;
                  Data := S;
                end;
              ^Y:
                begin
                  Data := '';
                  if not FReadOnly then CurPos := 0;
                end;
            else
              Exit;
            end
          end;
          if ExtendBlock then
            AdjustSelectBlock
          else
          begin
            SelStart := CurPos;
            SelEnd := CurPos;
          end;
          if FirstPos > CurPos then FirstPos := CurPos;
          I := CurPos - Size.X + 2;
          if FirstPos < I then FirstPos := I;
          DrawView;
          ClearEvent(Event);
        end;
    end;
  end
  else inherited HandleEvent(Event);
end;

procedure TViInputLine.SelectAll(Enable: Boolean);
begin
  CurPos := 0;
  FirstPos := 0;
  SelStart := 0;
  if Enable then
    SelEnd := Length(Data)
  else
    SelEnd := 0;
  DrawView;
end;

procedure TViInputLine.SetState(AState: Word; Enable: Boolean);
begin
  inherited SetState(AState, Enable);
  if (AState = sfSelected) or ((AState = sfActive) and
    (State and sfSelected <> 0)) then
  begin
    if Enable and AutoSelect then SelectAll(Enable);
  end
  else if AState = sfFocused then
    DrawView;
end;

function TViInputLine.Valid(Command: Word): Boolean;

(*
  function AppendError(Validator: PValidator): Boolean;
  begin
    AppendError := False;
    with Validator^ do
      if (Options and voOnAppend <> 0) and (CurPos <> Length(Data^))
          and not IsValidInput(Data^, True) then
      begin
        Error;
        AppendError := True;
      end;
  end;
*)

begin
  Valid := inherited Valid(Command);
(*
  if (Validator <> nil) and (State and sfDisabled = 0) then
    if Command = cmValid then
      Valid := Validator^.Status = vsOk
    else if Command <> cmCancel then
      if AppendError(Validator) or not Validator^.Valid(Data^) then
      begin
        Select;
        Valid := False;
      end;
*)
end;

procedure TViInputLine.DoTextChanged;
begin
  if Assigned(OnTextChanged) then OnTextChanged(Self);
end;

procedure TViInputLine.SetData(const Value: string);
begin
  if not FReadOnly and (FData <> Value) then
  begin
    FData := Value;
    DoTextChanged;
  end;
end;

procedure TViInputLine.InsertData(const Source: string; Index: Integer);
var
  S: string;
begin
  S := Data;
  Insert(Source, S, Index);
  Data := S;
end;

procedure TViInputLine.DeleteData(Index, Count: Integer);
var
  S: string;
begin
  S := Data;
  Delete(S, Index, Count);
  Data := S;
end;

function TViInputLine.GetText: string;
begin
  Result := FData;
end;

procedure TViInputLine.SetText(const Value: string);
begin
  FData := Value;
end;


{ TButton }

constructor TViButton.Create(var Bounds: TRect; ATitle: TTitleStr; ACommand: 
    Word; AFlags: Word; ViMode: Boolean = False);
begin
  inherited Create(Bounds);
  Palette := CButton;
  Options := Options or (ofSelectable + ofFirstClick +
    ofPreProcess + ofPostProcess);
  if ViMode then
    Options := Options or ofViInternal;
  EventMask := EventMask or evBroadcast;
  if not CommandEnabled(ACommand) then State := State or sfDisabled;
  Flags := AFlags;
  if AFlags and bfDefault <> 0 then
    AmDefault := True
  else
    AmDefault := False;
  Title := ATitle;
  Command := ACommand;
  Down := False;
end;

procedure TViButton.Draw;
begin
  DrawState(Down);
end;

procedure TViButton.DrawState(Down: Boolean);
var
  CButton, CShadow: Word;
  Ch: Char;
  I, S, Y, T: Integer;
  B: TDrawBuffer;

  procedure DrawTitle;
  var
    L, SCOff: Integer;
  begin
    if Flags and bfLeftJust <> 0 then
      L := 1
    else
    begin
      L := (S - CStrLen(Title) - 1) div 2;
      if L < 1 then L := 1;
    end;
    MoveCStr(B[I + L], Title, CButton);
    if ShowMarkers and not Down then
    begin
      if State and sfSelected <> 0 then
        SCOff := 0
      else if AmDefault then
        SCOff := 2
      else
        SCOff := 4;
      WordRec(B[0]).Lo := Byte(SpecialChars[SCOff]);
      WordRec(B[S]).Lo := Byte(SpecialChars[SCOff + 1]);
    end;
  end;

begin
  if State and sfDisabled <> 0 then
    CButton := GetColor($0404)
  else
  begin
    CButton := GetColor($0501);
    if State and sfActive <> 0 then
      if State and sfSelected <> 0 then
        CButton := GetColor($0703)
      else if AmDefault then
        CButton := GetColor($0602);
  end;
  CShadow := GetColor(8);
  S := Size.X - 1;
  T := Size.Y div 2 - 1;
  Ch := ' ';
  for Y := 0 to Size.Y - 2 do
  begin
    MoveChar(B, ' ', Byte(CButton), Size.X);
//    WordRec(B[0]).Hi := CShadow;
    Ch := ' ';
    if Down then
    begin
      WordRec(B[0]).Hi := CShadow;
      I := 1;
    end
    else
    begin
      WordRec(B[S]).Hi := Byte(CShadow);
      if not ShowMarkers then
      begin
        if Y = 0 then
          WordRec(B[S]).Lo := Byte(ldBlockBottom)
        else
          WordRec(B[S]).Lo := Byte(ldBlockFull);
        Ch := ldBlockTop;
      end;
      I := 0;
    end;
    if (Y = T) and (Title <> '') then DrawTitle;
    if ShowMarkers and not Down then
    begin
      WordRec(B[1]).Lo := Byte('[');
      WordRec(B[S - 1]).Lo := Byte(']');
    end;
    WriteLine(0, Y, Size.X, 1, B);
  end;
  MoveChar(B[0], ' ', Byte(CShadow), 2);
  MoveChar(B[1], Ch, Byte(CShadow), S);
  WriteLine(0, Size.Y - 1, Size.X, 1, B);
end;

procedure TViButton.HandleEvent(var Event: TEvent);
var
  C: Char;
  Mouse: TPoint;
  ClickRect: TRect;
begin
  GetExtent(ClickRect);
//  Inc(ClickRect.A.X);
  Dec(ClickRect.B.X);
  Dec(ClickRect.B.Y);
  if Event.What = evMouseDown then
  begin
    MakeLocal(Event.Where, Mouse);
    if not RectContains(ClickRect, Mouse) then ClearEvent(Event);
  end;
  if Flags and bfGrabFocus <> 0 then
    inherited HandleEvent(Event);
  case Event.What of
    evMouseMove:
      inherited HandleEvent(Event);
    evMouseDown:
      begin
        if State and sfDisabled = 0 then
        begin
          Inc(ClickRect.B.X);
          Down := False;
          repeat
            inherited HandleEvent(Event);
            MakeLocal(Event.Where, Mouse);
            if Down <> RectContains(ClickRect, Mouse) then
            begin
              Down := not Down;
              DrawState(Down);
            end;
          until not MouseEvent(Event, evMouseMove);
          if Down then
          begin
            Press;
            Down := False;
            DrawState(Down);
          end;
          inherited HandleEvent(Event);
        end;
        ClearEvent(Event);
      end;
    evKeyDown:
    begin
      inherited HandleEvent(Event);
      if Event.What = evNothing then Exit;
      if Title <> '' then
      begin
        C := HotKey(Title);
        if Options and ofViInternal <> 0 then
        begin
          if (Event.KeyCode = GetAltCode(C)) or
              (Owner.Phase = phPostProcess) and (C <> #0) and
              (Upcase(Event.CharCode) = C) or
              (State and sfFocused <> 0)
              and ((Event.KeyCode = kbEnter) or (Event.CharCode = ' ')) then
          begin
            Press;
            ClearEvent(Event);
          end;
        end
        else
        begin
          if (State and sfFocused <> 0) and (Event.CharCode = ' ') then
          begin
            Press;
            { Simulate a KeyUp event for space key on buttons.
              This behaviour is different to Windows CLX where
              the event fires only once when the space key fires
              a keyup event. Unfortunately we don't have any keyup
              events with DOS and WDosX :-(
            }
            Event.What := evKeyUp;
            inherited HandleEvent(Event);
          end;
        end;
      end;
      if Options and ofViInternal = 0 then
        ClearEvent(Event);
    end;
    evBroadcast:
      case Event.Command of
        cmDefault:
          if AmDefault and (State and sfDisabled = 0) then
          begin
            Press;
            ClearEvent(Event);
          end;
        cmGrabDefault, cmReleaseDefault:
          if Flags and bfDefault <> 0 then
          begin
            AmDefault := Event.Command = cmReleaseDefault;
            DrawView;
          end;
        cmCommandSetChanged:
          begin
//            SetState(sfDisabled, not CommandEnabled(Command));
//            DrawView;
          end;
      end;
    evCommand:
      inherited HandleEvent(Event);
  end;
end;

procedure TViButton.MakeDefault(Enable: Boolean);
var
  C: Word;
begin
  if Flags and bfDefault = 0 then
  begin
    if Enable then
      C := cmGrabDefault
    else
      C := cmReleaseDefault;
    SendMessage(Owner, evBroadcast, C, Self);
    AmDefault := Enable;
    DrawView;
  end;
end;

procedure TViButton.Press;
var
  E: TEvent;
begin
  SendMessage(Owner, evBroadcast, cmRecordHistory, nil);
  if Flags and bfBroadcast <> 0 then
    SendMessage(Owner, evBroadcast, Command, Self)
  else
  begin
    E.What := evCommand;
    E.Command := Command;
    E.InfoPtr := Self;
    PutEvent(E);
  end;
end;

procedure TViButton.SetState(AState: Word; Enable: Boolean);
begin
  inherited SetState(AState, Enable);
  if AState and (sfFocused + sfSelected + sfActive) <> 0 then DrawView;
//  if AState and sfFocused <> 0 then MakeDefault(Enable);
end;

{ TCluster }

constructor TViCluster.Create(var Bounds: TRect; AStrings: TStrings);
begin
  inherited Create(Bounds);
  Palette := CCluster;
  Options := Options or (ofSelectable + ofFirstClick + ofPreProcess +
    ofPostProcess + ofVersion20);
  Strings := TStringList.Create;
  Strings.Assign(AStrings);
  Value := 0;
  Sel := 0;
  SetCursor(2, 0);
//  ShowCursor;
  EnableMask := $FFFFFFFF;
end;

destructor TViCluster.Destroy;
begin
  Strings.Destroy;
  inherited Destroy;
end;

function TViCluster.ButtonState(Item: Integer): Boolean;
begin
  if Item > 31
    then
    ButtonState := False
  else
    ButtonState := ((1 shl Item) and EnableMask) <> 0;
end;

procedure TViCluster.DrawBox(const Icon: string; Marker: Char);
begin
  DrawMultiBox(Icon, ' ' + Marker);
end;

procedure TViCluster.DrawMultiBox(const Icon, Marker: string);
var
  I, J, Cur, Col: Integer;
  CNorm, CSel, CDis, Color: Word;
  B: TDrawBuffer;
begin
  CNorm := GetColor($0301);
  CSel := GetColor($0402);
  CDis := GetColor($0505);
  for I := 0 to Size.Y do
  begin
    MoveChar(B, ' ', Byte(CNorm), Size.X);
    for J := 0 to (Strings.Count - 1) div Size.Y + 1 do
    begin
      Cur := J * Size.Y + I;
      if Cur < Strings.Count then
      begin
        Col := Column(Cur);
        if (Col + CStrLen(Strings[Cur]) + 5 <
          Sizeof(TDrawBuffer) div Sizeof(Word)) and (Col < Size.X) then
        begin
          if not ButtonState(Cur) then
            Color := CDis
          else if (Cur = Sel) and (State and sfFocused <> 0) then
            Color := CSel
          else
            Color := CNorm;
          MoveChar(B[Col], ' ', Byte(Color), Size.X - Col);
          MoveStr(B[Col], Icon, Byte(Color));
          WordRec(B[Col + 2]).Lo := Byte(Marker[MultiMark(Cur) + 1]);
          MoveCStr(B[Col + 5], Strings[Cur], Color);
          if ShowMarkers and (State and sfFocused <> 0) and (Cur = Sel) then
          begin
            WordRec(B[Col]).Lo := Byte(SpecialChars[0]);
            WordRec(B[Column(Cur + Size.Y) - 1]).Lo := Byte(SpecialChars[1]);
          end;
        end;
      end;
    end;
    WriteBuf(0, I, Size.X, 1, B);
  end;
  SetCursor(Column(Sel) + 2, Row(Sel));
end;

function TViCluster.GetHelpCtx: Word;
begin
  if HelpCtx = hcNoContext then
    GetHelpCtx := hcNoContext
  else
    GetHelpCtx := HelpCtx + Sel;
end;

procedure TViCluster.HandleEvent(var Event: TEvent);
var
  Mouse: TPoint;
  I, S: Integer;
//  C: Char;

  procedure MoveSel;
  begin
    if I <= Strings.Count then
    begin
      Sel := S;
      MovedTo(Sel);
      DrawView;
    end;
  end;

begin
  if (Options and ofSelectable) = 0 then
  begin
    inherited HandleEvent(Event);
    Exit;
  end;
  if Event.What = evMouseMove then
    inherited HandleEvent(Event)
  else if Event.What = evCommand then
    inherited HandleEvent(Event)
  else if Event.What = evMouseDown then
  begin
    MakeLocal(Event.Where, Mouse);
    I := FindSel(Mouse);
    if I <> -1 then
      if ButtonState(I) then Sel := I;
    DrawView;
    repeat
      inherited HandleEvent(Event);
//      MakeLocal(Event.Where, Mouse);
//      if FindSel(Mouse) = Sel then
//        ShowCursor
//      else
//        HideCursor;
    until not MouseEvent(Event, evMouseMove); {Wait for mouse up}
//    ShowCursor;
    MakeLocal(Event.Where, Mouse);
    if (FindSel(Mouse) = Sel) and ButtonState(Sel) then
    begin
      Press(Sel);
      DoClicked;
      DrawView;
    end;
    inherited HandleEvent(Event);
    ClearEvent(Event);
  end
  else if Event.What = evKeyDown then
  begin
    inherited HandleEvent(Event);
    if Event.What = evNothing then Exit;
    S := Sel;

    case CtrlToArrow(Event.KeyCode) of
      kbUp:;
{
        if State and sfFocused <> 0 then
        begin
          I := 0;
          repeat
            Inc(I);
            Dec(S);
            if S < 0 then S := Strings.Count - 1;
          until ButtonState(S) or (I > Strings.Count);
          MoveSel;
          ClearEvent(Event);
        end;
      kbDown:
        if State and sfFocused <> 0 then
        begin
          I := 0;
          repeat
            Inc(I);
            Inc(S);
            if S >= Strings.Count then S := 0;
          until ButtonState(S) or (I > Strings.Count);
          MoveSel;
          ClearEvent(Event);
        end;
      kbRight:
        if State and sfFocused <> 0 then
        begin
          I := 0;
          repeat
            Inc(I);
            Inc(S, Size.Y);
            if S >= Strings.Count then
            begin
              S := (S + 1) mod Size.Y;
              if S >= Strings.Count then S := 0;
            end;
          until ButtonState(S) or (I > Strings.Count);
          MoveSel;
          ClearEvent(Event);
        end;
      kbLeft:
        if State and sfFocused <> 0 then
        begin
          I := 0;
          repeat
            Inc(I);
            if S > 0 then
            begin
              Dec(S, Size.Y);
              if S < 0 then
              begin
                S := ((Strings.Count + Size.Y - 1) div Size.Y) * Size.Y + S - 1;
                if S >= Strings.Count then S := Strings.Count - 1;
              end;
            end
            else
              S := Strings.Count - 1;
          until ButtonState(S) or (I > Strings.Count);
          MoveSel;
          ClearEvent(Event);
        end;
}
    else
      begin
//        for I := 0 to Strings.Count - 1 do
//        begin
//          C := HotKey(Strings[I]);
//          if (GetAltCode(C) = Event.KeyCode) or
//            (((Owner.Phase = phPostProcess) or (State and sfFocused <> 0))
//            and (C <> #0) and (UpCase(Event.CharCode) = C)) then
//          begin
//            if ButtonState(I) then
//            begin
//              if Focus then
//              begin
//                Sel := I;
//                MovedTo(Sel);
//                Press(Sel);
//                DrawView;
//              end;
//              ClearEvent(Event);
//            end;
//            Exit;
//          end;
//        end;
        if (Event.CharCode = ' ') and (State and sfFocused <> 0)
          and ButtonState(Sel) then
        begin
          Press(Sel);
          DoClicked;
          DrawView;
          ClearEvent(Event);
        end;
      end
    end
  end;
end;

procedure TViCluster.SetButtonState(AMask: LongWord; Enable: Boolean);
var
  I, M: Longint;
begin
  if Enable then
    EnableMask := EnableMask or AMask
  else
    EnableMask := EnableMask and not AMask;
  if Strings.Count <= 32 then
  begin
    M := 1;
    for I := 1 to Strings.Count do
    begin
      if (M and EnableMask) <> 0 then
      begin
        Options := Options or ofSelectable;
        Exit;
      end;
      M := M shl 1;
    end;
    Options := Options and not ofSelectable;
  end;
end;

procedure TViCluster.SetState(AState: Word; Enable: Boolean);
begin
  inherited SetState(AState, Enable);
  if AState = sfFocused then DrawView;
end;

function TViCluster.Mark(Item: Integer): Boolean;
begin
  Mark := False;
end;

function TViCluster.MultiMark(Item: Integer): Byte;
begin
  MultiMark := Byte(Mark(Item) = True);
end;

procedure TViCluster.MovedTo(Item: Integer);
begin
end;

procedure TViCluster.Press(Item: Integer);
begin
end;

function TViCluster.Column(Item: Integer): Integer;
var
  I, Col, Width, L: Integer;
begin
  if Item < Size.Y then
    Column := 0
  else
  begin
    Width := 0;
    Col := -6;
    for I := 0 to Item do
    begin
      if I mod Size.Y = 0 then
      begin
        Inc(Col, Width + 6);
        Width := 0;
      end;
      if I < Strings.Count then
        L := CStrLen(Strings[I])
      else L := 0;
      if L > Width then Width := L;
    end;
    Column := Col;
  end;
end;

function TViCluster.FindSel(P: TPoint): Integer;
var
  I, S: Integer;
  R: TRect;
begin
  GetExtent(R);
  if not RectContains(R, P) then
    FindSel := -1
  else
  begin
    I := 0;
    while P.X >= Column(I + Size.Y) do
      Inc(I, Size.Y);
    S := I + P.Y;
    if S >= Strings.Count then
      FindSel := -1
    else
      FindSel := S;
  end;
end;

function TViCluster.Row(Item: Integer): Integer;
begin
  Row := Item mod Size.Y;
end;

procedure TViCluster.DoClicked;
begin
  if Assigned(OnClicked) then OnClicked(Self);
end;


{ TRadioButtons }

procedure TViRadioButtons.Draw;
const
  Button = ' ( ) ';
begin
  DrawMultiBox(Button, #32#7);
end;

function TViRadioButtons.Mark(Item: Integer): Boolean;
begin
  Mark := Item = Value;
end;

procedure TViRadioButtons.Press(Item: Integer);
begin
  Value := Item;
end;

procedure TViRadioButtons.MovedTo(Item: Integer);
begin
  Value := Item;
end;

{ TCheckBoxes }

procedure TViCheckBoxes.Draw;
const
  Button = ' [ ] ';
begin
  DrawMultiBox(Button, ' X');
end;

function TViCheckBoxes.Mark(Item: Integer): Boolean;
begin
  Mark := Value and (1 shl Item) <> 0;
end;

procedure TViCheckBoxes.Press(Item: Integer);
begin
  Value := Value xor (1 shl Item);
end;

{ TMultiCheckBoxes }

constructor TViMultiCheckBoxes.Create(var Bounds: TRect; AStrings:
  TStrings; ASelRange: Byte; AFlags: Word; const AStates: string);
begin
  inherited Create(Bounds, AStrings);
  SelRange := ASelRange;
  Flags := AFlags;
  States := AStates;
end;

procedure TViMultiCheckBoxes.Draw;
const
  Button = ' [ ] ';
begin
  DrawMultiBox(Button, States);
end;

function TViMultiCheckBoxes.MultiMark(Item: Integer): Byte;
begin
  MultiMark := (Value shr (Word(Item) * WordRec(Flags).Hi))
    and WordRec(Flags).Lo;
end;

procedure TViMultiCheckBoxes.Press(Item: Integer);
var
  CurState: ShortInt;
begin
  CurState := (Value shr (Word(Item) * WordRec(Flags).Hi))
    and WordRec(Flags).Lo;

  Dec(CurState);
  if (CurState >= SelRange) or (CurState < 0) then
    CurState := SelRange - 1;
  Value := (Value and not (Longint(WordRec(Flags).Lo)
    shl (Word(Item) * WordRec(Flags).Hi))) or
    (Longint(CurState) shl (Word(Item) * WordRec(Flags).Hi));
end;

{ TListBox }

constructor TViListBox.Create(var Bounds: TRect; ANumCols: Word;
  AScrollBar: TViScrollBar);
begin
  inherited Create(Bounds, ANumCols, nil, AScrollBar);
  List := TStringList.Create;
  SetRange(0);
end;

function TViListBox.GetText(Item: Integer; MaxLen: Integer): string;
begin
  Result := '';
  if Item < List.Count then Result := List[Item];
end;

procedure TViListBox.AssignList(AList: TStrings);
begin
  List.Assign(AList);
  SetRange(List.Count);
  if Range > 0 then FocusItem(0);
  DrawView;
end;

destructor TViListBox.Destroy;
begin
  List.Free;
  inherited Destroy;
end;


{ TStaticText }

constructor TViStaticText.Create(var Bounds: TRect; const AText: string);
begin
  inherited Create(Bounds);
//  Palette := CStaticText;
  Palette := CLabel;
  Text := AText;
end;

procedure TViStaticText.Draw;
var
  Color: Byte;
  Center: Boolean;
  I, J, L, P, Y: Integer;
  B: TDrawBuffer;
  S: string;
begin
  Color := GetColor(1);
  GetText(S);
  L := Length(S);
  P := 1;
  Y := 0;
  Center := False;
  while Y < Size.Y do
  begin
    MoveChar(B, ' ', Color, Size.X);
    if P <= L then
    begin
      if S[P] = #3 then
      begin
        Center := True;
        Inc(P);
      end;
      I := P;
      repeat
        J := P;
        while (P <= L) and (S[P] = ' ') do
          Inc(P);
        while (P <= L) and (S[P] <> ' ') and (S[P] <> #13) do
          Inc(P);
      until (P > L) or (P >= I + Size.X) or (S[P] = #13);
      if P > I + Size.X then
        if J > I then
          P := J
        else
          P := I + Size.X;
      if Center then
        J := (Size.X - P + I) div 2
      else
        J := 0;
      MoveBuf(B[J], S[I], Color, P - I);
      while (P <= L) and (S[P] = ' ') do
        Inc(P);
      if (P <= L) and (S[P] = #13) then
      begin
        Center := False;
        Inc(P);
        if (P <= L) and (S[P] = #10) then Inc(P);
      end;
    end;
    WriteLine(0, Y, Size.X, 1, B);
    Inc(Y);
  end;
end;

procedure TViStaticText.GetText(var S: string);
begin
  S := Text;
end;

{ TLabel }

constructor TViLabel.Create(var Bounds: TRect; const AText: string; ALink:
  TViView);
begin
  inherited Create(Bounds, AText);
  Palette := CLabel;
  Link := ALink;
  Options := Options or (ofPreProcess + ofPostProcess);
  EventMask := EventMask or evBroadcast;
end;

procedure TViLabel.Draw;
var
  Color: Word;
  B: TDrawBuffer;
begin
  if Light then
    Color := GetColor($0402)
  else
    Color := GetColor($0301);
  MoveChar(B[0], ' ', Byte(Color), Size.X);
  if Text <> '' then MoveCStr(B[0], Text, Color);
  WriteLine(0, 0, Size.X, 1, B);
end;

procedure TViLabel.HandleEvent(var Event: TEvent);
var
  C: Char;

  procedure FocusLink;
  begin
    if (Link <> nil) and (Link.Options and ofSelectable <> 0) then
    begin
      Link.Focus;
      ClearEvent(Event);
    end;
  end;

begin
  inherited HandleEvent(Event);
  { if Event.What = evMouseDown then
    FocusLink
  else } if Event.What = evKeyDown then
  begin
    C := HotKey(Text);
    if (GetAltCode(C) = Event.KeyCode) or
      ((C <> #0) and (Owner.Phase = phPostProcess) and
      (UpCase(Event.CharCode) = C)) then
      FocusLink
  end
  else if Event.What = evBroadcast then
    if ((Event.Command = cmReceivedFocus) or
      (Event.Command = cmReleasedFocus)) and
      (Link <> nil) then
    begin
      {
      Light := Link.State and sfFocused <> 0;
      DrawView;
      }
    end;
end;

{ THistoryViewer }

constructor TViHistoryViewer.Create(var Bounds: TRect; AHScrollBar, 
    AVScrollBar: TViScrollBar; AStrings: TStrings);
begin
  inherited Create(Bounds, 1, AHScrollBar, AVScrollBar);
  Palette := CHistoryViewer;
  FStrings := AStrings;
  SetRange(FStrings.Count);
  if Range > 0 then FocusItem(0);
  HScrollBar.SetRange(1, HistoryWidth - Size.X + 3);
end;

function TViHistoryViewer.GetText(Item: Integer; MaxLen: Integer): string;
begin
  if (Item < 0) or (Item >= FStrings.Count) then
    Result := ''
  else
    Result := FStrings[Item];
end;

procedure TViHistoryViewer.HandleEvent(var Event: TEvent);
begin
  if ((Event.What = evMouseDown) and (Event.Double)) or
    ((Event.What = evKeyDown) and (Event.KeyCode = kbEnter)) then
  begin
    EndModal(cmOk);
    ClearEvent(Event);
  end
  else if ((Event.What = evKeyDown) and (Event.KeyCode = kbEsc)) or
    ((Event.What = evCommand) and (Event.Command = cmCancel)) then
  begin
    EndModal(cmCancel);
    ClearEvent(Event);
  end
  else
    inherited HandleEvent(Event);
end;

function TViHistoryViewer.HistoryWidth: Integer;
var
  Width, Count, I: Integer;
  T: Integer;
begin
  Width := 0;
  Count := FStrings.Count;
  for I := 0 to Count - 1 do
  begin
    T := Length(FStrings[I]);
    if T > Width then Width := T;
  end;
  HistoryWidth := Width;
end;

{ THistoryWindow }

constructor TViHistoryWindow.Create(var Bounds: TRect; Strings: TStrings);
begin
  inherited Create(Bounds, '', wnNoNumber);
  Palette := CHistoryWindow;
  Flags := wfClose;
  InitViewer(Strings);
end;

function TViHistoryWindow.GetSelection: string;
begin
  GetSelection := Viewer.GetText(Viewer.Focused, 255);
end;

procedure TViHistoryWindow.InitViewer(Strings: TStrings);
var
  R: TRect;
begin
  GetExtent(R);
  RectGrow(R, -1, -1);
  Viewer := TViHistoryViewer.Create(R,
    StandardScrollBar(sbHorizontal + sbHandleKeyboard),
    StandardScrollBar(sbVertical + sbHandleKeyboard),
    Strings);
  Viewer.OnChanged := DoChanged;
  Insert(Viewer);
end;

procedure TViHistoryWindow.DoChanged(Sender: TObject);
begin
  if Assigned(OnChanged) then OnChanged(Sender);
end;


{ THistory }

constructor TViHistory.Create(var Bounds: TRect; ALink: TViInputLine; AList: 
    TStrings);
begin
  inherited Create(Bounds);
  Palette := CHistory;
  Options := Options or ofPostProcess;
  EventMask := EventMask or evBroadcast;
  Link := ALink;
  List := AList;
end;

procedure TViHistory.Draw;
var
  B: TDrawBuffer;
begin
  MoveCStr(B, ldHistoryDropDown, GetColor($0102));
  WriteLine(0, 0, Size.X, Size.Y, B);
end;

procedure TViHistory.HandleEvent(var Event: TEvent);
var
  HistoryWindow: TViHistoryWindow;
  R, P: TRect;
  C: Word;
  Rslt: string;
  OwnerWindow: TViGroup;
  OldLinkData: string;
  OldReadOnly: Boolean;
begin
  inherited HandleEvent(Event);
  if (Event.What = evMouseDown) or
    ((Event.What = evKeyDown) and (CtrlToArrow(Event.KeyCode) = kbDown) and
    (Link.State and sfFocused <> 0)) then
  begin
    if not Link.Focus then
    begin
      ClearEvent(Event);
      Exit;
    end;
    Link.GetBounds(R);
    Dec(R.A.X);
    Inc(R.B.X);
    Inc(R.B.Y, 7);
    OwnerWindow := Owner;
    while not (OwnerWindow is TViWindow) do
    begin
      OwnerWindow := OwnerWindow.Owner;
      if OwnerWindow is TViApplication then
      begin
        OwnerWindow := Owner;
        Break;
      end;
    end;
    OwnerWindow.GetExtent(P);
    RectIntersect(R, P);
    RectMove(R, 1, 0);
    HistoryWindow := InitHistoryWindow(R);
    if HistoryWindow <> nil then
    begin
      OldLinkData := Link.Data;
      C := OwnerWindow.ExecView(HistoryWindow);
      if C = cmOk then
      begin
        Rslt := HistoryWindow.GetSelection;
        if Length(Rslt) > Link.MaxLen then
            SetLength(Rslt, Link.MaxLen);
        Link.Data := Rslt;
        Link.SelectAll(True);
        Link.DrawView;
        DoSelected(Self);
      end
      else
      begin
        OldReadOnly := Link.ReadOnly;
        Link.ReadOnly := False;
        Link.Data := OldLinkData;
        Link.ReadOnly := OldReadOnly;
        Link.SelectAll(True);
        Link.DrawView;
      end;
      HistoryWindow.Free;
    end;
    ClearEvent(Event);
  end;
end;

function TViHistory.InitHistoryWindow(var Bounds: TRect): TViHistoryWindow;
begin
  Result := TViHistoryWindow.Create(Bounds, List);
  Result.HelpCtx := Link.HelpCtx;
  Result.OnChanged := DoChanged;
end;

procedure TViHistory.RecordHistory(const S: string);
begin
  if List.IndexOf(S) < 0 then
    List.Add(S);
end;

procedure TViHistory.DoChanged(Sender: TObject);
begin
  if Assigned(OnChanged) then OnChanged(Sender);
end;

procedure TViHistory.DoSelected(Sender: TObject);
begin
  if Assigned(OnSelected) then OnSelected(Sender);
end;

end.

