
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViEditors.pas                                         }
{                                                       }
{ for Delphi 4, 5, 6, 7                                 }
{ Version 2.6.0                                         }
{ Copyright (c) 2003 by Team-DWP                        }
{ e-mail: team-dwp@dwp42.org                            }
{ www: http://www.dwp42.org                             }
{                                                       }
{ The initial developer of the code is                  }
{   Oleg A. Bachin, bachin@cis-kiev.com                 }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViEditors.pas,v 1.5 2004/03/31 20:38:15 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViEditors.pas,v $
  Revision 1.5  2004/03/31 20:38:15  iwache
  Bugfix: FClipBoard was freed twice in finalization section.
  This is fixed.

  Revision 1.4  2004/03/30 20:25:19  iwache
  - TViEditor and TViMemo modified for connection
  with WQ layer.
  - Buffer pointers in TViEditor changed from Word
  to Integer to prepare support for buffer size > 64k Byte.

  Revision 1.3  2003/11/19 18:52:34  iwache
  Some minor changes to make it work in Vi only mode.

  Revision 1.2  2003/11/07 19:20:30  iwache
  Missing type PWord = ^Word for Delphi 5 compatibility added.

  Revision 1.1  2003/11/06 20:28:39  iwache
  For DWPL translated ViEditors unit by Oleg A. Bachin added.
  Thanks to Oleg for this cool translation.

  ----------------------------------------------------------------------------
}

unit ViEditors;

{$I-}

interface

uses
  SysUtils, ViTypes, ViDrivers, ViClasses, ViViews, ViDialogs, ViUtils,
  RTLConsts, Classes;

const
  cmFind = 82;
  cmReplace = 83;
  cmSearchAgain = 84;

const
  cmCharLeft = 500;
  cmCharRight = 501;
  cmWordLeft = 502;
  cmWordRight = 503;
  cmLineStart = 504;
  cmLineEnd = 505;
  cmLineUp = 506;
  cmLineDown = 507;
  cmPageUp = 508;
  cmPageDown = 509;
  cmTextStart = 510;
  cmTextEnd = 511;
  cmNewLine = 512;
  cmBackSpace = 513;
  cmDelChar = 514;
  cmDelWord = 515;
  cmDelStart = 516;
  cmDelEnd = 517;
  cmDelLine = 518;
  cmInsMode = 519;
  cmStartSelect = 520;
  cmHideSelect = 521;
  cmIndentMode = 522;
  cmUpdateTitle = 523;

const
  edOutOfMemory = 0;
  edReadError = 1;
  edWriteError = 2;
  edCreateError = 3;
  edSaveModify = 4;
  edSaveUntitled = 5;
  edSaveAs = 6;
  edFind = 7;
  edSearchFailed = 8;
  edReplace = 9;
  edReplacePrompt = 10;

const
  efCaseSensitive = $0001;
  efWholeWordsOnly = $0002;
  efPromptOnReplace = $0004;
  efReplaceAll = $0008;
  efDoReplace = $0010;
  efBackupFiles = $0100;

const
  CIndicator = #2#3;
  CEditor = #6#7;
  CMemo = #26#27;

const
  MaxLineLength = 4096;
  MinBufLength = $1000;
  MaxBufLength = $7FFFFF00;
  NotFoundValue = - 1;
  LineInfoGrow = 1024;
  MaxLines = $7FFFFFF;

type
  TEditorDialog = function(Dialog: Integer; Info: Pointer): Word;

type
  TViIndicator = class(TViView)
  private
    Location: TPoint;
    Modified: Boolean;
  public
    constructor Create(Bounds: TRect);
    procedure Draw; override;
    function GetPalette: TPalette; override;
    procedure SetState(AState: Word; Enable: Boolean); override;
    procedure SetValue(ALocation: TPoint; AModified: Boolean);
  end;

type
  PEditBuffer = ^TEditBuffer;
  TEditBuffer = array[0..512 * 1024 * 1024 - 1] of Char;

type
  EViMemoError = class(Exception);

type
  TViEditor = class(TViView)
  private
    //fColorer: TColorerClass;
    FLockCount: Byte;
    FUpdateFlags: Byte;
    FKeyState: Integer;
    FOnChanged: TChangedEvent;
    FOnReturnPressed: TNotifyEvent;
    function CharPos(P, Target: Integer): Integer;
    function CharPtr(P: Word; Target: Integer): Integer;
    function ClipCopy: Boolean;
    procedure ClipCut;
    procedure ClipPaste;
    procedure DeleteRange(StartPtr, EndPtr: Integer; DelSelect: Boolean);
    procedure DoUpdate;
    procedure DoSearchReplace;
    procedure DrawLines(Y, Count: Integer; LinePtr: Integer);
    procedure FormatLine(var DrawBuf; LinePtr: Integer; Width: Integer; Colors:
        Word);
    procedure Find;
    function GetMousePtr(Mouse: TPoint): Integer;
    function HasSelection: Boolean;
    procedure HideSelect;
    function IsClipboard: Boolean;
    function LineEnd(P: Integer): Integer;
    function LineMove(P: Integer; Count: Integer): Integer;
    function LineStart(P: Integer): Integer;
    procedure Lock;
    function NextChar(P: Integer): Integer;
    function NextLine(P: Integer): Integer;
    function NextWord(P: Integer): Integer;
    function PrevChar(P: Integer): Integer;
    function PrevLine(P: Integer): Integer;
    function PrevWord(P: Integer): Integer;
    procedure Replace;
    procedure SetBufLen(Length: Integer);
    procedure StartSelect;
    procedure ToggleInsMode;
    procedure Unlock;
    procedure Update(AFlags: Byte);
  protected
    procedure DoChanged;
    procedure DoReturnPressed;
  public
    HScrollBar: TViScrollBar;
    VScrollBar: TViScrollBar;
    Indicator: TViIndicator;
    Buffer: PEditBuffer;
    BufSize: Integer;
    BufLen: Integer;
    GapLen: Integer;
    SelStart: Integer;
    SelEnd: Integer;
    CurPtr: Integer;
    CurPos: TPoint;
    Delta: TPoint;
    Limit: TPoint;
    DrawLine: Integer;
    DrawPtr: Integer;
    DelCount: Integer;
    InsCount: Integer;
    IsValid: Boolean;
    CanUndo: Boolean;
    Modified: Boolean;
    Selecting: Boolean;
    Overwrite: Boolean;
    AutoIndent: Boolean;
    TabSize: Integer;
    ReadOnly: Boolean;
    constructor Create(Bounds: TRect; AHScrollBar, AVScrollBar: TViScrollBar;
        AIndicator: TViIndicator; ViMode: Boolean = False);
    destructor Destroy; override;
    {
        constructor Load(var S: TStream);
    {}
    function BufChar(P: Integer): Char;
    function BufPtr(P: Integer): Word;
    procedure ChangeBounds(var Bounds: TRect); override;
    procedure ConvertEvent(var Event: TEvent); virtual;
    function CursorVisible: Boolean;
    procedure DeleteSelect;
    procedure DoneBuffer; virtual;
    procedure Draw; override;
    function GetPalette: TPalette; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure InitBuffer; virtual;
    function InsertBuffer(var P: PEditBuffer; Offset, Length: Integer; AllowUndo,
        SelectText: Boolean): Boolean;
    function InsertFrom(Editor: TViEditor): Boolean; virtual;
    function InsertText(Text: Pointer; Length: Integer; SelectText: Boolean):
        Boolean;
    procedure ScrollTo(X, Y: Integer);
    function Search(const FindStr: string; Opts: Word): Boolean;
    function SetBufSize(NewSize: Integer): Boolean; virtual;
    procedure SetCmdState(Command: Word; Enable: Boolean);
    procedure SetSelect(NewStart, NewEnd: Integer; CurStart: Boolean);
    procedure SetState(AState: Word; Enable: Boolean); override;

    procedure SetCurPtr(P: Integer; SelectMode: Byte);
    procedure NewLine;

    {TODO    procedure Store(var S: TStream);
    {}
    procedure TrackCursor(Center: Boolean);
    procedure Undo;
    procedure UpdateCommands; virtual;
    function Valid(Command: Word): Boolean; override;
    property OnChanged: TChangedEvent read FOnChanged write FOnChanged;
    property OnReturnPressed: TNotifyEvent read FOnReturnPressed write
        FOnReturnPressed;
  end;

type
  TMemoData = record
    Length: Word;
    Buffer: TEditBuffer;
  end;

type
  TViMemo = class(TViEditor)
  protected
    procedure Error(const Msg: string; Data: Integer); overload;
    procedure Error(Msg: PResStringRec; Data: Integer); overload;
  public
    constructor Create(Bounds: TRect; AHScrollBar, AVScrollBar: TViScrollBar;
        ViMode: Boolean = False);
    function GetPalette: TPalette; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure Clear;
    function NumLines: Integer;
    procedure RemoveLine(Index: Integer);
    function TextLine(Index: Integer): string;
    procedure InsertLine(Index: Integer; const Line: string);
    function IndexLineStart(Index: Integer): Integer;
    procedure Append(const Value: string);
    function GetText: string;
    procedure SetText(const Value: string);
    procedure Deselect;
    function HasMarkedText: Boolean;
    procedure Insert(const Value: string; Mark: Boolean);
    procedure Copy;
    procedure Cut;
    procedure Paste;
    procedure GetMarkedRegion(var Line1, Column1, Line2, Column2: Integer);
    procedure SetEdited(const Value: Boolean); virtual;
    function MarkedText: string;
    procedure SetCursorPosition(Line, Column: Integer);
    procedure InsertAt(const Text: string; Column, Row: Integer; Mark: Boolean = 
        False); overload;
  end;

type
  TFileEditor = class(TViEditor)
  private
    FStrings: TStrings;
  public
    FileName: string;
    constructor Create(Bounds: TRect; AHScrollBar, AVScrollBar: TViScrollBar;
        AIndicator: TViIndicator; ViMode: Boolean = False);
    destructor Destroy; override;

    {TODO    constructor Create(Bounds: TRect);
    {}
    {TODO    constructor Load(var S: TStream);
    {}
    procedure DoneBuffer; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure InitBuffer; override;
    function LoadFile: Boolean;
    function Save: Boolean;
    function SaveAs: Boolean;
    function SaveFile: Boolean;
    function SetBufSize(NewSize: Integer): Boolean; override;
    {TODO    procedure Store(var S: TStream);
    {}
    procedure UpdateCommands; override;
    function Valid(Command: Word): Boolean; override;
  end;

type
  TViEditWindow = class(TViWindow)
    Editor: TFileEditor;
    HScrollBar, VScrollBar: TViScrollBar;
    Indicator: TViIndicator;

    //Scroller: TViScroller;
    constructor Create(Bounds: TRect;
      AFileName: string; ANumber: Integer);
    {TODO    constructor Load(var S: TStream);
    {}
    procedure Close; override;
    function GetTitle: TTitleStr; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure SizeLimits(var Min, Max: TPoint); override;
    {TODO    procedure Store(var S: TStream);
    {}
  public
    destructor Destroy; override;
  end;

function DefEditorDialog(Dialog: Integer; Info: Pointer): Word;
function CreateFindDialog: TViDialog;
function CreateReplaceDialog: TViDialog;
function StdEditorDialog(Dialog: Integer; Info: Pointer): Word;

var
  WordChars: set of Char = ['0'..'9', 'A'..'Z', '_', 'a'..'z'];
  EditorFlags: Word = efBackupFiles + efPromptOnReplace;
  FindStr: string[80] = '';
  ReplaceStr: string[80] = '';
var
  EditorDialog: TEditorDialog = DefEditorDialog;

function Clipboard: TViEditWindow;

type
  TFindDialogRec = record
    Find: string[80];
    Options: Word;
  end;

type
  TReplaceDialogRec = record
    Find: string[80];
    Replace: string[80];
    Options: Word;
  end;

implementation

uses
  ViApp, ViScreens, ViStdDlg, ViMsgBox, ViEventsClass;

type 
  PWord = ^Word;

const
  ufUpdate = $01;
  ufLine = $02;
  ufView = $04;

const
  smExtend = $01;
  smDouble = $02;

const
  sfSearchFailed = -1;

const
  FirstKeys: array[0..37 * 2] of Word = (37,
    Ord(^A), cmWordLeft, Ord(^C), cmPageDown,
    Ord(^D), cmCharRight, Ord(^E), cmLineUp,
    Ord(^F), cmWordRight, Ord(^G), cmDelChar,
    Ord(^H), cmBackSpace, Ord(^K), $FF02,
    Ord(^L), cmSearchAgain, Ord(^M), cmNewLine,
    Ord(^O), cmIndentMode, Ord(^Q), $FF01,
    Ord(^R), cmPageUp, Ord(^S), cmCharLeft,
    Ord(^T), cmDelWord, Ord(^U), cmUndo,
    Ord(^V), cmInsMode, Ord(^X), cmLineDown,
    Ord(^Y), cmDelLine, kbLeft, cmCharLeft,
    kbRight, cmCharRight, kbCtrlLeft, cmWordLeft,
    kbCtrlRight, cmWordRight, kbHome, cmLineStart,
    kbEnd, cmLineEnd, kbUp, cmLineUp,
    kbDown, cmLineDown, kbPgUp, cmPageUp,
    kbPgDn, cmPageDown, kbCtrlPgUp, cmTextStart,
    kbCtrlPgDn, cmTextEnd, kbIns, cmInsMode,
    kbDel, cmDelChar, kbShiftIns, cmPaste,
    kbShiftDel, cmCut, kbCtrlIns, cmCopy,
    kbCtrlDel, cmClear);
  QuickKeys: array[0..8 * 2] of Word = (8,
    Ord('A'), cmReplace, Ord('C'), cmTextEnd,
    Ord('D'), cmLineEnd, Ord('F'), cmFind,
    Ord('H'), cmDelStart, Ord('R'), cmTextStart,
    Ord('S'), cmLineStart, Ord('Y'), cmDelEnd);
  BlockKeys: array[0..5 * 2] of Word = (5,
    Ord('B'), cmStartSelect, Ord('C'), cmPaste,
    Ord('H'), cmHideSelect, Ord('K'), cmCopy,
    Ord('Y'), cmCut);
  KeyMap: array[0..2] of Pointer = (@FirstKeys, @QuickKeys, @BlockKeys);

function ViRect(AX, AY, BX, BY: LongInt): TRect;
begin
  with Result do
  begin
    A.X := AX;
    A.Y := AY;
    B.X := BX;
    B.Y := BY;
  end;
end;

procedure FileSplit(const Path: string; var Dir: string; var Name: string; var
  Ext: string);
var
  I, NamePos, ExtPos: Integer;
begin
  NamePos := 0;
  ExtPos := 256;
  for I := 1 to Length(Path) do
    case Path[I] of
      ':', {$IFDEF LINUX} '/'{$ELSE} '\'{$ENDIF}:
        begin
          NamePos := I;
          ExtPos := 256;
        end;
      '.': ExtPos := I;
    end;
  Dir := Copy(Path, 1, NamePos);
  Name := Copy(Path, NamePos + 1, ExtPos - NamePos - 1);
  Ext := Copy(Path, ExtPos, 255);
end;

function DefEditorDialog(Dialog: Integer; Info: Pointer): Word;
begin
  DefEditorDialog := cmCancel;
end;

function CreateFindDialog: TViDialog;
var
  D: TViDialog;
  Control: TViView;
  R: TRect;
  cbItems: TStrings;
begin
  R := ViRect(0, 0, 38, 12);
  D := TViDialog.Create(R, 'Find');
  with D do
  begin
    Options := Options or ofCentered;
    R := ViRect(3, 3, 32, 4);
    Control := TViInputLine.Create(R, 80);
    Insert(Control);
    R := ViRect(2, 2, 15, 3);
    Insert(TViLabel.Create(R, '~T~ext to find', Control));
    R := ViRect(32, 3, 35, 4);
    // todo    Insert(TViHistory.Create(R, TViInputLine(Control), 10));

    R := ViRect(3, 5, 35, 7);
    cbItems := TStringList.Create;
    with cbItems do
    begin
      Add('~C~ase sensitive');
      Add('~W~hole words only');
    end;
    Insert(TViCheckBoxes.Create(R, cbItems));

    R := ViRect(14, 9, 24, 11);
    Insert(TViButton.Create(R, 'O~K~', cmOk, bfDefault));
    Inc(R.A.X, 12);
    Inc(R.B.X, 12);
    Insert(TViButton.Create(R, 'Cancel', cmCancel, bfNormal));

    SelectNext(False);
  end;
  CreateFindDialog := D;
end;

function CreateReplaceDialog: TViDialog;
var
  D: TViDialog;
  Control: TViView;
  R: TRect;
  cbItems: TStrings;
begin
  R := ViRect(0, 0, 40, 16);
  D := TViDialog.Create(R, 'Replace');
  with D do
  begin
    Options := Options or ofCentered;

    R := ViRect(3, 3, 34, 4);
    Control := TViInputLine.Create(R, 80);
    Insert(Control);
    R := ViRect(2, 2, 15, 3);
    Insert(TViLabel.Create(R, '~T~ext to find', Control));
    R := ViRect(34, 3, 37, 4);
    Insert(TViHistory.Create(R, TViInputLine(Control), nil));

    R := ViRect(3, 6, 34, 7);
    Control := TViInputLine.Create(R, 80);
    Insert(Control);
    R := ViRect(2, 5, 12, 6);
    Insert(TViLabel.Create(R, '~N~ew text', Control));
    R := ViRect(34, 6, 37, 7);
    Insert(TViHistory.Create(R, TViInputLine(Control), nil));

    R := ViRect(3, 8, 37, 12);
    cbItems := TStringList.Create;
    with cbItems do
    begin
      Add('~C~ase sensitive');
      Add('~W~hole words only');
      Add('~P~rompt on replace');
      Add('~R~eplace all');
    end;
    Insert(TViCheckBoxes.Create(R, cbItems));

    R := ViRect(17, 13, 27, 15);
    Insert(TViButton.Create(R, 'O~K~', cmOk, bfDefault));
    R := ViRect(28, 13, 38, 15);
    Insert(TViButton.Create(R, 'Cancel', cmCancel, bfNormal));

    SelectNext(False);
  end;
  CreateReplaceDialog := D;
end;

function StdEditorDialog(Dialog: Integer; Info: Pointer): Word;
var
  Strings: TStrings;
  R: TRect;
  //  T: TPoint;
begin
  Strings := TStringList.Create;
  try
    case Dialog of
      edOutOfMemory:
        StdEditorDialog := MessageBox('Not enough memory for this operation.',
          mfError + mfOkButton);
      edReadError:
        StdEditorDialog := MessageBox('Error reading file %s.',
          mfError + mfOkButton);
      edWriteError:
        StdEditorDialog := MessageBox('Error writing file %s.',
          mfError + mfOkButton);
      edCreateError:
        StdEditorDialog := MessageBox('Error creating file %s.',
          mfError + mfOkButton);
      edSaveModify:
        StdEditorDialog := MessageBox('%s has been modified. Save?',
          mfInformation + mfYesNoCancel);
      edSaveUntitled:
        StdEditorDialog := MessageBox('Save untitled file?',
          mfInformation + mfYesNoCancel);
      edSaveAs:
        StdEditorDialog :=
          ViApplication.ExecuteDialog(TViFileDialog.Create('*.*',
          'Save file as', '~N~ame', fdOkButton, Strings), Info);
      edFind:
        StdEditorDialog :=
          ViApplication.ExecuteDialog(CreateFindDialog, Info);
      edSearchFailed:
        StdEditorDialog := MessageBox('Search string not found.',
          mfError + mfOkButton);
      edReplace:
        StdEditorDialog :=
          ViApplication.ExecuteDialog(CreateReplaceDialog, Info);
      edReplacePrompt:
        begin
          // Avoid placing the dialog on the same line as the cursor
          R := ViRect(0, 1, 40, 8);
          {TODO
                  R.Move((Desktop^.Size.X - R.B.X) div 2, 0);
                  ViDesktop.MakeGlobal(R.B, T);
                  Inc(T.Y);
                  if TPoint(Info^).Y <= T.Y then
                    R.Move(0, ViDesktop.Size.Y - R.B.Y - 2);
          {}
          StdEditorDialog := MessageBoxRect(R, 'Replace this occurence?',
            mfYesNoCancel + mfInformation);
        end;
    else
      Result := cmCancel;
    end;
  finally
    Strings.Free;
  end;
end;

function Min(X, Y: Integer): Integer;
begin
  if X <= Y then
    Min := X
  else
    Min := Y;
end;

function Max(X, Y: Integer): Integer;
begin
  if X >= Y then
    Max := X
  else
    Max := Y;
end;

function CountLines(var Buf; Count: Word): Integer;
var
  P: PChar;
  Lines: Cardinal;
begin
  P := PChar(@Buf);
  Lines := 0;
  while (Count > 0) do
  begin
    if P^ in [#10, #13] then
    begin
      Inc(Lines);
      if Ord((P + 1)^) + Ord(P^) = 23 then
      begin
        Inc(P);
        Dec(Count);
        if Count = 0 then
          Break;
      end;
    end;
    Inc(P);
    Dec(Count);
  end;
  CountLines := Lines;
end;

function ScanKeyMap(KeyMap: Pointer; KeyCode: Word): Word;
var
  I, Key: Word;
  KeyTable: PWordArray absolute KeyMap;
begin
  for I := 1 to KeyTable^[0] do
  begin
    Key := KeyTable^[I * 2 - 1];
    if (Lo(Key) = Lo(KeyCode)) and ((Hi(Key) = 0) or (Hi(Key) = Hi(KeyCode)))
      then
    begin
      ScanKeyMap := KeyTable^[I * 2];
      Exit;
    end;
  end;
  ScanKeyMap := 0;
end;

type
  Btable = array[0..255] of Byte;

procedure BMMakeTable(const S: string; var T: Btable);
// Makes a Boyer-Moore search table. s = the search String t = the table
var
  X: LongInt;
begin
  FillChar(T, SizeOf(T), Length(S));
  for X := Length(S) downto 1 do
    if (T[Ord(S[X])] = Length(S)) then
      T[Ord(S[X])] := Length(S) - X;
end;

function Scan(var Block; Size: Word; const Str: string): Integer;
var
  Buffer: array[0..MaxBufLength - 1] of Byte absolute Block;
  S2: string;
  Len,
    Numb: Cardinal;
  Found: Boolean;
  Bt: Btable;
begin
  BMMakeTable(Str, Bt);
  Len := Length(Str);
  { sets the length to that of the search String }
  SetLength(S2, Len);
  Found := False;
  Numb := Pred(Len);
  while (not Found) and (Numb < (Size - Len)) do
  begin
    { partial match }
    if Buffer[Numb] = Ord(Str[Len]) then
    begin
      { less partial! }
      if Buffer[Numb - Pred(Len)] = Ord(Str[1]) then
      begin
        Move(Buffer[Numb - Pred(Len)], S2[1], Len);
        if (Str = S2) then
        begin
          Found := True;
          Break;
        end;
      end;
      Inc(Numb);
    end
    else
      Inc(Numb, Bt[Buffer[Numb]]);
  end;
  if not Found then
    Scan := NotFoundValue
  else
    Scan := Numb - Pred(Len);
end;

function IScan(var Block; Size: Integer; const Str: string): Integer;
var
  Buffer: array[0..MaxBufLength - 1] of Char absolute Block;
  S: string;
  Len, Numb, X: Integer;
  Found: Boolean;
  Bt: Btable;
  P: PChar;
  C: Char;
begin
  Len := Length(Str);
  { create uppercased string }
  SetLength(S, Len);
  for X := 1 to Len do
  begin
    if Str[X] in ['a'..'z'] then
      S[X] := Chr(Ord(Str[X]) - 32)
    else
      S[X] := Str[X];
  end;
  BMMakeTable(S, Bt);
  Found := False;
  Numb := Pred(Len);
  while (not Found) and (Numb < (Size - Len)) do
  begin
    { partial match }
    C := Buffer[Numb];
    if C in ['a'..'z'] then
      C := Chr(Ord(C) - 32);
    if (C = S[Len]) then
    begin
      { less partial! }
      P := @Buffer[Numb - Pred(Len)];
      X := 1;
      while (X <= Len) do
      begin
        if not (((P^ in ['a'..'z']) and (Chr(Ord(P^) - 32) = S[X])) or
          (P^ = S[X])) then
          Break;
        Inc(P);
        Inc(X);
      end;
      if (X > Len) then
      begin
        Found := True;
        Break;
      end;
      Inc(Numb);
    end
    else
      Inc(Numb, Bt[Ord(C)]);
  end;
  if not Found then
    IScan := NotFoundValue
  else
    IScan := Numb - Pred(Len);
end;

{ TViIndicator }

constructor TViIndicator.Create(Bounds: TRect);
begin
  inherited Create(Bounds);
  GrowMode := gfGrowLoY + gfGrowHiY;
end;

procedure TViIndicator.Draw;
var
  Color: Byte;
  Frame: Char;
  S: string[15];
  B: TDrawBuffer;
begin
  if State and sfDragging = 0 then
  begin
    Color := GetColor(1);
    Frame := ldDblHorizontalBar;
  end
  else
  begin
    Color := GetColor(2);
    Frame := ldHorizontalBar;
  end;
  MoveChar(B, Frame, Color, Size.X);
  if Modified then
    WordRec(B[0]).Lo := Ord(ldModified);
  S := Format(' %d:%d ', [Location.Y + 1, Location.X + 1]);
  MoveStr(B[8 - Pos(':', S)], S, Color);
  WriteBuf(0, 0, Size.X, 1, B);
end;

function TViIndicator.GetPalette: TPalette;
begin
  GetPalette := CIndicator;
end;

procedure TViIndicator.SetState(AState: Word; Enable: Boolean);
begin
  inherited SetState(AState, Enable);
  if AState = sfDragging then
    DrawView;
end;

procedure TViIndicator.SetValue(ALocation: TPoint; AModified: Boolean);
begin
  if (Location.X <> ALocation.X) or (Location.Y <> ALocation.Y) or
    (Modified <> AModified) then
  begin
    Location := ALocation;
    Modified := AModified;
    DrawView;
  end;
end;

{ TViEditor }

constructor TViEditor.Create(Bounds: TRect; AHScrollBar, AVScrollBar:
  TViScrollBar; AIndicator: TViIndicator; ViMode: Boolean = False);
begin
  inherited Create(Bounds);
  if not ViMode then
    GrowMode := gfGrowHiX + gfGrowHiY;
  Options := Options or ofSelectable;
  if ViMode then
    Options := Options or ofViInternal;
  EventMask := evMouseDown + evKeyDown + evCommand + evBroadcast;
  ShowCursor;
  HScrollBar := AHScrollBar;
  VScrollBar := AVScrollBar;
  Indicator := AIndicator;

  BufSize := $FFFF; //ABufSize;
  CanUndo := True;
  InitBuffer;

  if Buffer <> nil then
    IsValid := True
  else
  begin
    EditorDialog(edOutOfMemory, nil);
    BufSize := 0;
  end;

  TabSize := 8;
  SetBufLen(0);
end;

destructor TViEditor.Destroy;
begin
  DoneBuffer;
  inherited;
end;

{
constructor TViEditor.Load(var S: TStream);
begin
  inherited Load(S);
  GetPeerViewPtr(S, HScrollBar);
  GetPeerViewPtr(S, VScrollBar);
  GetPeerViewPtr(S, Indicator);
  S.Read(BufSize, SizeOf(Word));
  S.Read(CanUndo, SizeOf(Boolean));
  InitBuffer;
  if Buffer <> nil then IsValid := True else
  begin
    EditorDialog(edOutOfMemory, nil);
    BufSize := 0;
  end;
  Lock;
  SetBufLen(0);
end;

destructor TViEditor.Done;
begin
  DoneBuffer;
  inherited Done;
end;
{}

function TViEditor.BufChar(P: Integer): Char;
begin
  if P >= CurPtr then
    Inc(P, GapLen);
  BufChar := Buffer^[P];
end;

function TViEditor.BufPtr(P: Integer): Word;
begin
  if P >= CurPtr then
    BufPtr := P + GapLen
  else
    BufPtr := P;
end;

procedure TViEditor.ChangeBounds(var Bounds: TRect);
begin
  SetBounds(Bounds);
  Delta.X := Max(0, Min(Delta.X, Limit.X - Size.X));
  Delta.Y := Max(0, Min(Delta.Y, Limit.Y - Size.Y));
  Update(ufView);
end;

function TViEditor.CharPos(P, Target: Integer): Integer;
var
  Pos: Integer;
begin
  Pos := 0;
  while P < Target do
  begin
    if BufChar(P) = #9 then
      Pos := Pos or 7;
    Inc(Pos);
    Inc(P);
  end;
  CharPos := Pos;
end;

function TViEditor.CharPtr(P: Word; Target: Integer): Integer;
var
  Pos: Integer;
begin
  Pos := 0;
  while (Pos < Target) and (P < BufLen) and (BufChar(P) <> #13) do
  begin
    if BufChar(P) = #9 then
      Pos := Pos or 7;
    Inc(Pos);
    Inc(P);
  end;
  if Pos > Target then
    Dec(P);
  CharPtr := P;
end;

function TViEditor.ClipCopy: Boolean;
begin
  ClipCopy := False;
  if (Clipboard <> nil) and (Clipboard <> @Self) then
  begin
    ClipCopy := Clipboard.Editor.InsertFrom(Self);
    Selecting := False;
    Update(ufUpdate);
  end;
end;

procedure TViEditor.ClipCut;
begin
  if ClipCopy then
    DeleteSelect;
end;

procedure TViEditor.ClipPaste;
begin
  if (Clipboard <> nil) and (Clipboard <> @Self) then
    InsertFrom(Clipboard.Editor);
end;

procedure TViEditor.ConvertEvent(var Event: TEvent);
var
  Key: Word;
begin
  if Event.What = evKeyDown then
  begin
    if (GetShiftState and $03 <> 0) and
      (Event.ScanCode >= $47) and (Event.ScanCode <= $51) then
      Event.CharCode := #0;
    Key := Event.KeyCode;
    if FKeyState <> 0 then
    begin
      if (Lo(Key) >= $01) and (Lo(Key) <= $1A) then
        Inc(Key, $40);
      if (Lo(Key) >= $61) and (Lo(Key) <= $7A) then
        Dec(Key, $20);
    end;
    Key := ScanKeyMap(KeyMap[FKeyState], Key);
    FKeyState := 0;
    if Key <> 0 then
      if Hi(Key) = $FF then
      begin
        FKeyState := Lo(Key);
        ClearEvent(Event);
      end
      else
      begin
        Event.What := evCommand;
        Event.Command := Key;
      end;
  end;
end;

function TViEditor.CursorVisible: Boolean;
begin
  CursorVisible := (CurPos.Y >= Delta.Y) and (CurPos.Y < Delta.Y + Size.Y);
end;

procedure TViEditor.DeleteRange(StartPtr, EndPtr: Integer; DelSelect: Boolean);
begin
  if HasSelection and DelSelect then
    DeleteSelect
  else
  begin
    SetSelect(CurPtr, EndPtr, True);
    DeleteSelect;
    SetSelect(StartPtr, CurPtr, False);
    DeleteSelect;
  end;
end;

procedure TViEditor.DeleteSelect;
begin
  InsertText(nil, 0, False);
end;

procedure TViEditor.DoneBuffer;
begin
  if Buffer <> nil then
  begin
    FreeMem(Buffer, BufSize);
    Buffer := nil;
  end;
end;

procedure TViEditor.DoSearchReplace;
var
  I: Word;
  C: TPoint;
begin
  repeat
    I := cmCancel;
    if not Search(FindStr, EditorFlags) then
    begin
      if EditorFlags and (efReplaceAll + efDoReplace) <>
        (efReplaceAll + efDoReplace) then
        StdEditorDialog(edSearchFailed, nil)
    end
    else if EditorFlags and efDoReplace <> 0 then
    begin
      I := cmYes;
      if EditorFlags and efPromptOnReplace <> 0 then
      begin
        MakeGlobal(Cursor, C);
        I := StdEditorDialog(edReplacePrompt, @C);
      end;
      if I = cmYes then
      begin
        Lock;
        InsertText(@ReplaceStr[1], Length(ReplaceStr), False);
        TrackCursor(False);
        Unlock;
      end;
    end;
  until (I = cmCancel) or (EditorFlags and efReplaceAll = 0);
end;

procedure TViEditor.DoUpdate;
begin
  if FUpdateFlags <> 0 then
  begin
    SetCursor(CurPos.X - Delta.X, CurPos.Y - Delta.Y);
    if FUpdateFlags and ufView <> 0 then
      DrawView
    else if FUpdateFlags and ufLine <> 0 then
      DrawLines(CurPos.Y - Delta.Y, 1, LineStart(CurPtr));
    if HScrollBar <> nil then
      HScrollBar.SetParams(Delta.X, 0, Limit.X - Size.X, Size.X div 2, 1);
    if VScrollBar <> nil then
      VScrollBar.SetParams(Delta.Y, 0, Limit.Y - Size.Y, Size.Y - 1, 1);
    if Indicator <> nil then
      Indicator.SetValue(CurPos, Modified);
    if State and sfActive <> 0 then
      UpdateCommands;
    FUpdateFlags := 0;
  end;
end;

procedure TViEditor.Draw;
begin
  if DrawLine <> Delta.Y then
  begin
    DrawPtr := LineMove(DrawPtr, Delta.Y - DrawLine);
    DrawLine := Delta.Y;
  end;
  DrawLines(0, Size.Y, DrawPtr);
end;

procedure TViEditor.DrawLines(Y, Count: Integer; LinePtr: Integer);
var
  Color: Word;
  B: array[0..MaxLineLength - 1] of Word;
begin
  Color := GetColor($0201);
  while Count > 0 do
  begin
    FormatLine(B, LinePtr, Delta.X + Size.X, Color);
    WriteBuf(0, Y, Size.X, 1, B[Delta.X]);
    LinePtr := NextLine(LinePtr);
    Inc(Y);
    Dec(Count);
  end;
end;

procedure TViEditor.Find;
var
  FindRec: TFindDialogRec;
begin
  with FindRec do
  begin
    Find := FindStr;
    Options := EditorFlags;
    if StdEditorDialog(edFind, @FindRec) <> cmCancel then
    begin
      FindStr := FindRec.Find;
      EditorFlags := Options and not efDoReplace;
      DoSearchReplace;
    end;
  end;
end;

procedure TViEditor.FormatLine(var DrawBuf; LinePtr: Integer; Width: Integer; 
    Colors: Word);
var
  OutPtr: PWord;
  OutCnt, IdxPos: Integer;
  Attr: Word;

  procedure FillSpace(I: Integer);
  var
    W: Word;
  begin
    Inc(OutCnt, I);
    W := 32 or Attr;
    while (I > 0) do
    begin
      OutPtr^ := W;
      Inc(OutPtr);
      Dec(I);
    end;
  end;

  function FormatUntil(EndPos: Integer): Boolean;
  var
    P: PChar;
  begin
    FormatUntil := False;
    P := PChar(Buffer) + IdxPos;
    while EndPos > IdxPos do
    begin
      if OutCnt >= Width then
        Exit;
      case P^ of
        #9:
          FillSpace(TabSize - (OutCnt mod TabSize));
        #10, #13:
          begin
            FillSpace(Width - OutCnt);
            FormatUntil := True;
            Exit;
          end;
      else
        begin
          Inc(OutCnt);
          { Get color from Colorer }
          OutPtr^ := Ord(P^) or Attr;
          Inc(OutPtr);
        end;
      end; { case }
      Inc(P);
      Inc(IdxPos);
    end;
  end;
begin
  OutCnt := 0;
  OutPtr := @DrawBuf;
  IdxPos := LinePtr;
  Attr := Lo(Colors) shl 8;
  if FormatUntil(SelStart) then
    Exit;
  Attr := Hi(Colors) shl 8;
  if FormatUntil(CurPtr) then
    Exit;
  Inc(IdxPos, GapLen);
  if FormatUntil(SelEnd + GapLen) then
    Exit;
  Attr := Lo(Colors) shl 8;
  if FormatUntil(BufSize) then
    Exit;
  { fill up until width }
  FillSpace(Width - OutCnt);
end; {TViEditor.FormatLine}

function TViEditor.GetMousePtr(Mouse: TPoint): Integer;
begin
  MakeLocal(Mouse, Mouse);
  Mouse.X := Max(0, Min(Mouse.X, Size.X - 1));
  Mouse.Y := Max(0, Min(Mouse.Y, Size.Y - 1));
  GetMousePtr := CharPtr(LineMove(DrawPtr, Mouse.Y + Delta.Y - DrawLine),
    Mouse.X + Delta.X);
end;

function TViEditor.GetPalette: TPalette;
begin
  GetPalette := CEditor;
end;

procedure TViEditor.HandleEvent(var Event: TEvent);
var
  CenterCursor: Boolean;
  SelectMode: Byte;
  D, Mouse: TPoint;

  procedure CheckScrollBar(P: TViScrollBar; var D: Integer);
  begin
    if (Event.InfoPtr = P) and (P.Value <> D) then
    begin
      D := P.Value;
      Update(ufView);
    end;
  end;

begin
  if Event.What = evMouseDown then
    if (State and (sfSelected + sfDisabled) = 0) and
        (Options and ofSelectable <> 0) then
      inherited HandleEvent(Event);

  if Event.What = evKeyDown then
    inherited HandleEvent(Event);

  ConvertEvent(Event);
  CenterCursor := not CursorVisible;
  SelectMode := 0;
  if Selecting or (GetShiftState and $03 <> 0) then
    SelectMode := smExtend;
  case Event.What of
    evMouseDown:
      begin
        if Event.Double then
          SelectMode := SelectMode or smDouble;
        repeat
          inherited HandleEvent(Event);
          Lock;
          if Event.What = evMouseAuto then
          begin
            MakeLocal(Event.Where, Mouse);
            D := Delta;
            if Mouse.X < 0 then
              Dec(D.X);
            if Mouse.X >= Size.X then
              Inc(D.X);
            if Mouse.Y < 0 then
              Dec(D.Y);
            if Mouse.Y >= Size.Y then
              Inc(D.Y);
            ScrollTo(D.X, D.Y);
          end;
          SetCurPtr(GetMousePtr(Event.Where), SelectMode);
          SelectMode := SelectMode or smExtend;
          Unlock;
        until not MouseEvent(Event, evMouseMove + evMouseAuto);
        inherited HandleEvent(Event);
      end;
    evMouseUp, evMouseMove:
      inherited HandleEvent(Event);
    evKeyDown:
    begin
      case Event.CharCode of
        #9, #32..#255:
          begin
            Lock;
            if Overwrite and not HasSelection then
              if CurPtr <> LineEnd(CurPtr) then
                SelEnd := NextChar(CurPtr);
            InsertText(@Event.CharCode, 1, False);
            TrackCursor(CenterCursor);
            Unlock;
          end;
      else
        Exit;
      end;
    end;
    evCommand:
      case Event.Command of
        cmFind: Find;
        cmReplace: Replace;
        cmSearchAgain: DoSearchReplace;
      else
        begin
          Lock;
          case Event.Command of
            cmCut: ClipCut;
            cmCopy: ClipCopy;
            cmPaste: ClipPaste;
            cmUndo: Undo;
            cmClear: DeleteSelect;
            cmCharLeft: SetCurPtr(PrevChar(CurPtr), SelectMode);
            cmCharRight: SetCurPtr(NextChar(CurPtr), SelectMode);
            cmWordLeft: SetCurPtr(PrevWord(CurPtr), SelectMode);
            cmWordRight: SetCurPtr(NextWord(CurPtr), SelectMode);
            cmLineStart: SetCurPtr(LineStart(CurPtr), SelectMode);
            cmLineEnd: SetCurPtr(LineEnd(CurPtr), SelectMode);
            cmLineUp: SetCurPtr(LineMove(CurPtr, -1), SelectMode);
            cmLineDown: SetCurPtr(LineMove(CurPtr, 1), SelectMode);
            cmPageUp: SetCurPtr(LineMove(CurPtr, -(Size.Y - 1)), SelectMode);
            cmPageDown: SetCurPtr(LineMove(CurPtr, Size.Y - 1), SelectMode);
            cmTextStart: SetCurPtr(0, SelectMode);
            cmTextEnd: SetCurPtr(BufLen, SelectMode);
            cmNewLine:
              begin
                NewLine;
                DoReturnPressed;
              end;
            cmBackSpace: DeleteRange(PrevChar(CurPtr), CurPtr, True);
            cmDelChar: DeleteRange(CurPtr, NextChar(CurPtr), True);
            cmDelWord: DeleteRange(CurPtr, NextWord(CurPtr), False);
            cmDelStart: DeleteRange(LineStart(CurPtr), CurPtr, False);
            cmDelEnd: DeleteRange(CurPtr, LineEnd(CurPtr), False);
            cmDelLine: DeleteRange(LineStart(CurPtr), NextLine(CurPtr), False);
            cmInsMode: ToggleInsMode;
            cmStartSelect: StartSelect;
            cmHideSelect: HideSelect;
            cmIndentMode: AutoIndent := not AutoIndent;
          else
            Unlock;
            Exit;
          end;
          TrackCursor(CenterCursor);
          Unlock;
        end;
      end;
    evBroadcast:
      case Event.Command of
        cmScrollBarChanged:
          if (Event.InfoPtr = HScrollBar) or
            (Event.InfoPtr = VScrollBar) then
          begin
            CheckScrollBar(HScrollBar, Delta.X);
            CheckScrollBar(VScrollBar, Delta.Y);
          end
          else
            Exit;
      else
        Exit;
      end;
  else
    Exit;
  end;
  ClearEvent(Event);
end;

function TViEditor.HasSelection: Boolean;
begin
  HasSelection := SelStart <> SelEnd;
end;

procedure TViEditor.HideSelect;
begin
  Selecting := False;
  SetSelect(CurPtr, CurPtr, False);
end;

procedure TViEditor.InitBuffer;
begin
  Buffer := AllocMem(BufSize);
end;

function TViEditor.InsertBuffer(var P: PEditBuffer; Offset, Length: Integer; 
    AllowUndo, SelectText: Boolean): Boolean;
var
  SelLen, DelLen, SelLines, Lines: Word;
  NewSize: LongInt;
begin
  if ReadOnly then
  begin
    InsertBuffer := False;
    Exit;
  end;  
  InsertBuffer := True;
  Selecting := False;
  SelLen := SelEnd - SelStart;
  if (SelLen = 0) and (Length = 0) then
    Exit;
  DelLen := 0;
  if AllowUndo then
    if CurPtr = SelStart then
      DelLen := SelLen
    else if SelLen > InsCount then
      DelLen := SelLen - InsCount;
  NewSize := LongInt(BufLen + DelCount - SelLen + DelLen) + Length;
  if NewSize > BufLen + DelCount then
    if not SetBufSize(NewSize) then
    begin
      EditorDialog(edOutOfMemory, nil);
      InsertBuffer := False;
      SelEnd := SelStart;
      Exit;
    end;
  SelLines := CountLines(Buffer^[BufPtr(SelStart)], SelLen);
  if CurPtr = SelEnd then
  begin
    if AllowUndo then
    begin
      if DelLen > 0 then
        Move(Buffer^[SelStart],
          Buffer^[CurPtr + GapLen - DelCount - DelLen], DelLen);
      Dec(InsCount, SelLen - DelLen);
    end;
    CurPtr := SelStart;
    Dec(CurPos.Y, SelLines);
  end;
  if Delta.Y > CurPos.Y then
  begin
    Dec(Delta.Y, SelLines);
    if Delta.Y < CurPos.Y then
      Delta.Y := CurPos.Y;
  end;
  if Length > 0 then
    Move(P^[Offset], Buffer^[CurPtr], Length);
  Lines := CountLines(Buffer^[CurPtr], Length);
  Inc(CurPtr, Length);
  Inc(CurPos.Y, Lines);
  DrawLine := CurPos.Y;
  DrawPtr := LineStart(CurPtr);
  CurPos.X := CharPos(DrawPtr, CurPtr);
  if not SelectText then
    SelStart := CurPtr;
  SelEnd := CurPtr;
  Inc(BufLen, Length - SelLen);
  Dec(GapLen, Length - SelLen);
  if AllowUndo then
  begin
    Inc(DelCount, DelLen);
    Inc(InsCount, Length);
  end;
  Inc(Limit.Y, Lines - SelLines);
  Delta.Y := Max(0, Min(Delta.Y, Limit.Y - Size.Y));
  if not IsClipboard then
  begin
    Modified := True;
    DoChanged;
  end;
  SetBufSize(BufLen + DelCount);
  if (SelLines = 0) and (Lines = 0) then
    Update(ufLine)
  else
    Update(ufView);
end;

function TViEditor.InsertFrom(Editor: TViEditor): Boolean;
begin
  Result := InsertBuffer(Editor.Buffer,
    Editor.BufPtr(Editor.SelStart),
    Editor.SelEnd - Editor.SelStart, CanUndo, IsClipboard);
end;

function TViEditor.InsertText(Text: Pointer; Length: Integer; SelectText: 
    Boolean): Boolean;
begin
  InsertText := InsertBuffer(PEditBuffer(Text),
    0, Length, CanUndo, SelectText);
end;

function TViEditor.IsClipboard: Boolean;
begin
  IsClipboard := (Clipboard.Editor = Self);
end;

function TViEditor.LineEnd(P: Integer): Integer;
var
  Start, I: Integer;
  PC: PChar;
begin
  if P < CurPtr then
  begin
    I := CurPtr - P;
    PC := PChar(Buffer) + P;
    while (I > 0) do
    begin
      if PC^ in [#10, #13] then
      begin
        LineEnd := PC - PChar(Buffer);
        Exit;
      end;
      Inc(PC);
      Dec(I);
    end;
    Start := CurPtr;
  end
  else
    Start := P;
  I := BufLen - Start;
  PC := PChar(Buffer) + GapLen + Start;
  while (I > 0) do
  begin
    if PC^ in [#10, #13] then
    begin
      LineEnd := PC - (PChar(Buffer) + GapLen);
      Exit;
    end;
    Inc(PC);
    Dec(I);
  end;
  LineEnd := PC - (PChar(Buffer) + GapLen);
end;

function TViEditor.LineMove(P: Integer; Count: Integer): Integer;
var
  Pos: Integer;
  I: Word;
begin
  I := P;
  P := LineStart(P);
  Pos := CharPos(P, I);
  while Count <> 0 do
  begin
    I := P;
    if Count < 0 then
    begin
      P := PrevLine(P);
      Inc(Count);
    end
    else
    begin
      P := NextLine(P);
      Dec(Count);
    end;
  end;
  if P <> I then
    P := CharPtr(P, Pos);
  LineMove := P;
end;

function TViEditor.LineStart(P: Integer): Integer;
var
  I: Cardinal;
  Start, PC: PChar;
  //  oc : char;
begin
  PC := nil;
  Start := nil;
  if P > CurPtr then
  begin
    Start := PChar(Buffer) + GapLen;
    PC := Start;
    I := P - CurPtr;
    Dec(PC);
    while (I > 0) do
    begin
      if PC^ in [#10, #13] then
        Break;
      Dec(PC);
      Dec(I);
    end;
  end
  else
    I := 0;
  if I = 0 then
  begin
    Start := PChar(Buffer);
    I := P;
    PC := Start + P;
    Dec(PC);
    while (I > 0) do
    begin
      if PC^ in [#10, #13] then
        Break;
      Dec(PC);
      Dec(I);
    end;
    if I = 0 then
    begin
      LineStart := 0;
      Exit;
    end;
  end;
  //  oc:=pc^;
  if Assigned(PC) and Assigned(Start) then
    Result := PC - Start + 1
  else
    Result := 0;
end; { TViEditor.LineStart }

procedure TViEditor.Lock;
begin
  Inc(FLockCount);
end;

procedure TViEditor.NewLine;
const
  CrLf: array[1..2] of Char = #13#10;
var
  I, P: Word;
begin
  P := LineStart(CurPtr);
  I := P;
  while (I < CurPtr) and ((Buffer^[I] = ' ') or (Buffer^[I] = #9)) do
    Inc(I);
  InsertText(@CrLf, 2, False);
  if AutoIndent then
    InsertText(@Buffer^[P], I - P, False);
end;

function TViEditor.NextChar(P: Integer): Integer;
var
  PC: PChar;
begin
  if P <> BufLen then
  begin
    Inc(P);
    if P <> BufLen then
    begin
      PC := PChar(Buffer);
      if P >= CurPtr then
        Inc(PC, GapLen);
      Inc(PC, P - 1);
      if Ord(PC^) + Ord((PC + 1)^) = 23 then
        Inc(P);
    end;
  end;
  NextChar := P;
end; { TViEditor.NextChar }

function TViEditor.NextLine(P: Integer): Integer;
begin
  NextLine := NextChar(LineEnd(P));
end;

function TViEditor.NextWord(P: Integer): Integer;
begin
  while (P < BufLen) and (BufChar(P) in WordChars) do
    P := NextChar(P);
  while (P < BufLen) and not (BufChar(P) in WordChars) do
    P := NextChar(P);
  NextWord := P;
end;

function TViEditor.PrevChar(P: Integer): Integer;
var
  PC: PChar;
begin
  if P <> 0 then
  begin
    Dec(P);
    if P <> 0 then
    begin
      PC := PChar(Buffer);
      if P >= CurPtr then
        Inc(PC, GapLen);
      Inc(PC, P - 1);
      if Ord(PC^) + Ord((PC + 1)^) = 23 then
        Dec(P);
    end;
  end;
  PrevChar := P;
end; { TViEditor.PrevChar }

function TViEditor.PrevLine(P: Integer): Integer;
begin
  PrevLine := LineStart(PrevChar(P));
end;

function TViEditor.PrevWord(P: Integer): Integer;
begin
  while (P > 0) and not (BufChar(PrevChar(P)) in WordChars) do
    P := PrevChar(P);
  while (P > 0) and (BufChar(PrevChar(P)) in WordChars) do
    P := PrevChar(P);
  PrevWord := P;
end;

procedure TViEditor.Replace;
var
  ReplaceRec: TReplaceDialogRec;
begin
  with ReplaceRec do
  begin
    Find := FindStr;
    Replace := ReplaceStr;
    Options := EditorFlags;
    if EditorDialog(edReplace, @ReplaceRec) <> cmCancel then
    begin
      FindStr := Find;
      ReplaceStr := Replace;
      EditorFlags := Options or efDoReplace;
      DoSearchReplace;
    end;
  end;
end;

procedure TViEditor.ScrollTo(X, Y: Integer);
begin
  X := Max(0, Min(X, Limit.X - Size.X));
  Y := Max(0, Min(Y, Limit.Y - Size.Y));
  if (X <> Delta.X) or (Y <> Delta.Y) then
  begin
    Delta.X := X;
    Delta.Y := Y;
    Update(ufView);
  end;
end;

function TViEditor.Search(const FindStr: string; Opts: Word): Boolean;
var
  I, Pos: Integer;
begin
  Search := False;
  Pos := CurPtr;
  repeat
    if Opts and efCaseSensitive <> 0 then
      I := Scan(Buffer^[BufPtr(Pos)], BufLen - Pos, FindStr)
    else
      I := IScan(Buffer^[BufPtr(Pos)], BufLen - Pos, FindStr);
    if (I <> sfSearchFailed) then
    begin
      Inc(I, Pos);
      if (Opts and efWholeWordsOnly = 0) or
        not (((I <> 0) and (BufChar(I - 1) in WordChars)) or
        ((I + Length(FindStr) <> BufLen) and
        (BufChar(I + Length(FindStr)) in WordChars))) then
      begin
        Lock;
        SetSelect(I, I + Length(FindStr), False);
        TrackCursor(not CursorVisible);
        Unlock;
        Search := True;
        Exit;
      end
      else
        Pos := I + 1;
    end;
  until I = sfSearchFailed;
end;

procedure TViEditor.SetBufLen(Length: Integer);
begin
  BufLen := Length;
  GapLen := BufSize - Length;
  SelStart := 0;
  SelEnd := 0;
  CurPtr := 0;
  CurPos.X := 0;
  CurPos.Y := 0;
  Delta.X := 0;
  Delta.Y := 0;
  Limit.X := MaxLineLength;
  Limit.Y := CountLines(Buffer^[GapLen], BufLen) + 1;
  DrawLine := 0;
  DrawPtr := 0;
  DelCount := 0;
  InsCount := 0;
  Modified := False;
  Update(ufView);
end;

function TViEditor.SetBufSize(NewSize: Integer): Boolean;
begin
  SetBufSize := NewSize <= BufSize;
end;

procedure TViEditor.SetCmdState(Command: Word; Enable: Boolean);
var
  S: TCommandSet;
begin
  S := [Command];
  if Enable and (State and sfActive <> 0) then
    EnableCommands(S)
  else
    DisableCommands(S);
end;

procedure TViEditor.SetCurPtr(P: Integer; SelectMode: Byte);
var
  Anchor: Word;
begin
  if SelectMode and smExtend = 0 then
    Anchor := P
  else if CurPtr = SelStart then
    Anchor := SelEnd
  else
    Anchor := SelStart;
  if P < Anchor then
  begin
    if SelectMode and smDouble <> 0 then
    begin
      P := PrevLine(NextLine(P));
      Anchor := NextLine(PrevLine(Anchor));
    end;
    SetSelect(P, Anchor, True);
  end
  else
  begin
    if SelectMode and smDouble <> 0 then
    begin
      P := NextLine(P);
      Anchor := PrevLine(NextLine(Anchor));
    end;
    SetSelect(Anchor, P, False);
  end;
end;

procedure TViEditor.SetSelect(NewStart, NewEnd: Integer; CurStart: Boolean);
var
  Flags: Byte;
  P, L: Word;
begin
  if CurStart then
    P := NewStart
  else
    P := NewEnd;
  Flags := ufUpdate;
  if (NewStart <> SelStart) or (NewEnd <> SelEnd) then
    if (NewStart <> NewEnd) or (SelStart <> SelEnd) then
      Flags := ufView;
  if P <> CurPtr then
  begin
    if P > CurPtr then
    begin
      L := P - CurPtr;
      Move(Buffer^[CurPtr + GapLen], Buffer^[CurPtr], L);
      Inc(CurPos.Y, CountLines(Buffer^[CurPtr], L));
      CurPtr := P;
    end
    else
    begin
      L := CurPtr - P;
      CurPtr := P;
      Dec(CurPos.Y, CountLines(Buffer^[CurPtr], L));
      Move(Buffer^[CurPtr], Buffer^[CurPtr + GapLen], L);
    end;
    DrawLine := CurPos.Y;
    DrawPtr := LineStart(P);
    CurPos.X := CharPos(DrawPtr, P);
    DelCount := 0;
    InsCount := 0;
    SetBufSize(BufLen);
  end;
  SelStart := NewStart;
  SelEnd := NewEnd;
  Update(Flags);
end;

procedure TViEditor.SetState(AState: Word; Enable: Boolean);
begin
  inherited SetState(AState, Enable);
  case AState of
    sfExposed:
      if Enable then
        Unlock;
    sfFocused: if (State and sfSelected) = sfSelected then
      begin
        if (Options and ofViInternal) <> ofViInternal then
        begin
          if HScrollBar <> nil then
            HScrollBar.SetState(sfVisible, Enable);
          if VScrollBar <> nil then
            VScrollBar.SetState(sfVisible, Enable);
          if Indicator <> nil then
            Indicator.SetState(sfVisible, Enable);
        end;
        UpdateCommands;
      end;
    sfSelected:
      begin
        if (Options and ofViInternal) <> ofViInternal then
        begin
          if HScrollBar <> nil then
            HScrollBar.SetState(sfVisible, Enable);
          if VScrollBar <> nil then
            VScrollBar.SetState(sfVisible, Enable);
          if Indicator <> nil then
            Indicator.SetState(sfVisible, Enable);
        end;
        UpdateCommands;
      end;
  end;
end;

procedure TViEditor.StartSelect;
begin
  HideSelect;
  Selecting := True;
end;

{TODO
procedure TViEditor.Store(var S: TStream);
begin
  inherited Store(S);
  PutPeerViewPtr(S, HScrollBar);
  PutPeerViewPtr(S, VScrollBar);
  PutPeerViewPtr(S, Indicator);
  S.Write(BufSize, SizeOf(Word));
  S.Write(CanUndo, SizeOf(Boolean));
end;
{}

procedure TViEditor.ToggleInsMode;
begin
  Overwrite := not Overwrite;
  SetState(sfCursorIns, not GetState(sfCursorIns));
end;

procedure TViEditor.TrackCursor(Center: Boolean);
begin
  if Center then
    ScrollTo(CurPos.X - Size.X + 1, CurPos.Y - Size.Y div 2)
  else
    ScrollTo(Max(CurPos.X - Size.X + 1, Min(Delta.X, CurPos.X)),
      Max(CurPos.Y - Size.Y + 1, Min(Delta.Y, CurPos.Y)));
end;

procedure TViEditor.Undo;
var
  Length: Word;
begin
  if (DelCount <> 0) or (InsCount <> 0) then
  begin
    SelStart := CurPtr - InsCount;
    SelEnd := CurPtr;
    Length := DelCount;
    DelCount := 0;
    InsCount := 0;
    InsertBuffer(Buffer, CurPtr + GapLen - Length, Length, False, True);
  end;
end;

procedure TViEditor.Unlock;
begin
  if FLockCount > 0 then
  begin
    Dec(FLockCount);
    if FLockCount = 0 then
      DoUpdate;
  end;
end;

procedure TViEditor.Update(AFlags: Byte);
begin
  FUpdateFlags := FUpdateFlags or AFlags;
  if FLockCount = 0 then
    DoUpdate;
end;

procedure TViEditor.UpdateCommands;
begin
  SetCmdState(cmUndo, (DelCount <> 0) or (InsCount <> 0));
  if not IsClipboard then
  begin
    SetCmdState(cmCut, HasSelection);
    SetCmdState(cmCopy, HasSelection);
    SetCmdState(cmPaste, (Clipboard <> nil) and
      (Clipboard.Editor.HasSelection));
  end;
  SetCmdState(cmClear, HasSelection);
  SetCmdState(cmFind, True);
  SetCmdState(cmReplace, True);
  SetCmdState(cmSearchAgain, True);
end;

function TViEditor.Valid(Command: Word): Boolean;
begin
  Valid := IsValid;
end;

procedure TViEditor.DoChanged;
begin
  if Assigned(OnChanged) then OnChanged(Self);
end;

procedure TViEditor.DoReturnPressed;
begin
  if Assigned(OnReturnPressed) then OnReturnPressed(Self);
end;


{ TViMemo }

constructor TViMemo.Create(Bounds: TRect; AHScrollBar, AVScrollBar:
    TViScrollBar; ViMode: Boolean = False);
begin
  inherited Create(Bounds, AHScrollBar, AVScrollBar, nil, ViMode);
end;

function TViMemo.GetPalette: TPalette;
begin
  GetPalette := CMemo;
end;

procedure TViMemo.HandleEvent(var Event: TEvent);
begin
  if (Event.What <> evKeyDown) or (Event.KeyCode <> kbTab) then
    inherited HandleEvent(Event);
end;

procedure TViMemo.Clear;
begin
  SetBufLen(0);
end;

function TViMemo.NumLines: Integer;
begin
  if BufLen > 0 then
    Result := CountLines(Buffer[0], BufLen) + 1
  else
    NumLines := 0;
end;

procedure TViMemo.RemoveLine(Index: Integer);
var
  P: Integer;
begin
  P := IndexLineStart(Index);
  SetCurPtr(P, 0);
  DeleteRange(P, NextLine(P), False);
  if CurPtr = BufLen then
    DeleteRange(PrevChar(CurPtr), CurPtr, True);
end;

function TViMemo.TextLine(Index: Integer): string;
var
  OldCurPtr: Integer;
  OldStart, OldEnd: Integer;
  CurStart: Boolean;
  P, Q: Integer;
  Len: Integer;
begin
  OldCurPtr := CurPtr;
  OldStart := SelStart;
  OldEnd := SelEnd;
  CurStart := (OldStart = OldCurPtr);
  Lock;
  try
    SetCurPtr(BufLen, 0);
    P := IndexLineStart(Index);
    Q := LineEnd(P);
    Len := Q - P;
    if Len = 0 then Exit;
    SetLength(Result, Len);
    Move(Buffer[P], Result[1], Len);
  finally
    SetCurPtr(OldCurPtr, 0);
    SetSelect(OldStart, OldEnd, CurStart);
    Unlock;
  end;
end;

procedure TViMemo.InsertLine(Index: Integer; const Line: string);
var
  OldCurPtr: Integer;
  OldStart, OldEnd: Integer;
  CurStart: Boolean;
  P: Integer;
begin
  if Index = NumLines then
  begin
    Append(Line);
    Exit;
  end;
  P := IndexLineStart(Index);
  OldCurPtr := CurPtr;
  OldStart := SelStart;
  OldEnd := SelEnd;
  CurStart := (OldStart = OldCurPtr);
  try
    SetCurPtr(P, 0);
    InsertText(@Line[1], Length(Line), False);
    NewLine;
  finally
    SetCurPtr(OldCurPtr, 0);
    SetSelect(OldStart, OldEnd, CurStart);
  end;
end;

function TViMemo.IndexLineStart(Index: Integer): Integer;
var
  I: Integer;
begin
  if (Index < 0) or (Index >= NumLines) then Error(@SListIndexError, Index);
  Result := 0;
  for I := 1 to Index do
    Result := NextLine(Result);
end;

procedure TViMemo.Error(const Msg: string; Data: Integer);

  function ReturnAddr: Pointer;
  asm
          MOV     EAX,[EBP+4]
  end;

begin
  raise EViMemoError.CreateFmt(Msg, [Data]) at ReturnAddr;
end;

procedure TViMemo.Error(Msg: PResStringRec; Data: Integer);
begin
  Error(LoadResString(Msg), Data);
end;

procedure TViMemo.Append(const Value: string);
var
  OldCurPtr: Integer;
  OldStart, OldEnd: Integer;
  CurStart: Boolean;
begin
  OldCurPtr := CurPtr;
  OldStart := SelStart;
  OldEnd := SelEnd;
  CurStart := (OldStart = OldCurPtr);
  try
    SetCurPtr(BufLen, 0);
    if BufLen > 0 then
      NewLine;
    InsertText(@Value[1], Length(Value), False);
  finally
    SetCurPtr(OldCurPtr, 0);
    SetSelect(OldStart, OldEnd, CurStart);
  end;
end;

function TViMemo.GetText: string;
var
  OldCurPtr: Integer;
  OldStart, OldEnd: Integer;
  CurStart: Boolean;
begin
  OldCurPtr := CurPtr;
  OldStart := SelStart;
  OldEnd := SelEnd;
  CurStart := (OldStart = OldCurPtr);
  Lock;
  try
    SetCurPtr(BufLen, 0);
    SetLength(Result, BufLen);
    Move(Buffer[0], Result[1], BufLen);
  finally
    SetCurPtr(OldCurPtr, 0);
    SetSelect(OldStart, OldEnd, CurStart);
    Unlock;
  end;
end;

procedure TViMemo.SetText(const Value: string);
begin
  SetBufLen(0);
  InsertText(@Value[1], Length(Value), False);
end;

procedure TViMemo.Deselect;
begin
  HideSelect;
end;

function TViMemo.HasMarkedText: Boolean;
begin
  Result := HasSelection;
end;

procedure TViMemo.Insert(const Value: string; Mark: Boolean);
begin
  InsertText(@Value[1], Length(Value), Mark);
end;

procedure TViMemo.Copy;
begin
  ClipCopy;
end;

procedure TViMemo.Cut;
begin
  ClipCut;
end;

procedure TViMemo.Paste;
begin
  ClipPaste;
end;

procedure TViMemo.GetMarkedRegion(var Line1, Column1, Line2, Column2: Integer);
begin
  Line1 := CountLines(Buffer[0], SelStart);
  Column1 := CharPos(LineStart(SelStart), SelStart);
  Line2 := CountLines(Buffer[0], SelEnd);
  Column2 := CharPos(LineStart(SelEnd), SelEnd);
end;

procedure TViMemo.SetEdited(const Value: Boolean);
begin
  Modified := Value;
  Update(ufUpdate);
end;

function TViMemo.MarkedText: string;
var
  Len: Integer;
begin
  Result := '';
  Len := SelEnd - SelStart;
  if Len < 1 then Exit;
  SetLength(Result, Len);
  Move(Buffer[SelStart], Result[1], Len);
end;

procedure TViMemo.SetCursorPosition(Line, Column: Integer);
var
  P, I: Integer;
begin
  P := IndexLineStart(Line);
  for I := 1 to Column do
    P := NextChar(P);
  SetCurPtr(P, 0);
  TrackCursor(False);
end;

procedure TViMemo.InsertAt(const Text: string; Column, Row: Integer; Mark:
    Boolean = False);
var
  OldCurPtr: Integer;
  OldStart, OldEnd: Integer;
  CurStart: Boolean;
  P, I: Integer;
  Len: Integer;
begin
  P := 0;
  for I := 0 to Row - 1 do
    P := NextLine(P);
  for I := 1 to Column do
    P := NextChar(P);
  OldCurPtr := CurPtr;
  OldStart := SelStart;
  OldEnd := SelEnd;
  CurStart := (OldStart = OldCurPtr);
  Len := Length(Text);
  try
    SetCurPtr(P, 0);
    InsertText(@Text[1], Len, Mark);
  finally
    if not Mark then
    begin
      SetCurPtr(OldCurPtr, 0);
      SetSelect(OldStart + Len, OldEnd + Len, CurStart);
    end;  
  end;
end;


(*
constructor TMemo.Load(var S: TStream);
var
  Length: Word;
begin
  inherited Load(S);
  S.Read(Length, SizeOf(Word));
  if IsValid then
  begin
    S.Read(Buffer^[BufSize - Length], Length);
    SetBufLen(Length);
  end
  else S.Seek(S.GetPos + Length);
end;

function TMemo.DataSize: Word;
begin
  DataSize := BufSize + SizeOf(Word);
end;

procedure TMemo.GetData(var Rec);
var
  Data: TMemoData absolute Rec;
begin
  Data.Length := BufLen;
  Move(Buffer^, Data.Buffer, CurPtr);
  Move(Buffer^[CurPtr + GapLen], Data.Buffer[CurPtr], BufLen - CurPtr);
  FillChar(Data.Buffer[BufLen], BufSize - BufLen, 0);
end;

procedure TMemo.SetData(var Rec);
var
  Data: TMemoData absolute Rec;
begin
  Move(Data.Buffer, Buffer^[BufSize - Data.Length], Data.Length);
  SetBufLen(Data.Length);
end;

{TODO
procedure TMemo.Store(var S: TStream);
begin
  inherited Store(S);
  S.Write(BufLen, SizeOf(Word));
  S.Write(Buffer^, CurPtr);
  S.Write(Buffer^[CurPtr + GapLen], BufLen - CurPtr);
end;
{}
*)

{ TFileEditor }

constructor TFileEditor.Create(Bounds: TRect; AHScrollBar, AVScrollBar: 
    TViScrollBar; AIndicator: TViIndicator; ViMode: Boolean = False);
begin
  inherited;
  FStrings := TStringList.Create;
end;

destructor TFileEditor.Destroy;
begin
  FStrings.Free;
  inherited;
end;

(*TODO
constructor TFileEditor.Create(Bounds: TRect);
begin
  inherited Create(Bounds);
end;
*)
{
constructor TFileEditor.Init(var Bounds: TRect;
  AHScrollBar, AVScrollBar: PScrollBar;
  AIndicator: PIndicator; AFileName: FNameStr);
begin
  inherited Init(Bounds, AHScrollBar, AVScrollBar, AIndicator, 0);
  if AFileName <> '' then
  begin
    FileName := FExpand(AFileName);
    if IsValid then IsValid := LoadFile;
  end;
end;

constructor TFileEditor.Load(var S: TStream);
var
  SStart, SEnd, Curs: Word;
begin
  inherited Load(S);
  BufSize := 0;
  S.Read(FileName[0], SizeOf(Char));
  S.Read(Filename[1], Length(FileName));
  if IsValid then IsValid := LoadFile;
  S.Read(SStart, SizeOf(Word));
  S.Read(SEnd, SizeOf(Word));
  S.Read(Curs, SizeOf(Word));
  if IsValid and (SEnd <= BufLen) then
  begin
    SetSelect(SStart, SEnd, Curs = SStart);
    TrackCursor(True);
  end;
end;
{}

procedure TFileEditor.DoneBuffer;
begin
  inherited DoneBuffer;
end;

procedure TFileEditor.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  case Event.What of
    evCommand:
      case Event.Command of
        cmSave: Save;
        cmSaveAs: SaveAs;
      else
        Exit;
      end;
  else
    Exit;
  end;
  ClearEvent(Event);
end;

procedure TFileEditor.InitBuffer;
begin
  {TODO  NewBuffer(Pointer(Buffer), $1000);
  {}
  inherited;
end;

function TFileEditor.LoadFile: Boolean;
var
  Length: Word;
  FSize: LongInt;
  F: file;
begin
  LoadFile := True;
  if FileName = '' then
    Exit;
  LoadFile := False;
  Length := 0;
  Assign(F, FileName);
  Reset(F, 1);
  if IOResult <> 0 then
    EditorDialog(edReadError, @FileName)
  else
  begin
    FSize := FileSize(F);
    if not SetBufSize(Word(FSize)) then
      EditorDialog(edOutOfMemory, nil)
    else
    begin
      BlockRead(F, Buffer^[BufSize - Word(FSize)], FSize);
      if IOResult <> 0 then
        EditorDialog(edReadError, @FileName)
      else
      begin
        LoadFile := True;
        Length := FSize;
      end;
    end;
    Close(F);
  end;
  SetBufLen(Length);
end;

function TFileEditor.Save: Boolean;
begin
  if FileName = '' then
    Save := SaveAs
  else
    Save := SaveFile;
end;

function TFileEditor.SaveAs: Boolean;
var
  Dlg: TViFileDialog;
begin
  Result := False;
  Dlg := TViFileDialog.Create('*.*', 'Save as...', 'Filename',
      fdOkButton, FStrings);
  case ViDesktop.ExecView(Dlg) of
    cmFileOpen:
      begin
        FileName := Dlg.FileName.Data;
        SendMessage(Owner, evBroadcast, cmUpdateTitle, nil);
        Result := Save;
      end;
    cmNo: Modified := False;
    cmCancel: Result := False;
  end;
  Exit;
  if StdEditorDialog(edSaveAs, @FileName) <> cmCancel then
  begin
    FileName := ExpandFileName(FileName);
    { critical }
    SendMessage(Owner, evBroadcast, cmUpdateTitle, nil);
    SaveAs := SaveFile;
    if IsClipboard then
      FileName := '';
  end;
end;

function TFileEditor.SaveFile: Boolean;
var
  F: file;
  BackupName: string;
  D, N, E: string;
begin
  SaveFile := False;
  if ((EditorFlags and efBackupFiles) = efBackupFiles) and FileExists(FileName)
    then
  begin
    FileSplit(FileName, D, N, E);
    BackupName := D + N + '.BAK';
    Assign(F, BackupName);
    Erase(F);
    Assign(F, FileName);
    Rename(F, BackupName);
  end;
  Assign(F, FileName);
  Rewrite(F, 1);
  if IOResult <> 0 then
    EditorDialog(edCreateError, @FileName)
  else
  begin
    BlockWrite(F, Buffer^, CurPtr);
    BlockWrite(F, Buffer^[CurPtr + GapLen], BufLen - CurPtr);
    if IOResult <> 0 then
      EditorDialog(edWriteError, @FileName)
    else
    begin
      Modified := False;
      Update(ufUpdate);
      SaveFile := True;
    end;
    Close(F);
  end;
end;

function TFileEditor.SetBufSize(NewSize: Integer): Boolean;
var
  N: Word;
  P: Pointer;
begin
  SetBufSize := False;
  {COMPILER
    if NewSize > SizeOf(TEditBuffer) then Exit;
  {}
  if NewSize = 0 then
    NewSize := $1000
  else
    NewSize := (NewSize + $0FFF) and $FFFF000;
  if NewSize <> BufSize then
  begin
    if NewSize > BufSize then
    begin
      P := AllocMem(NewSize);
      if P = nil then
        Exit;
      Move(Buffer^, P^, BufSize);
      FreeMem(Buffer, BufSize);
      Buffer := P;
    end;
    N := BufLen - CurPtr + DelCount;
    Move(Buffer^[BufSize - N], Buffer^[NewSize - N], N);
    if NewSize < BufSize then
    begin
      P := AllocMem(NewSize);
      if P = nil then
        NewSize := BufSize
      else
      begin
        Move(Buffer^, P^, NewSize);
        FreeMem(Buffer, BufSize);
        Buffer := P;
      end;
    end;
    BufSize := NewSize;
    GapLen := BufSize - BufLen;
  end;
  SetBufSize := True;
end;

{TODO
procedure TFileEditor.Store(var S: TStream);
begin
  inherited Store(S);
  S.Write(FileName, Length(FileName) + 1);
  S.Write(SelStart, SizeOf(Word) * 3);
end;
{}

procedure TFileEditor.UpdateCommands;
begin
  inherited UpdateCommands;
  SetCmdState(cmSave, True);
  SetCmdState(cmSaveAs, True);
end;

function TFileEditor.Valid(Command: Word): Boolean;
begin
  if Command = cmValid then
    Result := IsValid
  else if Modified then
    Result := Save
  else
    Result := True;
end;

{ TViEditWindow }

constructor TViEditWindow.Create(Bounds: TRect;
  AFileName: string; ANumber: Integer);
var
  R: TRect;
begin
  inherited Create(Bounds, '', ANumber);
  Options := Options or ofTileable;
  R := ViRect(18, Size.Y - 1, Size.X - 2, Size.Y);
  HScrollBar := TViScrollBar.Create(R, True);
  R := ViRect(Size.X - 1, 1, Size.X, Size.Y - 1);
  VScrollBar := TViScrollBar.Create(R, True);
  R := ViRect(2, Size.Y - 1, 17, Size.Y + 1);
  Indicator := TViIndicator.Create(R);
  Editor := TFileEditor.Create(
    ViRect(1, 1, Size.X - 1, Size.Y - 1),
    HScrollBar, VScrollBar, Indicator);
  Insert(Editor);
  Insert(HScrollBar);
  Insert(VScrollBar);
  Insert(Indicator);
end;

{TODO
constructor TViEditWindow.Load(var S: TStream);
begin
  inherited Load(S);
  GetSubViewPtr(S, Editor);
end;
{}

destructor TViEditWindow.Destroy;
begin
  inherited Destroy;
end;

procedure TViEditWindow.Close;
begin
  if Editor.IsClipboard then
    Hide
  else
  begin
    Delete(Editor.HScrollBar);
    Delete(Editor.VScrollBar);
    Delete(Editor.Indicator);
    Delete(Editor);
    inherited Close;
  end;
end;

function TViEditWindow.GetTitle: TTitleStr;
begin
  if Editor.IsClipboard then
    GetTitle := 'Clipboard'
  else if Editor.FileName = '' then
    GetTitle := 'Untitled'
  else
    GetTitle := Editor.FileName;
end;

procedure TViEditWindow.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if (Event.What = evBroadcast) and (Event.Command = cmUpdateTitle) then
  begin
    Frame.DrawView;
    ClearEvent(Event);
  end;
end;

procedure TViEditWindow.SizeLimits(var Min, Max: TPoint);
begin
  inherited SizeLimits(Min, Max);
  Min.X := 23;
end;

{TODO
procedure TViEditWindow.Store(var S: TStream);
begin
  inherited Store(S);
  PutSubViewPtr(S, Editor);
end;
{}
var
  FClipBoard: TViEditWindow;

function Clipboard: TViEditWindow;
begin
  if not Assigned(FClipBoard) then
  begin
    FClipBoard := TViEditWindow.Create(ViRect(0, 0, Trunc(ViScreen.Width / 2),
      Trunc(ViScreen.Height / 2) - 1), 'Clipboard', 0);
    ViDesktop.Insert(FClipBoard);
    FClipBoard.Hide;
  end;
  Result := FClipBoard;
end;

end.

