
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViMenus.pas                                           }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViMenus.pas,v 1.6 2003/10/16 19:13:35 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViMenus.pas,v $
  Revision 1.6  2003/10/16 19:13:35  iwache
  Event handling for direct Vi-Layer usage in old TurboVision
  style programs added. TViMenuBar constructor Create has now
  a optional parameter ViMode: Boolean which must be True for
  old TurboVision programs.

  Revision 1.5  2003/03/04 10:06:51  iwache
  TViMenuPopup:
  - procedure DoSelect added

  Revision 1.4  2003/03/03 15:54:56  iwache
  function TopMenu:
  - uses now ExecMenuView
  procedure DoSelect
  - uses now ExecMenuView

  Revision 1.3  2002/10/27 22:25:18  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-05-24 Version 2.0.1.0 released
}

unit ViMenus;

{$X+,I-}

interface

uses
  ViDrivers, ViTypes, ViUtils, ViClasses, ViViews;

const

{ Color palettes }

  CMenuView   = #2#3#4#5#6#7;
  CStatusLine = #2#3#4#5#6#7;

type

{ TMenu types }

  TMenuStr = string;

  PMenu = ^TMenu;

  PMenuItem = ^TMenuItem;
  TMenuItem = record
    Next: PMenuItem;
    Name: string;
    Command: Word;
    Disabled: Boolean;
    Checked: Boolean;
    KeyCode: Word;
    HelpCtx: Word;
    Param: string;
    SubMenu: PMenu;
  end;

  TMenu = record
    Items: PMenuItem;
    Default: PMenuItem;
  end;

{ TMenuView }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Disabled text }
  { 3 = Shortcut text }
  { 4 = Normal selection }
  { 5 = Disabled selection }
  { 6 = Shortcut selection }

  TViMenuView = class(TViView)
  private
    FOnActivated: TActivatedEvent;
    FID: Integer;
    FViMode: Boolean;
  protected
    procedure DoActivated(Sender: TObject; ID: Integer); virtual;
  public
    ParentMenu: TViMenuView;
    Menu: PMenu;
    Current: PMenuItem;
    constructor Create(var Bounds: TRect; ViMode: Boolean = False);
    function Execute: Word; override;
    function FindItem(Ch: Char): PMenuItem;
    procedure GetItemRect(Item: PMenuItem; var R: TRect); virtual;
    function GetHelpCtx: Word; override;
    procedure HandleEvent(var Event: TEvent); override;
    function HotKey(KeyCode: Word): PMenuItem;
    function NewSubView(var Bounds: TRect; AMenu: PMenu;
      AParentMenu: TViMenuView): TViMenuView; virtual;
    property OnActivated: TActivatedEvent read FOnActivated write FOnActivated;
  end;

{ TMenuBar object }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Disabled text }
  { 3 = Shortcut text }
  { 4 = Normal selection }
  { 5 = Disabled selection }
  { 6 = Shortcut selection }

  TViMenuBar = class(TViMenuView)
  public
    constructor Create(var Bounds: TRect; AMenu: PMenu; ViMode: Boolean = False);
    destructor Destroy; override;
    procedure Draw; override;
    procedure GetItemRect(Item: PMenuItem; var R: TRect); override;
  end;

{ TMenuBox object }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Disabled text }
  { 3 = Shortcut text }
  { 4 = Normal selection }
  { 5 = Disabled selection }
  { 6 = Shortcut selection }

  TViMenuBox = class(TViMenuView)
  public
    constructor Create(var Bounds: TRect; AMenu: PMenu;
      AParentMenu: TViMenuView);
    procedure Draw; override;
    procedure GetItemRect(Item: PMenuItem; var R: TRect); override;
  end;

{ TMenuPopup object }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Disabled text }
  { 3 = Shortcut text }
  { 4 = Normal selection }
  { 5 = Disabled selection }
  { 6 = Shortcut selection }

  TViMenuPopup = class(TViMenuBox)
  public
    constructor Create(var Bounds: TRect; AMenu: PMenu);
    procedure HandleEvent(var Event: TEvent); override;
    destructor Destroy; override;
    procedure DoSelect;
  end;

{ TStatusItem }

  PStatusItem = ^TStatusItem;
  TStatusItem = record
    Next: PStatusItem;
    Text: string;
    KeyCode: Word;
    Command: Word;
  end;

{ TStatusDef }

  PStatusDef = ^TStatusDef;
  TStatusDef = record
    Next: PStatusDef;
    Min, Max: Word;
    Items: PStatusItem;
  end;

{ TStatusLine }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Disabled text }
  { 3 = Shortcut text }
  { 4 = Normal selection }
  { 5 = Disabled selection }
  { 6 = Shortcut selection }

  TViStatusLine = class(TViView)
  private
    procedure DrawSelect(Selected: PStatusItem);
    procedure FindItems;
  public
    Items: PStatusItem;
    Defs: PStatusDef;
    constructor Create(var Bounds: TRect; ADefs: PStatusDef);
    destructor Destroy; override;
    procedure Draw; override;
    procedure HandleEvent(var Event: TEvent); override;
    function Hint(AHelpCtx: Word): string; virtual;
    procedure Update; virtual;
  end;

{ TMenuItem routines }

function NewItem(Name, Param: TMenuStr; KeyCode: Word; Command: Word;
  AHelpCtx: Word; Next: PMenuItem): PMenuItem;
function NewLine(Next: PMenuItem): PMenuItem;
function NewSubMenu(Name: TMenuStr; AHelpCtx: Word; SubMenu: PMenu;
  Next: PMenuItem): PMenuItem;

{ TMenu routines }

function NewMenu(Items: PMenuItem): PMenu;
procedure DisposeMenu(Menu: PMenu);

{ TStatusLine routines }

function NewStatusDef(AMin, AMax: Word; AItems: PStatusItem;
  ANext: PStatusDef): PStatusDef;
function NewStatusKey(const AText: string; AKeyCode: Word; ACommand: Word;
  ANext: PStatusItem): PStatusItem;

implementation

uses
  ViApp;

{ TMenuItem routines }

function NewItem(Name, Param: TMenuStr; KeyCode: Word; Command: Word;
  AHelpCtx: Word; Next: PMenuItem): PMenuItem;
//const
//  T: TViView = nil;
var
  P: PMenuItem;
begin
  if (Name <> '') and (Command <> 0) then
  begin
    New(P);
    P^.Next := Next;
    P^.Name := Name;
    P^.Command := Command;
//    P^.Disabled := not T.CommandEnabled(Command);
    P^.Disabled := False;
    P^.Checked := False;
    P^.KeyCode := KeyCode;
    P^.HelpCtx := AHelpCtx;
    P^.Param := Param;
    NewItem := P;
  end else
  NewItem := Next;
end;

function NewLine(Next: PMenuItem): PMenuItem;
var
  P: PMenuItem;
begin
  New(P);
  P^.Next := Next;
  P^.Name := '';
  P^.HelpCtx := hcNoContext;
  NewLine := P;
end;

function NewSubMenu(Name: TMenuStr; AHelpCtx: Word; SubMenu: PMenu;
  Next: PMenuItem): PMenuItem;
var
  P: PMenuItem;
begin
  if (Name <> '') and (SubMenu <> nil) then
  begin
    New(P);
    P^.Next := Next;
    P^.Name := Name;
    P^.Command := 0;
    P^.Disabled := False;
    P^.Checked := False;
    P^.HelpCtx := AHelpCtx;
    P^.SubMenu := SubMenu;
    NewSubMenu := P;
  end else
  NewSubMenu := Next;
end;

{ TMenu routines }

function NewMenu(Items: PMenuItem): PMenu;
var
  P: PMenu;
begin
  New(P);
  P^.Items := Items;
  P^.Default := Items;
  NewMenu := P;
end;

procedure DisposeMenu(Menu: PMenu);
var
  P, Q: PMenuItem;
begin
  if Menu <> nil then
  begin
    P := Menu^.Items;
    while P <> nil do
    begin
      if P^.Name <> '' then
      begin
        if P^.Command = 0 then
          DisposeMenu(P^.SubMenu);
      end;
      Q := P;
      P := P^.Next;
      Dispose(Q);
    end;
    Dispose(Menu);
  end;
end;

{ TMenuView }

constructor TViMenuView.Create(var Bounds: TRect; ViMode: Boolean = False);
begin
  inherited Create(Bounds);
  FViMode := ViMode;
  Palette := CMenuView;
  EventMask := EventMask or evBroadcast;
end;

function TViMenuView.Execute: Word;
type
  MenuAction = (DoNothing, DoSelect, DoReturn);
var
  AutoSelect: Boolean;
  Action: MenuAction;
  Ch: Char;
  ItemShown, P: PMenuItem;
  Target: TViMenuView;
  R: TRect;
  E: TEvent;
  MouseActive: Boolean;

procedure TrackMouse;
var
  Mouse: TPoint;
  R: TRect;
begin
  MakeLocal(E.Where, Mouse);
  Current := Menu^.Items;
  while Current <> nil do
  begin
    GetItemRect(Current, R);
    if RectContains(R, Mouse) then
    begin
      MouseActive := True;
      Exit;
    end;
    Current := Current^.Next;
  end;
end;

procedure TrackKey(FindNext: Boolean);

procedure NextItem;
begin
  Current := Current^.Next;
  if Current = nil then Current := Menu^.Items;
end;

procedure PrevItem;
var
  P: PMenuItem;
begin
  P := Current;
  if P = Menu^.Items then P := nil;
  repeat NextItem until Current^.Next = P;
end;

begin
  if Current <> nil then
    repeat
      if FindNext then NextItem else PrevItem;
    until Current^.Name <> '';
end;

function MouseInOwner: Boolean;
var
  Mouse: TPoint;
  R: TRect;
begin
  MouseInOwner := False;
  if (ParentMenu <> nil) and (ParentMenu.Size.Y = 1) then
  begin
    ParentMenu.MakeLocal(E.Where, Mouse);
    ParentMenu.GetItemRect(ParentMenu.Current, R);
    MouseInOwner := RectContains(R, Mouse);
  end;
end;

function MouseInMenus: Boolean;
var
  P: TViMenuView;
begin
  P := ParentMenu;
  while (P <> nil) and not P.MouseInView(E.Where) do P := P.ParentMenu;
  MouseInMenus := P <> nil;
end;

function TopMenu: TViMenuView;
begin
  Result := Self;
  while Result.ParentMenu <> nil do Result := Result.ParentMenu;
end;

begin
  AutoSelect := False;
  Result := 0;
  FID := 0;
  ItemShown := nil;
  Current := Menu^.Default;
  MouseActive := False;
  repeat
    Action := DoNothing;
    GetEvent(E);
    case E.What of
      evMouseDown:
        if MouseInView(E.Where) or MouseInOwner then
        begin
          TrackMouse;
          if Size.Y = 1 then AutoSelect := True;
        end else Action := DoReturn;
      evMouseUp:
        begin
          TrackMouse;
          if MouseInOwner then
            Current := Menu^.Default
          else
            if (Current <> nil) and (Current^.Name <> '') then
              Action := DoSelect
            else
              if MouseActive or MouseInView(E.Where) then Action := DoReturn
              else
              begin
                Current := Menu^.Default;
                if Current = nil then Current := Menu^.Items;
                Action := DoNothing;
              end;
        end;
      evMouseMove:
        if E.Buttons <> 0 then
        begin
          TrackMouse;
          if not (MouseInView(E.Where) or MouseInOwner) and
            MouseInMenus then Action := DoReturn;
        end;
      evKeyDown:
        case CtrlToArrow(E.KeyCode) of
          kbUp, kbDown:
            if Size.Y <> 1 then
              TrackKey(CtrlToArrow(E.KeyCode) = kbDown) else
              if E.KeyCode = kbDown then AutoSelect := True;
          kbLeft, kbRight:
            if ParentMenu = nil then
              TrackKey(CtrlToArrow(E.KeyCode) = kbRight) else
              Action := DoReturn;
          kbHome, kbEnd:
            if Size.Y <> 1 then
            begin
              Current := Menu^.Items;
              if E.KeyCode = kbEnd then TrackKey(False);
            end;
          kbEnter:
            begin
              if Size.Y = 1 then AutoSelect := True;
              Action := DoSelect;
            end;
          kbEsc:
            begin
              Action := DoReturn;
              if (ParentMenu = nil) or (ParentMenu.Size.Y <> 1) then
                ClearEvent(E);
            end;
        else
          Target := Self;
          Ch := GetAltChar(E.KeyCode);
          if Ch = #0 then Ch := E.CharCode else Target := TopMenu;
          P := Target.FindItem(Ch);
          if P = nil then
          begin
            P := TopMenu.HotKey(E.KeyCode);
            if (P <> nil) and CommandEnabled(P^.Command) then
            begin
              Result := P^.Command;
              FID := Integer(P);
              Action := DoReturn;
            end
          end else
            if Target = Self then
            begin
              if Size.Y = 1 then AutoSelect := True;
              Action := DoSelect;
              Current := P;
            end else
              if (ParentMenu <> Target) or (ParentMenu.Current <> P) then
                Action := DoReturn;
        end;
      evCommand:
        if E.Command = cmMenu then
        begin
          AutoSelect := False;
          if ParentMenu <> nil then Action := DoReturn;
        end
        else if E.Command = cmValid then
          // do nothing
        else Action := DoReturn;
    end;
    if ItemShown <> Current then
    begin
      ItemShown := Current;
      DrawView;
    end;
    if (Action = DoSelect) or ((Action = DoNothing) and AutoSelect) then
      if Current <> nil then with Current^ do if Name <> '' then
        if Command = 0 then
        begin
          if E.What and (evMouseDown + evMouseMove) <> 0 then PutEvent(E);
          GetItemRect(Current, R);
          R.A.X := R.A.X + Origin.X;
          R.A.Y := R.B.Y + Origin.Y;
          R.B := Owner.Size;
          if Size.Y = 1 then Dec(R.A.X);
          Target := TopMenu.NewSubView(R, SubMenu, Self);
          Result := Owner.ExecMenuView(Target);
          FID := Target.FID;
          Target.Free;
        end else if Action = DoSelect then
        begin
          Result := Command;
          FID := Integer(Current);
        end;
//    if (Result <> 0) and CommandEnabled(Result) then
    if (FID <> 0) then
    begin
      Action := DoReturn;
      ClearEvent(E);
    end
    else
    begin
      Result := 0;
      FID := 0;
    end;
  until Action = DoReturn;
  if E.What <> evNothing then
    if (ParentMenu <> nil) or (E.What = evCommand) then PutEvent(E);
  if Current <> nil then
  begin
    Menu^.Default := Current;
    Current := nil;
    DrawView;
  end;
  Execute := Result;
end;

function TViMenuView.FindItem(Ch: Char): PMenuItem;
var
  P: PMenuItem;
  I: Integer;
begin
  Ch := UpCase(Ch);
  P := Menu^.Items;
  while P <> nil do
  begin
    if (P^.Name <> '') and not P^.Disabled then
    begin
      I := Pos('~', P^.Name);
      if (I <> 0) and (Ch = UpCase(P^.Name[I + 1])) then
      begin
        FindItem := P;
        Exit;
      end;
    end;
    P := P^.Next;
  end;
  FindItem := nil;
end;

procedure TViMenuView.GetItemRect(Item: PMenuItem; var R: TRect);
begin
end;

function TViMenuView.GetHelpCtx: Word;
var
  C: TViMenuView;
begin
  C := Self;
  while (C <> nil) and
     ((C.Current = nil) or (C.Current^.HelpCtx = hcNoContext) or
      (C.Current^.Name = '')) do
    C := C.ParentMenu;
  if C <> nil then GetHelpCtx := C.Current^.HelpCtx
  else GetHelpCtx := hcNoContext;
end;

procedure TViMenuView.HandleEvent(var Event: TEvent);
var
  CallDraw: Boolean;
  P: PMenuItem;

procedure UpdateMenu(Menu: PMenu);
var
  P: PMenuItem;
  CommandState: Boolean;
begin
  P := Menu^.Items;
  while P <> nil do
  begin
    if P^.Name <> '' then
      if P^.Command = 0 then UpdateMenu(P^.SubMenu)
      else
      begin
        CommandState := CommandEnabled(P^.Command);
        if P^.Disabled = CommandState then
        begin
//          P^.Disabled := not CommandState;
          CallDraw := True;
        end;
      end;
    P := P^.Next;
  end;
end;

procedure DoSelect;
begin
  PutEvent(Event);
  Event.Command := Owner.ExecMenuView(Self);
//  if (Event.Command <> 0) and CommandEnabled(Event.Command) then
  if (Event.Command <> 0) then
  begin
    DoActivated(Self, FID);
    if FViMode then
    begin
      Event.What := evCommand;
      Event.InfoPtr := nil;
      PutEvent(Event);
    end;
  end;
  ClearEvent(Event);
end;

begin
  if Menu <> nil then
    case Event.What of
      evMouseDown:
        DoSelect;
      evKeyDown:
        if (FindItem(GetAltChar(Event.KeyCode)) <> nil) then
          DoSelect
        else
        begin
          P := HotKey(Event.KeyCode);
          if (P <> nil) and (CommandEnabled(P^.Command)) then
          begin
            Event.What := evCommand;
            Event.Command := P^.Command;
            Event.InfoPtr := nil;
            PutEvent(Event);
            ClearEvent(Event);
          end;
        end;
      evCommand:
        if Event.Command = cmMenu then DoSelect;
      evBroadcast:
        if Event.Command = cmCommandSetChanged then
        begin
          CallDraw := False;
          UpdateMenu(Menu);
          if CallDraw then DrawView;
        end;
    end;
end;

function TViMenuView.HotKey(KeyCode: Word): PMenuItem;

function FindHotKey(P: PMenuItem): PMenuItem;
var
  T: PMenuItem;
begin
  while P <> nil do
  begin
    if P^.Name <> '' then
      if P^.Command = 0 then
      begin
        T := FindHotKey(P^.SubMenu^.Items);
        if T <> nil then
        begin
          FindHotKey := T;
          Exit;
        end;
      end
      else if not P^.Disabled and (P^.KeyCode <> kbNoKey) and
        (P^.KeyCode = KeyCode) then
      begin
        FindHotKey := P;
        Exit;
      end;
    P := P^.Next;
  end;
  FindHotKey := nil;
end;

begin
  HotKey := FindHotKey(Menu^.Items);
end;

function TViMenuView.NewSubView(var Bounds: TRect; AMenu: PMenu; 
    AParentMenu: TViMenuView): TViMenuView;
begin
  Result := TViMenuBox.Create(Bounds, AMenu, AParentMenu);
end;

procedure TViMenuView.DoActivated(Sender: TObject; ID: Integer);
begin
  if Assigned(OnActivated) then OnActivated(Self, ID);
end;


{ TMenuBar }

constructor TViMenuBar.Create(var Bounds: TRect; AMenu: PMenu; ViMode: Boolean 
    = False);
begin
  inherited Create(Bounds, ViMode);
  GrowMode := gfGrowHiX;
  Menu := AMenu;
  Options := Options or ofPreProcess;
end;

destructor TViMenuBar.Destroy;
begin
  DisposeMenu(Menu);
  inherited Destroy;
end;

procedure TViMenuBar.Draw;
var
  X, L: Integer;
  CNormal, CSelect, CNormDisabled, CSelDisabled, Color: Word;
  P: PMenuItem;
  B: TDrawBuffer;
begin
  CNormal := GetColor($0301);
  CSelect := GetColor($0604);
  CNormDisabled := GetColor($0202);
  CSelDisabled := GetColor($0505);
  MoveChar(B, ' ', Byte(CNormal), Size.X);
  if Menu <> nil then
  begin
    X := 1;
    P := Menu^.Items;
    while P <> nil do
    begin
      if P^.Name <> '' then
      begin
        L := CStrLen(P^.Name);
        if X + L < Size.X then
        begin
          if P^.Disabled then
            if P = Current then
              Color := CSelDisabled else
              Color := CNormDisabled else
            if P = Current then
              Color := CSelect else
              Color := CNormal;
          if P^.Checked then
            MoveChar(B[X], '*', Byte(Color), 1)
          else
            MoveChar(B[X], ' ', Byte(Color), 1);
          MoveCStr(B[X + 1], P^.Name, Color);
          MoveChar(B[X + L + 1], ' ', Byte(Color), 1);
        end;
        Inc(X, L + 2);
      end;
      P := P^.Next;
    end;
  end;
  WriteBuf(0, 0, Size.X, 1, B);
end;

procedure TViMenuBar.GetItemRect(Item: PMenuItem; var R: TRect);
var
  P: PMenuItem;
begin
  RectAssign(R, 1, 0, 1, 1);
  P := Menu^.Items;
  while True do
  begin
    R.A.X := R.B.X;
    if P^.Name <> '' then Inc(R.B.X, CStrLen(P^.Name)+2);
    if P = Item then Exit;
    P := P^.Next;
  end;
end;

{ TMenuBox }

constructor TViMenuBox.Create(var Bounds: TRect; AMenu: PMenu; AParentMenu: 
    TViMenuView);
var
  W, H, L: Integer;
  P: PMenuItem;
  R: TRect;
begin
  W := 10;
  H := 2;
  if AMenu <> nil then
  begin
    P := AMenu^.Items;
    while P <> nil do
    begin
      if P^.Name <> '' then
      begin
        L := CStrLen(P^.Name) + 6;
        if P^.Command = 0 then Inc(L, 3) else
          if P^.Param <> '' then Inc(L, CStrLen(P^.Param) + 2);
        if L > W then W := L;
      end;
      Inc(H);
      P := P^.Next;
    end;
  end;
  RectCopy(R, Bounds);
  if R.A.X + W < R.B.X then R.B.X := R.A.X + W else R.A.X := R.B.X - W;
  if R.A.Y + H < R.B.Y then R.B.Y := R.A.Y + H else R.A.Y := R.B.Y - H;
  inherited Create(R);
  State := State or sfShadow;
  Options := Options or ofPreProcess;
  Menu := AMenu;
  ParentMenu := AParentMenu;
end;

procedure TViMenuBox.Draw;
var
  CNormal, CSelect, CNormDisabled, CSelDisabled, Color: Word;
  Y: Integer;
  P: PMenuItem;
  B: TDrawBuffer;

procedure FrameLine(N: Integer);
begin
  MoveBuf(B[0], ldMenuFrameChars[N], Byte(CNormal), 2);
  MoveChar(B[2], ldMenuFrameChars[N + 2], Byte(Color), Size.X - 4);
  MoveBuf(B[Size.X - 2], ldMenuFrameChars[N + 3], Byte(CNormal), 2);
end;

procedure DrawLine;
begin
  WriteBuf(0, Y, Size.X, 1, B);
  Inc(Y);
end;

begin
  CNormal := GetColor($0301);
  CSelect := GetColor($0604);
  CNormDisabled := GetColor($0202);
  CSelDisabled := GetColor($0505);
  Y := 0;
  Color := CNormal;
  FrameLine(0);
  DrawLine;
  if Menu <> nil then
  begin
    P := Menu^.Items;
    while P <> nil do
    begin
      Color := CNormal;
      if P^.Name = '' then FrameLine(15) else
      begin
        if P^.Disabled then
          if P = Current then
            Color := CSelDisabled else
            Color := CNormDisabled else
          if P = Current then Color := CSelect;
        FrameLine(10);
        MoveCStr(B[3], P^.Name, Color);
        if P^.Checked then
          MoveChar(B[2], #7, Byte(Color), 1);
        if P^.Command = 0 then
          MoveChar(B[Size.X - 4], ldSubmenuArrow, Byte(Color), 1) else
          if P^.Param <> '' then
            MoveStr(B[Size.X - 3 - Length(P^.Param)],
              P^.Param, Byte(Color));
      end;
      DrawLine;
      P := P^.Next;
    end;
  end;
  Color := CNormal;
  FrameLine(5);
  DrawLine;
end;

procedure TViMenuBox.GetItemRect(Item: PMenuItem; var R: TRect);
var
  Y: Integer;
  P: PMenuItem;
begin
  Y := 1;
  P := Menu^.Items;
  while P <> Item do
  begin
    Inc(Y);
    P := P^.Next;
  end;
  RectAssign(R, 2, Y, Size.X - 2, Y + 1);
end;


constructor TViMenuPopup.Create(var Bounds: TRect; AMenu: PMenu);
begin
  inherited Create(Bounds, AMenu, nil);
end;

procedure TViMenuPopup.HandleEvent(var Event: TEvent);
var
  P: PMenuItem;
begin
  case Event.What of
    evKeyDown:
      begin
        P := FindItem(GetCtrlChar(Event.KeyCode));
        if P = nil then
          P := HotKey(Event.KeyCode);
        if (P <> nil) and (CommandEnabled(P^.Command)) then
        begin
          Event.What := evCommand;
          Event.Command := P^.Command;
          Event.InfoPtr := nil;
          PutEvent(Event);
          ClearEvent(Event);
        end
        else
          if GetAltChar(Event.KeyCode) <> #0 then
            ClearEvent(Event);
      end;
  end;
  inherited HandleEvent(Event);
end;

destructor TViMenuPopup.Destroy;
begin
  DisposeMenu(Menu);
  inherited Destroy;
end;

procedure TViMenuPopup.DoSelect;
var
  Command: Word;
begin
//  PutEvent(Event);
  Command := ViApplication.ExecMenuView(Self);
  if (Command <> 0) then
  begin
    DoActivated(Self, FID);
  end;
//  ClearEvent(Event);
end;


{ TStatusLine }

constructor TViStatusLine.Create(var Bounds: TRect; ADefs: PStatusDef);
begin
  inherited Create(Bounds);
  Palette := CStatusLine;
  Options := Options or ofPreProcess;
  EventMask := EventMask or evBroadcast;
  GrowMode := gfGrowLoY + gfGrowHiX + gfGrowHiY;
  Defs := ADefs;
  FindItems;
end;

destructor TViStatusLine.Destroy;
var
  T: PStatusDef;

  procedure DisposeItems(Item: PStatusItem);
  var
    T: PStatusItem;
  begin
    while Item <> nil do
    begin
      T := Item;
      Item := Item^.Next;
      Dispose(T);
    end;
  end;

begin
  while Defs <> nil do
  begin
    T := Defs;
    Defs := Defs^.Next;
    DisposeItems(T^.Items);
    Dispose(T);
  end;
  inherited Destroy;
end;

procedure TViStatusLine.Draw;
begin
  DrawSelect(nil);
end;

procedure TViStatusLine.DrawSelect(Selected: PStatusItem);
var
  B: TDrawBuffer;
  T: PStatusItem;
  I, L: Integer;
  CSelect, CNormal, CSelDisabled, CNormDisabled: Word;
  Color: Word;
  HintBuf: string;
begin
  CNormal := GetColor($0301);
  CSelect := GetColor($0604);
  CNormDisabled := GetColor($0202);
  CSelDisabled := GetColor($0505);
  MoveChar(B, ' ', Byte(CNormal), Size.X);
  T := Items;
  I := 0;
  while T <> nil do
  begin
    if T^.Text <> '' then
    begin
      L := CStrLen(T^.Text);
      if I + L - 1 < Size.X then
      begin
        if CommandEnabled(T^.Command) then
          if T = Selected then
            Color := CSelect else
            Color := CNormal else
          if T = Selected then
            Color := CSelDisabled else
            Color := CNormDisabled;
        MoveCStr(B[I], T^.Text, Color);
        MoveChar(B[I + L], ' ', Byte(CNormal), 1);
      end;
      Inc(I, L + 1);
    end;
    T := T^.Next;
  end;
  if I < Size.X - 2 then
  begin
    HintBuf := Hint(HelpCtx);
    if HintBuf <> '' then
    begin
      MoveChar(B[I], ldVerticalBar, Byte(CNormal), 1);
      Inc(I, 2);
      if I + Length(HintBuf) > Size.X then SetLength(HintBuf, Size.X - I);
      MoveStr(B[I], HintBuf, Byte(CNormal));
    end;
  end;
  WriteLine(0, 0, Size.X, 1, B);
end;

procedure TViStatusLine.FindItems;
var
  P: PStatusDef;
begin
  P := Defs;
  while (P <> nil) and ((HelpCtx < P^.Min) or (HelpCtx > P^.Max)) do
    P := P^.Next;
  if P = nil then Items := nil else Items := P^.Items;
end;

procedure TViStatusLine.HandleEvent(var Event: TEvent);
var
  Mouse: TPoint;
  T: PStatusItem;

function ItemMouseIsIn: PStatusItem;
var
  I,K: Word;
  T: PStatusItem;
begin
  ItemMouseIsIn := nil;
  if Mouse.Y <> 0 then Exit;
  I := 0;
  T := Items;
  while T <> nil do
  begin
    if T^.Text <> '' then
    begin
      K := I + CStrLen(T^.Text) + 1;
      if (Mouse.X >= I) and (Mouse.X < K) then
      begin
        ItemMouseIsIn := T;
        Exit;
      end;
      I := K;
    end;
    T := T^.Next;
  end;
end;

begin
  inherited HandleEvent(Event);
  case Event.What of
    evMouseDown:
      begin
        T := nil;
        repeat
          MakeLocal(Event.Where, Mouse);
          if T <> ItemMouseIsIn then
          begin
            T := ItemMouseIsIn;
            DrawSelect(T);
          end;
        until not MouseEvent(Event, evMouseMove);
        if (T <> nil) and CommandEnabled(T^.Command) then
        begin
          Event.What := evCommand;
          Event.Command := T^.Command;
          Event.InfoPtr := nil;
          PutEvent(Event);
        end;
        ClearEvent(Event);
        DrawView;
      end;
    evKeyDown:
      begin
        T := Items;
        while T <> nil do
        begin
          if (Event.KeyCode = T^.KeyCode) and
            CommandEnabled(T^.Command) then
          begin
            Event.What := evCommand;
            Event.Command := T^.Command;
            Event.InfoPtr := nil;
            Exit;
          end;
          T := T^.Next;
        end;
      end;
    evBroadcast:
      if Event.Command = cmCommandSetChanged then DrawView;
  end;
end;

function TViStatusLine.Hint(AHelpCtx: Word): string;
begin
  Hint := '';
end;

procedure TViStatusLine.Update;
var
  H: Word;
  P: TViView;
begin
  P := TopView;
  if P <> nil then
    H := P.GetHelpCtx else
    H := hcNoContext;
  if HelpCtx <> H then
  begin
    HelpCtx := H;
    FindItems;
    DrawView;
  end;
end;

function NewStatusDef(AMin, AMax: Word; AItems: PStatusItem;
  ANext:PStatusDef): PStatusDef;
var
  T: PStatusDef;
begin
  New(T);
  with T^ do
  begin
    Next := ANext;
    Min := AMin;
    Max := AMax;
    Items := AItems;
  end;
  NewStatusDef := T;
end;

function NewStatusKey(const AText: string; AKeyCode: Word; ACommand: Word;
  ANext: PStatusItem): PStatusItem;
var
  T: PStatusItem;
begin
  New(T);
  T^.Text := AText;
  T^.KeyCode := AKeyCode;
  T^.Command := ACommand;
  T^.Next := ANext;
  NewStatusKey := T;
end;

end.
