
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViMessages.pas                                        }
{                                                       }
{ for Delphi 4 - 7                                      }
{ Version 2.2                                           }
{ Copyright (c) 2003 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViMessages.pas,v 1.1 2003/02/04 17:24:36 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViMessages.pas,v $
  Revision 1.1  2003/02/04 17:24:36  iwache
  Message loop to handle Thread synchronization added.
  This is needed because TurboVision is not always thread
  safe.

  ----------------------------------------------------------------------------
}
unit ViMessages;

interface

uses
  SysUtils, Classes;

const
  WM_USER             = $0400;

type

{ Generic window message record }

  PMessage = ^TMessage;
  TMessage = packed record
    Msg: Cardinal;
    case Integer of
      0: (
        WParam: Longint;
        LParam: Longint;
        Result: Longint);
      1: (
        WParamLo: Word;
        WParamHi: Word;
        LParamLo: Word;
        LParamHi: Word;
        ResultLo: Word;
        ResultHi: Word);
  end;

function ProcessMessage: Boolean;

function RegisterHandleObject(AObject: TObject): THandle;

procedure UnregisterHandleObject(AHandle: THandle);

function PostMessage(AHandle: THandle; AMsg: Cardinal; AWParam: LongInt;
    ALParam: LongInt): Boolean;

function SendMessage(AHandle: THandle; AMsg: Cardinal; AWParam: LongInt;
    ALParam: LongInt): LongInt;

procedure ProcessMessages;

implementation

resourcestring
  SHandleObjectNotRegistered = 'Handle object not registered';
  SMsgEventObjectAlreadyRegistered = 'MsgEvent object already registered';
  SMsgEventObjectNotRegistered = 'MsgEvent object not registered';

const
  MIN_HANDLE = 1024;

  MsgBufferSize =64;

type
  THandle = Cardinal;

  { Generic window message record }

  THandleMessage = record
    Handle: THandle;
    Msg: TMessage;
  end;

  PHandleObject = ^THandleObject;
  THandleObject = record
    Obj: TObject;
    Handle: THandle;
  end;

var
  FHandleObjects: TList;
  FMsgBuffer: array[0..MsgBufferSize -1] of THandleMessage;
  FMsgCount: Integer;
  FMsgHead: Integer;
  FMsgTail: Integer;
  FNextHandle: THandle;

{ local functions }

procedure FreeMsgBuffer;
begin
  { Ende des MsgBuffers freigeben }
  Inc(FMsgTail);
  if FMsgTail =MsgBufferSize then FMsgTail :=0;
  { Puffer-Zhler verringern }
  Dec(FMsgCount);
end;

procedure AllocMsgBuffer;
begin
  { Msgzeiger fr nchsten freien Platz erhhen }
  Inc(FMsgHead);
  if FMsgHead =MsgBufferSize then FMsgHead :=0;
  { Puffer-Zhler erhhen }
  Inc(FMsgCount);
end;

function GetNextHandle: THandle;
begin
  Result := FNextHandle;
  Inc(FNextHandle);
  if (FNexthandle < MIN_HANDLE) or (FNextHandle = High(THandle)) then
    FNextHandle := MIN_HANDLE;
end;

function FindHandleObject(AObject: TObject): Integer; overload;
var
  I: Integer;
  AHandleObject: PHandleObject;
begin
  Result := - 1;
  for I := 0 to FHandleObjects.Count - 1 do
  begin
    AHandleObject := FHandleObjects.Items[I];
    if AHandleObject.Obj = AObject then
    begin
      Result := I;
      Break;
    end;
  end;
end;

function FindHandleObject(AHandle: THandle): Integer; overload;
var
  I: Integer;
  AHandleObject: PHandleObject;
begin
  Result := - 1;
  for I := 0 to FHandleObjects.Count - 1 do
  begin
    AHandleObject := FHandleObjects.Items[I];
    if AHandleObject.Handle = AHandle then
    begin
      Result := I;
      Break;
    end;
  end;
end;

function PeekMessage(var AHandle: THandle; var AMsg: TMessage): Boolean;
begin
  Result :=False;
  if FMsgCount >0 then { get Message }
  begin
    with FMsgBuffer[FMsgTail] do
    begin
      AHandle :=Handle;
      AMsg :=Msg;
    end;
    Result :=True;
    FreeMsgBuffer;
  end
end;

procedure DispatchMessage(AHandle: THandle; var AMsg: TMessage);
var
  Index: Integer;
  AHandleObject: PHandleObject;
begin
  Index := FindHandleObject(AHandle);
  if Index <> -1 then
  begin
    AHandleObject := FHandleObjects.Items[Index];
    AHandleObject.Obj.Dispatch(AMsg);
  end;
end;

function ProcessMessage: Boolean;
var
  Msg: TMessage;
  Handle: THandle;
begin
  Result := PeekMessage(Handle, Msg);
  if Result then
    DispatchMessage(Handle, Msg);
end;

function RegisterHandleObject(AObject: TObject): THandle;
var
  AHandleObject: PHandleObject;
begin
  Result := High(THandle);
  if not Assigned(AObject) then Exit;
  if FindHandleObject(AObject) <> -1 then
    raise Exception.Create(SMsgEventObjectAlreadyRegistered)
  else
  begin
    Result := GetNextHandle;
    New(AHandleObject);
    AHandleObject.Obj := AObject;
    AHandleObject.Handle := Result;
    FHandleObjects.Add(AHandleObject);
  end;
end;

procedure UnregisterHandleObject(AHandle: THandle);
var
  Index: Integer;
  AHandleObject: PHandleObject;
begin
  Index := FindHandleObject(AHandle);
  if Index = -1 then
    raise Exception.Create(SHandleObjectNotRegistered);
  AHandleObject := FHandleObjects.Items[Index];
  Dispose(AHandleObject);
  FHandleObjects.Delete(Index);
end;

function PostMessage(AHandle: THandle; AMsg: Cardinal; AWParam: LongInt;
    ALParam: LongInt): Boolean;
begin
  Result := False;
  { Prfung, ob Msg-Puffer voll ist }
  if FMsgCount =MsgBufferSize then Exit;
  { Msg-Puffer fllen }
  with FMsgBuffer[FMsgHead] do
  begin
    Handle :=AHandle;
    with Msg do
    begin
      Msg :=AMsg;
      WParam :=AWParam;
      LParam :=ALParam;
      Result :=0;
    end;
  end;
  { Position im Msg-Puffer belegen }
  AllocMsgBuffer;
  Result :=True;
end;

function SendMessage(AHandle: THandle; AMsg: Cardinal; AWParam: LongInt;
    ALParam: LongInt): LongInt;
var
  AMessage: TMessage;
begin
  with AMessage do
  begin
    Msg :=AMsg;
    WParam :=AWParam;
    LParam :=ALParam;
    Result :=0;
  end;
  DispatchMessage(AHandle, AMessage);
  Result :=AMessage.Result;
end;

procedure DeleteHandleObjects;
var
  I: Integer;
  AHandleObject: PHandleObject;
begin
  for I := 0 to FHandleObjects.Count - 1 do
  begin
    AHandleObject := FHandleObjects.Items[I];
    Dispose(AHandleObject);
  end;
end;

procedure ProcessMessages;
begin
  while ProcessMessage do;
end;

initialization
  FNextHandle := MIN_HANDLE;
  FHandleObjects := TList.Create;

finalization
  DeleteHandleObjects;
  FreeAndNil(FHandleObjects);
end.



