
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViMouses.pas                                          }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViMouses.pas,v 1.4 2003/10/28 20:16:51 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViMouses.pas,v $
  Revision 1.4  2003/10/28 20:16:51  iwache
  Procedure TViMouse.Init reads now the current mouse position
  into property Where. This avoids invisible mouse cursor at
  program startup.

  Revision 1.3  2002/10/27 22:25:19  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-05-24 Version 2.0.1.0 released
}

unit ViMouses;

interface

{$IFDEF WIN}
uses
  SysUtils, WinMouses, ViTypes;
{$ENDIF}
{$IFDEF WDOSX}
uses
  SysUtils, WDosMouses, ViTypes;
{$ENDIF}

type
  TViMouse = class(TObject)
  private
    FLastButtons: Integer;
    FDownButtons: Integer;
    FLastDouble: Boolean;
    FDownWhere: TPoint;
    FDownTicks: Integer;
    FAutoTicks: Integer;
    FAutoDelay: Integer;
    FReverse: Boolean;
    FDoubleDelay: Integer;
    FWhere: TPoint;
    FButtonCount: Integer;
    FButtons: Byte;
    FDetected: Boolean;
    FRepeatDelay: Integer;
  public
    procedure Hide;
    procedure Show;
    constructor Create;
    destructor Destroy; override;
    procedure GetEvent(var Event: TEvent);
    procedure ProcessEvents;
    procedure Done;
    procedure Init;
    property ButtonCount: Integer read FButtonCount;
    property Reverse: Boolean read FReverse write FReverse;
    property DoubleDelay: Integer read FDoubleDelay write FDoubleDelay;
    property Where: TPoint read FWhere;
    property Buttons: Byte read FButtons;
    property Detected: Boolean read FDetected;
    property RepeatDelay: Integer read FRepeatDelay write FRepeatDelay;
  end;

var
  ViMouse: TViMouse;

implementation

procedure TViMouse.Hide;
begin
  if FDetected then HideMouse;
end;

procedure TViMouse.Show;
begin
  if FDetected then ShowMouse;
end;

constructor TViMouse.Create;
begin
  inherited Create;
  FReverse := False;
  FDoubleDelay := 8;
  FRepeatDelay := 8;
end;

destructor TViMouse.Destroy;
begin
  inherited Destroy;
end;

procedure TViMouse.GetEvent(var Event: TEvent);
var
  CurTicks: Integer;

  procedure StoreEvent(EventWhat: Word);
  begin
    FLastButtons := FButtons;
    FWhere.X := MouseWhereX;
    FWhere.Y := MouseWhereY;
    with Event do
    begin
      What := EventWhat;
      Buttons := FButtons;
      Double := FLastDouble;
      Where.X := FWhere.X;
      Where.Y := FWhere.Y;
    end;
  end;

begin
  if not Detected then
    Event.What := evNothing
  else
  begin
    if FReverse then
    begin
      FButtons := 0;
      if (MouseButtons and 1) <> 0 then
        Inc(FButtons, 2);
      if (MouseButtons and 2) <> 0 then
        Inc(FButtons, 1);
    end
    else FButtons := MouseButtons;
    // ms -> ticks: 1 DOS timer tick = 55ms
    CurTicks := MouseTime div 55;
    // Process mouse event
    if (FLastButtons <> 0) and (Buttons = 0) then
      StoreEvent(evMouseUp) // button is released
    else if FLastButtons = FButtons then
    begin
      if (MouseWhereY <> FWhere.Y) or (MouseWhereX <> FWhere.X) then
        StoreEvent(evMouseMove)
      else if (FButtons <> 0) and ((CurTicks - FAutoTicks) >= FAutoDelay) then
      begin
        FAutoTicks := CurTicks;
        FAutoDelay := 1;
        StoreEvent(evMouseAuto);
      end
      else
        StoreEvent(evNothing);
    end
    else // CurButton <> 0, LastButton = 0
    begin
      FLastDouble := False;
      if (FButtons = FDownButtons)
          and (MouseWhereY = FDownWhere.Y)
          and (MouseWhereX = FDownWhere.X)
          and ((CurTicks - FDownTicks) < FDoubleDelay) then
        FLastDouble := True;
      FDownButtons := FButtons;
      FDownWhere.Y := MouseWhereY;
      FDownWhere.X := MouseWhereX;
      FDownTicks := CurTicks;
      FAutoTicks := CurTicks;
      FAutoDelay := FRepeatDelay;
      StoreEvent(evMouseDown);
    end;
  end;
end;

procedure TViMouse.ProcessEvents;
begin
  ProcessConsoleEvents;
end;

procedure TViMouse.Done;
begin
  DoneMouse;
end;

procedure TViMouse.Init;
begin
  InitMouse;
  FButtonCount := MouseButtonCount;
  FDetected := FButtonCount > 0;
  ProcessConsoleEvents;
  FWhere.X := MouseWhereX;
  FWhere.Y := MouseWhereY;
end;


initialization
  ViMouse := TViMouse.Create;

finalization
  FreeAndNil(ViMouse);

end.

