
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViMsgBox.pas                                          }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViMsgBox.pas,v 1.4 2003/12/01 21:02:21 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViMsgBox.pas,v $
  Revision 1.4  2003/12/01 21:02:21  iwache
  Function MessageBoxRect now handles correct buttons
  with actions for cmAbort, cmRetry and cmIgnore.

  Revision 1.3  2003/01/23 19:21:18  iwache
  - MessageBoxRect and InputBoxRect create now
  TViDialog and TViButtons with the new option ofViInternal
  because these objects are not connected to VisualCLX
  components. So the terribly enter key bug should be fixed
  here too.

  Revision 1.2  2002/10/27 22:25:19  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-05-24 Version 2.0.1.0 released
}

unit ViMsgBox;

{$X+,I-}

interface

uses
  ViTypes, ViUtils, ViClasses; 

const

{ Message box classes }

  mfWarning      = $0000;       { Display a Warning box }
  mfError        = $0001;       { Dispaly a Error box }
  mfInformation  = $0002;       { Display an Information Box }
  mfConfirmation = $0003;       { Display a Confirmation Box }

  mfInsertInApp  = $0080;       { Insert message box into application }
                                { instead of the Desktop }

{ Message box button flags }

  mfYesButton    = $0100;       { Put a Yes button into the dialog }
  mfNoButton     = $0200;       { Put a No button into the dialog }
  mfOKButton     = $0400;       { Put an OK button into the dialog }
  mfCancelButton = $0800;       { Put a Cancel button into the dialog }

  mfAbort        = $1000;
  mfRetry        = $2000;
  mfIgnore       = $4000;

  mfYesNoCancel  = mfYesButton + mfNoButton + mfCancelButton;
                                { Standard Yes, No, Cancel dialog }
  mfOKCancel     = mfOKButton + mfCancelButton;
                                { Standard OK, Cancel dialog }

{ MessageBox displays the given string in a standard sized      }
{ dialog box. The string is displayed as a TStaticText view     }
{ in the dialog.                                                }

function MessageBox(const Msg: string; AOptions: Word): Word;

{ MessageBoxRec allows the specification of a TRect for the     }
{ message box to occupy.                                        }

function MessageBoxRect(var R: TRect; const Msg: string; AOptions: Word): 
    Word;

{ InputBox displays a simple dialog that allows the user to     }
{ type in a string.                                             }

function InputBox(const Title, ALabel: string; var S: string; Limit: 
    Integer): Word;

{ InputBoxRect is like InputBox but allows the specification of }
{ a rectangle.                                                  }

function InputBoxRect(var Bounds: TRect; const Title, ALabel: string; var 
    S: string;  Limit: Integer): Word;

implementation

uses
  ViDrivers, ViViews, ViDialogs, ViApp;

function MessageBox(const Msg: string; AOptions: Word): Word;
var
  R: TRect;
begin
  RectAssign(R, 0, 0, 40, 9);
  if AOptions and mfInsertInApp = 0 then
    RectMove(R, (ViDesktop.Size.X - R.B.X) div 2,
        (ViDesktop.Size.Y - R.B.Y) div 2)
  else RectMove(R, (ViApplication.Size.X - R.B.X) div 2,
      (ViApplication.Size.Y - R.B.Y) div 2);
  MessageBox := MessageBoxRect(R, Msg, AOptions);
end;

function MessageBoxRect(var R: TRect; const Msg: string; AOptions: Word): 
    Word;
const
  ButtonName: array[0..6] of string =
    ('~Y~es', '~N~o', 'O~K~', 'Cancel', '~A~bort', '~R~etry', '~I~gnore');
  Commands: array[0..6] of Word =
    (cmYes, cmNo, cmOK, cmCancel, cmAbort, cmRetry, cmIgnore);
  Titles: array[0..3] of string =
    ('Warning', 'Error', 'Information', 'Confirm');
var
  I, X, ButtonCount: Integer;
  Dialog: TViDialog;
  Control: TViView;
  ButtonList: array[0..4] of TViView;
begin
  Dialog := TViDialog.Create(R, Titles[AOptions and 3]);
  Dialog.Options := Dialog.Options + ofViInternal;
  try
    with Dialog do
    begin
      RectAssign(R, 3, 2, Size.X - 2, Size.Y - 3);
      Control := TViStaticText.Create(R, Msg);
      Insert(Control);
      X := -2;
      ButtonCount := 0;
      for I := 0 to 6 do
        if AOptions and ($0100 shl I) <> 0 then
        begin
          RectAssign(R, 0, 0, 10, 2);
          Control := TViButton.Create(R, ButtonName[I], Commands[i],
            bfNormal);
          Control.Options := Control.Options + ofViInternal;
          Inc(X, Control.Size.X + 2);
          ButtonList[ButtonCount] := Control;
          Inc(ButtonCount);
        end;
      X := (Size.X - X) shr 1;
      for I := 0 to ButtonCount - 1 do
      begin
        Control := ButtonList[I];
        Insert(Control);
        Control.MoveTo(X, Size.Y - 3);
        Inc(X, Control.Size.X + 2);
      end;
      SelectNext(False);
    end;
    if AOptions and mfInsertInApp = 0 then
      MessageBoxRect := ViDeskTop.ExecView(Dialog)
    else MessageBoxRect := ViApplication.ExecView(Dialog);
  finally
    Dialog.Free;
  end;
end;

function InputBox(const Title, ALabel: string; var S: string; Limit: 
    Integer): Word;
var
  R: TRect;
begin
  RectAssign(R, 0, 0, 60, 8);
  RectMove(R, (ViDesktop.Size.X - R.B.X) div 2,
      (ViDesktop.Size.Y - R.B.Y) div 2);
  InputBox := InputBoxRect(R, Title, ALabel, S, Limit);
end;

function InputBoxRect(var Bounds: TRect; const Title, ALabel: string; var 
    S: string;  Limit: Integer): Word;
var
  Dialog: TViDialog;
  Control: TViInputLine;
  Button: TViButton;
  R: TRect;
  C: Word;
begin
  Dialog := TViDialog.Create(Bounds, Title);
  Dialog.Options := Dialog.Options + ofViInternal;
  try
    with Dialog do
    begin
      RectAssign(R, 4 + CStrLen(ALabel), 2, Size.X - 3, 3);
      Control := TViInputLine.Create(R, Limit);
      Insert(Control);
      RectAssign(R, 2, 2, 3 + CStrLen(ALabel), 3);
      Insert(TViLabel.Create(R, ALabel, Control));
      RectAssign(R, Size.X - 24, Size.Y - 4, Size.X - 14, Size.Y - 2);
      Button := TViButton.Create(R, 'O~K~', cmOk, bfDefault);
      Button.Options := Button.Options + ofViInternal;
      Insert(Button);
      Inc(R.A.X, 12); Inc(R.B.X, 12);
      Button := TViButton.Create(R, 'Cancel', cmCancel, bfNormal);
      Button.Options := Button.Options + ofViInternal;
      Insert(Button);
      Inc(R.A.X, 12); Inc(R.B.X, 12);
      SelectNext(False);
    end;
    Control.Data := S;
    C := ViDeskTop.ExecView(Dialog);
    if C <> cmCancel then S := Control.Data;
  finally
    Dialog.Free;
  end;
  InputBoxRect := C;
end;

end.
