
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViTypes.pas                                           }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViTypes.pas,v 1.6 2003/01/25 11:21:16 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViTypes.pas,v $
  Revision 1.6  2003/01/25 11:21:16  iwache
  DWPL-Bugs-673813:
  Description:
  The dblclick event of TEdit and TListBox is invalid.
  The ReadOnly property of TEdit is invalid.

  Solution for Doubleclick:
  DblClick Event handler added in TWidgetControl.EventFilter.
  TCustomListBox.DblClick modified to fit behaviour with
  the TV-ListBox on doubleclick. An other solution would
  be fit behaiour of TV-ListBox to the VisualCLX one. But this
  would need much more work and tests.

  Solution for ReadOnly property of TEdit:
  TViInputLine added two new properties 'ReadOnly' and 'Text'.
  TViInputLine.HandleEvent and TViInputLine.SetData modified
  to handle ReadOnly property right.

  WQ.pas: modified for both bugs.
  ViTypes.pas: event code evMouseDbl  = $0800; added.
  This is a dummy TV event code.

  Revision 1.5  2003/01/20 22:09:26  iwache
  New Vi const kbSpace and WQ const Key_Space added.
  Both for handling space bar with focused buttons.

  Revision 1.4  2002/10/27 22:25:22  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-05-24 Version 2.0.1.0 released
}

unit ViTypes;

interface

uses
  Windows;
  
const

{ echo modes }
  emNormal   = 0;
  emNone     = 1;
  emPassword = 2;

{ Event codes }

  evMouseDown = $0001;
  evMouseUp   = $0002;
  evMouseMove = $0004;
  evMouseAuto = $0008;
  evKeyDown   = $0010;
  evKeyUp     = $0020;
  evCommand   = $0100;
  evBroadcast = $0200;
  evKeyPress  = $0400;

  evMouseDbl  = $0800;  // this is a dummy TV event code

{ Event masks }

  evNothing   = $0000;
  evMouse     = $000F;
  evKeyboard  = $0030;
  evMessage   = $FF00;

{ Extended key codes }

  kbEsc       = $011B;  kbAltSpace  = $0200;  kbCtrlIns   = $0400;
  kbShiftIns  = $0500;  kbCtrlDel   = $0600;  kbShiftDel  = $0700;
  kbBack      = $0E08;  kbCtrlBack  = $0E7F;  kbShiftTab  = $0F00;
  kbTab       = $0F09;  kbAltQ      = $1000;  kbAltW      = $1100;
  kbAltE      = $1200;  kbAltR      = $1300;  kbAltT      = $1400;
  kbAltY      = $1500;  kbAltU      = $1600;  kbAltI      = $1700;
  kbAltO      = $1800;  kbAltP      = $1900;  kbCtrlEnter = $1C0A;
  kbEnter     = $1C0D;  kbAltA      = $1E00;  kbAltS      = $1F00;
  kbAltD      = $2000;  kbAltF      = $2100;  kbAltG      = $2200;
  kbAltH      = $2300;  kbAltJ      = $2400;  kbAltK      = $2500;
  kbAltL      = $2600;  kbAltZ      = $2C00;  kbAltX      = $2D00;
  kbAltC      = $2E00;  kbAltV      = $2F00;  kbAltB      = $3000;
  kbAltN      = $3100;  kbAltM      = $3200;  kbF1        = $3B00;
  kbF2        = $3C00;  kbF3        = $3D00;  kbF4        = $3E00;
  kbF5        = $3F00;  kbF6        = $4000;  kbF7        = $4100;
  kbF8        = $4200;  kbF9        = $4300;  kbF10       = $4400;
  kbHome      = $4700;  kbUp        = $4800;  kbPgUp      = $4900;
  kbGrayMinus = $4A2D;  kbLeft      = $4B00;  kbRight     = $4D00;
  kbGrayPlus  = $4E2B;  kbEnd       = $4F00;  kbDown      = $5000;
  kbPgDn      = $5100;  kbIns       = $5200;  kbDel       = $5300;
  kbShiftF1   = $5400;  kbShiftF2   = $5500;  kbShiftF3   = $5600;
  kbShiftF4   = $5700;  kbShiftF5   = $5800;  kbShiftF6   = $5900;
  kbShiftF7   = $5A00;  kbShiftF8   = $5B00;  kbShiftF9   = $5C00;
  kbShiftF10  = $5D00;  kbCtrlF1    = $5E00;  kbCtrlF2    = $5F00;
  kbCtrlF3    = $6000;  kbCtrlF4    = $6100;  kbCtrlF5    = $6200;
  kbCtrlF6    = $6300;  kbCtrlF7    = $6400;  kbCtrlF8    = $6500;
  kbCtrlF9    = $6600;  kbCtrlF10   = $6700;  kbAltF1     = $6800;
  kbAltF2     = $6900;  kbAltF3     = $6A00;  kbAltF4     = $6B00;
  kbAltF5     = $6C00;  kbAltF6     = $6D00;  kbAltF7     = $6E00;
  kbAltF8     = $6F00;  kbAltF9     = $7000;  kbAltF10    = $7100;
  kbCtrlPrtSc = $7200;  kbCtrlLeft  = $7300;  kbCtrlRight = $7400;
  kbCtrlEnd   = $7500;  kbCtrlPgDn  = $7600;  kbCtrlHome  = $7700;
  kbAlt1      = $7800;  kbAlt2      = $7900;  kbAlt3      = $7A00;
  kbAlt4      = $7B00;  kbAlt5      = $7C00;  kbAlt6      = $7D00;
  kbAlt7      = $7E00;  kbAlt8      = $7F00;  kbAlt9      = $8000;
  kbAlt0      = $8100;  kbAltMinus  = $8200;  kbAltEqual  = $8300;
  kbCtrlPgUp  = $8400;  kbAltBack   = $0800;  kbNoKey     = $0000;

{ Additional keyboard codes that Borland forgot to define }

  kbCtrlA     = $1E01;  kbCtrlB     = $3002;  kbCtrlC     = $2E03;
  kbCtrlD     = $2004;  kbCtrlE     = $1205;  kbCtrlF     = $2106;
  kbCtrlG     = $2207;  kbCtrlH     = $2308;  kbCtrlI     = $1709;
  kbCtrlJ     = $240A;  kbCtrlK     = $250B;  kbCtrlL     = $260C;
  kbCtrlM     = $320D;  kbCtrlN     = $310E;  kbCtrlO     = $180F;
  kbCtrlP     = $1910;  kbCtrlQ     = $1011;  kbCtrlR     = $1312;
  kbCtrlS     = $1F13;  kbCtrlT     = $1414;  kbCtrlU     = $1615;
  kbCtrlV     = $2F16;  kbCtrlW     = $1117;  kbCtrlX     = $2D18;
  kbCtrlY     = $1519;  kbCtrlZ     = $2C1A;

  kbSpace     = $3920;

{ 101-key AT keyboard }

  kbAltTab     = $A500; kbAltDel     = $A300;  kbAltIns    = $A200;
  kbAltPgDn    = $A100; kbAltDown    = $A000;  kbAltEnd    = $9F00;
  kbAltRight   = $9D00; kbAltLeft    = $9B00;  kbAltPgUp   = $9900;
  kbAltUp      = $9800; kbAltHome    = $9700;  kbCtrlTab   = $9400;
  kbCtrlGreyPlus=$9000; kbCtrlCenter = $8F00;  kbCtrlMinus = $8E00;
  kbCtrlUp     = $8D00; kbAltF12     = $8C00;  kbAltF11    = $8B00;
  kbCtrlF12    = $8A00; kbCtrlF11    = $8900;  kbShiftF12  = $8800;
  kbShiftF11   = $8700; kbF12        = $8600;  kbF11       = $8500;
  kbAltGrayPlus= $4E00; kbCenter     = $4C00;  kbAltGreyAst= $3700;
  kbAltSlash   = $3500; kbAltPeriod  = $3400;  kbAltComma  = $3300;
  kbAltBackSlash=$2B00; kbAltOpQuote = $2900;  kbAltQuote  = $2800;
  kbAltSemicolon=$2700; kbAltRgtBrack= $1B00;  kbAltLftBrack=$1A00;
  kbAltEsc     = $0100; kbCtrlDown   = $9100;

{ Special keys }

  kbAltShiftBack = $0900;

{ Keyboard state and shift masks }

  kbRightShift  = $0001;
  kbLeftShift   = $0002;
  kbCtrlShift   = $0004;
  kbAltShift    = $0008;
  kbScrollState = $0010;
  kbNumState    = $0020;
  kbCapsState   = $0040;
  kbInsState    = $0080;

{ Mouse button state masks }

  mbLeftButton  = $01;
  mbRightButton = $02;

{ Screen modes }

  smBW40        = $0000;
  smCO40        = $0001;
  smBW80        = $0002;
  smCO80        = $0003;
  smMono        = $0007;
  smNonStandard = $00FF;
  smFont8x8     = $0100;

{ Colors }

  Black = 0;
  Blue = 1;
  Green = 2;
  Cyan = 3;
  Red = 4;
  Magenta = 5;
  Brown = 6;
  LightGray = 7;
  DarkGray = 8;
  LightBlue = 9;
  LightGreen = 10;
  LightCyan = 11;
  LightRed = 12;
  LightMagenta = 13;
  Yellow = 14;
  White = 15;
  Blink = 128;

type

{ Type conversion records }

  WordRec = packed record
    Lo, Hi: Byte;
  end;

  LongRec = packed record
    Lo, Hi: Word;
  end;

{ Character set type }

  PCharSet = ^TCharSet;
  TCharSet = set of Char;

{ General arrays }

  PByteArray = ^TByteArray;
  TByteArray = array[0..32767] of Byte;

  PWordArray = ^TWordArray;
  TWordArray = array[0..16383] of Word;

{ Point }

  TPoint = record
    X, Y: Integer;
  end;

{ Rectangle }

  TRect = record
    A, B: TPoint;
  end;

{ Scrollbar orientation }

  TOrientation = (orHorizontal, orVertical);

{ Cursor styles }

  TCursorStyle = (csNorm, csBig);

{ Video buffer }

  PVideoBuf = ^TVideoBuf;
  TVideoBuf = array[0..3999] of WordRec;

{ Screen buffer }

{$IFDEF WDOSX}
  PCharInfo = ^TCharInfo;
  TCharInfo = record
    AsciiChar: Char;
    Attributes: Byte;
  end;
{$ENDIF}

  TScr = record
    X1, Y1, X2, Y2: Integer;
    Data: array[0..3999] of TCharInfo;
    OldX, OldY: Integer;
    OldAttr: Byte;
  end;

{ Event record }

  PEvent = ^TEvent;
  TEvent = record
    What: Word;
    case Word of
      evNothing: ();
      evMouse: (
        Buttons: Byte;
        Double: Boolean;
        Where: TPoint);
      evKeyDown: (
        case Integer of
          0: (KeyCode: Word;
              ShiftState: Byte);
          1: (CharCode: Char;
              ScanCode: Byte));
      evMessage: (
        Command: Word;
        case Word of
          0: (InfoPtr: Pointer);
          1: (InfoLong: Longint);
          2: (InfoWord: Word);
          3: (InfoInt: Integer);
          4: (InfoByte: Byte);
          5: (InfoChar: Char));
  end;

implementation

end.


