
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViUtils.pas                                           }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViUtils.pas,v 1.2 2002/10/27 22:25:23 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViUtils.pas,v $
  Revision 1.2  2002/10/27 22:25:23  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-05-24 Version 2.0.1.0 released
}

unit ViUtils;

interface

uses
  ViTypes;

{ Buffer move routines }

{
  MoveBuf:
  Moves text and video attributes into a buffer. Count bytes are moved
  from Source into the low bytes of corresponding words in Dest. The
  high bytes of the words in Dest are set to Attr, or remain unchanged
  if Attr is zero.
}
procedure MoveBuf(var Dest; var Source; Attr: Byte; Count: Word);

{
  MoveChar:
  Moves characters into a buffer. The low bytes of the first Count
  words of Dest are set to C, or remain unchanged if C = #0. The high
  bytes of the words are set to Attr, or remain unchanged if Attr is
  zero.
}
procedure MoveChar(var Dest; C: Char; Attr: Byte; Count: Word);

{
  MoveCStr:
  Moves a two-colored string into a buffer. The characters in Str are
  moved into the low bytes of corresponding words in Dest. The high
  bytes of the words are set to Lo(Attr) or Hi(Attr). Tilde characters
  (~) in the string toggle between the two attribute bytes passed in
  the Attr word.
}
procedure MoveCStr(var Dest; const Str: string; Attrs: Word);

{
  MoveStr:
  Moves a string into a buffer. The characters in Str are moved into
  the low bytes of corresponding words in Dest. The high bytes of the
  words are set to Attr, or remain unchanged if Attr is zero.
}
procedure MoveStr(var Dest; const Str: string; Attr: Byte);

{
  CStrLen:
  Returns the length of string S, where S is a control string using
  tilde characters ('~') to designate shortcut characters. The tildes
  are excluded from the length of the string, as they will not appear
  on the screen.
}
function CStrLen(const S: string): Integer;

procedure MoveVideoBuf(var Dest; var Source; Attr: Byte; Count: Word);

{ Keyboard support routines }

function GetAltChar(KeyCode: Word): Char;

function GetAltCode(Ch: Char): Word;

function GetCtrlChar(KeyCode: Word): Char;

function GetCtrlCode(Ch: Char): Word;

function CtrlToArrow(KeyCode: Word): Word;

implementation

uses
  ViClasses;

procedure MoveBuf(var Dest; var Source; Attr: Byte; Count: Word);
var
  SourceBuffer: PByteArray;
  DestBuffer: PVideoBuf;
  AWordRec: WordRec;
  I: Integer;
begin
  SourceBuffer := Addr(Source);
  DestBuffer := Addr(Dest);
  if Attr = 0 then
    for I := 0 to Count -1 do
      DestBuffer[I].Lo := SourceBuffer[I]
  else
    for I := 0 to Count -1 do
    begin
      AWordRec.Lo := SourceBuffer[I];
      AWordRec.Hi := Attr;
      DestBuffer[I] := AWordRec;
    end;
end;

procedure MoveChar(var Dest; C: Char; Attr: Byte; Count: Word);
var
  Buffer: PVideoBuf;
  AWordRec: WordRec;
  I: Integer;
begin
  Buffer := Addr(Dest);
  if C = #0 then
  begin
    if Attr = 0 then Exit;
    for I := 0 to Count -1 do
      Buffer[I].Hi := Attr;
  end
  else
  begin
    if Attr = 0 then
      for I := 0 to Count -1 do
        Buffer[I].Lo := Ord(C)
    else
    begin
      AWordRec.Lo := Ord(C);
      AWordRec.Hi := Attr;
      for I := 0 to Count -1 do
        Buffer[I] := AWordRec;
    end;
  end;
end;

procedure MoveCStr(var Dest; const Str: string; Attrs: Word);
var
  Buffer: PVideoBuf;
  AWordRec: WordRec;
  I: Integer;
  Index: Integer;
begin
  Buffer := Addr(Dest);
  Index := 0;
  AWordRec.Hi := WordRec(Attrs).Lo;
  for I := 1 to Length(Str) do
  begin
    if Str[I] = '~' then
    begin
      Attrs := Swap(Attrs);
      AWordRec.Hi := WordRec(Attrs).Lo;
    end
    else
    begin
      AWordRec.Lo := Byte(Str[I]);
      Buffer[Index] := AWordRec;
      Inc(Index);
    end;
  end;
end;

procedure MoveStr(var Dest; const Str: string; Attr: Byte);
var
  Buffer: PVideoBuf;
  AWordRec: WordRec;
  I: Integer;
begin
  Buffer := Addr(Dest);
  if Attr = 0 then
    for I := 1 to Length(Str) do
      Buffer[I - 1].Lo := Byte(Str[I])
  else
  begin
    AWordRec.Hi := Attr;
    for I := 1 to Length(Str) do
    begin
      AWordRec.Lo := Byte(Str[I]);
      Buffer[I - 1] := AWordRec;
    end;
  end;
end;

function CStrLen(const S: string): Integer;
var
  I: Integer;
begin
  Result := Length(S);
  for I := 1 to Length(S) do
    if S[I] = '~' then Dec(Result);
end;

procedure MoveVideoBuf(var Dest; var Source; Attr: Byte; Count: Word);
var
  SourceBuffer, DestBuffer: PVideoBuf;
  AWordRec: WordRec;
  I: Integer;
begin
  SourceBuffer := Addr(Source);
  DestBuffer := Addr(Dest);
  if Attr = 0 then
    for I := 0 to Count -1 do
      DestBuffer[I] := SourceBuffer[I]
  else
    for I := 0 to Count -1 do
    begin
      AWordRec.Lo := SourceBuffer[I].Lo;
      AWordRec.Hi := Attr;
      DestBuffer[I] := AWordRec;
    end;
end;

// Keyboard support routines

const

  AltCodes1: array[$10..$32] of Char =
    'QWERTYUIOP'#0#0#0#0'ASDFGHJKL'#0#0#0#0#0'ZXCVBNM';

  AltCodes2: array[$78..$83] of Char =
    '1234567890-=';

function GetAltChar(KeyCode: Word): Char;
begin
  GetAltChar := #0;
  if Lo(KeyCode) = 0 then
    case Hi(KeyCode) of
      $02: GetAltChar := #240;
      $10..$32: GetAltChar := AltCodes1[Hi(KeyCode)];
      $78..$83: GetAltChar := AltCodes2[Hi(KeyCode)];
    end;
end;

function GetAltCode(Ch: Char): Word;
var
  I: Word;
begin
  GetAltCode := 0;
  if Ch = #0 then Exit;
  Ch := UpCase(Ch);
  if Ch = #240 then
  begin
    GetAltCode := $0200;
    Exit;
  end;
  for I := $10 to $32 do
    if AltCodes1[I] = Ch then
    begin
      GetAltCode := I shl 8;
      Exit;
    end;
  for I := $78 to $83 do
    if AltCodes2[I] = Ch then
    begin
      GetAltCode := I shl 8;
      Exit;
    end;
end;

function GetCtrlChar(KeyCode: Word): Char;
begin
  GetCtrlChar := #0;
  if (Lo(KeyCode) <> 0) and (Lo(KeyCode) <= Byte('Z') - Byte('A') + 1) then
    GetCtrlChar := Char(Lo(KeyCode) + Byte('A') - 1);
end;

function GetCtrlCode(Ch: Char): Word;
begin
  GetCtrlCode := GetAltCode(Ch) or (Byte(UpCase(Ch)) - Byte('A') + 1);
end;

function CtrlToArrow(KeyCode: Word): Word;
const
  NumCodes = 11;
  CtrlCodes: array[0..NumCodes-1] of Char = ^S^D^E^X^A^F^G^V^R^C^H;
  ArrowCodes: array[0..NumCodes-1] of Word =
    (kbLeft, kbRight, kbUp, kbDown, kbHome, kbEnd, kbDel, kbIns,
     kbPgUp, kbPgDn, kbBack);
var
  I: Integer;
begin
  CtrlToArrow := KeyCode;
  for I := 0 to NumCodes - 1 do
    if WordRec(KeyCode).Lo = Byte(CtrlCodes[I]) then
    begin
      CtrlToArrow := ArrowCodes[I];
      Exit;
    end;
end;

end.
