
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ WDosMouses.pas                                        }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - CRT unit from Free Pascal RTL                     }
{     Copyright (c) 1999 by                             }
{     The Free Pascal Organization                      }
{     www: http://www.freepascal.org                    }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/wdosx/WDosMouses.pas,v 1.2 2002/10/27 22:25:25 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosMouses.pas,v $
  Revision 1.2  2002/10/27 22:25:25  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-05-24 Version 2.0.1.0 released
}

unit WDosMouses;

interface

uses
  Windows, SysUtils;

procedure ShowMouse;

procedure HideMouse;

procedure ProcessConsoleEvents;

procedure InitMouse;

procedure DoneMouse;

var
  MouseButtonCount: Integer;
  MouseButtons: Integer;
  MouseWhereX: Integer;
  MouseWhereY: Integer;
  MouseTime: Cardinal;

  KeyboardShiftState: Byte;
  KeyboardScanCode: Byte;
  KeyboardCharCode: Char;

implementation

uses
//  Windows,
  WDosScreens, ViTypes;

type
  dpmiRealCallRegs = packed record
    case Integer of
      0: (edi, esi, ebp, espres, ebx, edx, ecx, eax: Integer;
        flags, es, ds, fs, gs, ip, cs, sp, ss: Word);
      1: (di, diu, si, siu, bp, bpu, spres, spu, bx, bxu, dx, dxu, cx, cxu, ax,
        axu: Word);
      2: (dil, dih, di2, di3, sil, sih, si2, si3, bpl, bph, bp2, bp3, spl, sph,
        sp2, sp3, bl,
        bh, bx2, bx3, dl, dh, dx2, dx3, cl, Ch, cx2, cx3, al, ah, ax2, ax3:
          Byte);
  end;

var
  MR: dpmiRealCallRegs;
  MouseInited: Boolean = False;
  mxshift: Byte = 3;
  myshift: Byte = 3;
  LastMouseButtons: Integer;

function dpmiRealModeInt(intnr: Integer; var cregs: dpmiRealCallRegs): Integer;
  register; assembler;
asm
  push	ebx
  push	edi
  xor   ecx,ecx
  mov	ebx,eax
  mov	edi,edx
  mov	eax,300h
  int	31h
  sbb	eax,eax
  pop	edi
  pop	ebx
end;

procedure ShowMouse;
begin
  if mouseinited then
  begin
    mr.ax := 1;
    dpmirealmodeint($33, mr);
  end;
end;

procedure HideMouse;
begin
  if mouseinited then
  begin
    mr.ax := 2;
    dpmirealmodeint($33, mr);
  end;
end;

procedure GetMouseState;
begin
  if MouseInited then
  begin
    mr.ax := 3;
    dpmirealmodeint($33, mr);
    MouseWhereX := mr.cx shr mxshift;
    MouseWhereY := mr.dx shr myshift;
    MouseButtons := MR.BX;

    if MouseButtons <> LastMouseButtons then
    begin
      LastMouseButtons := MouseButtons;
      MouseTime := GetTickCount;
    end;
  end;
end;

procedure MouseXY(X, Y: Word);
begin
  if mouseinited then
  begin
    mr.ax := 4;
    mr.cx := X shl mxshift;
    mr.dx := Y shl myshift;
    dpmirealmodeint($33, mr);
    MouseWhereX := X;
    MouseWhereY := Y;
  end;
end;

procedure MouseArea(X1, Y1, X2, Y2: Word);
begin
  if mouseinited then
  begin
    mr.ax := 7;
    mr.cx := X1 shl mxshift;
    mr.dx := X2 shl mxshift;
    dpmirealmodeint($33, mr);
    mr.ax := 8;
    mr.cx := Y1 shl myshift;
    mr.dx := Y2 shl myshift;
    dpmirealmodeint($33, mr);
  end;
end;

procedure InitMouse;
begin
  if not MouseInited then
    with MR do
    begin
      ax := 0;
      dpmirealmodeint($33, mr);
      if AX = 0 then Exit;
      case bx of
        $0000: MouseButtonCount := 3; //other than two
        $FFFF: MouseButtonCount := 2; //two
      else
        MouseButtonCount := bx;
      end;
      MouseArea(0, 0, 79, 24);
      MouseXY(0, 0);
      MouseInited := True;
    end;
end;

procedure DoneMouse;
begin
  mouseinited := False;
end;

procedure GetKeyState;
asm

        MOV    AH,1
        INT    16H
        JZ     @@1
        SUB    AH,AH
        INT    16H
        MOV    KeyboardScanCode, AH
        MOV    KeyboardCharCode, AL
@@1:
end;

procedure GetShiftState;
begin
  KeyboardShiftState := PByte($417)^;
end;

procedure ProcessConsoleEvents;
begin
  GetMouseState;
  GetKeyState;
  GetShiftState;
end;

end.

