{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/win/WinTimers.pas,v 1.3 2002/10/27 22:25:26 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WinTimers.pas,v $
  Revision 1.3  2002/10/27 22:25:26  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
unit WinTimers;

interface

uses
  Windows, Classes, WinMessages;


type
  TIntFreq =(ifNone, if32768, if16384, if8192,
             if4096, if2048,  if1024,  if512,
             if256,  if128,   if64,    if32,
             if16,   if8,     if4,     if2);

  TViTimer = class(TObject)
  private
    FInterval: Cardinal;
    FWindowHandle: HWND;
    FOnTimer: TNotifyEvent;
    FEnabled: Boolean;
    procedure UpdateTimer;
    procedure SetEnabled(Value: Boolean);
    procedure SetInterval(Value: Cardinal);
    procedure SetOnTimer(Value: TNotifyEvent);
    procedure WndProc(var Msg: TMessage);
  protected
    procedure Timer; dynamic;
  public
    constructor Create;
    destructor Destroy; override;
    property Enabled: Boolean read FEnabled write SetEnabled;
    property Interval: Cardinal read FInterval write SetInterval;
    property OnTimer: TNotifyEvent read FOnTimer write SetOnTimer;
  end;

  TViRtcTimer = class(TViTimer)
  private
    function GetIntFreq: TIntFreq;
    procedure SetIntFreq(Value: TIntFreq);
  public
    property IntFreq: TIntFreq read GetIntFreq write SetIntFreq;
  end;

implementation

resourcestring
  SNoTimers = 'Not enough timers available';

var
  FIntFreq: TIntFreq;

function SetTimer(hWnd: HWND; nIDEvent, uElapse: UINT;
    lpTimerFunc: TFNTimerProc): UINT; stdcall;
    external user32 name 'SetTimer';

function KillTimer(hWnd: HWND; uIDEvent: UINT): BOOL; stdcall;
    external user32 name 'KillTimer';

function DefWindowProc(hWnd: HWND; Msg: UINT; wParam: WPARAM;
    lParam: LPARAM): LRESULT; stdcall;
    external user32 name 'DefWindowProcA';


constructor TViTimer.Create;
begin
  inherited Create;
  FEnabled := True;
  FInterval := 1000;
  FWindowHandle := AllocateHWnd(WndProc);
end;

destructor TViTimer.Destroy;
begin
  FEnabled := False;
  UpdateTimer;
  DeallocateHWnd(FWindowHandle);
  inherited Destroy;
end;

procedure TViTimer.SetEnabled(Value: Boolean);
begin
  if Value <> FEnabled then
  begin
    FEnabled := Value;
    UpdateTimer;
  end;
end;

procedure TViTimer.SetInterval(Value: Cardinal);
begin
  if Value <> FInterval then
  begin
    FInterval := Value;
    UpdateTimer;
  end;
end;

procedure TViTimer.SetOnTimer(Value: TNotifyEvent);
begin
  FOnTimer := Value;
  UpdateTimer;
end;

procedure TViTimer.Timer;
begin
  if Assigned(FOnTimer) then FOnTimer(Self);
end;

procedure TViTimer.UpdateTimer;
begin
  KillTimer(FWindowHandle, 1);
  if (FInterval <> 0) and FEnabled and Assigned(FOnTimer) then
    if SetTimer(FWindowHandle, 1, FInterval, nil) = 0 then
      raise EOutOfResources.Create(SNoTimers);
end;

procedure TViTimer.WndProc(var Msg: TMessage);
begin
  with Msg do
    if Msg = WM_TIMER then
      try
        Timer;
      except
        // do nothing
      end
    else
      Result := DefWindowProc(FWindowHandle, Msg, wParam, lParam);
end;

function TViRtcTimer.GetIntFreq: TIntFreq;
begin
  Result := FIntFreq;
end;

procedure TViRtcTimer.SetIntFreq(Value: TIntFreq);
begin
  FIntFreq := Value;
end;



end.
