
{*******************************************************}
{                                                       }
{ DWPL Delphi WDosX and Cross Platform Library (rtl)    }
{ WDosMM.pas                                            }
{                                                       }
{ for Delphi 6                                          }
{ Version 2.0                                           }
{ Copyright (c) 2002 by Team-DWP                        }
{ e-mail: team-dwp@dwp42.org                            }
{ www: http://www.dwp42.org                             }
{                                                       }
{ The initial developer of the code is                  }
{   Laszlo Agocs alp@dwp42.org                          }
{                                                       }
{ Portions created by Laszlo Agocs are                  }
{ Copyright (C) 2002 Laszlo Agocs.                      }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ Licensees holding a valid Team-DWP No-Nonsense        }
{ License for this Software may use this file in        }
{ accordance with such license, which appears in        }
{ the file license.txt that came with this software.    }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/memmgr/WDosMM.pas,v 1.1 2002/11/05 19:22:11 lagocs Exp $
  ----------------------------------------------------------------------------

  $Log: WDosMM.pas,v $
  Revision 1.1  2002/11/05 19:22:11  lagocs
  WDosMM moved to memmgr. Common getmem.inc for D5 and 6. Crt units are renamed to
  Crt.old.

  Revision 1.2  2002/10/27 22:53:12  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
unit WDosMM;

interface

function SysGetMem(Size: Integer): Pointer;
function SysFreeMem(P: Pointer): Integer;
function SysReallocMem(P: Pointer; Size: Integer): Pointer;
function GetHeapStatus: THeapStatus;

implementation

const
  MemMgr = 'memmgr.dll';

function Alloc(Size: Integer): Pointer; stdcall; external MemMgr
    name 'alp_malloc';
procedure Free(P: Pointer); stdcall; external MemMgr
    name 'alp_free';
function ReAlloc(P: Pointer; NewSize: Integer): Pointer; stdcall;
    external MemMgr name 'alp_realloc';

const

// Heap error codes

  cHeapOk           = 0;          // everything's fine

function SysGetMem(Size: Integer): Pointer;
begin
  Result := Alloc(Size);
end;

function SysFreeMem(P: Pointer): Integer;
begin
  Free(P);
  Result := 0;
end;

function SysReallocMem(P: Pointer; Size: Integer): Pointer;
begin
  Result := ReAlloc(P, Size);
end;

function GetHeapStatus: THeapStatus;
begin
  Result.TotalAddrSpace   := 0;
  Result.TotalUncommitted := 0;
  Result.TotalCommitted   := 0;
  Result.TotalAllocated   := 0;
  Result.TotalFree        := 0;
  Result.FreeSmall        := 0;
  Result.FreeBig          := 0;
  Result.Unused           := 0;
  Result.Overhead         := 0;
  Result.HeapErrorCode    := cHeapOk;
end;

procedure InitMemoryManager;
var
  MemoryManager: TMemoryManager;
begin
  MemoryManager.GetMem := SysGetMem;
  MemoryManager.FreeMem := SysFreeMem;
  MemoryManager.ReallocMem := SysReallocMem;
  SetMemoryManager(MemoryManager);
end;

initialization
  if not IsMemoryManagerSet then
    InitMemoryManager;
end.

