
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ WDosAudioCD.pas                                       }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 1.3                                           }
{ Copyright (c) 2001 by Laszlo Agocs                    }
{ e-mail: alp@dwp42.org                                 }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ The GWO_CD unit for Free Pascal by Peter Santing was  }
{ a great source of info                                }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosAudioCD.pas,v 1.3 2002/10/27 22:53:12 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosAudioCD.pas,v $
  Revision 1.3  2002/10/27 22:53:12  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:

    ALP 2001-09-06 Version 1.3 released
      - added cdaNumDrives
      - added cdaOpenDoor, cdaCloseDoor, cdaLockDoor, cdaUnlockDoor
      - changed the return codes of cdaLoadCDPlayerIni
      - added cdaCurrentDrive, cdaDrives
      - all functions are using cdaCurrentDrive instead cdaFirstDrive
      - so now we support systems having more than one CD-ROM drives

    ALP 2001-06-04 Version 1.2 released
}

unit WDosAudioCD;

interface

uses SysUtils,WDosPorts,WDosDpmi;

type
  TRedBook=packed record
    frame:byte;
    second:byte;
    minute:byte;
    reserved:byte;
  end;

  TDiscInfo=packed record
    reserved:byte;
    first:byte;
    last:byte;
    leadout:TRedBook;
  end;

var
  cdaFirstDrive:byte;
  cdaNumDrives:byte;
  cdaCurrentDrive:byte;
  cdaReady:boolean;
  cdaStatus:word; //updated by IOCtlInput ONLY!
  cdaDrives:array [1..26] of byte; //index must be in range 1..cdaNumDrives
  cdaArtist,cdaTitle:string;
  cdaTracks:array of string;

function cdaGetDiscInfo:TDiscInfo; //first track, last track, leadout position
function cdaGetTrackPos(track:byte):TRedBook; //position of the track
function cdaGetDiscLength:TRedBook; //length of the disc
function cdaGetTrackLength(track:byte):TRedBook; //length of the track

procedure cdaOpenDoor; //open door
procedure cdaCloseDoor; //close door
procedure cdaLockDoor; //lock door
procedure cdaUnlockDoor; //unlock door

//*** these are only available when playing
function cdaGetTrack:byte; //current track
function cdaGetTrackTime:TRedBook; //elapsed time from track
function cdaGetDiscTime:TRedBook; //elapsed time from disc
function cdaGetTrackTimeRem:TRedBook; //remaining time from track
function cdaGetDiscTimeRem:TRedBook; //remaining time from disc
//***

function cdaIsTrack(track:byte):boolean; //does the given track exist?
function cdaIsAudioTrack(track:byte):boolean; //is it an audio track?
function cdaIsPlaying:boolean; //is the drive busy?
function cdaIsDiscInDrive:boolean; //is the drive ready?

procedure cdaPause; //pause
procedure cdaResume; //resume
procedure cdaStop; //stop
procedure cdaPlayFromTrack(track:byte); //play disc from given track
procedure cdaPlayTrack(track:byte); //play one track only

function cdaCalcID:string; //get the disc ID (same format as cdplayer.ini)
function cdaLoadCDPlayerIni(fn:string):byte; //load cdplayer.ini
//0=no error, 1=file error, 2=no data

implementation

var
  r:dpmirmregisters;
  bufsel,bufseg,reqsel,reqseg:integer;
  req:array [0..21] of byte;
  buf:array [0..10] of byte;
  ReqPtr, BufPtr: Pointer;

procedure IOCtlOutput;
begin
  req[2]:=$0C;
  req[14]:=0;
  req[15]:=0;
  req[16]:=lo(bufseg);
  req[17]:=hi(bufseg);
  move(buf, BufPtr^, sizeof(buf));
  move(req, ReqPtr^, sizeof(req));
  r.ax:=$1510;
  r.es:=reqseg;
  r.bx:=0;
  r.cx:=cdaCurrentDrive;
  dpmiRealModeInt($2F,r);
  move(ReqPtr^, req,sizeof(req));
  move(BufPtr^, buf,sizeof(buf));
end;

procedure IOCtlInput;
begin
  req[2]:=$03;
  req[14]:=0;
  req[15]:=0;
  req[16]:=lo(bufseg);
  req[17]:=hi(bufseg);
  move(buf, BufPtr^, sizeof(buf));
  move(req, ReqPtr^, sizeof(req));
  r.ax:=$1510;
  r.es:=reqseg;
  r.bx:=0;
  r.cx:=cdaCurrentDrive;
  dpmiRealModeInt($2F,r);
  move(ReqPtr^, req,sizeof(req));
  move(BufPtr^, buf,sizeof(buf));
  move(req[3],cdaStatus,2);
end;

function cdaGetDiscInfo:TDiscInfo;
begin
  result.first:=0;result.last:=0;
  if not cdaReady then exit;
  buf[0]:=$0A;
  IOCtlInput;
  move(buf,result,sizeof(TDiscInfo));
end;

function cdaGetDiscLength:TRedBook;
begin
  result.minute:=0;result.second:=0;result.frame:=0;
  if not cdaReady then exit;
  result:=cdaGetDiscInfo.leadout;
end;

function cdaGetTrackPos(track:byte):TRedBook;
begin
  result.minute:=0;result.second:=0;result.frame:=0;
  if not cdaReady then exit;
  buf[0]:=$0B;
  buf[1]:=track;
  IOCtlInput;
  move(buf[2],result,sizeof(TRedBook));
end;

function cdaGetTrackLength(track:byte):TRedBook;
var
  trackpos,nextpos:TRedBook;
  iframe,isecond,iminute:integer;
begin
  result.minute:=0;result.second:=0;result.frame:=0;
  if not cdaReady then exit;
  trackpos:=cdaGetTrackPos(track);
  if track=cdaGetDiscInfo.last then nextpos:=cdaGetDiscLength else
    nextpos:=cdaGetTrackPos(track+1);
  with result do begin
    iframe:=nextpos.frame-trackpos.frame;
    isecond:=nextpos.second-trackpos.second;
    iminute:=nextpos.minute-trackpos.minute;
    if iframe<0 then begin
      inc(iframe,75);
      dec(isecond);
    end;
    if isecond<0 then begin
      inc(isecond,60);
      dec(iminute);
    end;
    if iminute<0 then begin
      iframe:=0;
      isecond:=0;
      iminute:=0;
    end;
    frame:=iframe;
    second:=isecond;
    minute:=iminute;
  end;
end;

procedure cdaOpenDoor;
begin
  if not cdaReady then exit;
  buf[0]:=$00;
  IOCTLOutput;
end;

procedure cdaCloseDoor;
begin
  if not cdaReady then exit;
  buf[0]:=$05;
  IOCTLOutput;
end;

procedure cdaLockDoor;
begin
  if not cdaReady then exit;
  buf[0]:=1;
  buf[1]:=1;
  IOCTLOutput;
end;

procedure cdaUnlockDoor;
begin
  if not cdaReady then exit;
  buf[0]:=1;
  buf[1]:=0;
  IOCTLOutput;
end;

function cdaGetTrack:byte;
begin
  result:=0;
  if not cdaReady then exit;
  buf[0]:=$0C;
  IOCtlInput;
  result:=(buf[2] shr 4)*10+(buf[2] and $0F); // ???
end;

function cdaGetTrackTime:TRedBook;
begin
  result.minute:=0;result.second:=0;result.frame:=0;
  if not cdaReady then exit;
  buf[0]:=$0C;
  IOCtlInput;
  result.frame:=buf[6];
  result.second:=buf[5];
  result.minute:=buf[4];
end;

function cdaGetDiscTime:TRedBook;
begin
  result.minute:=0;result.second:=0;result.frame:=0;
  if not cdaReady then exit;
  buf[0]:=$0C;
  IOCtlInput;
  result.frame:=buf[10];
  result.second:=buf[9];
  result.minute:=buf[8];
end;

function cdaGetTrackTimeRem:TRedBook;
var
  curtime,endtime:TRedBook;
  iframe,isecond,iminute:integer;
begin
  result.minute:=0;result.second:=0;result.frame:=0;
  if not cdaReady then exit;
  curtime:=cdaGetTrackTime;
  endtime:=cdaGetTrackLength(cdaGetTrack);
  with result do begin
    iframe:=endtime.frame-curtime.frame;
    isecond:=endtime.second-curtime.second;
    iminute:=endtime.minute-curtime.minute;
    if iframe<0 then begin
      inc(iframe,75);
      dec(isecond);
    end;
    if isecond<0 then begin
      inc(isecond,60);
      dec(iminute);
    end;
    if iminute<0 then begin
      iframe:=0;
      isecond:=0;
      iminute:=0;
    end;
    frame:=iframe;
    second:=isecond;
    minute:=iminute;
  end;
end;

function cdaGetDiscTimeRem:TRedBook;
var
  curtime,endtime:TRedBook;
  iframe,isecond,iminute:integer;
begin
  result.minute:=0;result.second:=0;result.frame:=0;
  if not cdaReady then exit;
  curtime:=cdaGetDiscTime;
  endtime:=cdaGetDiscLength;
  with result do begin
    iframe:=endtime.frame-curtime.frame;
    isecond:=endtime.second-curtime.second;
    iminute:=endtime.minute-curtime.minute;
    if iframe<0 then begin
      inc(iframe,75);
      dec(isecond);
    end;
    if isecond<0 then begin
      inc(isecond,60);
      dec(iminute);
    end;
    if iminute<0 then begin
      iframe:=0;
      isecond:=0;
      iminute:=0;
    end;
    frame:=iframe;
    second:=isecond;
    minute:=iminute;
  end;
end;

function cdaIsTrack(track:byte):boolean;
var
  discinfo:TDiscInfo;
begin
  result:=false;
  if not cdaReady then exit;
  discinfo:=cdaGetDiscInfo;
  result:=((track>=discinfo.first) and (track<=discinfo.last));
end;

function cdaIsAudioTrack(track:byte):boolean;
begin
  result:=false;
  if not cdaReady then exit;
  result:=cdaIsTrack(track);
  if not result then exit;
  buf[0]:=$0B;
  buf[1]:=track;
  IOCtlInput;
  result:=((buf[6] shr 6)<>1);
end;

function cdaIsPlaying:boolean;
begin
  result:=false;
  if not cdaReady then exit;
  cdaGetDiscInfo; //to update cdaStatus (by ioctlinput)
  result:=((cdaStatus and $200)<>0);
end;

function cdaIsDiscInDrive:boolean;
begin
  result:=false;
  if not cdaReady then exit;
  cdaGetDiscInfo; //to update cdaStatus
  result:=((cdaStatus shr 15)<>1);
end;

procedure cdaPause;
begin
  if not cdaReady then exit;
  req[2]:=$85;
  move(req, ReqPtr^,sizeof(req));
  r.ax:=$1510;
  r.es:=reqseg;
  r.bx:=0;
  r.cx:=cdaCurrentDrive;
  dpmiRealModeInt($2F,r);
end;

procedure cdaResume;
begin
  if not cdaReady then exit;
  req[2]:=$88;
  move(req, ReqPtr^, sizeof(req));
  r.ax:=$1510;
  r.es:=reqseg;
  r.bx:=0;
  r.cx:=cdaCurrentDrive;
  dpmiRealModeInt($2F,r);
end;

procedure cdaStop;
begin
  if not cdaReady then exit;
  req[2]:=133;
  move(req, ReqPtr^, sizeof(req));
  r.ax:=$1510;
  r.es:=reqseg;
  r.bx:=0;
  r.cx:=cdaCurrentDrive;
  dpmiRealModeInt($2F,r);
end;

procedure cdaPlay;
begin
  cdaPause;
  req[2]:=$84;
  req[13]:=0;
  move(req, ReqPtr^, sizeof(req));
  r.ax:=$1510;
  r.es:=reqseg;
  r.bx:=0;
  r.cx:=cdaCurrentDrive;
  dpmiRealModeInt($2F,r);
end;

procedure cdaPlayFromTrack(track:byte);
var
  trkstart,disclen:TRedBook;
  l1,l2:longint;
begin
  if not cdaReady then exit;
  trkstart:=cdaGetTrackPos(track);
  disclen:=cdaGetDiscLength;
  l1:=(trkstart.minute*4500)+(trkstart.second*75)+trkstart.frame-150;
  move(l1,req[14],4);
  l2:=(disclen.minute*4500)+(disclen.second*75)+disclen.frame-150-l1;
  move(l2,req[18],4);
  cdaPlay;
end;

procedure cdaPlayTrack(track:byte);
var
  trkstart,trklen:TRedBook;
  l1,l2:longint;
begin
  if not cdaReady then exit;
  trkstart:=cdaGetTrackPos(track);
  trklen:=cdaGetTrackLength(track);
  l1:=(trkstart.minute*4500)+(trkstart.second*75)+trkstart.frame-150;
  move(l1,req[14],4);
  l2:=(trklen.minute*4500)+(trklen.second*75)+trklen.frame-150;
  move(l2,req[18],4);
  cdaPlay;
end;

function cdaCalcID:string;
var
  discinfo:TDiscInfo;
  tracknum:integer;
  trackpos,disclen:TRedBook;
  min,sec,frm,mag:byte;
  id:longint;
begin
  result:='';
  if not cdaReady then exit;
  if not cdaIsDiscInDrive then exit;
  discinfo:=cdaGetDiscInfo;
  id:=0;mag:=0;
  for tracknum:=1 to discinfo.last do begin
    trackpos:=cdaGetTrackPos(tracknum);
    min:=trackpos.minute;
    sec:=trackpos.second;
    frm:=trackpos.frame;
    inc(id,(min shl 16)+(sec shl 8)+(frm));
    if tracknum=1 then mag:=frm;
  end;
  if discinfo.last<3 then begin
    disclen:=cdaGetDiscLength;
    inc(id,mag+(disclen.minute*60*75)+(disclen.second*75)+disclen.frame);
  end;
  result:=inttohex(id,1);
end;

function toupper(s:string):string;
var
  i:integer;
begin
  result:=s;
  for i:=1 to length(result) do result[i]:=upcase(result[i]);
end;

function AStrToInt(src:string;var dest:integer):boolean;
var
  i:integer;
begin
  Val(src,dest,i);
  result:=(i=0);
end;

function AFStrToInt(src:string):integer;
var
  i:integer;
begin
  Val(src,result,i);
  if i<>0 then result:=0;
end;

function cdaLoadCDPlayerIni(fn:string):byte;
var
  f:text;
  i,j:integer;
  tmp:longint;
  cdid,s,s2:shortstring;
  found,b:boolean;
begin
  result:=1;
  cdaArtist:='?';
  cdaTitle:='?';
  assign(f,fn);
  {$I-}
  reset(f);
  {$I+}
  if ioresult<>0 then exit;
  cdid:=cdaCalcID;
  readln(f,s);
  found:=false;
  repeat
    if copy(s,2,length(s)-2)=cdid then begin
      readln(f,s); {EntryType}
      readln(f,s);
      if toupper(copy(s,1,6))='ARTIST' then cdaArtist:=copy(s,8,255);
      readln(f,s);
      if toupper(copy(s,1,5))='TITLE' then cdaTitle:=copy(s,7,255);
      readln(f,s);
      j:=afstrtoint(copy(s,11,255));
      b:=true;
      for i:=1 to j do if b then if not eof(f) then begin
        readln(f,s);
        s2:=copy(s,1,pos('=',s)-1);
        if not astrtoint(s2,tmp) then b:=false else begin
          delete(s,1,pos('=',s));
          cdaTracks[afstrtoint(s2)+1]:=s;
        end;
      end;
      found:=true;
      if not eof(f) then readln(f,s);
    end else begin
      s[0]:=#0;
      s[1]:=#0;
      while s[1]<>'[' do
        if eof(f) then s:='[' else readln(f,s);
    end;
  until eof(f);
  close(f);
  if not found then result:=2 else result:=0;
end;

var
  i:integer;

initialization
  r.ss:=0;r.sp:=0;
  r.ax:=$1500;
  r.bx:=0;
  dpmiRealModeInt($2F,r);
  cdaReady:=(r.bx<>0);
  cdaFirstDrive:=r.cx;
  cdaCurrentDrive:=cdaFirstDrive;
  cdaNumDrives:=r.bx;
  dpmiAllocDosMem(16,bufseg,bufsel); //256 bytes
  dpmiAllocDosMem(16,reqseg,reqsel); //256 bytes
  r.ax:=$150D;
  r.es:=bufseg;
  r.bx:=0;
  dpmiRealModeInt($2F,r);
  for i:=1 to cdaNumDrives do cdaDrives[i]:=mem[bufseg shl 4+i-1];
  SetLength(cdaTracks,99);
  BufPtr := Pointer(BufSeg shl 4);
  ReqPtr := Pointer(ReqSeg shl 4);

finalization
  dpmiFreeDosMem(bufsel);
  dpmiFreeDosMem(reqsel);
  cdaTracks:=nil;

end.
