
{*******************************************************}
{                                                       }
{ DWPL Delphi WDosX and Cross Platform Library (rtl)    }
{ WDosComCtrl.pas                                       }
{                                                       }
{ for Delphi  5, 6, 7                                   }
{ Version 2.6.0                                         }
{ Copyright (c) 2001-2003 by Thomas@Bornhaupt           }
{ e-mail: Thomas@Bornhaupt.de                           }
{ www: http://www.Bornhaupt.de                          }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosComCtrl.pas,v 1.1 2003/09/22 18:49:33 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosComCtrl.pas,v $
  Revision 1.1  2003/09/22 18:49:33  iwache
  A WDOSX only implementation for serial port access.
  Thanks a lot for the code to Thomas.

  ----------------------------------------------------------------------------
}

unit WDosComCtrl;

interface

{$IFDEF DESIGNPACKAGE}
uses
   SysUtils, Classes, QForms;
{$ELSE}
uses
   SysUtils, Classes, QForms, WDosPorts, WDosInterrupts;
{$ENDIF}

type
   TwdxComCtrl = class;

   TComDevice = (Com1, Com2);
   TBaudType = (b110, b150, b300, b600, b1200, b2400,
      b4800, b9600, b19200, b38400, b57600, b115200);
   TParityType = (None, Odd, Even, Mark, Space);
   TDataBitsType = (d5, d6, d7, d8);
   TStopBitsType = (s1, s2);

   TReciveByteProc = procedure(Sender: TwdxComCtrl);

   TwdxComCtrl = class(TComponent)
   private
      FActive: Boolean;
      FStreamedActive: Boolean;

      FComDevice: TComDevice;
      FBaudrate: TBaudType;
      FParity: TParityType;
      FDataBits: TDataBitsType;
      FStopBits: TStopBitsType;
      FRTS: Boolean;
      FDTR: Boolean;
      FBufferSize: integer;

      FBufferHead: integer;
      FBufferTail: integer;
      FBuffer: PByteArray;
      FComDeviceBaseAddr: Word;
      FOnReciveByteProc: TReciveByteProc;

      procedure InitCom;
      procedure ClearPendingInterrupts;
      procedure ResetCom;
      procedure AllocBuffer;
      procedure FreeBuffer;
      procedure ReadBytes;

      procedure SetComDeviceStr(AComDevice: string);
      procedure SetBaudRateStr(ABaudrate: string);
      procedure SetParityStr(AParity: string);
      procedure SetDataBitsStr(ADataBits: string);
      procedure SetStopBitsStr(AStopBits: string);
      procedure SetRTSStr(ARTS: string);
      procedure SetDTRStr(ADTR: string);

      function GetComDeviceStr: string;
      function GetBaudrateStr: string;
      function GetParityStr: string;
      function GetDataBitsStr: string;
      function GetStopBitsStr: string;
      function GetRTSStr: string;
      function GetDTRStr: string;
      procedure SetActive(const Value: Boolean);
   public
      constructor Create(Owner: TComponent); override;
      destructor destroy; override;
      procedure Loaded; override;

      procedure Close;
      procedure Open;

      class function ComDeviceBaseAddr(ComDevice: TComDevice): Word;

      function DataAvail: BOOLEAN;
      function GetByte: BYTE;
      
      procedure Send(Data: BYTE); overload;
      procedure Send(Data: string); overload;

      property Buffer: PByteArray read FBuffer;

      property ComDeviceStr: string read GetComDeviceStr write SetComDeviceStr;
      property BaudrateStr: string read GetBaudrateStr write SetBaudrateStr;
      property ParityStr: string read GetParityStr write SetParityStr;
      property DataBitsStr: string read GetDataBitsStr write SetDataBitsStr;
      property StopBitsStr: string read GetStopBitsStr write SetStopBitsStr;
      property RTSStr: string read GetRTSStr write SetRTSStr;
      property DTRStr: string read GetDTRStr write SetDTRStr;

      property OnReciveByteProc: TReciveByteProc read FOnReciveByteProc write
         FOnReciveByteProc;
   published
      property Active: Boolean read FActive write SetActive default False;
      property ComDevice: TComDevice read FComDevice write FComDevice default
         Com1;
      property Baudrate: TBaudType read FBaudrate write FBaudrate default b9600;
      property Parity: TParityType read FParity write FParity default none;
      property DataBits: TDataBitsType read FDataBits write FDataBits default
         d8;
      property StopBits: TStopBitsType read FStopBits write FStopBits default
         s1;
      property RTS: boolean read FRTS write FRTS default true;
      property DTR: boolean read FDTR write FDTR default true;
      property BufferSize: integer read FBufferSize write FBufferSize default
         4096;
   end;

   EComCtrlError = class(Exception)
   end;

implementation

resourcestring
   { WDosComCtrls }
   sComCtrlNotOpen = 'Device %s is not open';
   sComCtrlOnOpenError = 'Operation on opened device %s not allowed';
   sComCtrlOpenError = 'Device %s can''t be opened';
   sComCtrlOverrun = 'Device %s overrun';
   sComCtrlTimeOut = 'Device %s timeout';

const
   BaudConst: array[TBaudType] of WORD =
   ($0417, $0300, $0180, $00C0,
      $0060, $0030, $0018, $000C,
      $0006, $0003, $0002, $0001);

   ParityConst: array[TParityType] of BYTE = ($00, $08, $18, $28, $38);
   BitsConst: array[TDataBitsType] of BYTE = ($00, $01, $02, $03);
   StopConst: array[TStopBitsType] of BYTE = ($00, $04);
   IntMasks: array[TComDevice] of WORD = ($EF, $F7);
   IntVect: array[TComDevice] of BYTE = ($0C, $0B);

   RBR = $00; {xF8 Receive Buffer Register            }
   THR = $00; {xF8 Transmitter Holding Register       }
   IER = $01; {xF9 Interrupt Enable Register          }
   IIR = $02; {xFA Interrupt Identification Register  }
   LCR = $03; {xFB Line Control Register              }
   MCR = $04; {xFC Modem Control Register             }
   LSR = $05; {xFD Line Status Register               }
   MSR = $06; {xFE Modem Status Register              }
   {--- if LCR Bit 7 = 1  ---              }
   DLL = $00; {xF8 Divisor Latch Low Byte             }
   DLH = $01; {xF9 Divisor Latch Hi  Byte             }
   CMD8259 = $20; {Interrupt Controller Command Register  }
   IMR8259 = $21; {Interrupt Controller Mask Register     }
   EOI = $20; {End of Interrupt}

   {Should be evaluated by any higher-level send-routine}
   LoopLimit = 1000; {When does a timeout-error occur        }

type
   TComOpenDevices = set of TComDevice;
var
   OldComInts: array[TComDevice] of DpmiPmVector;
   ComOpenDevices: TComOpenDevices;
   ActiveDevices: array[TComDevice] of TwdxComCtrl;

procedure SaveOldInts;
var
   i: TComDevice;
begin
   for i := Low(TComDevice) to High(TComDevice) do begin
      Exclude(ComOpenDevices, i);
      ActiveDevices[i] := nil;
      GetIntVec(IntVect[i], OldComInts[i]);
   end;
end;

procedure ResetOldInts;
var
   i: TComDevice;
begin
   for i := Low(TComDevice) to High(TComDevice) do begin
      if Assigned(ActiveDevices[i]) and
         Assigned(OldComInts[i].PmOffset) then begin
         ActiveDevices[i].Close;
         SetIntVec(IntVect[i], OldComInts[i]);
      end;
   end;
end;

procedure V24Int1;
begin
   if Assigned(ActiveDevices[Com1]) then begin
      ActiveDevices[Com1].ReadBytes;
   end;
   CallInterrupt(OldComInts[Com1]);
end;

procedure V24Int2;
begin
   if Assigned(ActiveDevices[Com2]) then begin
      ActiveDevices[Com2].ReadBytes;
   end;
   CallInterrupt(OldComInts[Com2]);
end;

{ TwdxComCtrl }

class function TwdxComCtrl.ComDeviceBaseAddr(ComDevice: TComDevice): Word;
const
   Offset: array[TComDevice] of Byte = (0, 2);
begin
   result := MemW[$400 + Offset[ComDevice]];
end;

constructor TwdxComCtrl.Create(Owner: TComponent);
begin
   inherited create(Owner);
   FStreamedActive := False;
   FActive := False;
   FComDevice := Com1;
   FBaudrate := b9600;
   FParity := none;
   FDataBits := d8;
   FStopBits := s1;
   FRTS := true;
   FDTR := true;
   FBufferSize := 4096;
   FBuffer := nil;
end;

destructor TwdxComCtrl.destroy;
begin
   Close;
end;

procedure TwdxComCtrl.Open;
const
   IntProc: array[TComDevice] of TProcedure = (V24Int1, V24Int2);
begin
   if not FActive then begin
      if (FComDevice in ComOpenDevices) then begin
         raise EComCtrlError.CreateFmt(sComCtrlOpenError, [ComDeviceStr]);
      end;

      Include(ComOpenDevices, FComDevice);
      ActiveDevices[FComDevice] := Self;
      FActive := True;

      AllocBuffer;

      FComDeviceBaseAddr := ComDeviceBaseAddr(FComDevice);

      ClearPendingInterrupts;

      SetIntProc(IntVect[FComDevice], IntProc[FComDevice]);

      InitCom;
   end;
end;

procedure TwdxComCtrl.Close;
begin
   if FActive then begin
      ResetCom;

      SetIntVec(IntVect[ComDevice], OldComInts[ComDevice]);

      FreeBuffer;

      Exclude(ComOpenDevices, FComDevice);
      ActiveDevices[FComDevice] := nil;
      FActive := False;
   end;
end;

procedure TwdxComCtrl.AllocBuffer;
begin
   FreeBuffer;
   FBufferTail := 0;
   FBufferHead := 0;
   FBuffer := AllocMem(FBufferSize);
end;

procedure TwdxComCtrl.FreeBuffer;
begin
   if assigned(FBuffer) then begin
      FreeMem(FBuffer);
      FBuffer := nil;
   end;
end;

procedure TwdxComCtrl.InitCom;
var
   b: byte;
begin
   asm
      cli
   end;
   PORT[FComDeviceBaseAddr + LCR] := $80; {Adress Divisor Latch}
   PORT[FComDeviceBaseAddr + DLH] := Hi(BaudConst[Baudrate]); {Set Baud rate}
   PORT[FComDeviceBaseAddr + DLL] := Lo(BaudConst[Baudrate]);
   PORT[FComDeviceBaseAddr + LCR] := ($00 or ParityConst[Parity] {Setup Parity}
      or BitsConst[DataBits] {Setup number of databits}
      or StopConst[StopBits]); {Setup number of stopbits}
   b := $08;
   if DTR then
      b := b or 1;
   if RTS then
      b := b or 2;
   PORT[FComDeviceBaseAddr + MCR] := B;
   PORT[FComDeviceBaseAddr + IER] := $01; {Enable Data-Available Interrupts}
   PORT[IMR8259] := PORT[IMR8259] and IntMasks[ComDevice]; {Enable V24-Interrups}
   asm
      STI
   end;
end;

procedure TwdxComCtrl.ResetCom;
begin
   asm
      CLI
   end;
   PORT[FComDeviceBaseAddr + MCR] := 00; {Disable Interrupts, Reset MCR}
   PORT[IMR8259] := PORT[IMR8259] or not (IntMasks[ComDevice]);
   {Disable Interrupt Level 3 and 4}
   PORT[FComDeviceBaseAddr + IER] := 0; {Disable 8250-Interrupts}
   ClearPendingInterrupts; {Clean up}
   asm
      STI
   end;
end;

procedure TwdxComCtrl.ClearPendingInterrupts;
begin {ClearPendingInterrupts}
   while (PORT[FComDeviceBaseAddr + IIR] and 1) = 0 do
   {While Interrupts are pending} begin
      {N := } PORT[FComDeviceBaseAddr + LSR]; {Read Line Status}
      {N := } PORT[FComDeviceBaseAddr + MSR]; {Read Modem Status}
      {N := } PORT[FComDeviceBaseAddr + RBR]; {Read Receive Buffer Register}
      PORT[CMD8259] := EOI; {End of Interrupt}
   end;
end; {ClearPendingInterrupts}

function TwdxComCtrl.DataAvail: BOOLEAN;
begin
   Result := FBufferHead <> FBufferTail;
end;

function TwdxComCtrl.GetByte: BYTE;
begin
   if FActive then begin
      Result := FBuffer[FBufferTail];
      Inc(FBufferTail);
      if FbufferTail >= BufferSize then begin
         FBuffertail := 0;
      end;
   end
   else begin
      raise EComCtrlError.CreateFmt(sComCtrlNotOpen, [ComDeviceStr]);
   end;
end;

procedure TwdxComCtrl.Send(Data: BYTE);
var
   Count: integer;
begin
   if FActive then begin
      Count := 0;
      repeat
         Count := SUCC(Count);
      until ((PORT[FComDeviceBaseAddr + LSR] and $20) <> 0) or (Count >
         LoopLimit);
      if Count > LoopLimit then begin
         raise EComCtrlError.CreateFmt(sComCtrlTimeOut, [ComDeviceStr]);
      end
      else begin
         PORT[FComDeviceBaseAddr + THR] := Data;
      end;
   end
   else begin
      raise EComCtrlError.CreateFmt(sComCtrlNotOpen, [ComDeviceStr]);
   end;
end;

procedure TwdxComCtrl.Send(Data: string);
var
   i: integer;
begin
   for i:=1 to Length(Data) do begin
     send(Ord(Data[i]));
   end;
end;


procedure TwdxComCtrl.SetComDeviceStr(AComDevice: string);
var
   s: string;
begin
   s := UpperCase(AComDevice);
   if s = 'COM1' then
      FComDevice := COM1
   else if s = 'COM2' then
      FComDevice := COM2
   else
      FComDevice := COM1;
end;

procedure TwdxComCtrl.SetBaudRateStr(ABaudrate: string);
begin
   if ABaudRate = '110' then
      FBaudrate := b110
   else if ABaudRate = '150' then
      FBaudrate := b150
   else if ABaudRate = '300' then
      FBaudrate := b300
   else if ABaudRate = '600' then
      FBaudrate := b600
   else if ABaudRate = '1200' then
      FBaudrate := b1200
   else if ABaudRate = '2400' then
      FBaudrate := b2400
   else if ABaudRate = '4800' then
      FBaudrate := b4800
   else if ABaudRate = '9600' then
      FBaudrate := b9600
   else if ABaudRate = '19200' then
      FBaudrate := b19200
   else if ABaudRate = '38400' then
      FBaudrate := b38400
   else if ABaudRate = '57600' then
      FBaudrate := b57600
   else if ABaudRate = '115200' then
      FBaudrate := b115200
   else
      FBaudrate := b9600;
end;

procedure TwdxComCtrl.SetParityStr(AParity: string);
var
   s: string;
begin
   s := UpperCase(AParity);
   if s = 'SPACE' then
      FParity := Space
   else if s = 'ODD' then
      FParity := Odd
   else if s = 'MARK' then
      FParity := Mark
   else if s = 'EVEN' then
      FParity := Even
   else if s = 'NONE' then
      FParity := None
   else
      FParity := None;
end;

procedure TwdxComCtrl.SetDataBitsStr(ADataBits: string);
begin
   if ADataBits = '5' then
      FDataBits := d5
   else if ADataBits = '6' then
      FDataBits := d6
   else if ADataBits = '7' then
      FDataBits := d7
   else if ADataBits = '8' then
      FDataBits := d8
   else
      FDataBits := d8;
end;

procedure TwdxComCtrl.SetStopBitsStr(AStopBits: string);
begin
   if AStopBits = '1' then
      FStopBits := s1
   else if AStopBits = '2' then
      FStopBits := s2
   else
      FStopBits := s1;
end;

procedure TwdxComCtrl.SetRTSStr(ARTS: string);
begin
   FRTS := Uppercase(ARTS) = 'RTS';
end;

procedure TwdxComCtrl.SetDTRStr(ADTR: string);
begin
   FDTR := Uppercase(ADTR) = 'DTR'
end;

function TwdxComCtrl.GetComDeviceStr: string;
begin
   case FComDevice of
      Com1: result := 'COM1';
      Com2: result := 'COM2';
   else
      ComDeviceStr := '';
   end;
end;

function TwdxComCtrl.GetBaudrateStr: string;
begin
   case FBaudrate of
      b110: result := '110';
      b150: result := '150';
      b300: result := '300';
      b600: result := '600';
      b1200: result := '1200';
      b2400: result := '2400';
      b4800: result := '4800';
      b9600: result := '9600';
      b19200: result := '19200';
      b38400: result := '38400';
      b57600: result := '57600';
      b115200: result := '115200';
   else
      result := '';
   end;
end;

function TwdxComCtrl.GetParityStr: string;
begin
   case FParity of
      None: result := 'NONE';
      Odd: result := 'ODD';
      Even: result := 'EVEN';
      Mark: result := 'MARK';
      Space: result := 'SPACE';
   else
      result := '';
   end;
end;

function TwdxComCtrl.GetDataBitsStr: string;
begin
   case FDataBits of
      d5: result := '5';
      d6: result := '6';
      d7: result := '7';
      d8: result := '8';
   else
      result := '';
   end;
end;

function TwdxComCtrl.GetStopBitsStr: string;
begin
   case FStopBits of
      s1: result := '1';
      s2: result := '2';
   else
      result := '';
   end;
end;

function TwdxComCtrl.GetRTSStr: string;
begin
   if FRTS then begin
      result := 'RTS';
   end
   else begin
      result := 'No RTS';
   end;
end;

function TwdxComCtrl.GetDTRStr: string;
begin
   if FDTR then begin
      result := 'DTR';
   end
   else begin
      result := 'No DTR';
   end;
end;

procedure TwdxComCtrl.SetActive(const Value: Boolean);
begin
   if csLoading in ComponentState then begin
      FStreamedActive := Value
   end
   else if Value then begin
      writeln('bin hier');
      Open;
   end
   else begin
      Close;
   end;
end;

procedure TwdxComCtrl.Loaded;
begin
   inherited Loaded;
   try
      Active := FStreamedActive;
   except
      { let Delphi handle exceptions in design modus }
      if csDesigning in ComponentState then
         Application.HandleException(Self)
      else
         raise;
   end;
end;

procedure TwdxComCtrl.ReadBytes;
begin
   Buffer[FBufferHead] := Port[FComDeviceBaseAddr];
   Inc(FBufferHead);
   if FBufferHead > FBufferSize then begin
      FBufferHead := 0;
   end;
   Port[CMD8259] := EOI;

   if (FBufferHead = FBufferTail) then begin
      raise EComCtrlError.CreateFmt(sComCtrlOverrun, [ComDeviceStr]);
   end;

   if Assigned(OnReciveByteProc) then begin
      OnReciveByteProc(Self);
   end;
end;

initialization
   SaveOldInts;

finalization
   ResetOldInts;

end.

