
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosDHCP.pas                                          }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.6.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache   immo.wache@dwp42.org                   }
{   www:         http://www.dwp42.org                   }
{                                                       }
{   Joe Da Silva joe.dasilva@emailmetering.com          }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosDHCP.pas,v 1.2 2003/11/05 18:15:27 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosDHCP.pas,v $
  Revision 1.2  2003/11/05 18:15:27  iwache
  Changes by Joe:
  - New feature for adjustable DHCP timeout.
  - Error handling "safety measures" in ReadOption.
  - The use of a temporary DHCP record, Rsp, in SendReceive.
  - Minor changes to comments and messages.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosDHCP;

interface

uses
  SysUtils, WDosClock, WDosUtil, WDosNetwork, WDosIP, WDosSock;

function DHCP_ConfigAdapter(Adapter: TNetworkAdapter; var IPAddress,
  NetMask, Gateway, DHCPServer: TIPAddress; var Timeout: DWORD): Boolean;
function DHCP_Release(Adapter: TNetworkAdapter;
  IPAddress, ServerAddress: TIPAddress): Boolean;
function DHCP_GetInfo(Adapter: TNetworkAdapter;
  IPAddress, DHCPServer: TIPAddress; Option: BYTE;
  var Buffer; var BuffLen: Integer): Boolean;
function DHCP_RenewIP(Adapter: TNetworkAdapter; IPAddress,
  ServerAddress: TIPAddress; var Timeout: DWORD): Boolean;

var
  DHCP_Timeout: Integer = 3000;

implementation

type
  TDHCPRec = packed record
    op: BYTE;
    htype: BYTE;
    hlen: BYTE;
    hops: BYTE;
    xid: DWORD;
    secs: WORD;
    flags: WORD;
    ciaddr: DWORD;
    yiaddr: DWORD;
    siaddr: DWORD;
    giaddr: DWORD;
    chaddr: packed array[0..15] of BYTE;
    sname: packed array[0..63] of CHAR;
    cfile: packed array[0..127] of CHAR;
    options: packed array[0..311] of BYTE;
  end;

procedure InitRec(var Rec: TDHCPRec; op: Integer; HAddr: TEthernetAddress);
begin
  // Set basic Fields
  FillChar(Rec, SizeOf(Rec), 0);
  Rec.op := op;
  Rec.htype := 1;
  Rec.hlen := 6;
  Rec.secs := 60;
  Rec.flags := 0;
  Move(HAddr, Rec.chaddr, 6);
  // Clear Options
  Rec.options[0] := 99;
  Rec.options[1] := 130;
  Rec.options[2] := 83;
  Rec.options[3] := 99;
  Rec.options[4] := 255;
end;

procedure AddOption(var Rec: TDHCPRec; OpType: BYTE; Data: AnsiString);
var
  i, j: Integer;
begin
  for i := High(Rec.options) downto 0 do
  begin
    if Rec.options[i] = $FF then
    begin
      Rec.options[i] := OpType;
      Rec.options[i + 1] := Length(Data);
      for j := 1 to Length(Data) do
      begin
        Rec.options[i + j + 1] := BYTE(Data[j]);
      end;
      Rec.options[i + Length(Data) + 2] := $FF;
      Exit;
    end;
  end;
end;

function ReadOption(var Rec: TDHCPRec; OpType: BYTE): AnsiString;
var
  P: PChar;
  i, Len: Integer;
begin
  (*
  P := @Rec.options;
  for i := 0 to sizeof(rec.options) - 1 do
  begin
    Write(BYTE(P[i]), ' ');
  end;
  WriteLn;
  ReadLn;
  *)
  Result := '';
  P := @Rec.options[4];
  repeat
    // We have received a record
    if Integer(P) - Integer(@Rec.options) >= SizeOf(Rec.options) - 1 then
      Exit;
    if BYTE(P[0]) = 255 then
      Exit;
    Len := BYTE(P[1]);
    if BYTE(P[0]) = OpType then
    begin
      // Prevent access violations
      if Integer(P) - Integer(@Rec.options) + Len + 2 <= SizeOf(Rec.options)
        then
      begin
        for i := 1 to Len do
          Result := Result + P[1 + i];
      end;
      Exit;
    end;
    Inc(P, Len + 2);
  until False;
end;

function CalcSize(var Rec: TDHCPRec): Integer;
var
  i: Integer;
begin
  Result := SizeOf(TDHCPRec);
  for i := High(Rec.options) downto 0 do
  begin
    if Rec.options[i] = $FF then
      break;
    Dec(Result);
  end;
end;

function GetClientID(HAddress: TEthernetAddress): AnsiString;
var
  i: Integer;
begin
  Result := #1;
  for i := 0 to 5 do
    Result := Result + Chr(HAddress[i]);
end;

function SendReceive(HSocket: Integer; var Rec: TDHCPRec;
  EsperarRespuesta: Boolean): Boolean;
var
  Rsp: TDHCPRec;
  Timeout: Int64;
  i, ID, RcvLen, attempts, time_out: DWORD;
begin
  Result := False;
  ID := Random($8FFFFFF);
  Rec.xid := ID;
  // Calculate the number of retries and the interval between them
  attempts := 10;
  time_out := DHCP_Timeout div Int64(attempts);
  if (time_out < 500) then
  begin
    time_out := 500;
    attempts := DHCP_Timeout div 500;
    if attempts < 1 then
      attempts := 1;
  end;
  for i := 1 to attempts do
  begin
    if send(HSocket, Rec, CalcSize(Rec), 0) <= 0 then
      Exit;
    if not EsperarRespuesta then
      Result := True
    else
    begin
      Timeout := GetMSCounter + time_out;
      while GetMSCounter < Timeout do
      begin
        ioctlsocket(HSocket, FIONREAD, @RcvLen);
        if (RcvLen > 0) then
        begin
          recv(HSocket, Rsp, RcvLen, 0);
          if (Rsp.xid = ID) then
          begin
            (*
            SOpt := ReadOption(Rsp, 53);
            if Length(SOpt) > 0 then
            begin
              Move(SOpt[1], B, 1);
              WriteLn('Response: ', B);
            end
            else
            begin
              WriteLn('No response');
            end;
            *)
            Rec := Rsp;
            Result := True;
            Break;
          end;
        end;
      end;
    end;
    if Result then
      Break;
  end;
end;

function DHCP_Discover(HAddress: TEthernetAddress; var NewIPAddress,
  ServerAddress: TIPAddress): Boolean;
var
  SOpt: AnsiString;
  HSocket: Integer;
  ClientRec: TDHCPRec;
  SockAddr: TSockAddr;
  ClientID: AnsiString;
begin
  Result := False;

  { Build the Client ID based on the Hardware Address of the Adapter }
  ClientID := GetClientID(HAddress);

  { Create a UDP socket for connecting to the DHCP Server }
  HSocket := socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
  if HSocket <= 0 then
  begin
    // WriteLn('DHCP: Error in socket()');
    Exit;
  end;

  { Bind to the "DHCP Client" port }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := INADDR_ANY;
  SockAddr.sin_port := Swap(Word(68));
  if bind(HSocket, SockAddr, SizeOf(SockAddr)) <> 0 then
  begin
    // WriteLn('DHCP: Error in bind()');
    CloseSocket(HSocket);
    Exit;
  end;

  { Connect to the "DHCP Server" port }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := INADDR_BROADCAST;
  SockAddr.sin_port := Swap(Word(67));
  if connect(HSocket, SockAddr, SizeOf(SockAddr)) <> 0 then
  begin
    // WriteLn('DHCP: Error in connect()');
    CloseSocket(HSocket);
    Exit;
  end;

  { Set-Up a record with the DHCP-Discover information }
  InitRec(ClientRec, 1, HAddress);

  // DHCP message Type = 1 (DHCPDISCOVER)
  AddOption(ClientRec, 53, #1);

  // Client identifier
  AddOption(ClientRec, 61, ClientID);

  { Send a DHCP-Discover and wait for the first reply }
  if not SendReceive(HSocket, ClientRec, True) then
  begin
    // No server replied to our request
    // WriteLn('DHCP: Error in SendReceive()');
    CloseSocket(HSocket);
    Exit;
  end;
  CloseSocket(HSocket);

  // Check if this is a DHCPOFFER
  if ReadOption(ClientRec, 53) = #2 then
  begin
    // WriteLn('DHCPOFFER');
  end
  else
  begin
    // WriteLn('DHCP Error: message type = ', Ord(SOpt[1]),' (should be 2)');
    Exit;
  end;

  // Retrieve the IP Address
  NewIPAddress := ClientRec.yiaddr;

  // Retrieve option 54 (Server IP Address)
  SOpt := ReadOption(ClientRec, 54);
  if Length(SOpt) <> 4 then
  begin // No Server IP ?
    // WriteLn('DHCP: No server IP');
    Exit;
  end;
  Move(SOpt[1], ServerAddress, 4);

  // Listo
  Result := True;
end;

function DHCP_Request(HAddress: TEthernetAddress; IPAddress, DHCPServer:
  TIPAddress;
  var NetMask, Gateway: TIPAddress; Renew: Boolean; var Timeout: DWORD):
  Boolean;
var
  SOpt: AnsiString;
  HSocket: Integer;
  ClientRec: TDHCPRec;
  ClientID: AnsiString;
  SockAddr: TSockAddr;
begin
  Result := False;

  { Build the Client ID based on the Hardware Address of the Adapter }
  ClientID := GetClientID(HAddress);

  { Create a UDP socket for connecting to the DHCP Server }
  HSocket := socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
  if HSocket <= 0 then
  begin
    Exit;
  end;

  { Bind to the "DHCP Client" port }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := INADDR_ANY;
  SockAddr.sin_port := Swap(Word(68));
  if bind(HSocket, SockAddr, SizeOf(SockAddr)) <> 0 then
  begin
    CloseSocket(HSocket);
    Exit;
  end;

  { Connect to the "DHCP Server" port }
  SockAddr.sin_family := AF_INET;
  if Renew then
    SockAddr.sin_addr.S_addr := DHCPServer
  else
    SockAddr.sin_addr.S_addr := INADDR_BROADCAST;
  SockAddr.sin_port := Swap(Word(67));
  if connect(HSocket, SockAddr, SizeOf(SockAddr)) <> 0 then
  begin
    CloseSocket(HSocket);
    Exit;
  end;

  // Set-Up a record with the DHCP-Request information
  InitRec(ClientRec, 1, HAddress);

  // DHCP message Type = 3 (DHCPREQUEST)
  AddOption(ClientRec, 53, #3);

  // Client Identifier
  AddOption(ClientRec, 61, ClientID);

  // Requested Options =
  //  Gateway                (3)
  //  Net Mask               (4)
  //  IP Address Lease Time (51)
  AddOption(ClientRec, 55, #1#3#51);

  // Server Identifier
  SetLength(SOpt, 4);
  Move(DHCPServer, SOpt[1], 4);
  AddOption(ClientRec, 54, SOpt);

  // Host Name
  if Net_GetHostName <> '' then
    AddOption(ClientRec, 12, Net_GetHostName + #0);

  // Other options
  if Renew then
  begin
    Clientrec.ciaddr := IPAddress;
  end
  else
  begin
    // Requested IP Address
    SetLength(SOpt, 4);
    Move(IPAddress, SOpt[1], 4);
    AddOption(ClientRec, 50, SOpt);
  end;

  if not SendReceive(HSocket, ClientRec, True) then
  begin
    // No server replied our request
    CloseSocket(HSocket);
    Exit;
  end;
  CloseSocket(HSocket);

  // Check if this is a DHCPACK
  if ReadOption(ClientRec, 53) = #5 then
  begin
    // WriteLn('DHCPACK');
  end
  else
  begin
    // WriteLn('DHCP Error: message type = ', Ord(SOpt[1]),' (should be 5)');
    Exit;
  end;

  // Retrieve the NetMask
  SOpt := ReadOption(ClientRec, 1);
  if Length(SOpt) = 4 then
  begin
    Move(SOpt[1], NetMask, 4);
  end
  else
  begin
    NetMask := 0;
    // WriteLn('DHCP: No NetMask');
  end;

  // Retrieve the Gateway
  SOpt := ReadOption(ClientRec, 3);
  if Length(SOpt) = 4 then
  begin
    Move(SOpt[1], Gateway, 4);
  end
  else
  begin
    Gateway := 0;
    // WriteLn('DHCP: No gateway');
  end;

  // Retrieve the Lease duration
  SOpt := ReadOption(ClientRec, 51);
  if Length(SOpt) = 4 then
  begin
    Move(SOpt[1], Timeout, 4);
    Timeout := SwapDWord(Timeout);
  end
  else
  begin
    Timeout := 0;
    // WriteLn('DHCP: No Lease Time');
  end;

  { Done }
  Result := True;
end;

function DHCP_GetInfo(Adapter: TNetworkAdapter;
  IPAddress, DHCPServer: TIPAddress; Option: BYTE;
  var Buffer; var BuffLen: Integer): Boolean;
var
  SOpt: AnsiString;
  HSocket: Integer;
  ClientRec: TDHCPRec;
  ClientID: AnsiString;
  SockAddr: TSockAddr;
  HAddress: TEthernetAddress;
begin
  Result := False;

  { Build the Client ID based on the Hardware Address of the Adapter }
  Adapter.GetAddress(HAddress);
  ClientID := GetClientID(HAddress);

  { Create a UDP socket for connecting to the DHCP Server }
  HSocket := socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
  if HSocket <= 0 then
  begin
    Exit;
  end;

  { Bind to the "DHCP Client" port }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := INADDR_ANY;
  SockAddr.sin_port := Swap(Word(68));
  if bind(HSocket, SockAddr, SizeOf(SockAddr)) <> 0 then
  begin
    CloseSocket(HSocket);
    Exit;
  end;

  { Connect to the "DHCP Server" port }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := DHCPServer;
  SockAddr.sin_port := Swap(Word(67));
  if connect(HSocket, SockAddr, SizeOf(SockAddr)) <> 0 then
  begin
    CloseSocket(HSocket);
    Exit;
  end;

  // Set-Up a record with the DHCP-Request information
  SetLength(SOpt, 4);
  InitRec(ClientRec, 1, HAddress);
  AddOption(ClientRec, 53, #3);
  AddOption(ClientRec, 61, ClientID);
  Move(DHCPServer, SOpt[1], 4);
  AddOption(ClientRec, 54, SOpt);
  Move(IPAddress, SOpt[1], 4);
  AddOption(ClientRec, 50, SOpt);
  AddOption(ClientRec, 55, Chr(Option));
  if Net_GetHostName <> '' then
    AddOption(ClientRec, 12, Net_GetHostName + #0);
  if not SendReceive(HSocket, ClientRec, True) then
  begin
    // No server replied our request
    CloseSocket(HSocket);
    Exit;
  end;
  CloseSocket(HSocket);

  { Get the Option }
  SOpt := ReadOption(ClientRec, Option);
  if (SOpt <> '') and (BuffLen >= Length(SOpt)) then
  begin
    BuffLen := Length(SOpt);
    Move(SOpt[1], Buffer, BuffLen);
    Result := True;
  end
  else
  begin
    Result := False;
  end;

end;

function DHCP_Release(Adapter: TNetworkAdapter;
  IPAddress, ServerAddress: TIPAddress): Boolean;
var
  SOpt: AnsiString;
  HSocket: Integer;
  ClientRec: TDHCPRec;
  ClientID: AnsiString;
  SockAddr: TSockAddr;
  HAddress: TEthernetAddress;
begin
  Result := False;

  { Build the Client ID based on the Hardware Address of the Adapter }
  Adapter.GetAddress(HAddress);
  ClientID := GetClientID(HAddress);

  { Create a UDP socket for connecting to the DHCP Server }
  HSocket := socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
  if HSocket <= 0 then
  begin
    Exit;
  end;

  { Bind to the "DHCP Client" port }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := INADDR_ANY;
  SockAddr.sin_port := Swap(Word(68));
  if bind(HSocket, SockAddr, SizeOf(SockAddr)) <> 0 then
  begin
    CloseSocket(HSocket);
    Exit;
  end;

  { Connect to the "DHCP Server" port }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := ServerAddress;
  SockAddr.sin_port := Swap(Word(67));
  if connect(HSocket, SockAddr, SizeOf(SockAddr)) <> 0 then
  begin
    CloseSocket(HSocket);
    Exit;
  end;

  // Set-Up a record with the DHCP-Release information
  SetLength(SOpt, 4);
  InitRec(ClientRec, 1, HAddress);
  ClientRec.ciaddr := IPAddress;
  //
  AddOption(ClientRec, 53, #7); // DHCP Option
  AddOption(ClientRec, 61, ClientID); // Client Identifier
  Move(ServerAddress, SOpt[1], 4);
  AddOption(ClientRec, 54, SOpt); // Server Identifier

  Result := SendReceive(HSocket, ClientRec, False);
  CloseSocket(HSocket);
end;

function DHCP_ConfigAdapter(Adapter: TNetworkAdapter; var IPAddress,
  NetMask, Gateway, DHCPServer: TIPAddress; var Timeout: DWORD): Boolean;
var
  EthAddr: TEthernetAddress;
begin
  Result := False;
  { Get Adapter Hardware Address and build ClientID }
  if not Adapter.GetAddress(EthAddr) then
  begin
    // WriteLn('DHCP: Error in GetAddress');
    Exit;
  end;
  // Send a DCHP-Discover and wait for a reply
  if not DHCP_Discover(EthAddr, IPAddress, DHCPServer) then
  begin
    // WriteLn('Error in DHCP_Discover');
    Exit;
  end;
  // WriteLn('DHCP Server: ', IP_AddrToString(DHCPServer), ' IP: ', IP_AddrToString(IPAddress));
  // Send a DHCP-Request and wait for the reply
  if not DHCP_Request(EthAddr, IPAddress, DHCPServer, NetMask, Gateway, False,
    Timeout) then
  begin
    // WriteLn('Error in DHCP_Request');
    Exit;
  end;
  // Validations
  if (IPAddress = 0) or (NetMask = 0) then
    Exit;
  // Adjustments
  if Timeout = 0 then
    Timeout := 60 * 60 * 24; // 1 day in seconds
  // Done
  Result := True;
end;

function DHCP_RenewIP(Adapter: TNetworkAdapter; IPAddress,
  ServerAddress: TIPAddress; var Timeout: DWORD): Boolean;
var
  EthAddr: TEthernetAddress;
  NetMask, Gateway: TIPAddress;
begin
  Result := False;
  { Get Adapter Hardware Address and build ClientID }
  if not Adapter.GetAddress(EthAddr) then
    Exit;
  // Send a DHCP-Request and wait for the reply
  if not DHCP_Request(EthAddr, IPAddress, ServerAddress, NetMask, Gateway, True,
    Timeout) then
    Exit;
  // Adjustments
  if Timeout = 0 then
    Timeout := 60 * 60 * 24; // 1 day in seconds
  { Done }
  Result := True;
end;

end.