
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosDNS.pas                                           }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosDNS.pas,v 1.2 2003/10/25 22:00:07 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosDNS.pas,v $
  Revision 1.2  2003/10/25 22:00:07  iwache
  Minor changes in coding format added. No functional change.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosDNS;

interface

uses
  SysUtils, WDosSock, WDosNetwork, WDosIP, WDosUDP, Threads, WDosClock;

const
  DNS_DEFAULT_PORT = Swap(WORD(53));

procedure SetDNSServer(ServerIP: TIPAddress; Port: WORD);
function GetHostByName(HostName: AnsiString; var IPAddress: TIPAddress):
  Boolean;

implementation

type
  PWORD = ^WORD;

  PDNSHeader = ^TDNSHeader;
  TDNSHeader = packed record
    ID: WORD;
    Param0: BYTE;
    Param1: BYTE;
    RequestCount: WORD;
    AnswerCount: WORD;
    AuthCount: WORD;
    AdditCount: WORD;
  end;

  TDNSServers = record
    Count: Integer;
    Server: array[1..20] of string;
  end;

  TDNSCacheEntry = record
    HostName: string;
    IPAddress: TIPAddress;
    Expires: Int64;
  end;

var
  DNS_PORT: WORD;
  DNS_IP: TIPAddress;
  Cache: array[1..20] of TDNSCacheEntry;

procedure SetDNSServer(ServerIP: TIPAddress; Port: WORD);
begin
  DNS_IP := ServerIP;
  DNS_PORT := Port;
end;

{**************************************************************************}
{                                                                          }
{                          DNS Cache handling                              }
{                                                                          }
{**************************************************************************}

function Cache_GetHostByName(HostName: AnsiString; var IPAddress: TIPAddress):
  Boolean;
var
  i: Integer;
begin
  { We store host names as Upper Case }
  HostName := UpperCase(HostName);
  { Go through all cache entries, ignoring those that are expired }
  for i := Low(Cache) to High(Cache) do
  begin
    if (Cache[i].HostName = HostName) and (Cache[i].Expires > GetMSCounter) then
    begin
      { Found it! }
      IPAddress := Cache[i].IPAddress;
      Cache_GetHostByName := True;
      Exit;
    end;
  end;
  { No entry found for the given host name }
  Cache_GetHostByName := False;
end;

procedure Cache_AddHostByName(HostName: AnsiString; IPAddress: TIPAddress);
var
  i, Entry: Integer;
  TmpIP: TIPAddress;
begin
  { If it's already in the cache, exit }
  if Cache_GetHostByName(HostName, TmpIP) then
    Exit;
  { Chage to Upper Case }
  HostName := UpperCase(HostName);
  { Remove expired entries }
  for i := Low(Cache) to High(Cache) do
  begin
    if (Cache[i].HostName <> '') and (Cache[i].Expires < GetMSCounter) then
    begin
      Cache[i].HostName := '';
    end;
  end;
  { Search for an empty entry }
  Entry := Random(High(Cache)) + 1;
  for i := Low(Cache) to High(Cache) do
  begin
    if (Cache[i].HostName = '') then
    begin
      Entry := i;
      Break;
    end;
  end;
  { Copy the given information to the selected entry }
  Cache[Entry].IPAddress := IPAddress;
  Cache[Entry].HostName := HostName;
  Cache[Entry].Expires := GetMSCounter + 40000; { valid for the next 40 seconds }
end;

{**************************************************************************}
{                                                                          }
{                     DNS Server message processing                        }
{                                                                          }
{**************************************************************************}

function GetResourceName(Org: PChar; var P: PChar): AnsiString;
var
  b: BYTE;
  q: PChar;
  s: string;
  i: Integer;
  Comp: Boolean;
begin
  s := '';
  Comp := False;
  while P[0] <> #0 do
  begin
    B := BYTE(P[0]);
    if s <> '' then
      s := s + '.';
    if (B and 192) = 192 then
    begin
      Comp := True;
      Q := Org + (B and 63) * $100 + BYTE(P[1]);
      s := s + GetResourceName(Org, Q);
      Inc(P, 2);
      Break;
    end
    else
    begin
      for i := 1 to B do
        s := s + P[i];
      Inc(P, B + 1);
    end;
  end;
  if not Comp then
    Inc(P);
  GetResourceName := s;
end;

function MakeResourceName(s: AnsiString): AnsiString;
var
  i, n, q: Integer;
begin
  s := '.' + s;
  q := 1;
  n := 0;
  for i := 1 to Length(s) do
  begin
    if (s[i] = '.') then
    begin
      s[q] := Chr(n);
      q := i;
      n := 0;
    end
    else
    begin
      Inc(n);
    end;
  end;
  s[q] := Chr(n);
  MakeResourceName := s;
end;

procedure ProcessReq(Org: PChar; var P: PChar);
begin
  GetResourceName(Org, P);
  Inc(P, 4);
end;

procedure ProcessAns(Org: PChar; var DNSServers: TDNSServers; var CanonicalName:
  AnsiString; var P: PChar);
var
  //	TTL: Cardinal;
  DNS2: PChar;
  ResourceName: string;
  ResType, {ResClass, } DataLen: WORD;
begin
  (*	WriteLn('Answer');	*)
      { Process the resource name }
  ResourceName := GetResourceName(Org, P);
  { Process the resource's type, class, TTL and DataLen }
  ResType := Swap(PWORD(P)^);
  Inc(P, 2);
  //    ResClass := Swap(PWORD(P)^);
  Inc(P, 2);
  //    TTL      := SwapCardinal(PCardinal(P)^);
  Inc(P, 4);
  DataLen := Swap(PWORD(P)^);
  Inc(P, 2);
  { Obtain other resources's data }
{	WriteLn('Resource Name    : ', ResourceName);
WriteLn('Respuesta de tipo: ', ResType);
WriteLn('Clase            : ', ResClass);
  WriteLn('DataLen          : ', DataLen);
}
  case ResType of
    1: { Host IP Address }
      begin
        //          	WriteLn('***** IP *****');
        Cache_AddHostByName(ResourceName, PIPAddress(P)^);
      end;

    2: { Domain Server }
      begin
        //                WriteLn('***** DNS *****');
        if DNSServers.Count < High(DNSServers.Server) then
        begin
          DNS2 := P;
          Inc(DNSServers.Count);
          DNSServers.Server[DNSServers.Count] := GetResourceName(Org, DNS2);
        end;
      end;
    5: { Canonical name for a given Alias }
      begin
        //         		WriteLn('***** Canonical Name *****');
        DNS2 := P;
        CanonicalName := GetResourceName(Org, DNS2);
        //              WriteLn('Canonical Name: ', CanonicalName);
      end;
  end;
  P := @P[DataLen];
end;

procedure ProcessDNSRequest(P: PDNSHeader; var DNSServers: TDNSServers; var
  CanonicalName: AnsiString);
var
  //	c: Char;
  PC: PChar;
  i: Integer;
begin
  (*	for i := 0 to 150 do begin
    c := PChar(P)[i];
    case c of
     #0..#32: Write('<', ord(c), '>');
     else Write(c);
    end;
   end;
   WriteLn;
  *)
   { Procesamos the DNS Reply packet }
  PC := PChar(P) + SizeOf(TDNSHeader);
  (*	WriteLn('Param     : ', P^.Param0, ',', P^.Param1);
   WriteLn('Preguntas : ', Swap(P^.RequestCount));
   WriteLn('Respuestas: ', Swap(P^.AnswerCount));
   WriteLn('Auth      : ', Swap(P^.AuthCount));
   WriteLn('Additional: ', Swap(P^.AdditCount));
  *)

      { Procesamos las Preguntas }
  for i := 1 to Swap(P^.RequestCount) do
    ProcessReq(PChar(P), PC);
  { Procesamos las Respuestas }
  for i := 1 to Swap(P^.AnswerCount) do
    ProcessAns(PChar(P), DNSServers, CanonicalName, PC);
  { Procesamos las autorizaciones }
  for i := 1 to Swap(P^.AuthCount) do
    ProcessAns(PChar(P), DNSServers, CanonicalName, PC);
  { Procesamos los Adicionales }
  for i := 1 to Swap(P^.AdditCount) do
    ProcessAns(PChar(P), DNSServers, CanonicalName, PC);
end;

function QueryDNSServers(NS_IP: TIPAddress; HostName: AnsiString; var
  DNSServers: TDNSServers): Boolean;
{
  This function returns a list of DNS Servers capable of resolving a given
  Host Name. It returns False if there are no DNS Servers which can deal with
  the given Host Name.
}
var
  P: PChar;
  i: Integer;
  Timeout: Int64;
  RcvLen: Cardinal;
  HSocket: Pointer;
  ServerOk: Boolean;
  SockAddr: TSockAddr;
  ResourceName: string;
  ReqType, ReqClass: WORD;
  CanonicalName: AnsiString;
  DNSRequest, DNSReply: PDNSHeader;
begin
  //	WriteLn('Querying Server ', NS_IP);
   { Allocate memory for the request }
  if not GetMemCheck(Pointer(DNSRequest), SizeOf(TDNSHeader) + 1 +
    Length(HostName) + 1 + 2 + 2) then
  begin
    Result := False;
    Exit;
  end;
  { Build the header for the request }
  DNSRequest.ID := Random($FFFF) + 1;
  DNSRequest.Param0 := 1;
  DNSRequest.Param1 := 0;
  DNSRequest.RequestCount := Swap(Word(1));
  DNSRequest.AnswerCount := 0;
  DNSRequest.AuthCount := 0;
  DNSRequest.AdditCount := 0;
  ReqType := Swap(Word(2)); { Ask for DNS Servers }
  ReqClass := Swap(Word(1));
  { Build the data for the request }
  P := PChar(DNSRequest) + SizeOf(TDNSHeader);
  ResourceName := MakeResourceName(HostName);
  Move(ResourceName[1], P[0], Length(ResourceName));
  Inc(P, Length(ResourceName));
  P[0] := #0;
  Inc(P, 1);
  Move(ReqType, P[0], SizeOf(ReqType));
  Inc(P, SizeOf(ReqType));
  Move(ReqClass, P[0], SizeOf(ReqClass));
  { Open an UDP socket to connect to the DNS Server }
  HSocket := UDP_Socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
  { Bind it to any local adapter }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := INADDR_ANY;
  SockAddr.sin_port := 0;
  UDP_Bind(HSocket, SockAddr, SizeOf(SockAddr));
  { Connect to the DNS Server }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := NS_IP;
  SockAddr.sin_port := DNS_PORT;
  UDP_Connect(HSocket, SockAddr, SizeOf(SockAddr));
  { Loop waiting for a reply }
  ServerOk := False;
  for i := 1 to 2 do
  begin { 2 tries, every 5 seconds }
    { Send the request }
    if UDP_Send(HSocket, DNSRequest^, SizeOf(TDNSHeader) + 1 +
      Length(HostName) + 1 + 2 + 2, 0) = SOCKET_ERROR then
      Break;
    //		WriteLn('Sent');
            { Wait for a reply }
    Timeout := GetMSCounter + 5000; // 5 seconds
    while (GetMSCounter < Timeout) and not ServerOk do
    begin
      UDP_ioctlsocket(HSocket, FIONREAD, @RcvLen);
      if RcvLen > 0 then
      begin
        { We received a reply. Process it }
//				WriteLn('Reply!');
        GetMem(DNSReply, RcvLen);
        UDP_Recv(HSocket, DNSReply^, RcvLen, 0);
        ProcessDNSRequest(DNSReply, DNSServers, CanonicalName);
        FreeMem(DNSReply, RcvLen);
        ServerOk := True;
      end
      else
      begin
        Yield;
      end;
    end;
    if ServerOk then
      Break;
  end;
  { Close the socket }
  UDP_Close(HSocket);
  { Free the allocated memory }
  FreeMem(DNSRequest, SizeOf(TDNSHeader) + 1 + Length(HostName) + 1 + 2 + 2);
  { Done }
  Result := (DNSServers.Count > 0);
end;

function QueryHostAddress(NS_IP: TIPAddress; HostName: AnsiString; var
  CanonicalName: AnsiString): Boolean;
var
  P: PChar;
  i: Integer;
  RcvLen: Cardinal;
  Timeout: Int64;
  HSocket: Pointer;
  ServerOk: Boolean;
  ResourceName: string;
  SockAddr: TSockAddr;
  DNSServers: TDNSServers;
  HostAddress: TIPAddress;
  ReqType, ReqClass: WORD;
  DNSRequest, DNSReply: PDNSHeader;
begin
  QueryHostAddress := False;
  { Pedimos memoria para el Pedido }
  if not GetMemCheck(Pointer(DNSRequest),
    SizeOf(TDNSHeader) + 1 + Length(HostName) + 1 + 2 + 2) then
    Exit;
  { Armamos el Header }
  DNSRequest^.ID := Random($FFFF) + 1;
  DNSRequest^.Param0 := 1;
  DNSRequest^.Param1 := 0;
  DNSRequest^.RequestCount := Swap(Word(1));
  DNSRequest^.AnswerCount := 0;
  DNSRequest^.AuthCount := 0;
  DNSRequest^.AdditCount := 0;
  ReqType := Swap(Word(1)); { Pedido de Address }
  ReqClass := Swap(Word(1));
  { Armamos los Datos }
  P := PChar(DNSRequest) + SizeOf(TDNSHeader);
  ResourceName := MakeResourceName(HostName);
  Move(ResourceName[1], P[0], Length(ResourceName));
  Inc(P, Length(ResourceName));
  P[0] := #0;
  Inc(P, 1);
  Move(ReqType, P[0], SizeOf(ReqType));
  Inc(P, SizeOf(ReqType));
  Move(ReqClass, P[0], SizeOf(ReqClass));
  { Abrimos un Socket para comunicarnos con el Servidor DNS }
  HSocket := UDP_Socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
  { Hacemos Bind }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := INADDR_ANY;
  SockAddr.sin_port := 0;
  UDP_Bind(HSocket, SockAddr, SizeOf(SockAddr));
  { Conectamos al equipo remoto }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := NS_IP;
  SockAddr.sin_port := DNS_PORT;
  UDP_Connect(HSocket, SockAddr, SizeOf(SockAddr));
  { Esperamos la respuesta }
  ServerOk := False;
  for i := 1 to 2 do
  begin { 2 intentos de 5 seg. con este server }
    { Enviamos el Pedido }
    if UDP_Send(HSocket, DNSRequest^, SizeOf(TDNSHeader) + 1 +
      Length(HostName) + 1 + 2 + 2, 0) = SOCKET_ERROR then
      Break;
    { Esperamos la respuesta }
    Timeout := GetMSCounter + 5000;
    while (GetMSCounter < Timeout) and not ServerOk do
    begin
      UDP_ioctlsocket(HSocket, FIONREAD, @RcvLen);
      if RcvLen > 0 then
      begin
        { Lleg la respuesta. Armamos un Buffer, y la copiamos }
        if GetMemCheck(Pointer(DNSReply), RcvLen) then
        begin
          UDP_Recv(HSocket, DNSReply^, RcvLen, 0);
          CanonicalName := '';
          ProcessDNSRequest(DNSReply, DNSServers, CanonicalName);
          { Recibimos una respuesta }
          QueryHostAddress := (CanonicalName <> '')
            or Cache_GetHostByName(HostName, HostAddress);
          FreeMem(DNSReply, RcvLen);
        end;
        ServerOk := True;
      end
      else
      begin
        Yield;
      end;
    end;
    if ServerOk then
      Break;
  end;
  { Liberamos la memoria del Pedido }
  FreeMem(DNSRequest, SizeOf(TDNSHeader) + 1 + Length(HostName) + 1 + 2 + 2);
  { Cerramos el Socket }
  UDP_Close(HSocket);
end;

function GetHostByName(HostName: AnsiString; var IPAddress: TIPAddress):
  Boolean;
var
  i: Integer;
  DNSServers: TDNSServers;
  DNSServerIP: TIPAddress;
  CanonicalName, SearchName: AnsiString;
begin
  { If the given Host's address is already in the Cache, return it right away }
  if Cache_GetHostByName(HostName, IPAddress) then
  begin
    Result := True;
    Exit;
  end;
  { If there's no DNS server configured, we can't go on }
  if DNS_IP = 0 then
  begin
    Result := False;
    Exit;
  end;
  { Create a list of DNS servers which are likely to resolve the given Host Name }
  DNSServers.Count := 0;
  QueryDNSServers(DNS_IP, HostName, DNSServers);
  //	WriteLn('Available DNS Servers: ', DNSServers.Count);
   { Test from each DNS Server, trying to resolve the Host's name }
  i := 0;
  CanonicalName := '';
  SearchName := HostName;
  while not Cache_GetHostByName(SearchName, IPAddress) and (i <=
    DNSServers.Count) do
  begin
    if i = 0 then
      DNSServerIP := DNS_IP
    else
    begin
      if not Cache_GetHostByName(DNSServers.Server[i], DNSServerIP) then
        DNSServerIP := 0;
    end;
    if DNSServerIP <> 0 then
    begin
      { Ask this server for the IP Address of the host being searched }
//			if i > 0 then WriteLn('Testing from ', DNSServers.Server[i]);
      if QueryHostAddress(DNSServerIP, HostName, CanonicalName) then
      begin
        if CanonicalName <> '' then
        begin
          //					WriteLn('Canonical Name: ', CanonicalName);
          SearchName := CanonicalName;
          i := 0;
        end;
      end;
    end;
    Inc(i);
  end;
  { Done }
  if Cache_GetHostByName(SearchName, IPAddress) then
  begin
    Cache_AddHostByName(HostName, IPAddress);
    GetHostByName := True;
  end
  else
  begin
    GetHostByName := False;
  end;
end;

var
  i: Integer;
  
initialization
  DNS_PORT := 0;
  DNS_IP := 0;
  for i := Low(Cache) to High(Cache) do
    Cache[i].HostName := '';
end.

