
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosICMP.pas                                          }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosICMP.pas,v 1.1 2003/05/08 18:59:57 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosICMP.pas,v $
  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosICMP;

interface

uses
  WDosNetwork, WDosSock, WDosClock, WDosIP, Threads;

function ICMP_Ping(TargetIP: TIPAddress; Timeout: WORD): Integer;
function ICMP_Init: Boolean;
function ICMP_Deinit: Boolean;

implementation

const
  IPPROTO_ICMP = 1;
  ICMP_PINGREQ = 8;
  ICMP_PINGANS = 0;

type
  PICMPPingHeader = ^TICMPPingHeader;
  TICMPPingHeader = record
    _Type: BYTE;
    Code: BYTE;
    Checksum: WORD;
    ID: WORD;
    Secuence: WORD;
  end;

var
  PingAnswer: Boolean;
  PingSecuence: Word = 2200;

procedure ProcessIP(SourceIP, TargetIP: TIPAddress;
  var Data; DataLen: Word);
var
  Hdr: PICMPPingHeader;
begin
  Hdr := @Data;
{$IFDEF DEBUG}
  WriteLn('ICMP message from ', IP_AddrToString(SourceIP), ' to ',
    IP_AddrToString(TargetIP), ' type ', Hdr._Type, ' code ', Hdr.Code);
{$ENDIF}
  case Hdr^._Type of
    ICMP_PINGANS: { Respuesta de PING }
      begin
{$IFDEF DEBUG}
        WriteLn('PING Answer');
{$ENDIF}
        PingAnswer := True;
      end;

    ICMP_PINGREQ: { Pedido de Ping }
      begin
        { Lo contestamos }
{$IFDEF DEBUG}
        WriteLn('PING Request');
{$ENDIF}
        Hdr^._Type := ICMP_PINGANS;
        Hdr^.Checksum := 0;
        Hdr^.Checksum := IP_Checksum(@Hdr^, DataLen, nil, 0);
        IP_Send(0, SourceIP, IPPROTO_ICMP, Data, Datalen);
      end;
  end;
end;

function ICMP_Init: Boolean;
begin
  Result := IP_AddProcessor(IPPROTO_ICMP, @ProcessIP);
end;

function ICMP_Deinit: Boolean;
begin
  ICMP_Deinit := IP_RemoveProcessor(IPPROTO_ICMP);
end;

function ICMP_Ping(TargetIP: TIPAddress; Timeout: WORD): Integer;
const
  PingData: array[0..31] of char = 'abcdefghijklmnopqrstuvwabcdefghi';
type
  TPingRec = packed record
    Hdr: TICMPPingHeader;
    Data: packed array[0..31] of char; { 32 bytes de datos }
  end;
var
  Rec: TPingRec;
  Resu: Integer;
  TimeStop: Int64;
begin
  { Armamos el paquete }
  if PingSecuence < $FFFF then
    Inc(PingSecuence)
  else
    PingSecuence := 0;
  Rec.Hdr._Type := ICMP_PINGREQ;
  Rec.Hdr.Code := 0;
  Rec.Hdr.Checksum := 0;
  Rec.Hdr.ID := 1;
  Rec.Hdr.Secuence := PingSecuence;
  Move(PingData, Rec.Data, SizeOf(Rec.Data));
  Rec.Hdr.Checksum := IP_Checksum(@Rec, SizeOf(Rec), nil, 0);
  { Enviamos el paquete de PING }
  PingAnswer := False;
  Resu := IP_Send(0, TargetIP, IPPROTO_ICMP, Rec, SizeOf(Rec));
  if Resu <> 0 then
  begin
    ICMP_Ping := Resu;
    Exit;
  end;
  { Esperamos la respuesta }
  TimeStop := GetMSCounter + Timeout;
  while GetMSCounter < TimeStop do
  begin
    NET_Work_Low;
    Yield;
    if PingAnswer then
    begin
      ICMP_Ping := 0;
      Exit;
    end;
  end;
  { Timeout }
  ICMP_Ping := WSAETIMEDOUT;
end;

end.

