
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ WDosLfn                                               }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Laszlo Agocs                    }
{ e-mail: alp@dwp42.org                                 }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Long filename DOS unit, Arne de Bruijn,           }
{     19960402, Public Domain (taken from SWAG)         }
{                                                       }
{ Please inform yourself about special copyrights for   }
{ these portions on the appropriate web pages.          }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosLfn.pas,v 1.4 2002/10/27 22:53:12 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosLfn.pas,v $
  Revision 1.4  2002/10/27 22:53:12  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    ALP 2001-05-31 Version 1.2.1.1 released
    ALP 2001-05-30 Version 1.2.1.0 released
}

{$I WDOSX.INC}

unit WDosLfn;

{ All functions return 0 if successful or an error code if not }
{ Except: LOpenFile, LCreateFile => these return 0 if there was an error }
{ and the error code is stored in LIOResult (since the result is a handle) }
{ LReadFile, LWriteFile => these return the number of read/written bytes }

interface

uses
  SysUtils, WDosDpmi;

type
  str3=string[3];

const
  SEEK_SET=0; //constants for LSeekFile
  SEEK_CUR=1;
  SEEK_END=2;

  fOpenRead=0; //constants for LOpenFile
  fOpenWrite=1;
  fOpenReadWrite=2;

var
  IsLFNSupported:boolean=false;
  //on startup LFileSystemInfo for 'C:\' is called
  //IsLFNSupported will be true if LFNs are supported
  //but if major DOS version < 7 this var will be false
  //(except if WinNT is detected, since W2K supports int21/71xx,
  //while NT 4, 3.51, ... doesn't)

  LIOError:word; //used only if calling LOpenFile or LCreateFile

{
  rootname : 'C:\', 'D:\', ...
  returned flags:
    b0 : searches are case sensitive
    b1 : preserves case in directory entries
    b2 : uses unicode characters in file and directory names
    b14 : supports DOS long filename functions
    b15 : volume is compressed
  maxpathlen could be 0 for CD-ROM drives under Win95
}
function LFileSystemInfo(rootname:str3;var flags,maxfilenamelen,maxpathlen:word):word;

function LMkDir(dir:shortstring):word;
function LRmDir(dir:shortstring):word;
function LChDir(dir:shortstring):word;
function LGetDir(drive:byte;var s:shortstring):word;

function LErase(filename:shortstring):word;
function LRename(oldfilename,newfilename:shortstring):word;
function LGetAttr(filename:shortstring;var attr:word):word;
function LSetAttr(filename:shortstring;attr:word):word;

function LFindFirst(path:shortstring;attr:word;var sr:TSearchRec):word;
function LFindNext(var sr:TSearchRec):word;
procedure LFindClose(var sr:TSearchRec);

function LOpenFile(filename:shortstring;mode:byte):word;
function LCreateFile(filename:shortstring;attr:word):word;
procedure LCloseFile(handle:word);
function LReadFile(handle:word;var dest;count:integer):integer;
function LWriteFile(handle:word;var src;count:integer):integer;
procedure LSeekFile(handle:word;newpos:longint;origin:byte);
function LGetFilePos(handle:word):longint;
function LGetFileSize(handle:word):longint;

{
  same as ParamStr(n) but this supports long filenames specified between " "
  n=0 is also supported to get path of the exe
  only space is handled as delimiter!
}
function GetParamStr(n:integer):string;

implementation

uses
  WDosPorts;

var
(*  mem:array [0..$7ffffffe] of byte absolute 0;*)
  rseg,rsel:integer;
  r:DpmiRmRegisters;

function LFileSystemInfo(rootname:str3;var flags,maxfilenamelen,maxpathlen:word):word;
var
  Dest: Pointer;
begin
  rootname[1]:=upcase(rootname[1]);
  Dest := Pointer(RSeg shl 4);
  Move(RootName[1], Dest^, 3);
(*  move(rootname[1],mem[rseg shl 4],3);*)
  Mem[RSeg shl 4 + 3] := 0;
  r.ds:=rseg;
  r.dx:=0;
  r.es:=rseg;
  r.di:=4;
  r.cx:=64;
  r.ax:=$71A0;
  dpmirealmodeint($21,r);
  flags:=r.bx;
  maxfilenamelen:=r.cx;
  maxpathlen:=r.dx;
  if (r.flags and 1)<>0 then result:=r.ax else result:=0;
end;

function LMkDir(dir:shortstring):word;
var
  Dest: Pointer;
begin
  Dest := Pointer(RSeg shl 4);
  Move(Dir[1], Dest^, Length(Dir));
(*  move(dir[1],mem[rseg shl 4],length(dir));*)
  Mem[RSeg shl 4 + Length(Dir)] := 0;
  r.ds:=rseg;
  r.dx:=0;
  r.ax:=$7139;
  dpmirealmodeint($21,r);
  if (r.flags and 1)<>0 then result:=r.ax else result:=0;
end;

function LRmDir(dir:shortstring):word;
var
  Dest: Pointer;
begin
  Dest := Pointer(RSeg shl 4);
  Move(Dir[1], Dest^, Length(Dir));
(*  move(dir[1],mem[rseg shl 4],length(dir));*)
  Mem[RSeg shl 4 + Length(Dir)] := 0;
  r.ds:=rseg;
  r.dx:=0;
  r.ax:=$713A;
  dpmirealmodeint($21,r);
  if (r.flags and 1)<>0 then result:=r.ax else result:=0;
end;

function LChDir(dir:shortstring):word;
var
  Dest: Pointer;
begin
  Dest := Pointer(RSeg shl 4);
  Move(Dir[1], Dest^, Length(Dir));
(*  move(dir[1],mem[rseg shl 4],length(dir));*)
  Mem[RSeg shl 4 + Length(Dir)] := 0;
  r.ds:=rseg;
  r.dx:=0;
  r.ax:=$713B;
  dpmirealmodeint($21,r);
  if (r.flags and 1)<>0 then result:=r.ax else result:=0;
end;

function LGetDir(drive:byte;var s:shortstring):word;
var
  o:longint;
begin
  if drive=0 then begin
    r.ah:=$19;
    dpmirealmodeint($21,r);
    drive:=r.al+1;
  end;
  s:=chr(drive+64)+':\';
  r.dl:=drive;
  r.ds:=rseg;
  r.si:=0;
  r.ax:=$7147;
  dpmirealmodeint($21,r);
  if (r.flags and 1)<>0 then result:=r.ax else result:=0;
  s[0]:=chr(3);
  if result=0 then begin
    o:=rseg shl 4;
    while mem[o]<>0 do begin
      s[o-(rseg shl 4)+4]:=chr(mem[o]);
      inc(s[0]);
      inc(o);
    end;
  end;
end;

function LErase(FileName: ShortString): Word;
var
  Dest: Pointer;
begin
  Dest := Pointer(RSeg shl 4);
  Move(FileName[1], Dest^, Length(FileName));
(*  move(filename[1],mem[rseg shl 4],length(filename));*)
  Mem[RSeg shl 4 + Length(FileName)] := 0;
  r.ds:=rseg;
  r.dx:=0;
  r.ax:=$7141;
  dpmirealmodeint($21,r);
  if (r.flags and 1)<>0 then result:=r.ax else result:=0;
end;

function LRename(oldfilename,newfilename:shortstring):word;
var
  Dest: Pointer;
begin
  Dest := Pointer(RSeg shl 4);
  Move(OldFileName[1], Dest^, Length(OldFileName));
(*  move(oldfilename[1],mem[rseg shl 4],length(oldfilename));*)
  Mem[RSeg shl 4 + Length(OldFileName)] := 0;
  Dest := Pointer(RSeg shl 4 + Length(OldFileName) + 1);
  Move(NewFileName[1], Dest^, Length(NewFileName));
(*  move(newfilename[1],mem[rseg shl 4+length(oldfilename)+1],length(newfilename));*)
  Mem[RSeg shl 4 + Length(OldFileName) + 1 + Length(NewFileName)] := 0;
  r.ds:=rseg;
  r.dx:=0;
  r.es:=rseg;
  r.di:=length(oldfilename)+1;
  r.ax:=$7156;
  dpmirealmodeint($21,r);
  if (r.flags and 1)<>0 then result:=r.ax else result:=0;
end;

function LGetAttr(filename:shortstring;var attr:word):word;
var
  Dest: Pointer;
begin
  Dest := Pointer(RSeg shl 4);
  Move(FileName[1], Dest^, Length(FileName));
(*  move(filename[1],mem[rseg shl 4],length(filename));*)
  Mem[RSeg shl 4 + Length(FileName)] := 0;
  r.ds:=rseg;
  r.dx:=0;
  r.ax:=$7143;
  r.bl:=0;
  dpmirealmodeint($21,r);
  attr:=r.cx;
  if (r.flags and 1)<>0 then result:=r.ax else result:=0;
end;

function LSetAttr(filename:shortstring;attr:word):word;
var
  Dest: Pointer;
begin
  Dest := Pointer(RSeg shl 4);
  Move(FileName[1], Dest^, Length(FileName));
(*  move(filename[1],mem[rseg shl 4],length(filename));*)
  Mem[RSeg shl 4 + Length(FileName)] := 0;
  r.ds:=rseg;
  r.dx:=0;
  r.ax:=$7143;
  r.bl:=1;
  r.cx:=attr;
  dpmirealmodeint($21,r);
  if (r.flags and 1)<>0 then result:=r.ax else result:=0;
end;

function LFindFirst(path:shortstring;attr:word;var sr:TSearchRec):word;
var
  Src, Dest: Pointer;
begin
  r.ax:=$714E;
  r.cx:=attr;
  r.si:=0;
  r.ds:=rseg;
  r.dx:=0;

  Dest := Pointer(RSeg shl 4);
  Move(Path[1], Dest^, Length(Path));
  Mem[RSeg shl 4 + Length(Path)] := 0;
(*
  move(path[1],mem[rseg shl 4],length(path));
  mem[rseg shl 4+length(path)]:=0;
*)
  r.es:=rseg;
  r.di:=512;
  dpmirealmodeint($21,r);
  if (r.flags and 1)<>0 then result:=r.ax else begin
    result:=0;

    Src := Pointer(RSeg shl 4 + 512);
    Move(Src^, Sr.FindData, 320);
(*
    move(mem[rseg shl 4+512],sr.finddata,320);
*)
    with sr do begin
      findhandle:=r.ax;
      size:=finddata.nfilesizelow;
      attr:=finddata.dwfileattributes;
      name:=finddata.cfilename;
      r.ds:=rseg;
      r.si:=0;

      Dest := Pointer(RSeg shl 4);
      Move(FindData.ftLastWriteTime, Dest^, 8);
(*
      move(finddata.ftlastwritetime,mem[rseg shl 4],8);
*)
      r.bl:=0;
      r.ax:=$71A7;
      dpmirealmodeint($21,r);
      sr.time:=(r.dx shl 16) or r.cx;
    end;
  end;
end;

function LFindNext(var sr:TSearchRec):word;
var
  Src, Dest: Pointer;
begin
  r.ax:=$714F;
  r.bx:=sr.findhandle;
  r.si:=0;
  r.es:=rseg;
  r.di:=0;
  dpmirealmodeint($21,r);
  if (r.flags and 1)<>0 then result:=r.ax else begin
    result:=0;

    Src := Pointer(RSeg shl 4);
    Move(Src^, Sr.FindData, 320);

(*
    move(mem[rseg shl 4],sr.finddata,320);
*)
    with sr do begin
      size:=finddata.nfilesizelow;
      attr:=finddata.dwfileattributes;
      name:=finddata.cfilename;
      r.ds:=rseg;
      r.si:=0;

      Dest := Pointer(RSeg shl 4);
      Move(FindData.ftLastWriteTime, Dest^, 8);
(*
      move(finddata.ftlastwritetime,mem[rseg shl 4],8);
*)
      r.bl:=0;
      r.ax:=$71A7;
      dpmirealmodeint($21,r);
      sr.time:=(r.dx shl 16) or r.cx;
    end;
  end;
end;

procedure LFindClose(var sr:TSearchRec);
begin
  r.ax:=$71A1;
  r.bx:=sr.findhandle;
  dpmirealmodeint($21,r);
end;

function LOpenFile(filename:shortstring;mode:byte):word;
var
  Dest: Pointer;
begin
  r.ax:=$716C;
  r.bx:=mode;
  r.cx:=0;
  r.dx:=1;

  Dest := Pointer(RSeg shl 4);
  Move(FileName[1], Dest^, Length(FileName));
  Mem[RSeg shl 4 + Length(FileName)] := 0;
(*
  move(filename[1],mem[rseg shl 4],length(filename));
  mem[rseg shl 4+length(filename)]:=0;
*)
  r.ds:=rseg;
  r.si:=0;
  r.di:=1;
  dpmirealmodeint($21,r);
  LIOError:=0;
  if (r.flags and 1)=0 then result:=r.ax else begin
    LIOError:=r.ax;
    result:=0;
  end;
end;

function LCreateFile(filename:shortstring;attr:word):word;
var
  Dest: Pointer;
begin
  r.ax:=$716C;
  r.bx:=fOpenReadWrite;
  r.cx:=attr;
  r.dx:=18;

  Dest := Pointer(RSeg shl 4);
  Move(FileName[1], Dest^, Length(FileName));
  Mem[RSeg shl 4 + Length(FileName)] := 0;
(*
  move(filename[1],mem[rseg shl 4],length(filename));
  mem[rseg shl 4+length(filename)]:=0;
*)
  r.ds:=rseg;
  r.si:=0;
  r.di:=1;
  dpmirealmodeint($21,r);
  LIOError:=0;
  if (r.flags and 1)=0 then result:=r.ax else begin
    LIOError:=r.ax;
    result:=0;
  end;
end;

procedure LCloseFile(handle:word);
begin
  asm
    push ebx
    mov ah,3eh
    mov bx,handle
    int 21h
    pop ebx
  end;
end;

function LReadFile(handle:word;var dest;count:integer):integer;
begin
  asm
    push ebx
    mov edx,dest
    mov ah,3fh
    mov bx,handle
    mov ecx,count
    int 21h
    jnc @end
    xor eax,eax
  @end:
    mov result,eax
    pop ebx
  end;
end;

function LWriteFile(handle:word;var src;count:integer):integer;
begin
  asm
    push ebx
    mov edx,src
    mov ah,40h
    mov bx,handle
    mov ecx,count
    int 21h
    jnc @end
    xor eax,eax
  @end:
    mov result,eax
    pop ebx
  end;
end;

procedure LSeekFile(handle:word;newpos:longint;origin:byte);
begin
  asm
    push ebx
    mov cx,word ptr [newpos+2]
    mov dx,word ptr [newpos]
    mov ah,42h
    mov al,origin
    mov bx,handle
    int 21h
    pop ebx
  end;
end;

function LGetFilePos(handle:word):longint;
begin
  asm
    push ebx
    xor cx,cx
    xor dx,dx
    mov ah,42h
    mov al,SEEK_CUR
    mov bx,handle
    int 21h
    shl edx,16
    mov dx,ax
    mov result,edx
    pop ebx
  end;
end;

function LGetFileSize(handle:word):longint;
var
  oldpos:longint;
begin
  oldpos:=LGetFilePos(handle);
  LSeekFile(handle,0,SEEK_END);
  result:=LGetFilePos(handle);
  LSeekFile(handle,oldpos,SEEK_SET);
end;

function GetParamStr(n:integer):string;
var
  exepath,s:string;
  in_lfn:boolean;
  spc_pos,i,pno:integer;
begin
  s:=strpas(cmdline);
  spc_pos:=pos(' ',s);
  if spc_pos>0 then begin
    exepath:=copy(s,1,spc_pos-1);
    delete(s,1,spc_pos);
  end else begin
    exepath:=s;
    s:='';
  end;
  if n=0 then begin
    result:=exepath;
    exit;
  end;
  in_lfn:=false; //between " " ?
  i:=1; //char index
  pno:=1; //index of current parameter
  result:='';
  while i<=length(s) do begin
    if s[i]='"' then begin
      in_lfn:=not in_lfn;
    end else if s[i]=' ' then begin
      if not in_lfn then begin
        inc(pno);
      end else if pno=n then result:=result+s[i];
    end else if pno=n then result:=result+s[i];
    inc(i);
  end;
end;

function DosVersion:word;assembler;
asm
  push ebx
  mov ah,30h
  int 21h
  pop ebx
end;

function GetEnvironmentVariable(name,buffer:pchar;size:longint):longint;stdcall;
  external 'kernel32.dll' name 'GetEnvironmentVariableA';

function GetEnv(name:string):string;
var
  pch,buf:array [0..512] of char;
begin
  strpcopy(pch,name);
  if GetEnvironmentVariable(pch,buf,512)>0 then
    result:=strpas(buf) else result:='';
end;

var
  flags,maxfnlen,maxplen:word;

initialization
  r.ss:=0;r.sp:=0;
  dpmiallocdosmem(512,rseg,rsel);
  if lfilesysteminfo('C:\',flags,maxfnlen,maxplen)=0 then begin
    IsLFNSupported:=((flags and $4000)<>0);
    if (lo(dosversion)<7) and (getenv('OS')<>'Windows_NT') then IsLFNSupported:=false;
  end;

finalization
  dpmifreedosmem(rsel);

end.
