
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosNetwork.pas                                       }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosNetwork.pas,v 1.2 2003/10/25 22:00:07 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosNetwork.pas,v $
  Revision 1.2  2003/10/25 22:00:07  iwache
  Minor changes in coding format added. No functional change.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosNetwork;

{$X+}

interface

uses
  Classes, SysUtils, WDosUtil, WDosClock;

type
  { 6 Byte ethernet address }
  PEthernetAddress = ^TEthernetAddress;
  TEthernetAddress = packed array[0..5] of Byte;

const
  ETH_BROADCASTADDRESS: TEthernetAddress = ($FF, $FF, $FF, $FF, $FF, $FF);
  ETH_NULLADDRESS: TEthernetAddress = (0, 0, 0, 0, 0, 0);

type
  { Ethernet Header }
  TEthernetHeader = packed record
    TargetAddr: TEthernetAddress;
    SourceAddr: TEthernetAddress;
    TypeLen: Word;
  end;

  { Ethernet Packet }
  PEthernetPacket = ^TEthernetPacket;
  TEthernetPacket = packed record
    Header: TEthernetHeader;
    Data: packed array[0..1500] of BYTE;
  end;

  TNetworkAdapter = class;
  TNetRoutine = procedure(Adapter: TNetworkAdapter; Packet: PEthernetPacket);
  TNetChangedRoutine = procedure;
  TNetIdleRoutine = procedure;

  TAdapterStatus = (asOk, asError);

  { Adapter Object }
  TNetworkAdapter = class
  private
    FProtocols: TList;
    procedure Process;
  protected
    FStatus: TAdapterStatus;
    FAdapterName: AnsiString;
    procedure CheckStatus; virtual; abstract;
    function IntAddProtocol(TypeLen: WORD): Boolean; virtual; abstract;
    function IntRemoveProtocol(TypeLen: WORD): Boolean; virtual; abstract;
    function GetPacket(var Packet: TEthernetPacket): Boolean; virtual; abstract;
  public
    constructor Create;
    destructor Destroy; override;
    function GetAddress(var Addr: TEthernetAddress): Boolean; virtual; abstract;
    function SendPkt(Buffer: Pointer; BufLen: Word): Boolean; virtual; abstract;
    function AddProtocol(TypeLen: WORD; Routine: TNetRoutine): Boolean; virtual;
    function RemoveProtocol(TypeLen: WORD; Routine: TNetRoutine): Boolean;
      virtual;
    property Status: TAdapterStatus read FStatus;
    property AdapterName: AnsiString read FAdapterName;
  end;

function Ethernet_AddrToString(const EtherAddr: TEthernetAddress): AnsiString;
function Ethernet_Compare(const Addr1, Addr2: TEthernetAddress): Boolean;
function GetMemCheck(var P: Pointer; Size: WORD): Boolean;
procedure NET_Work_Low;
procedure NET_Work;
procedure NET_Delay(MS: WORD);
procedure NET_AddIdleProc(Proc: TNetIdleRoutine);
procedure NET_RemoveIdleProc(Proc: TNetIdleRoutine);
procedure NET_NotifyNetChanged;
procedure NET_AddNetChangedProc(Proc: TNetChangedRoutine);
procedure NET_RemoveNetChangedProc(Proc: TNetChangedRoutine);
function Net_SetHostName(HostName: AnsiString): Boolean;
function Net_GetHostName: AnsiString;

implementation

type
  PProtocolEntry = ^TProtocolEntry;
  TProtocolEntry = record
    TypeLen: WORD;
    Routine: TNetRoutine;
  end;

var
  AdapterList: TList;
  IdleProcList: TList;
  FHostName: AnsiString;
  ChangedProcList: TList;

procedure NET_AddIdleProc(Proc: TNetIdleRoutine);
begin
  IdleProcList.Remove(@Proc);
  IdleProcList.Add(@Proc);
end;

procedure NET_RemoveIdleProc(Proc: TNetIdleRoutine);
begin
  IdleProcList.Remove(@Proc);
end;

function Net_SetHostName(HostName: AnsiString): Boolean;
begin
  FHostName := UpperCase(Copy(HostName, 1, 15));
  Net_NotifyNetChanged;
  Result := True;
end;

function Net_GetHostName: AnsiString;
begin
  Result := FHostName;
end;

procedure NET_Delay(MS: WORD);
var
  Timeout: Int64;
begin
  Timeout := GetMSCounter + MS;
  while GetMSCounter < Timeout do
  begin
    NET_Work_Low;
  end;
end;

procedure NET_Work;
var
  i: Integer;
  IdleProc: TNetIdleRoutine;
begin
  Net_Work_Low;
  for i := 0 to IdleProcList.Count - 1 do
  begin
    @IdleProc := IdleProcList[i];
    IdleProc;
  end;
end;

procedure NET_Work_Low;
var
  i: Integer;
begin
  for i := 0 to AdapterList.Count - 1 do
  begin
    TNetworkAdapter(AdapterList[i]).Process;
  end;
end;

function Ethernet_AddrToString(const EtherAddr: TEthernetAddress): AnsiString;
var
  tCount: Byte;
  tString: AnsiString;
begin
  tString := '';
  for tCount := 0 to 5 do
  begin
    tString := tString + IntToHex(EtherAddr[tCount], 2) + ':';
  end;
  Ethernet_AddrToString := Copy(tString, 1, 17);
end;

function Ethernet_Compare(const Addr1, Addr2: TEthernetAddress): Boolean;
begin
  Ethernet_Compare :=
    (Addr1[0] = Addr2[0])
    and (Addr1[1] = Addr2[1])
    and (Addr1[2] = Addr2[2])
    and (Addr1[3] = Addr2[3])
    and (Addr1[4] = Addr2[4])
    and (Addr1[5] = Addr2[5]);
end;

function GetMemCheck(var P: Pointer; Size: WORD): Boolean;
begin
  //    if MaxAvail >= Size then begin
  GetMem(P, Size);
  GetMemCheck := True;
  //	end else begin
  //		GetMemCheck := False;
  //	end;
end;

procedure NET_NotifyNetChanged;
var
  i: integer;
begin
  i := 0;
  while (i < ChangedprocList.Count) do
  begin
    TNetChangedRoutine(ChangedProcList[i]);
    Inc(i);
  end;
end;

procedure NET_AddNetChangedProc(Proc: TNetChangedRoutine);
begin
  ChangedProcList.Add(@Proc);
end;

procedure NET_RemoveNetChangedProc(Proc: TNetChangedRoutine);
begin
  ChangedProcList.Remove(@Proc);
end;

{ TNetworkAdapter }

constructor TNetworkAdapter.Create;
begin
  inherited Create;
  FAdapterName := 'None';
  { Create Protocol List. It holds the protocols bound to this adapter }
  FProtocols := TList.Create;
  { Include this adapter in the adapter list }
  AdapterList.Add(Self);
  { Check for an installed packet driver }
  CheckStatus;
end;

destructor TNetworkAdapter.Destroy;
var
  PE: PProtocolEntry;
begin
  { Quitamos este adaptador de la lista de adaptadores }
  AdapterList.Remove(Self);
  { Eliminamos los procesadores }
  while FProtocols.Count > 0 do
  begin
    PE := FProtocols[0];
    RemoveProtocol(PE.TypeLen, PE.Routine);
  end;
  FProtocols.Free;
  { Listo }
  inherited Destroy;
end;

procedure TNetworkAdapter.Process;
var
  i: Integer;
  Pkt, PktTmp: TEthernetPacket;
begin
  if (FStatus <> asOk) or not GetPacket(Pkt) then
    Exit;
  { Hay al menos un paquete. Lo descomponemos }
  for i := 0 to FProtocols.Count - 1 do
  begin
    if PProtocolEntry(FProtocols[i])^.TypeLen = Pkt.Header.TypeLen then
    begin
      PktTmp := Pkt;
{$IFDEF DEBUG}
      WriteLn('Ethernet Packet received: From ',
        Ethernet_AddrToString(Pkt.Header.SourceAddr), ' to ',
        Ethernet_AddrToString(Pkt.Header.TargetAddr), ' TypeLen ',
        IntToHex(Swap(Pkt.Header.TypeLen), 4));
{$ENDIF}
      PProtocolEntry(FProtocols[i])^.Routine(Self, @PktTmp);
    end;
  end;
end;

function TNetworkAdapter.AddProtocol(TypeLen: WORD; Routine: TNetRoutine):
  Boolean;
var
  PE: PProtocolEntry;
begin
  AddProtocol := False;
  if Status <> asOk then
    Exit;
  { First, try to register this protocol with the Packet Driver          }
  if not IntAddProtocol(TypeLen) then
    Exit;
  { Done with the Packet Driver. Include the new protocol in the list of }
  { bound protocols.													   }
  PE := New(PProtocolEntry);
  PE^.TypeLen := TypeLen;
  PE^.Routine := Routine;
  FProtocols.Add(PE);
  AddProtocol := True;
end;

function TNetworkAdapter.RemoveProtocol(TypeLen: WORD; Routine: TNetRoutine):
  Boolean;
var
  i: Integer;
  PE: PProtocolEntry;
begin
  RemoveProtocol := False;
  for i := 0 to FProtocols.Count - 1 do
  begin
    PE := FProtocols[i];
    if (PE^.TypeLen = TypeLen) and (@PE^.Routine = @Routine) then
    begin
      IntRemoveProtocol(PE^.TypeLen);
      FProtocols.Remove(PE);
      Dispose(PE);
      RemoveProtocol := True;
      Break;
    end;
  end;
end;

initialization
  // Create various Objetcs
  FHostName := '';
  AdapterList := TList.Create;
  IdleProcList := TList.Create;
  ChangedProcList := TList.Create;

finalization
  while ChangedProcList.Count > 0 do
    ChangedProcList.Delete(0);
  ChangedProcList.Free;
  while IdleProcList.Count > 0 do
    IdleProcList.Delete(0);
  IdleProcList.Free;
  while AdapterList.Count > 0 do
    TNetworkAdapter(AdapterList[0]).Free;
  AdapterList.Free;
end.

