
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosPktDrv.pas                                        }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosPktDrv.pas,v 1.4 2003/10/25 22:01:43 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosPktDrv.pas,v $
  Revision 1.4  2003/10/25 22:01:43  iwache
  Non used var I in function TDriverCallBack.GetPacket removed.

  Revision 1.3  2003/10/07 19:27:30  iwache
  TDriverCallBack.FPacketQueue is now a ring buffer.
  Old implementation produced packet lost with TCP.

  Revision 1.2  2003/05/13 19:45:50  iwache
  Packet driver completely rewritten to make it work with DWPL.
  The callback function now use the classes from WDosSysUtils.pas.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosPktDrv;

{$X+}

interface

uses
  Crt, SysUtils, WDosNetwork, WDosTypes, WDosSysUtils;

const
  QUEUE_SIZE = 64;

type
  TDriverCallBack = class;

  { Adapter Object }
  TPacketDriverAdapter = class(TNetworkAdapter)
  private
    FPDInt: BYTE;
    FProtCount: Integer;
    FAddress: TEthernetAddress;
    FDriverCallBack: TDriverCallBack;
    FDosBuffer: TDosBuf;
    FIfClass: BYTE;
    FTypeLen: WORD;
  protected
    function Handle: Word;
    procedure CheckStatus; override;
    function GetPacket(var Packet: TEthernetPacket): Boolean; override;
    function CheckForPD(Int: BYTE): Boolean;
    function IntAddProtocol(TypeLen: WORD): Boolean; override;
    function IntRemoveProtocol(TypeLen: WORD): Boolean; override;
    function GetAddressInternal(var Addr: TEthernetAddress): Boolean;
  public
    constructor Create;
    destructor Destroy; override;
    function SendPkt(Buffer: Pointer; BufLen: Word): Boolean; override;
    function GetAddress(var Addr: TEthernetAddress): Boolean; override;
  end;

  TDriverCallBack = class(TDosCallBack)
  private
    FOwner: TPacketDriverAdapter;
    FPacketQueue: array[0..QUEUE_SIZE - 1] of TEthernetPacket;
    FQueueHead: Integer;
    FQueueTail: Integer;
    FDosBuffer: TDosBuf;
    FHandle: Word;
    FInstalled: Boolean;
  protected
    procedure ISR; override;
    function  ReHook(CodePtr: Pointer): Boolean; override;
    function  UnHook: Boolean; override;
    function GetPacket(var Packet: TEthernetPacket): Boolean;
  public
    constructor Create(Owner: TPacketDriverAdapter);
    destructor Destroy; override;
    property Handle: Word read FHandle;
    property Installed: Boolean read FInstalled;
  end;

implementation


{*********************** TDriverCallBack **************************************}

constructor TDriverCallBack.Create(Owner: TPacketDriverAdapter);
begin
  inherited Create(PtrRec(@TDriverCallBack.ReHook).Ofs
      - PtrRec(@TDriverCallBack.ISR).Ofs, False);
  FOwner := Owner;
  FQueueHead := 0;
  FQueueTail := 0;
  GetDosMem(FDosBuffer, SizeOf(TEthernetPacket));
  InitHook;
end;

destructor TDriverCallBack.Destroy;
begin
  FreeDosMem(FDosBuffer);
  inherited Destroy;
end;

procedure TDriverCallBack.ISR; register;
begin
  if Regs.AX = 0 then
  begin
    { allocate buffer memory }
    Regs.ES := FDosBuffer.RealSeg;
    Regs.DI := FDosBuffer.RealOfs;
  end
  else if Regs.AX = 1 then
  begin
    { fill the next buffer at Head }
    if FPacketQueue[FQueueHead].Header.TypeLen = 0 then
    begin
      Move(FDosBuffer.Buf^, FPacketQueue[FQueueHead],
          SizeOf(TEthernetPacket));
      Inc(FQueueHead);
      if FQueueHead = QUEUE_SIZE then FQueueHead := 0;
    end;
  end;
end;

function TDriverCallBack.ReHook(CodePtr: Pointer): Boolean;
var
  FDosBuffer: TDosBuf;
  R: TRealRegs;
begin
  { Register with packet driver }
  ClearRealRegs(R);
  with R do
  begin
    AH := 2;
    AL := FOwner.FIfClass;
    BX := $FFFF;
    CX := 0;
    DL := 0;
    // TypeLen
    Move(FOwner.FTypeLen, FDosBuffer.Buf^, 2);
    DS := FDosBuffer.RealSeg;
    SI := FDosBuffer.RealOfs;

    // Receiver Routine
    ES := DosPtrRec(CodePtr).Seg;
    DI := DosPtrRec(CodePtr).Ofs;

    // Call Packet Driver
    Intr(FOwner.FPDInt, R);
    FInstalled := ((Flags and fCarry) <> fCarry);
    if FInstalled then
      FHandle := AX
    else
      FHandle := 0;
  end;
  Result := True;
end;

function TDriverCallBack.UnHook: Boolean; register;
var
  R: TRealRegs;
begin
  ClearRealRegs(R);
  with R do
  begin
    AH := 3;
    BX := FHandle;
    Intr(FOwner.FPDInt, R);
    Result := ((Flags and fCarry) <> fCarry);
    if not Result then Exit;
  end;
  Result := True;
end;

function TDriverCallBack.GetPacket(var Packet: TEthernetPacket): Boolean;
begin
  Result := False;
  if FQueueHead <> FQueueTail then
  begin
    Packet := FPacketQueue[FQueueTail];
    FPacketQueue[FQueueTail].Header.TypeLen := 0;
    Inc(FQueueTail);
    if FQueueTail = QUEUE_SIZE then FQueueTail := 0;
    Result := True;
  end;
end;


{*********************** TPacketDriverAdapter *********************************}

constructor TPacketDriverAdapter.Create;
begin
  inherited Create;
  GetDosMem(FDosBuffer, SizeOf(TEthernetPacket));
end;

destructor TPacketDriverAdapter.Destroy;
begin
  FDriverCallBack.Free;
  FreeDosMem(FDosBuffer);
  inherited;
end;

procedure TPacketDriverAdapter.CheckStatus;
var
  I: Integer;
begin
  for i := $60 to $80 do
  begin
    if CheckForPD(I) then
    begin
      FStatus := asOk;
      FPDInt := I;
      Exit;
    end;
  end;
  FStatus := asError;
end;

function TPacketDriverAdapter.CheckForPD(Int: Byte): Boolean;
const
  Signature = 'PKT DRVR';
var
  I: Integer;
  PktISR: DosPtr;
  ISRPtr: Pointer;
  P: PChar;
  Buff: array[0..20] of Char;
begin
  { Check to see if there's a packet driver installed at Int }
  PktISR := GetDosIntVec(Int);
  ISRPtr := MapDosPtr(PktISR);
  try
    P := ISRPtr;
    for I := 0 to 10 do
    begin
      FillChar(Buff, SizeOf(Buff), 0);
      Move(P^, Buff, Length(Signature));
      if StrPas(Buff) = Signature then
      begin
        Result := True;
        Exit;
      end;
      Inc(P);
    end;
  finally
    FreeDosPtr(ISRPtr);
  end;
  { No Packet Driver found }
  Result := False;
end;

function TPacketDriverAdapter.IntAddProtocol(TypeLen: WORD): Boolean;
var
  Regs: TRealRegs;
  Buffer: array[0..255] of Char;
begin
  if FProtCount <> 0 then
  begin
    Inc(FProtCount);
    Result := True;
    Exit;
  end;

  Result := False;

  { Obtain info for the packet driver }
  ClearRealRegs(Regs);
  with Regs do
  begin
    AH := 1;
    AL := $FF;
    BX := 0;
    Intr(FPDInt, Regs);
    if (Regs.Flags and 1) = 1 then
      Exit;
    FIfClass := CH;
    Move(Pointer(DS * 16 + SI)^, Buffer, 40);
    FAdapterName := StrPas(Buffer);
  end;

  { TODO : for what is TypeLen useful? }

  TypeLen := 0;
  FTypeLen := TypeLen;

  { Register Callback with packet driver }
  FDriverCallBack := TDriverCallback.Create(Self);
  if not FDriverCallBack.Installed then Exit;

  { Set receive mode to 3: receive packets for this board or broadcast }
  ClearRealRegs(Regs);
  with Regs do
  begin
    AH := 20;
    BX := Handle;
    CX := 3;
    Intr(FPDInt, Regs);
    if (Regs.Flags and 1) = 1 then
      Exit;
  end;
  { Done! }
  FProtCount := 1;
  Result := True;
end;

function TPacketDriverAdapter.IntRemoveProtocol(TypeLen: WORD): Boolean;
begin
  Result := True;
  if FProtCount <> 1 then
  begin
    Dec(FProtCount);
    Exit;
  end;
  { Unregister with packet driver }
  FDriverCallBack.Free;
  FDriverCallBack := nil;
  { Done! }
  FProtCount := 0;
end;

function TPacketDriverAdapter.SendPkt(Buffer: Pointer; BufLen: Word): Boolean;
var
  Regs: TRealRegs;
begin
  SendPkt := False;
  if FStatus <> asOk then
    Exit;
  if BufLen < 60 then
    BufLen := 60;
  Move(Buffer^, FDosBuffer.Buf^, BufLen);
  ClearRealRegs(Regs);
  with Regs do
  begin
    AH := 4;
    BX := Handle;
    DS := FDosBuffer.RealSeg;
    SI := FDosBuffer.RealOfs;
    ES := FDosBuffer.RealSeg;
    DI := FDosBuffer.RealOfs;
    CX := BufLen;
    Intr(FPDInt, Regs);
    if (Flags and 1) = 1 then
      Exit;
  end;
  SendPkt := True;
end;

function TPacketDriverAdapter.GetAddressInternal(var Addr: TEthernetAddress):
    Boolean;
var
  Regs: TRealRegs;
begin
  Result := False;
  if (FStatus <> asOk) then
    Exit;
  ClearRealRegs(Regs);
  with Regs do
  begin
    AH := 6;
    BX := Handle;
    CX := SizeOf(TEthernetAddress);
    ES := FDosBuffer.RealSeg;
    DI := FDosBuffer.RealOfs;
    Intr(FPDInt, Regs);
    if (Flags and 1) = 1 then Exit;
  end;
  Move(FDosBuffer.Buf^, Addr, SizeOf(Addr));
  Result := True;
end;

function TPacketDriverAdapter.GetAddress(var Addr: TEthernetAddress): Boolean;
begin
  if Ethernet_Compare(FAddress, ETH_NULLADDRESS)
      and Assigned(FDriverCallBack) then
    GetAddressInternal(FAddress);
  Addr := FAddress;
  Result := (Status = asOk);
end;

function TPacketDriverAdapter.GetPacket(var Packet: TEthernetPacket): Boolean;
begin
  Result := False;
  if Assigned(FDriverCallBack) then
    Result := FDriverCallBack.GetPacket(Packet);
end;

function TPacketDriverAdapter.Handle: Word;
begin
  Assert(Assigned(FDriverCallBack), 'packet driver not assigned');
  Result := FDriverCallBack.Handle;
end;

end.

