
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosPktDrvPM.pas                                      }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosPktDrvPM.pas,v 1.1 2003/05/08 18:59:57 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosPktDrvPM.pas,v $
  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosPktDrvPM;

{$X+}

interface

uses
  SysUtils, WDosUtil, WDosNetwork, WDosDPMI;

type
  { Adapter Object }
  TPKTPMAdapter = class(TNetworkAdapter)
  private
  protected
    procedure CheckStatus; override;
    function GetPacket(var Packet: TEthernetPacket): Boolean; override;
    function IntAddProtocol(TypeLen: WORD): Boolean; override;
    function IntRemoveProtocol(TypeLen: WORD): Boolean; override;
  public
    function GetAddress(var Addr: TEthernetAddress): Boolean; override;
    function SendPkt(Buffer: Pointer; BufLen: Word): Boolean; override;
  end;

implementation

procedure TPKTPMAdapter.CheckStatus;
var
  SegInfo: RealModePtr;
  Buff: array[0..10] of char;
begin
  { Revisamos si PKT-PM est instalado }
  DpmiGetRmVector($65, SegInfo);
  Move(Pointer(SegInfo.RmSegment * 16 + SegInfo.RmOffset + 10)^, Buff, 6);
  if Copy(StrPas(Buff), 1, 6) = 'PKT-PM' then
  begin
    FStatus := asOk;
    { Listo }
  end
  else
  begin
    FStatus := asError;
  end;
end;

function TPKTPMAdapter.IntAddProtocol(TypeLen: WORD): Boolean;
begin
  Result := True;
end;

function TPKTPMAdapter.IntRemoveProtocol(TypeLen: WORD): Boolean;
begin
  Result := True;
end;

function TPKTPMAdapter.SendPkt(Buffer: Pointer; BufLen: Word): Boolean;
var
  Regs: DpmiRmRegisters;
  Segment, Selector: Integer;
begin
  if (FStatus = asOk) then
  begin
    DpmiAllocDOSMem(128, Segment, Selector);
    Move(Buffer^, Pointer(Segment * 16)^, BufLen);
    Regs.AX := 2;
    Regs.DS := Segment;
    Regs.SI := 0;
    Regs.CX := BufLen;
    Regs.SS := 0;
    Regs.SP := 0;
    DpmiRealModeInt($65, Regs);
    SendPkt := (Regs.BL = 0);
    DpmiFreeDOSMem(Selector);
  end
  else
  begin
    SendPkt := False;
  end;
end;

function TPKTPMAdapter.GetAddress(var Addr: TEthernetAddress): Boolean;
var
  Regs: DpmiRmRegisters;
  Segment, Selector: Integer;
begin
  if (FStatus = asOk) then
  begin
    DpmiAllocDOSMem(1, Segment, Selector);
    Regs.AX := 4;
    Regs.CX := 6;
    Regs.ES := Segment;
    Regs.DI := 0;
    Regs.SS := 0;
    Regs.SP := 0;
    DpmiRealModeInt($65, Regs);
    Move(Pointer(Segment * 16)^, Addr, SizeOf(Addr));
    DpmiFreeDOSMem(Selector);
    GetAddress := True;
  end
  else
  begin
    GetAddress := False;
  end;
end;

function TPKTPMAdapter.GetPacket(var Packet: TEthernetPacket): Boolean;
var
  Regs: DpmiRmRegisters;
  Segment, Selector: Integer;
begin
  Result := False;
  if (FStatus <> asOk) then
    Exit;
  DpmiAllocDOSMem(128, Segment, Selector);
  Regs.AX := 3;
  Regs.DS := Segment;
  Regs.SI := 0;
  Regs.SS := 0;
  Regs.SP := 0;
  DpmiRealModeInt($65, Regs);
  if Regs.BL = 0 then
  begin
    Move(Pointer(Segment * 16)^, Packet, SizeOf(Packet));
    Result := True;
  end
  else
  begin
    Result := False;
  end;
  DpmiFreeDOSMem(Selector);
end;

end.

