
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosSMB.pas                                           }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosSMB.pas,v 1.4 2003/10/25 22:03:10 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosSMB.pas,v $
  Revision 1.4  2003/10/25 22:03:10  iwache
  Conditional compilation settings for compatibilty
  with Delphi 5 added by use of the JEDI.INC include file.
  Minor changes in coding format added.

  Revision 1.3  2003/08/15 21:09:19  iwache
  Missing type PWord for D5 compatibility added.
  Warning switch off removed for compatibility too.
  This warnings are useless for automatic DWPL
  build process.

  Revision 1.2  2003/06/17 18:56:10  iwache
  procedure Init added. This was missing in older versions.
  See demos\rtl\wdosx\sockets\smbdemo project for use of it.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosSMB;

{$INCLUDE WDOSX.INC}

{$IFDEF COMPILER6_UP}
  {$WARN SYMBOL_PLATFORM OFF}
{$ENDIF}

interface

uses
  Classes, SysUtils, WDosUtil, WDosClock, WDosNetwork, WDosSock,
  WDosIP, WDosNetBIOS;

function SMB_AddShare(ShareName, Directory: AnsiString): Boolean;

procedure Init;

implementation

const
  // Error Classes
  SUCCESS = $00; // The request was successful.
  ERRDOS = $01; // Error is from the core DOS operating system set.
  ERRSRV = $02; // Error is generated by the server network file manager.
  ERRHRD = $03; // Error is an hardware error.
  ERRCMD = $FF; // Command was not in the "SMB" format.

  // Error Codes (Class = ErrDOS)
  ERRbadfunc = 1;
    // Invalid function.  The server did not recognize or could not perform a system call generated by the server, e.g., set the DIRECTORY attribute on a data file, invalid seek mode.
  ERRbadfile = 2;
    // File not found.  The last component of a file's pathname could not be found.
  ERRbadpath = 3;
    // Directory invalid.  A directory component in a pathname could not be found.
  ERRnofids = 4;
    // Too many open files.  The server has no file handles available.
  ERRnoaccess = 5;
    // Access denied, the client's context does not permit the requested function.  This includes the following conditions:
  // invalid rename command
  // write to fid open for read only
        // read on fid open for write only
  // attempt to delete a non-empty directory
  ERRbadfid = 6;
    // Invalid file handle.  The file handle specified was not recognized by the server.

  // Error Codes (Class = ErrSrv)
  ERRerror = 1;
    // Non-specific error code.  It is returned under the following conditions:
  // resource other than disk space exhausted (e.g.  TIDs)
  // first SMB command was not negotiate
  // multiple negotiates attempted
  // internal server error
  ERRbadpw = 2;
    // Bad password - name/password pair in a Tree Connect or Session Setup are invalid.
  ERRaccess = 4;
    // The client does not have the necessary access rights within the specified context for the requested function.
  ERRinvnid = 5; // The Tid specified in a command was invalid.
  ERRinvnetname = 6; // Invalid network name in tree connect.
  ERRinvdevice = 7;
    // Invalid device - printer request made to non-printer connection or non-printer request made to printer connection.
  ERRqfull = 49; // Print queue full (files) -- returned by open print file.
  ERRqtoobig = 50; // Print queue full -- no space.
  ERRqeof = 51; // EOF on print queue dump.
  ERRinvpfid = 52; // Invalid print file FID.
  ERRsmbcmd = 64; // The server did not recognize the command received.
  ERRsrverror = 65;
    // The server encountered an internal error, e.g., system file unavailable.
  ERRfilespecs = 67;
    // The Fid and pathname parameters contained an invalid combination of values.
  ERRbadpermits = 69;
    // The access permissions specified for a file or directory are not a valid combination.  The server cannot set the requested attribute.
  ERRsetattrmode = 71;
    // The attribute mode in the Set File Attribute request is invalid.
  ERRpaused = 81; // Server is paused.  (reserved for messaging)
  ERRmsgoff = 82; // Not receiving messages.  (reserved for messaging).
  ERRnoroom = 83; // No room to buffer message.  (reserved for messaging).
  ERRrmuns = 87; // Too many remote user names.  (reserved for messaging).
  ERRtimeout = 88; // Operation timed out.
  ERRnoresource = 89; // No resources currently available for request.
  ERRtoomanyuids = 90; // Too many Uids active on this session.
  ERRbaduid = 91;
    // The Uid is not known as a valid user identifier on this session.
  ERRusempx = 250; // Temporarily unable to support Raw, use MPX mode.
  ERRusestd = 251; // Temporarily unable to support Raw, use standard read/write.
  ERRcontmpx = 252; // Continue in MPX mode.
  ERRnosupport = 65535; // Function not supported.

type
  TSessionHeader = packed record
    MsgType: BYTE;
    Flags: BYTE;
    Length: WORD;
  end;

  TSessionRequest = packed record
    Header: TSessionHeader;
    CalledName: TNetBIOSName;
    CallingName: TNetBIOSName;
  end;

  TSessionResponse = packed record
    Header: TSessionHeader;
    ErrorCode: BYTE;
  end;

  PSMBHeader = ^TSMBHeader;
  TSMBHeader = packed record
    Signature: array[0..3] of char;
    Command: BYTE;
    ErrorClass: BYTE;
    MustBeZero: BYTE;
    ErrorCode: WORD;
    Flags: BYTE;
    Flags2: WORD;
    Pad: array[1..12] of BYTE;
    TreeID: WORD;
    ProcessID: WORD;
    UserID: WORD;
    MultiplexID: WORD;
  end;

  TTreeSearchRecord = record
    Mask: ShortString;
    Attr: DWORD;
  end;

  TShareEntry = class
  public
    ShareName: AnsiString;
    Directory: AnsiString;
    Drive: BYTE;
    constructor Create;
    destructor Destroy; override;
  end;

  TServerSessionState =
    (ssWaitingSession, ssWaitingNeg, ssReady);
  TServerSession = class
    Timeout: Int64;
    Socket: Integer;
    Trees: TList;
    State: TServerSessionState;
    constructor Create(ASocket: Integer);
    destructor Destroy; override;
  end;

  TTreeEntry = class
  public
    TreeID: Integer;
    Share: TShareEntry;
    Session: TServerSession;
    Search: TTreeSearchRecord;
    Files: TList;
    constructor Create(ASession: TServerSession; AShare: TShareEntry);
    destructor Destroy; override;
  end;

  TFileEntry = class
    FileID: WORD;
    Handle: Integer;
    Tree: TTreeEntry;
    constructor Create(ATree: TTreeEntry; AHandle: Integer);
    destructor Destroy; override;
  end;

var
  ServerSocket: Integer;
  TreeIDCounter: Integer;
  FileIDCounter: Integer;
  Shares: TList;
  ServerSessions: TList;
  SMB_Started: Boolean;

function DWORDToStr(DW: DWORD): AnsiString;
begin
  SetLength(Result, 4);
  Move(DW, Result[1], 4);
end;

function WORDToStr(W: WORD): AnsiString;
begin
  SetLength(Result, 2);
  Move(W, Result[1], 2);
end;

function DOS_Attrib(A: DWORD): DWORD;
begin
  Result := 0;
  if (A and $0001) = $0001 then
    Result := Result or faReadOnly;
  if (A and $0002) = $0002 then
    Result := Result or faHidden;
  if (A and $0004) = $0004 then
    Result := Result or faSysFile;
  if (A and $0008) = $0008 then
    Result := Result or faVolumeID;
  if (A and $0010) = $0010 then
    Result := Result or faDirectory;
  if (A and $0020) = $0020 then
    Result := Result or faArchive;
end;

function DOS_AccessMode(A: DWORD; var Granted: WORD): DWORD;
var
  AM, SM: WORD;
begin
  AM := A and $0F;
  SM := (A shr 4) and $0F;

  Result := 0;
  case AM of
    0, 3: Result := fmOpenRead;
    1: Result := fmOpenWrite;
    2: Result := fmOpenReadWrite;
  end;
  case SM of
    0: Result := Result or fmShareCompat;
    1: Result := Result or fmShareExclusive;
    2: Result := Result or fmShareDenyWrite;
    3: Result := Result or fmShareDenyRead;
    4: Result := Result or fmShareDenyNone;
  end;
  Granted := A;
end;

function SMB_Attrib(A: DWORD): DWORD;
begin
  Result := 0;
  if (A and faReadOnly) = faReadOnly then
    Result := Result or $0001;
  if (A and faHidden) = faHidden then
    Result := Result or $0002;
  if (A and faSysFile) = faSysFile then
    Result := Result or $0004;
  if (A and faVolumeID) = faVolumeID then
    Result := Result or $0008;
  if (A and faDirectory) = faDirectory then
    Result := Result or $0010;
  if (A and faArchive) = faArchive then
    Result := Result or $0020;
end;

function SMB_UnixTime(Time: DWORD): DWORD;
var
  DTime: TDateTime;
begin
  DTime := FileDateToDateTime(Time);
  DTime := DTime - EncodeDate(1970, 1, 1);
  Result := Round(DTime * SecsPerDay);
end;

function GetFileAttribs(FileName: AnsiString;
  var Attribs: WORD; var Time: DWORD; var Size: DWORD): Boolean;
var
  R: Integer;
  Rec: TSearchRec;
begin
  if IsDir(FileName) then
  begin
    Result := False;
    R := FindFirst(GoodDir(FileName) + '*.*', faDirectory, Rec);
    while R = 0 do
    begin
      if Rec.Name = '.' then
      begin
        Result := True;
        Break;
      end;
      R := FindNext(Rec);
    end;
    if not Result then
    begin
      { Root Dir }
      Rec.Attr := faDirectory;
      Rec.Size := 0;
      Rec.Time := DateTimeToFileDate(Now);
      Result := True;
    end;
  end
  else
  begin
    Result := (FindFirst(FileName, faAnyFile, Rec) = 0);
  end;
  if Result then
  begin
    Attribs := SMB_Attrib(Rec.Attr);
    Size := Rec.Size;
    Time := SMB_UnixTime(Rec.Time);
  end;
  FindClose(Rec);
end;

procedure SendSMBRequest(Socket: Integer; Command, Flags, ErrorClass: BYTE;
  ErrorCode, TreeID, ProcessID, UserID, MultID: WORD; const Params, Data:
    AnsiString);

  procedure AddParams(var s: AnsiString);
  begin
    s := s + Chr(Length(Params) div 2) + Params;
  end;

  procedure AddData(var s: AnsiString);
  var
    W: WORD;
  begin
    SetLength(s, Length(s) + 2);
    W := Length(Data);
    Move(W, s[Length(s) - 1], 2);
    s := s + Data;
  end;

var
  P: PSMBHeader;
  S, s1: AnsiString;
  Header: TSessionHeader;
begin
  SetLength(S, SizeOf(TSMBHeader));
  P := @S[1];
  FillChar(P^, SizeOf(P^), 0);
  P.Signature[0] := #$FF;
  P.Signature[1] := 'S';
  P.Signature[2] := 'M';
  P.Signature[3] := 'B';
  P.Command := Command;
  P.Flags := Flags;
  P.ErrorClass := ErrorClass;
  P.ErrorCode := ErrorCode;
  P.TreeID := TreeID;
  P.ProcessID := ProcessID;
  P.UserID := UserID;
  P.MultiplexID := MultID;
  AddParams(S);
  AddData(S);
  { NetBIOS Header }
  Header.MsgType := 0;
  Header.Flags := 0;
  Header.Length := Swap(WORD(Length(s)));
  SetLength(s1, Sizeof(Header));
  Move(Header, s1[1], SizeOf(Header));
  s := s1 + s;
  send(Socket, s[1], Length(s), 0);
end;

function RecvSMBReply(Socket: Integer; var Header: TSMBHeader;
  var Params, Data: AnsiString): Boolean;
var
  B: BYTE;
  W: WORD;
  Timeout: Int64;
  RecvLen: DWORD;
begin
  Result := False;
  Params := '';
  Data := '';
  { Read Header }
  Timeout := GetMSCounter + 5000;
  repeat
    ioctlsocket(Socket, FIONREAD, @RecvLen);
    if RecvLen >= SizeOf(TSessionHeader) + SizeOf(TSMBHeader) then
    begin
      recv(Socket, Header, SizeOf(TSessionHeader), 0);
      recv(Socket, Header, SizeOf(Header), 0);
      Break;
    end;
    if GetMSCounter > Timeout then
      Exit;
  until False;
  { Read Params Length }
  Timeout := GetMSCounter + 5000;
  repeat
    ioctlsocket(Socket, FIONREAD, @RecvLen);
    if RecvLen > 0 then
    begin
      recv(Socket, B, 1, 0);
      W := WORD(B) * 2;
      Break;
    end;
    if GetMSCounter > Timeout then
      Exit;
  until False;
  { Read Params }
  if (W > 0) then
  begin
    Timeout := GetMSCounter + 5000;
    repeat
      ioctlsocket(Socket, FIONREAD, @RecvLen);
      if RecvLen >= w then
      begin
        SetLength(Params, w);
        recv(Socket, Params[1], w, 0);
        Break;
      end;
      if GetMSCounter > Timeout then
        Exit;
    until False;
  end;
  { Read Data Length }
  Timeout := GetMSCounter + 5000;
  repeat
    ioctlsocket(Socket, FIONREAD, @RecvLen);
    if RecvLen >= 2 then
    begin
      recv(Socket, w, 2, 0);
      Break;
    end;
    if GetMSCounter > Timeout then
      Exit;
  until False;
  { Read Data }
  if (w > 0) then
  begin
    Timeout := GetMSCounter + 5000;
    repeat
      ioctlsocket(Socket, FIONREAD, @RecvLen);
      if RecvLen >= w then
      begin
        SetLength(Data, w);
        recv(Socket, Data[1], w, 0);
        Break;
      end;
      if GetMSCounter > Timeout then
        Exit;
    until False;
  end;
  { Done }
  Result := True;
end;

function SMB_AddShare(ShareName, Directory: AnsiString): Boolean;
var
  Share: TShareEntry;
begin
  Share := TShareEntry.Create;
  Share.ShareName := ShareName;
  Share.Directory := Directory;
  { Calculate drive letter for this share }
  if (Length(Directory) < 2) or (Directory[2] <> ':') then
    GetDir(0, Directory);
  Share.Drive := Ord(UpCase(Directory[1])) - Ord('A') + 1;
  Result := True;
end;

function CheckTreeID(Session: TServerSession; Hdr: TSMBHeader): TTreeEntry;
var
  i: integer;
begin
  for i := 0 to Session.Trees.Count - 1 do
  begin
    Result := Session.Trees[i];
    if Result.TreeID = Hdr.TreeID then
      Exit;
  end;
  SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8, ERRSRV,
    ERRinvnid, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID,
    '', '');
  Result := nil;
end;

function CheckFileID(Tree: TTreeEntry; Hdr: TSMBHeader; FileID: WORD):
  TFileEntry;
var
  i: integer;
begin
  for i := 0 to Tree.Files.Count - 1 do
  begin
    Result := Tree.Files[i];
    if Result.FileID = FileID then
      Exit;
  end;
  SendSMBRequest(Tree.Session.Socket, Hdr.Command, $80 + $10 + $8, ERRSRV,
    ERRinvpfid, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID,
    '', '');
  Result := nil;
end;

procedure CheckServerWaitingSession(Session: TServerSession);
var
  Len: DWORD;
  Hdr: TSessionHeader;
  Req: TSessionRequest;
  Rsp: TSessionResponse;
begin
  ioctlsocket(Session.Socket, FIONREAD, @Len);
  if Len >= SizeOf(TSessionRequest) then
  begin
    recv(Session.Socket, Req, SizeOf(Req), 0);
    if Req.Header.MsgType = $81 then
    begin
      Hdr.MsgType := $82;
      Hdr.Flags := 0;
      Hdr.Length := 0;
      send(Session.Socket, Hdr, SizeOf(Hdr), 0);
      Session.State := ssWaitingNeg;
      Session.Timeout := GetMSCounter + 5000;
    end
    else
    begin
      Session.Timeout := 1; // To force removal
      Rsp.Header.MsgType := $83; // Negative Session Response
      Rsp.Header.Flags := 0;
      Rsp.Header.Length := 1;
      Rsp.ErrorCode := $8F; // Unspecified Error
      send(Session.Socket, Rsp, SizeOf(Rsp), 0);
    end;
  end;
end;

procedure CheckServerWaitingNeg(Session: TServerSession);
var
  Len: DWORD;
  Hdr: TSMBHeader;
  Params, Data: AnsiString;
begin
  ioctlsocket(Session.Socket, FIONREAD, @Len);
  if Len >= SizeOf(TSessionHeader) + SizeOf(TSMBHeader) then
  begin
    if RecvSMBReply(Session.Socket, Hdr, Params, Data)
      and (Hdr.Command = $72) then
    begin
      SendSMBRequest(Session.Socket, $72, $80 + $10 + $8,
        0, 0, 0, 0, 0, Hdr.MultiplexID,
        WordToStr($0000), // Dialect index $0000
        #1#2#3#4#5#6#7#8); // Key
      Session.State := ssReady;
      Session.Timeout := 0;
    end
    else
    begin
      Session.Timeout := 1; // To force removal
    end;
  end;
end;

procedure ServerTreeConnectRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);

  function CheckShare(Data: AnsiString): TShareEntry;
  var
    i: integer;
  begin
    Data := Copy(Data, 4, Length(Data));
    Data := UpperCase(StrPas(PChar(Data)));
    { Remove Server Part }
    i := Pos('\', Data);
    Data := Copy(Data, i + 1, Length(Data));
    { Search through the Share Names }
    for i := 0 to Shares.Count - 1 do
    begin
      Result := Shares[i];
      if UpperCase(Result.ShareName) = Data then
        Exit;
    end;
    { No share found }
    Result := nil;
  end;

var
  Tree: TTreeEntry;
  Share: TShareEntry;
begin
  Share := CheckShare(Data);
  if Assigned(Share) then
  begin
    // Create New Tree
    Tree := TTreeEntry.Create(Session, Share);
    SendSMBRequest(Session.Socket, $70, $80 + $10 + $8,
      0, 0, Tree.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID,
      WordToStr($0400) + // Max Msg Size $0400
      WordToStr(Tree.TreeID), // Tree ID
      '');
  end
  else
  begin
    SendSMBRequest(Session.Socket, $70, $80 + $10 + $8, ERRSRV,
      ERRInvNetName, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID,
      WordToStr($0400) + // Max Msg Size $0400
      WordToStr($0000), // Tree ID
      '');
  end;
end;

procedure ServerTreeDisconnectRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
var
  Tree: TTreeEntry;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Close Tree }
  Tree.Free;

  { Reply }
  SendSMBRequest(Session.Socket, $71, $80 + $10 + $8,
    0, 0, 0, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID, '', '');
end;

procedure ServerQueryInformationRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
var
  Attribs: WORD;
  Tree: TTreeEntry;
  Time, Size: DWORD;
  FileName: AnsiString;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Parse FileName }
  FileName := Copy(Data, 2, Length(Data));
  FileName := StrPas(PChar(FileName));
  if Copy(FileName, 1, 1) = '\' then
    FileName := Copy(FileName, 2, 255);
  FileName := GoodDir(Tree.Share.Directory) + FileName;

  { Get Attributes }
  if GetFileAttribs(FileName, Attribs, Time, Size) then
  begin
    SendSMBRequest(Session.Socket, $08, $80 + $10 + $8,
      0, 0, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID,
      WORDToStr(Attribs) + // File Attributes
      DWORDToStr(Time) + // File Time
      DWORDToStr(Size) + // File Size
      #00#00#00#00#00#00#00#00#00#00, // Reserved
      '');
  end
  else
  begin
    SendSMBRequest(Session.Socket, $08, $80 + $10 + $8,
      ERRDOS, ErrBadFile, Hdr.TreeID, Hdr.ProcessID,
      Hdr.UserID, Hdr.MultiplexID, '', '');
  end;
end;

procedure ServerSearchRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);

  function ResumeKey(FileName: AnsiString): AnsiString;
  var
    i: Integer;
    Name, Ext: AnsiString;
  begin
    FileName := ExtractFileName(FileName);
    if (FileName = '.') or (FileName = '..') then
    begin
      Name := FileName;
      Ext := '';
    end
    else
    begin
      i := Pos('.', FileName);
      if i > 0 then
        Name := Copy(FileName, 1, i - 1)
      else
        Name := FileName;
      if i > 0 then
        Ext := Copy(FileName, i + 1, 3)
      else
        Ext := '';
    end;
    Result := #0 + PadR(Name, 8, ' ') + PadR(Ext, 3, ' ') +
      #0#0#0#0#0 + #0#0#0#0;
  end;

  function Entry(FileName: AnsiString; Rec: TSearchRec): AnsiString;
  begin
    Result :=
      ResumeKey(FileName) +
      Chr(SMB_Attrib(Rec.Attr)) +
      WORDToStr(DWORD(Rec.Time) and $FFFF) +
      WORDToStr(DWORD(Rec.Time) shr 16) +
      DWORDToStr(DWORD(Rec.Size)) +
      PadR(ExtractFileName(FileName), 12, ' ') + #0;
  end;

  function IsSearchMatch(Attr, Mask: WORD): Boolean;
  begin
    if (Mask and faVolumeID) = faVolumeID then
    begin
      Result := ((Attr and faVolumeID) = faVolumeID);
    end
    else
    begin
      Result := False;
      if ((Attr and faDirectory) = faDirectory)
        and ((Mask and faDirectory) <> faDirectory) then
        Exit;
      if ((Attr and faHidden) = faDirectory)
        and ((Mask and faHidden) <> faHidden) then
        Exit;
      if ((Attr and faSysFile) = faSysFile)
        and ((Mask and faSysFile) <> faSysFile) then
        Exit;
      if ((Attr and faReadOnly) = faReadOnly)
        and ((Mask and faReadOnly) <> faReadOnly) then
        Exit;
      if ((Attr and faVolumeID) = faVolumeID)
        and ((Mask and faVolumeID) <> faVolumeID) then
        Exit;
      Result := True;
    end;
  end;

var
  R: Integer;
  MCnt, Cnt: WORD;
  Rec: TSearchRec;
  Tree: TTreeEntry;
  Mask, Path, FileName, FileStr, RKey: AnsiString;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Parse Data }
  Move(Params[1], MCnt, 2);
  Mask := Copy(Data, 2, Length(Data));
  Mask := StrPas(PChar(Mask));
  RKey := Copy(Data, Length(Mask) + 6, Length(Data));

  { New Search? Store Params }
  if RKey = '' then
  begin
    Move(Params[3], Tree.Search.Attr, 2);
    Tree.Search.Attr := DOS_Attrib(Tree.Search.Attr);
    if Tree.Search.Attr = faVolumeID then
    begin
      Mask := Chr(64 + Tree.Share.Drive) + ':\*.*';
    end
    else
    begin
      if Copy(Mask, 1, 1) = '\' then
        Mask := Copy(Mask, 2, 255);
      Tree.Search.Mask := GoodDir(Tree.Share.Directory) + Mask;
    end;
  end;

  { Search }
  Cnt := 0;
  FileStr := '';
  Path := GoodDir(ExtractFilePath(Tree.Search.Mask));
  R := FindFirst(Tree.Search.Mask, faAnyFile, Rec);
  while (R = 0) and (Cnt < MCnt) do
  begin
    FileName := Path + Rec.Name;
    if RKey = '' then
    begin
      { We're beyond the ResumeKey pointer (or there was no ResueKey) }
      if IsSearchMatch(Rec.Attr, Tree.Search.Attr) then
      begin
        FileStr := FileStr + Entry(FileName, Rec);
        Inc(Cnt);
      end;
    end
    else
    begin
      { We still haven't reached the ResumeKey pointer }
      if Copy(ResumeKey(FileName), 2, 11) = Copy(RKey, 2, 11) then
        RKey := '';
    end;
    R := FindNext(Rec);
  end;
  FindClose(Rec);

  { Send Response }
  SendSMBRequest(Session.Socket, $81, $80 + $10 + $8,
    0, 0, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID,
    WordToStr(Cnt),
    #5 + // Buffer Format 5 (Variable Block)
    WORDToStr(Length(FileStr)) + // Data Length
    FileStr); // Data
end;

procedure ServerQueryInformationDiskRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
var
  Tree: TTreeEntry;
  FreeAlloc, TotalAlloc: Int64;
  BlockSize, BlocksPerAlloc: WORD;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Calculate disk properties }
  BlockSize := 4096; // Fixed
  BlocksPerAlloc := 2048; // Fixed
  FreeAlloc := (DiskFree(Tree.Share.Drive) div (Int64(BlockSize) *
    Int64(BlocksPerAlloc)));
  TotalAlloc := (DiskSize(Tree.Share.Drive) div (Int64(BlockSize) *
    Int64(BlocksPerAlloc)));

  { Send Response }
  SendSMBRequest(Session.Socket, $80, $80 + $10 + $8,
    0, 0, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID,
    WordToStr(TotalAlloc) +
    WordToStr(BlocksPerAlloc) +
    WordToStr(BlockSize) +
    WordToStr(FreeAlloc) +
    WordToStr(0),
    '');
end;

procedure ServerOpenRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
var
  H: Integer;
  Tree: TTreeEntry;
  Time, Size: DWORD;
  _File: TFileEntry;
  FileName: AnsiString;
  AccessMode, GrantedAccess, Attribs: WORD;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Parse Parameters }
  Move(Params[1], AccessMode, 2);
  AccessMode := DOS_AccessMode(AccessMode, GrantedAccess);
  FileName := Copy(Data, 2, Length(Data));
  FileName := StrPas(PChar(FileName));
  if Copy(FileName, 1, 1) = '\' then
    FileName := Copy(FileName, 2, 255);
  FileName := GoodDir(Tree.Share.Directory) + FileName;

  { Check File Existence }
  if not FileExists(FileName) then
  begin
    SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
      ErrDOS, ERRbadfile, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID,
      Hdr.MultiplexID, '', '');
    Exit;
  end;

  { Get File Attributes }
  if not GetFileAttribs(FileName, Attribs, Time, Size) then
  begin
    SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
      ErrDOS, ERRnoaccess, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID,
      Hdr.MultiplexID, '', '');
    Exit;
  end;

  { Open }
  H := FileOpen(FileName, AccessMode);
  if H = -1 then
  begin
    SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
      ErrDOS, ERRnoaccess, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID,
      Hdr.MultiplexID, '', '');
    Exit;
  end;

  { All right. Create the file object }
  _File := TFileEntry.Create(Tree, H);
  SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
    0, 0, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID,
    WORDToStr(_File.FileID) +
    WORDToStr(Attribs) +
    DWORDToStr(Time) +
    DWORDToStr(Size) +
    WORDToStr(GrantedAccess),
    '');
end;

procedure ServerCreateNewRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
var
  H: Integer;
  Attribs: WORD;
  Tree: TTreeEntry;
  _File: TFileEntry;
  FileName: AnsiString;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Parse Parameters }
  Move(Params[1], Attribs, 2);
  Attribs := DOS_Attrib(Attribs);
  FileName := Copy(Data, 2, Length(Data));
  FileName := StrPas(PChar(FileName));
  if Copy(FileName, 1, 1) = '\' then
    FileName := Copy(FileName, 2, 255);
  FileName := GoodDir(Tree.Share.Directory) + FileName;

  { Open }
  H := FileCreate(FileName);
  if H = -1 then
  begin
    SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
      ErrDOS, ERRnoaccess, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID,
      Hdr.MultiplexID, '', '');
    Exit;
  end;

  { All right. Create the File Object }
  _File := TFileEntry.Create(Tree, H);
  SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
    0, 0, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID,
    WORDToStr(_File.FileID),
    '');
end;

procedure ServerCloseRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
var
  FileID: WORD;
  Tree: TTreeEntry;
  _File: TFileEntry;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Check File ID }
  Move(Params[1], FileID, 2);
  _File := CheckFileID(Tree, Hdr, FileID);
  if not Assigned(_File) then
    Exit;

  { Close File }
  _File.Free;

  { Done }
  SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
    0, 0, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID, '', '');
end;

procedure ServerReadRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
var
  Offset: DWORD;
  Tree: TTreeEntry;
  _File: TFileEntry;
  Buffer: AnsiString;
  FileID, Count: WORD;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Check File ID }
  Move(Params[1], FileID, 2);
  Move(Params[3], Count, 2);
  Move(Params[5], Offset, 4);
  _File := CheckFileID(Tree, Hdr, FileID);
  if not Assigned(_File) then
    Exit;

  { Read from File }
  if Count > 0 then
  begin
    FileSeek(_File.Handle, Offset, fsFromBeginning);
    SetLength(Buffer, Count);
    SetLength(Buffer, FileRead(_File.Handle, Buffer[1], Count));
  end
  else
  begin
    Buffer := '';
  end;

  { Done }
  SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
    0, 0, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID,
    WORDToStr(Length(Buffer)) + // Count
    #0#0#0#0#0#0#0#0, // Reserved
    #1 + // 0x01 = "Data Block"
    WORDToStr(Length(Buffer)) + // Data Length
    Buffer); // Data
end;

procedure ServerWriteRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
var
  Offset: DWORD;
  Tree: TTreeEntry;
  _File: TFileEntry;
  Buffer: AnsiString;
  FileID, Count: WORD;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Check File ID }
  Move(Params[1], FileID, 2);
  Move(Params[3], Count, 2);
  Move(Params[5], Offset, 4);
  _File := CheckFileID(Tree, Hdr, FileID);
  if not Assigned(_File) then
    Exit;

  { Get Data }
  Buffer := Copy(Data, 4, Length(Data));

  { Write to File }
  if Length(Buffer) > 0 then
  begin
    FileSeek(_File.Handle, Offset, fsFromBeginning);
    Count := FileWrite(_File.Handle, Buffer[1], Length(Buffer));
  end
  else
  begin
    FileTruncate(_File.Handle, Offset);
    Count := 0;
  end;

  { Done }
  SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
    0, 0, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID,
    WORDToStr(Length(Buffer)), // Count
    '');
end;

procedure ServerRenameRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
var
  Tree: TTreeEntry;
  OldName, NewName: AnsiString;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Parse FileNames }
  OldName := Copy(Data, 2, Length(Data));
  OldName := StrPas(PChar(OldName));
  NewName := Copy(Data, Length(OldName) + 4, Length(Data));
  NewName := StrPas(PChar(NewName));
  {}
  if Copy(OldName, 1, 1) = '\' then
    OldName := Copy(OldName, 2, 255);
  OldName := GoodDir(Tree.Share.Directory) + OldName;
  if Copy(NewName, 1, 1) = '\' then
    NewName := Copy(NewName, 2, 255);
  NewName := GoodDir(Tree.Share.Directory) + NewName;

  { Source exists? }
  if not FileExists(OldName) and not IsDir(OldName) then
  begin
    SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
      ErrDOS, ERRbadfile, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID,
      Hdr.MultiplexID, '', '');
    Exit;
  end;

  { Rename }
  if not RenameFile(OldName, NewName) then
  begin
    SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
      ErrDOS, ERRnoaccess, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID,
      Hdr.MultiplexID, '', '');
    Exit;
  end;

  { Done }
  SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
    0, 0, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID, '', '');
end;

procedure ServerDeleteRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
var
  Ok: Boolean;
  Tree: TTreeEntry;
  FileName: AnsiString;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Parse FileNames }
  FileName := Copy(Data, 2, 255);
  FileName := StrPas(PChar(FileName));
  if Copy(FileName, 1, 1) = '\' then
    FileName := Copy(FileName, 2, 255);
  FileName := GoodDir(Tree.Share.Directory) + FileName;

  { File exists? }
  if not FileExists(FileName) and not IsDir(FileName) then
  begin
    SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
      ErrDOS, ERRbadfile, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID,
      Hdr.MultiplexID, '', '');
    Exit;
  end;

  { Delete }
  if FileExists(FileName) then
  begin
    Ok := DeleteFile(FileName);
  end
  else
  begin
    Ok := RemoveDir(FileName);
  end;
  if not Ok then
  begin
    SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
      ErrDOS, ERRnoaccess, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID,
      Hdr.MultiplexID, '', '');
    Exit;
  end;

  { Done }
  SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
    0, 0, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID, '', '');
end;

procedure ServerCreateDirectoryRequest(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
var
  Tree: TTreeEntry;
  DirName: AnsiString;
begin
  { Check Tree ID }
  Tree := CheckTreeID(Session, Hdr);
  if not Assigned(Tree) then
    Exit;

  { Parse FileName }
  DirName := Copy(Data, 2, 255);
  DirName := StrPas(PChar(DirName));
  if Copy(DirName, 1, 1) = '\' then
    DirName := Copy(DirName, 2, 255);
  DirName := GoodDir(Tree.Share.Directory) + DirName;

  { Try to create directory }
  if not CreateDir(DirName) then
  begin
    SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
      ErrDOS, ERRnoaccess, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID,
      Hdr.MultiplexID, '', '');
    Exit;
  end;

  { Done }
  SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
    0, 0, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID, Hdr.MultiplexID, '', '');
end;

procedure ServerCommandNotFound(Session: TServerSession;
  Hdr: TSMBHeader; Params, Data: AnsiString);
begin
  { Send Response }
  SendSMBRequest(Session.Socket, Hdr.Command, $80 + $10 + $8,
    ErrSrv, ERRsmbcmd, Hdr.TreeID, Hdr.ProcessID, Hdr.UserID,
    Hdr.MultiplexID, '', '');
end;

procedure CheckServerOther(Session: TServerSession);
var
  W: WORD;
  Len: DWORD;
  Hdr: TSMBHeader;
  Params, Data: AnsiString;
begin
  // Connection is Ok?
  if recv(Session.Socket, W, 0, 0) <> 0 then
  begin
    Session.Timeout := 1;
    Exit;
  end;
  // Anything New to read?
  ioctlsocket(Session.Socket, FIONREAD, @Len);
  if Len >= SizeOf(TSessionHeader) + SizeOf(TSMBHeader) then
  begin
    if RecvSMBReply(Session.Socket, Hdr, Params, Data) then
    begin
      case Hdr.Command of
        $00: // CreateDirectoryRequest
          ServerCreateDirectoryRequest(Session, Hdr, Params, Data);
        $01: // DeleteDirectoryRequest
          ServerDeleteRequest(Session, Hdr, Params, Data);
        $02: // Open Request
          ServerOpenRequest(Session, Hdr, Params, Data);
        $04: // Close Request
          ServerCloseRequest(Session, Hdr, Params, Data);
        $06: // Delete Request
          ServerDeleteRequest(Session, Hdr, Params, Data);
        $07: // Read Request
          ServerRenameRequest(Session, Hdr, Params, Data);
        $08: // Query Information Request
          ServerQueryInformationRequest(Session, Hdr, Params, Data);
        $0A: // Read Request
          ServerReadRequest(Session, Hdr, Params, Data);
        $0B: // Write Request
          ServerWriteRequest(Session, Hdr, Params, Data);
        $0F: // Create New Request
          ServerCreateNewRequest(Session, Hdr, Params, Data);
        $70: // Tree Connnect Request
          ServerTreeConnectRequest(Session, Hdr, Params, Data);
        $71: // Tree Disconnnect Request
          ServerTreeDisconnectRequest(Session, Hdr, Params, Data);
        $80: // Query InformationDiskRequest
          ServerQueryInformationDiskRequest(Session, Hdr, Params, Data);
        $81: // Search Request
          ServerSearchrequest(Session, Hdr, Params, Data);
      else
        ServerCommandNotFound(Session, Hdr, Params, Data);
      end;
    end
    else
    begin
      Session.Timeout := 1; // To force removal
    end;
  end;
end;

procedure ServerSessionWork(Session: TServerSession);
begin
  if not SMB_Started then
    Exit;
  if (Session.Timeout > 0) and (GetMSCounter > Session.Timeout) then
  begin
    Session.Free;
  end
  else
  begin
    case Session.State of
      ssWaitingSession:
        CheckServerWaitingSession(Session);
      ssWaitingNeg:
        CheckServerWaitingNeg(Session);
      ssReady:
        CheckServerOther(Session);
    end;
  end;
end;

procedure SMBServerProc;
var
  Len: LongInt;
  Addr: TSockAddr;
  i, HSock: Integer;
begin
  if SMB_Started then
  begin
    Len := SizeOf(Addr);
    HSock := Accept(ServerSocket, Addr, Len);
    if HSock <> SOCKET_ERROR then
      TServerSession.Create(HSock);
    i := 0;
    while i < ServerSessions.Count do
    begin
      ServerSessionWork(TServerSession(ServerSessions[i]));
      Inc(i);
    end;
  end;
end;

{ TServerSession }

constructor TServerSession.Create(ASocket: Integer);
begin
  inherited Create;
  Socket := ASocket;
  State := ssWaitingSession;
  Timeout := GetMSCounter + 5000;
  Trees := TList.Create;
  ServerSessions.Add(Self);
end;

destructor TServerSession.Destroy;
begin
  //	CloseSocket(Socket);  hola
  ServerSessions.Remove(Self);
  while Trees.Count > 0 do
    TTreeEntry(Trees[0]).Free;
  Trees.Free;
  inherited;
end;

{ TTreeEntry }

constructor TTreeEntry.Create(ASession: TServerSession; AShare: TShareEntry);
begin
  inherited Create;
  Inc(TreeIDCounter);
  TreeID := Word(TreeIDCounter);
  Session := ASession;
  Share := AShare;
  Files := TList.Create;
  Session.Trees.Add(Self);
end;

destructor TTreeEntry.Destroy;
begin
  if Assigned(Session) then
    Session.Trees.Remove(Self);
  while Files.Count > 0 do
    TFileEntry(Files[0]).Free;
  Files.Free;
  inherited;
end;

{ TFileEntry }

constructor TFileEntry.Create(ATree: TTreeEntry; AHandle: Integer);
begin
  inherited Create;
  Inc(FileIDCounter);
  FileID := WORD(FileIDCounter);
  Tree := ATree;
  Handle := AHandle;
  Tree.Files.Add(Self);
end;

destructor TFileEntry.Destroy;
begin
  FileClose(Handle);
  if Assigned(Tree) then
    Tree.Files.Remove(Self);
  inherited;
end;

function SMB_Init: Boolean;
var
  Addr: TSockAddr;
begin
  if not SMB_Started then
  begin
    Addr.sin_family := PF_INET;
    Addr.sin_port := Swap(WORD(139));
    Addr.sin_addr.S_addr := INADDR_ANY;
    ServerSocket := Socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
    if (ServerSocket = SOCKET_ERROR)
      or (Bind(ServerSocket, Addr, SizeOf(Addr)) = SOCKET_ERROR)
      or (Listen(ServerSocket, 10) = SOCKET_ERROR) then
    begin
      if ServerSocket <> SOCKET_ERROR then
        CloseSocket(ServerSocket);
      ServerSocket := 0;
      Result := False;
      Exit;
    end;
    TreeIDCounter := 0;
    FileIDCounter := 0;
    ServerSessions := TList.Create;
    NET_AddIdleProc(SMBServerProc);
    SMB_Started := True;
  end;
  Result := True;
end;

function SMB_Deinit: Boolean;
begin
  if SMB_Started then
  begin
    NET_RemoveIdleProc(SMBServerProc);
    SMB_Started := False;
    CloseSocket(ServerSocket);
    while ServerSessions.Count > 0 do
      TServerSession(ServerSessions[0]).Free;
    ServerSessions.Free;
  end;
  Result := True;
end;

procedure NetChanged;
begin
  { The network changed. We should try to restart NetBIOS }
  SMB_DeInit;
  SMB_Init;
end;

procedure Init;
begin
  NET_NotifyNetChanged;
end;

{ TShareEntry }

constructor TShareEntry.Create;
begin
  inherited Create;
  Shares.Add(Self);
end;

destructor TShareEntry.Destroy;
begin
  Shares.Remove(Self);
  inherited Destroy;
end;

initialization
  SMB_Started := False;
  Shares := TList.Create;
  Net_AddNetChangedProc(NetChanged);
  
finalization
  NET_RemoveNetChangedProc(NetChanged);
  SMB_Deinit;
  while Shares.Count > 0 do
    TShareEntry(Shares[0]).Free;
  Shares.Free;
end.

