
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosSock.pas                                          }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Kylix and Delphi Cross-Platform Runtime Library   }
{     Copyright (C) 1996, 2001                          }
{     Borland Software Corporation                      }
{     www: http://www.borland.com                       }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}

{ *************************************************************************** }
{                                                                             }
{ Kylix and Delphi Cross-Platform Runtime Library                             }
{                                                                             }
{ Copyright (c) 1995-2001 Borland Software Corporation                        }
{                                                                             }
{ This file may be distributed and/or modified under the terms of the GNU     }
{ General Public License (GPL) version 2 as published by the Free Software    }
{ Software Foundation and appearing at http://www.borland.com/kylix/gpl.html. }
{                                                                             }
{ Licensees holding a valid Borland No-Nonsense License for this Software may }
{ use this file in accordance with such license, which appears in the file    }
{ license.txt that came with this Software.                                   }
{                                                                             }
{ *************************************************************************** }
{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosSock.pas,v 1.7 2003/10/28 21:22:04 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosSock.pas,v $
  Revision 1.7  2003/10/28 21:22:04  iwache
  Call of Halt(n) in procedure InitNetwork works fine. The Bug
  was because of NetworkAdapter.Free was called twice in
  procedure InitNetwork and again in the finalization section.
  This is fixed now. (Thanks to Diego for the hint.)

  Revision 1.6  2003/10/25 22:09:25  iwache
  Call of Halt(n) in procedure InitNetwork replaced with ExitProcess(n).
  The old solution did not worked with Windows 2000. If you now run
  a WDOSX socket in a Windows DOS box you get the error "Could not
  initialize network adapter."
  Halt seems not proper to work with DWPL in Windows DOS box.

  Revision 1.5  2003/10/07 19:25:42  iwache
  In function Shutdown now drop WSAENOTCONN
  "The socket is not connected" error.

  Revision 1.4  2003/06/17 18:53:53  iwache
  var NetworkAdapter: TNetworkAdapter is now public available.
  This is useful for get informations about the current adapter.

  Revision 1.3  2003/06/12 20:22:42  iwache
  function select modified and a bug fixed with wrong time format.

  Revision 1.2  2003/06/11 21:10:26  iwache
  Missing functions for compatibility with unit Sockets.pas added:
    function select()
    function inet_addr()
    function gethostbyaddr()
    function gethostbyname()
    function getservbyname()
    function inet_ntoa()
    function gethostname()

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosSock;

interface

uses
  Windows, SysUtils, Threads, WDosUtil, WDosNetwork;

{ HPPEMIT '#include <windows.h>'}

type
{$EXTERNALSYM u_char}
  u_char = Char;
{$EXTERNALSYM u_short}
  u_short = Word;
{$EXTERNALSYM u_int}
  u_int = Integer;
{$EXTERNALSYM u_long}
  u_long = LongWord;

  { The new type to be used in all
    instances which refer to sockets. }
{$EXTERNALSYM TSocket}
  TSocket = u_int;

const
{$EXTERNALSYM FD_SETSIZE}
  FD_SETSIZE = 64;

  // WinSock 2 extension -- manifest constants for shutdown()
{$EXTERNALSYM SD_RECEIVE}
  SD_RECEIVE = 0;
{$EXTERNALSYM SD_SEND}
  SD_SEND = 1;
{$EXTERNALSYM SD_BOTH}
  SD_BOTH = 2;

type
  // the following emits are a workaround to the name conflict with
  // procedure FD_SET and struct fd_set in winsock.h
  (*$HPPEMIT '#include <winsock.h>'*)
  (*$HPPEMIT 'namespace Winsock'*)
  (*$HPPEMIT '{'*)
  (*$HPPEMIT 'typedef fd_set *PFDSet;'*)// due to name conflict with procedure FD_SET
    (*$HPPEMIT 'typedef fd_set TFDSet;'*)// due to name conflict with procedure FD_SET
  (*$HPPEMIT '}'*)

{$NODEFINE PFDSet}
  PFDSet = ^TFDSet;
{$NODEFINE TFDSet}
  TFDSet = record
    fd_count: u_int;
    fd_array: array[0..FD_SETSIZE - 1] of TSocket;
  end;

  PTimeVal = ^TTimeVal;
{$EXTERNALSYM timeval}
  timeval = record
    tv_sec: Longint;
    tv_usec: Longint;
  end;
  TTimeVal = timeval;

const
{$EXTERNALSYM IOCPARM_MASK}
  IOCPARM_MASK = $7F;
{$EXTERNALSYM IOC_VOID}
  IOC_VOID = $20000000;
{$EXTERNALSYM IOC_OUT}
  IOC_OUT = $40000000;
{$EXTERNALSYM IOC_IN}
  IOC_IN = $80000000;
{$EXTERNALSYM IOC_INOUT}
  IOC_INOUT = (IOC_IN or IOC_OUT);

{$EXTERNALSYM FIONREAD}
  FIONREAD = IOC_OUT or { get # bytes to read }
  ((Longint(SizeOf(Longint)) and IOCPARM_MASK) shl 16) or
    (Longint(Byte('f')) shl 8) or 127;
{$EXTERNALSYM FIONBIO}
  FIONBIO = IOC_IN or { set/clear non-blocking i/o }
  ((Longint(SizeOf(Longint)) and IOCPARM_MASK) shl 16) or
    (Longint(Byte('f')) shl 8) or 126;
{$EXTERNALSYM FIOASYNC}
  FIOASYNC = IOC_IN or { set/clear async i/o }
  ((Longint(SizeOf(Longint)) and IOCPARM_MASK) shl 16) or
    (Longint(Byte('f')) shl 8) or 125;

type
  PHostEnt = ^THostEnt;
{$EXTERNALSYM hostent}
  hostent = record
    h_name: PChar;
    h_aliases: ^PChar;
    h_addrtype: Smallint;
    h_length: Smallint;
    case Byte of
      0: (h_addr_list: ^PChar);
      1: (h_addr: ^PChar)
  end;
  THostEnt = hostent;

  PNetEnt = ^TNetEnt;
{$EXTERNALSYM netent}
  netent = record
    n_name: PChar;
    n_aliases: ^PChar;
    n_addrtype: Smallint;
    n_net: u_long;
  end;
  TNetEnt = netent;

  PServEnt = ^TServEnt;
{$EXTERNALSYM servent}
  servent = record
    s_name: PChar;
    s_aliases: ^PChar;
    s_port: Word;
    s_proto: PChar;
  end;
  TServEnt = servent;

  PProtoEnt = ^TProtoEnt;
{$EXTERNALSYM protoent}
  protoent = record
    p_name: PChar;
    p_aliases: ^Pchar;
    p_proto: Smallint;
  end;
  TProtoEnt = protoent;

const

  { Protocols }

{$EXTERNALSYM IPPROTO_IP}
  IPPROTO_IP = 0; { dummy for IP }
{$EXTERNALSYM IPPROTO_ICMP}
  IPPROTO_ICMP = 1; { control message protocol }
{$EXTERNALSYM IPPROTO_IGMP}
  IPPROTO_IGMP = 2; { group management protocol }
{$EXTERNALSYM IPPROTO_GGP}
  IPPROTO_GGP = 3; { gateway^2 (deprecated) }
{$EXTERNALSYM IPPROTO_TCP}
  IPPROTO_TCP = 6; { tcp }
{$EXTERNALSYM IPPROTO_PUP}
  IPPROTO_PUP = 12; { pup }
{$EXTERNALSYM IPPROTO_UDP}
  IPPROTO_UDP = 17; { user datagram protocol }
{$EXTERNALSYM IPPROTO_IDP}
  IPPROTO_IDP = 22; { xns idp }
{$EXTERNALSYM IPPROTO_ND}
  IPPROTO_ND = 77; { UNOFFICIAL net disk proto }

{$EXTERNALSYM IPPROTO_RAW}
  IPPROTO_RAW = 255; { raw IP packet }
{$EXTERNALSYM IPPROTO_MAX}
  IPPROTO_MAX = 256;

  { Port/socket numbers: network standard functions}

{$EXTERNALSYM IPPORT_ECHO}
  IPPORT_ECHO = 7;
{$EXTERNALSYM IPPORT_DISCARD}
  IPPORT_DISCARD = 9;
{$EXTERNALSYM IPPORT_SYSTAT}
  IPPORT_SYSTAT = 11;
{$EXTERNALSYM IPPORT_DAYTIME}
  IPPORT_DAYTIME = 13;
{$EXTERNALSYM IPPORT_NETSTAT}
  IPPORT_NETSTAT = 15;
{$EXTERNALSYM IPPORT_FTP}
  IPPORT_FTP = 21;
{$EXTERNALSYM IPPORT_TELNET}
  IPPORT_TELNET = 23;
{$EXTERNALSYM IPPORT_SMTP}
  IPPORT_SMTP = 25;
{$EXTERNALSYM IPPORT_TIMESERVER}
  IPPORT_TIMESERVER = 37;
{$EXTERNALSYM IPPORT_NAMESERVER}
  IPPORT_NAMESERVER = 42;
{$EXTERNALSYM IPPORT_WHOIS}
  IPPORT_WHOIS = 43;
{$EXTERNALSYM IPPORT_MTP}
  IPPORT_MTP = 57;

  { Port/socket numbers: host specific functions }

{$EXTERNALSYM IPPORT_TFTP}
  IPPORT_TFTP = 69;
{$EXTERNALSYM IPPORT_RJE}
  IPPORT_RJE = 77;
{$EXTERNALSYM IPPORT_FINGER}
  IPPORT_FINGER = 79;
{$EXTERNALSYM IPPORT_TTYLINK}
  IPPORT_TTYLINK = 87;
{$EXTERNALSYM IPPORT_SUPDUP}
  IPPORT_SUPDUP = 95;

  { UNIX TCP sockets }

{$EXTERNALSYM IPPORT_EXECSERVER}
  IPPORT_EXECSERVER = 512;
{$EXTERNALSYM IPPORT_LOGINSERVER}
  IPPORT_LOGINSERVER = 513;
{$EXTERNALSYM IPPORT_CMDSERVER}
  IPPORT_CMDSERVER = 514;
{$EXTERNALSYM IPPORT_EFSSERVER}
  IPPORT_EFSSERVER = 520;

  { UNIX UDP sockets }

{$EXTERNALSYM IPPORT_BIFFUDP}
  IPPORT_BIFFUDP = 512;
{$EXTERNALSYM IPPORT_WHOSERVER}
  IPPORT_WHOSERVER = 513;
{$EXTERNALSYM IPPORT_ROUTESERVER}
  IPPORT_ROUTESERVER = 520;

  { Ports < IPPORT_RESERVED are reserved for
    privileged processes (e.g. root). }

{$EXTERNALSYM IPPORT_RESERVED}
  IPPORT_RESERVED = 1024;

  { Link numbers }

{$EXTERNALSYM IMPLINK_IP}
  IMPLINK_IP = 155;
{$EXTERNALSYM IMPLINK_LOWEXPER}
  IMPLINK_LOWEXPER = 156;
{$EXTERNALSYM IMPLINK_HIGHEXPER}
  IMPLINK_HIGHEXPER = 158;

type
{$EXTERNALSYM SunB}
  SunB = packed record
    s_b1, s_b2, s_b3, s_b4: u_char;
  end;

{$EXTERNALSYM SunW}
  SunW = packed record
    s_w1, s_w2: u_short;
  end;

  PInAddr = ^TInAddr;
{$EXTERNALSYM in_addr}
  in_addr = record
    case integer of
      0: (S_un_b: SunB);
      1: (S_un_w: SunW);
      2: (S_addr: u_long);
  end;
  TInAddr = in_addr;

  PSockAddrIn = ^TSockAddrIn;
{$EXTERNALSYM sockaddr_in}
  sockaddr_in = record
    case Integer of
      0: (sin_family: u_short;
        sin_port: u_short;
        sin_addr: TInAddr;
        sin_zero: array[0..7] of Char);
      1: (sa_family: u_short;
        sa_data: array[0..13] of Char)
  end;
  TSockAddrIn = sockaddr_in;

const
{$EXTERNALSYM INADDR_ANY}
  INADDR_ANY: u_long = $00000000;
{$EXTERNALSYM INADDR_LOOPBACK}
  // INADDR_LOOPBACK: u_long = $7F000001;
  INADDR_LOOPBACK: u_long = $0100007F;
  { TODO : The number value looks strange
           compared to the original Borland version }
{$EXTERNALSYM INADDR_BROADCAST}
   INADDR_BROADCAST: u_long = $FFFFFFFF;
{$EXTERNALSYM INADDR_NONE}
   INADDR_NONE: u_long = $FFFFFFFF;

{$EXTERNALSYM WSADESCRIPTION_LEN}
  WSADESCRIPTION_LEN = 256;
{$EXTERNALSYM WSASYS_STATUS_LEN}
  WSASYS_STATUS_LEN = 128;

type
  PWSAData = ^TWSAData;
{$EXTERNALSYM WSAData}
  WSAData = record // !!! also WSDATA
    wVersion: Word;
    wHighVersion: Word;
    szDescription: array[0..WSADESCRIPTION_LEN] of Char;
    szSystemStatus: array[0..WSASYS_STATUS_LEN] of Char;
    iMaxSockets: Word;
    iMaxUdpDg: Word;
    lpVendorInfo: PChar;
  end;
  TWSAData = WSAData;

{$EXTERNALSYM PTransmitFileBuffers}
  PTransmitFileBuffers = ^TTransmitFileBuffers;
{$EXTERNALSYM _TRANSMIT_FILE_BUFFERS}
  _TRANSMIT_FILE_BUFFERS = record
    Head: Pointer;
    HeadLength: DWORD;
    Tail: Pointer;
    TailLength: DWORD;
  end;
{$EXTERNALSYM TTransmitFileBuffers}
  TTransmitFileBuffers = _TRANSMIT_FILE_BUFFERS;
{$EXTERNALSYM TRANSMIT_FILE_BUFFERS}
  TRANSMIT_FILE_BUFFERS = _TRANSMIT_FILE_BUFFERS;

const
{$EXTERNALSYM TF_DISCONNECT}
  TF_DISCONNECT = $01;
{$EXTERNALSYM TF_REUSE_SOCKET}
  TF_REUSE_SOCKET = $02;
{$EXTERNALSYM TF_WRITE_BEHIND}
  TF_WRITE_BEHIND = $04;

  { Options for use with [gs]etsockopt at the IP level. }

{$EXTERNALSYM IP_OPTIONS}
  IP_OPTIONS = 1;
{$EXTERNALSYM IP_MULTICAST_IF}
  IP_MULTICAST_IF = 2; { set/get IP multicast interface   }
{$EXTERNALSYM IP_MULTICAST_TTL}
  IP_MULTICAST_TTL = 3; { set/get IP multicast timetolive  }
{$EXTERNALSYM IP_MULTICAST_LOOP}
  IP_MULTICAST_LOOP = 4; { set/get IP multicast loopback    }
{$EXTERNALSYM IP_ADD_MEMBERSHIP}
  IP_ADD_MEMBERSHIP = 5; { add  an IP group membership      }
{$EXTERNALSYM IP_DROP_MEMBERSHIP}
  IP_DROP_MEMBERSHIP = 6; { drop an IP group membership      }
{$EXTERNALSYM IP_TTL}
  IP_TTL = 7; { set/get IP Time To Live          }
{$EXTERNALSYM IP_TOS}
  IP_TOS = 8; { set/get IP Type Of Service       }
{$EXTERNALSYM IP_DONTFRAGMENT}
  IP_DONTFRAGMENT = 9; { set/get IP Don't Fragment flag   }

{$EXTERNALSYM IP_DEFAULT_MULTICAST_TTL}
  IP_DEFAULT_MULTICAST_TTL = 1; { normally limit m'casts to 1 hop  }
{$EXTERNALSYM IP_DEFAULT_MULTICAST_LOOP}
  IP_DEFAULT_MULTICAST_LOOP = 1; { normally hear sends if a member  }
{$EXTERNALSYM IP_MAX_MEMBERSHIPS}
  IP_MAX_MEMBERSHIPS = 20; { per socket; must fit in one mbuf }

  { This is used instead of -1, since the
    TSocket type is unsigned.}

{$EXTERNALSYM INVALID_SOCKET    =}
  INVALID_SOCKET = TSocket(not (0));
{$EXTERNALSYM SOCKET_ERROR      =}
  SOCKET_ERROR = -1;

  { Types }

{$EXTERNALSYM SOCK_STREAM}
  SOCK_STREAM = 1; { stream socket }
{$EXTERNALSYM SOCK_DGRAM}
  SOCK_DGRAM = 2; { datagram socket }
{$EXTERNALSYM SOCK_RAW}
  SOCK_RAW = 3; { raw-protocol interface }
{$EXTERNALSYM SOCK_RDM}
  SOCK_RDM = 4; { reliably-delivered message }
{$EXTERNALSYM SOCK_SEQPACKET}
  SOCK_SEQPACKET = 5; { sequenced packet stream }

  { Option flags per-socket. }

{$EXTERNALSYM SO_DEBUG}
  SO_DEBUG = $0001; { turn on debugging info recording }
{$EXTERNALSYM SO_ACCEPTCONN}
  SO_ACCEPTCONN = $0002; { socket has had listen() }
{$EXTERNALSYM SO_REUSEADDR}
  SO_REUSEADDR = $0004; { allow local address reuse }
{$EXTERNALSYM SO_KEEPALIVE}
  SO_KEEPALIVE = $0008; { keep connections alive }
{$EXTERNALSYM SO_DONTROUTE}
  SO_DONTROUTE = $0010; { just use interface addresses }
{$EXTERNALSYM SO_BROADCAST}
  SO_BROADCAST = $0020; { permit sending of broadcast msgs }
{$EXTERNALSYM SO_USELOOPBACK}
  SO_USELOOPBACK = $0040; { bypass hardware when possible }
{$EXTERNALSYM SO_LINGER}
  SO_LINGER = $0080; { linger on close if data present }
{$EXTERNALSYM SO_OOBINLINE}
  SO_OOBINLINE = $0100; { leave received OOB data in line }

{$EXTERNALSYM SO_DONTLINGER}
  SO_DONTLINGER = $FF7F;

  { Additional options. }

{$EXTERNALSYM SO_SNDBUF}
  SO_SNDBUF = $1001; { send buffer size }
{$EXTERNALSYM SO_RCVBUF}
  SO_RCVBUF = $1002; { receive buffer size }
{$EXTERNALSYM SO_SNDLOWAT}
  SO_SNDLOWAT = $1003; { send low-water mark }
{$EXTERNALSYM SO_RCVLOWAT}
  SO_RCVLOWAT = $1004; { receive low-water mark }
{$EXTERNALSYM SO_SNDTIMEO}
  SO_SNDTIMEO = $1005; { send timeout }
{$EXTERNALSYM SO_RCVTIMEO}
  SO_RCVTIMEO = $1006; { receive timeout }
{$EXTERNALSYM SO_ERROR}
  SO_ERROR = $1007; { get error status and clear }
{$EXTERNALSYM SO_TYPE}
  SO_TYPE = $1008; { get socket type }


  SO_CONNTIMEOUT  = 200;


  { Options for connect and disconnect data and options.  Used only by
    non-TCP/IP transports such as DECNet, OSI TP4, etc. }

{$EXTERNALSYM SO_CONNDATA}
  SO_CONNDATA = $7000;
{$EXTERNALSYM SO_CONNOPT}
  SO_CONNOPT = $7001;
{$EXTERNALSYM SO_DISCDATA}
  SO_DISCDATA = $7002;
{$EXTERNALSYM SO_DISCOPT}
  SO_DISCOPT = $7003;
{$EXTERNALSYM SO_CONNDATALEN}
  SO_CONNDATALEN = $7004;
{$EXTERNALSYM SO_CONNOPTLEN}
  SO_CONNOPTLEN = $7005;
{$EXTERNALSYM SO_DISCDATALEN}
  SO_DISCDATALEN = $7006;
{$EXTERNALSYM SO_DISCOPTLEN}
  SO_DISCOPTLEN = $7007;

  { Option for opening sockets for synchronous access. }

{$EXTERNALSYM SO_OPENTYPE}
  SO_OPENTYPE = $7008;

{$EXTERNALSYM SO_SYNCHRONOUS_ALERT}
  SO_SYNCHRONOUS_ALERT = $10;
{$EXTERNALSYM SO_SYNCHRONOUS_NONALERT}
  SO_SYNCHRONOUS_NONALERT = $20;

  { Other NT-specific options. }

{$EXTERNALSYM SO_MAXDG}
  SO_MAXDG = $7009;
{$EXTERNALSYM SO_MAXPATHDG}
  SO_MAXPATHDG = $700A;
{$EXTERNALSYM SO_UPDATE_ACCEPT_CONTEXT}
  SO_UPDATE_ACCEPT_CONTEXT = $700B;
{$EXTERNALSYM SO_CONNECT_TIME}
  SO_CONNECT_TIME = $700C;

  { TCP options. }

{$EXTERNALSYM TCP_NODELAY}
  TCP_NODELAY = $0001;
{$EXTERNALSYM TCP_BSDURGENT}
  TCP_BSDURGENT = $7000;

  { Address families. }

{$EXTERNALSYM AF_UNSPEC}
  AF_UNSPEC = 0; { unspecified }
{$EXTERNALSYM AF_UNIX}
  AF_UNIX = 1; { local to host (pipes, portals) }
{$EXTERNALSYM AF_INET}
  AF_INET = 2; { internetwork: UDP, TCP, etc. }
{$EXTERNALSYM AF_IMPLINK}
  AF_IMPLINK = 3; { arpanet imp addresses }
{$EXTERNALSYM AF_PUP}
  AF_PUP = 4; { pup protocols: e.g. BSP }
{$EXTERNALSYM AF_CHAOS}
  AF_CHAOS = 5; { mit CHAOS protocols }
{$EXTERNALSYM AF_IPX}
  AF_IPX = 6; { IPX and SPX }
{$EXTERNALSYM AF_NS}
  AF_NS = 6; { XEROX NS protocols }
{$EXTERNALSYM AF_ISO}
  AF_ISO = 7; { ISO protocols }
{$EXTERNALSYM AF_OSI}
  AF_OSI = AF_ISO; { OSI is ISO }
{$EXTERNALSYM AF_ECMA}
  AF_ECMA = 8; { european computer manufacturers }
{$EXTERNALSYM AF_DATAKIT}
  AF_DATAKIT = 9; { datakit protocols }
{$EXTERNALSYM AF_CCITT}
  AF_CCITT = 10; { CCITT protocols, X.25 etc }
{$EXTERNALSYM AF_SNA}
  AF_SNA = 11; { IBM SNA }
{$EXTERNALSYM AF_DECnet}
  AF_DECnet = 12; { DECnet }
{$EXTERNALSYM AF_DLI}
  AF_DLI = 13; { Direct data link interface }
{$EXTERNALSYM AF_LAT}
  AF_LAT = 14; { LAT }
{$EXTERNALSYM AF_HYLINK}
  AF_HYLINK = 15; { NSC Hyperchannel }
{$EXTERNALSYM AF_APPLETALK}
  AF_APPLETALK = 16; { AppleTalk }
{$EXTERNALSYM AF_NETBIOS}
  AF_NETBIOS = 17; { NetBios-style addresses }
{$EXTERNALSYM AF_VOICEVIEW}
  AF_VOICEVIEW = 18; { VoiceView }
{$EXTERNALSYM AF_FIREFOX}
  AF_FIREFOX = 19; { FireFox }
{$EXTERNALSYM AF_UNKNOWN1}
  AF_UNKNOWN1 = 20; { Somebody is using this! }
{$EXTERNALSYM AF_BAN}
  AF_BAN = 21; { Banyan }

{$EXTERNALSYM AF_MAX}
  AF_MAX = 22;

type
  { Structure used by kernel to store most addresses. }

{$EXTERNALSYM PSOCKADDR}
  PSOCKADDR = ^TSockAddr;
{$EXTERNALSYM TSockAddr}
  TSockAddr = sockaddr_in;

  { Structure used by kernel to pass protocol information in raw sockets. }
  PSockProto = ^TSockProto;
{$EXTERNALSYM sockproto}
  sockproto = record
    sp_family: u_short;
    sp_protocol: u_short;
  end;
  TSockProto = sockproto;

const
  { Protocol families, same as address families for now. }

{$EXTERNALSYM PF_UNSPEC}
  PF_UNSPEC = AF_UNSPEC;
{$EXTERNALSYM PF_UNIX}
  PF_UNIX = AF_UNIX;
{$EXTERNALSYM PF_INET}
  PF_INET = AF_INET;
{$EXTERNALSYM PF_IMPLINK}
  PF_IMPLINK = AF_IMPLINK;
{$EXTERNALSYM PF_PUP}
  PF_PUP = AF_PUP;
{$EXTERNALSYM PF_CHAOS}
  PF_CHAOS = AF_CHAOS;
{$EXTERNALSYM PF_NS}
  PF_NS = AF_NS;
{$EXTERNALSYM PF_IPX}
  PF_IPX = AF_IPX;
{$EXTERNALSYM PF_ISO}
  PF_ISO = AF_ISO;
{$EXTERNALSYM PF_OSI}
  PF_OSI = AF_OSI;
{$EXTERNALSYM PF_ECMA}
  PF_ECMA = AF_ECMA;
{$EXTERNALSYM PF_DATAKIT}
  PF_DATAKIT = AF_DATAKIT;
{$EXTERNALSYM PF_CCITT}
  PF_CCITT = AF_CCITT;
{$EXTERNALSYM PF_SNA}
  PF_SNA = AF_SNA;
{$EXTERNALSYM PF_DECnet}
  PF_DECnet = AF_DECnet;
{$EXTERNALSYM PF_DLI}
  PF_DLI = AF_DLI;
{$EXTERNALSYM PF_LAT}
  PF_LAT = AF_LAT;
{$EXTERNALSYM PF_HYLINK}
  PF_HYLINK = AF_HYLINK;
{$EXTERNALSYM PF_APPLETALK}
  PF_APPLETALK = AF_APPLETALK;
{$EXTERNALSYM PF_VOICEVIEW}
  PF_VOICEVIEW = AF_VOICEVIEW;
{$EXTERNALSYM PF_FIREFOX}
  PF_FIREFOX = AF_FIREFOX;
{$EXTERNALSYM PF_UNKNOWN1}
  PF_UNKNOWN1 = AF_UNKNOWN1;
{$EXTERNALSYM PF_BAN}
  PF_BAN = AF_BAN;

{$EXTERNALSYM PF_MAX}
  PF_MAX = AF_MAX;

type
  { Structure used for manipulating linger option. }
  PLinger = ^TLinger;
{$EXTERNALSYM linger}
  linger = record
    l_onoff: u_short;
    l_linger: u_short;
  end;
  TLinger = linger;

const
  { Level number for (get/set)sockopt() to apply to socket itself. }

{$EXTERNALSYM SOL_SOCKET}
  SOL_SOCKET = $FFFF; {options for socket level }

  { Maximum queue length specifiable by listen. }

{$EXTERNALSYM SOMAXCONN}
  SOMAXCONN = 5;

{$EXTERNALSYM MSG_OOB}
  MSG_OOB = $1; {process out-of-band data }
{$EXTERNALSYM MSG_PEEK}
  MSG_PEEK = $2; {peek at incoming message }
{$EXTERNALSYM MSG_DONTROUTE}
  MSG_DONTROUTE = $4; {send without using routing tables }

{$EXTERNALSYM MSG_MAXIOVLEN}
  MSG_MAXIOVLEN = 16;

{$EXTERNALSYM MSG_PARTIAL}
  MSG_PARTIAL = $8000; {partial send or recv for message xport }

  { Define constant based on rfc883, used by gethostbyxxxx() calls. }

{$EXTERNALSYM MAXGETHOSTSTRUCT}
  MAXGETHOSTSTRUCT = 1024;

  { Define flags to be used with the WSAAsyncSelect() call. }

{$EXTERNALSYM FD_READ}
  FD_READ = $01;
{$EXTERNALSYM FD_WRITE}
  FD_WRITE = $02;
{$EXTERNALSYM FD_OOB}
  FD_OOB = $04;
{$EXTERNALSYM FD_ACCEPT}
  FD_ACCEPT = $08;
{$EXTERNALSYM FD_CONNECT}
  FD_CONNECT = $10;
{$EXTERNALSYM FD_CLOSE}
  FD_CLOSE = $20;

  { All Windows Sockets error constants are biased by WSABASEERR from the "normal" }

{$EXTERNALSYM WSABASEERR}
  WSABASEERR = 10000;

  { Windows Sockets definitions of regular Microsoft C error constants }

{$EXTERNALSYM WSAEINTR}
  WSAEINTR = (WSABASEERR + 4);
{$EXTERNALSYM WSAEBADF}
  WSAEBADF = (WSABASEERR + 9);
{$EXTERNALSYM WSAEACCES}
  WSAEACCES = (WSABASEERR + 13);
{$EXTERNALSYM WSAEFAULT}
  WSAEFAULT = (WSABASEERR + 14);
{$EXTERNALSYM WSAEINVAL}
  WSAEINVAL = (WSABASEERR + 22);
{$EXTERNALSYM WSAEMFILE}
  WSAEMFILE = (WSABASEERR + 24);

  { Windows Sockets definitions of regular Berkeley error constants }

{$EXTERNALSYM WSAEWOULDBLOCK}
  WSAEWOULDBLOCK = (WSABASEERR + 35);
{$EXTERNALSYM WSAEINPROGRESS}
  WSAEINPROGRESS = (WSABASEERR + 36);
{$EXTERNALSYM WSAEALREADY}
  WSAEALREADY = (WSABASEERR + 37);
{$EXTERNALSYM WSAENOTSOCK}
  WSAENOTSOCK = (WSABASEERR + 38);
{$EXTERNALSYM WSAEDESTADDRREQ}
  WSAEDESTADDRREQ = (WSABASEERR + 39);
{$EXTERNALSYM WSAEMSGSIZE}
  WSAEMSGSIZE = (WSABASEERR + 40);
{$EXTERNALSYM WSAEPROTOTYPE}
  WSAEPROTOTYPE = (WSABASEERR + 41);
{$EXTERNALSYM WSAENOPROTOOPT}
  WSAENOPROTOOPT = (WSABASEERR + 42);
{$EXTERNALSYM WSAEPROTONOSUPPORT}
  WSAEPROTONOSUPPORT = (WSABASEERR + 43);
{$EXTERNALSYM WSAESOCKTNOSUPPORT}
  WSAESOCKTNOSUPPORT = (WSABASEERR + 44);
{$EXTERNALSYM WSAEOPNOTSUPP}
  WSAEOPNOTSUPP = (WSABASEERR + 45);
{$EXTERNALSYM WSAEPFNOSUPPORT}
  WSAEPFNOSUPPORT = (WSABASEERR + 46);
{$EXTERNALSYM WSAEAFNOSUPPORT}
  WSAEAFNOSUPPORT = (WSABASEERR + 47);
{$EXTERNALSYM WSAEADDRINUSE}
  WSAEADDRINUSE = (WSABASEERR + 48);
{$EXTERNALSYM WSAEADDRNOTAVAIL}
  WSAEADDRNOTAVAIL = (WSABASEERR + 49);
{$EXTERNALSYM WSAENETDOWN}
  WSAENETDOWN = (WSABASEERR + 50);
{$EXTERNALSYM WSAENETUNREACH}
  WSAENETUNREACH = (WSABASEERR + 51);
{$EXTERNALSYM WSAENETRESET}
  WSAENETRESET = (WSABASEERR + 52);
{$EXTERNALSYM WSAECONNABORTED}
  WSAECONNABORTED = (WSABASEERR + 53);
{$EXTERNALSYM WSAECONNRESET}
  WSAECONNRESET = (WSABASEERR + 54);
{$EXTERNALSYM WSAENOBUFS}
  WSAENOBUFS = (WSABASEERR + 55);
{$EXTERNALSYM WSAEISCONN}
  WSAEISCONN = (WSABASEERR + 56);
{$EXTERNALSYM WSAENOTCONN}
  WSAENOTCONN = (WSABASEERR + 57);
{$EXTERNALSYM WSAESHUTDOWN}
  WSAESHUTDOWN = (WSABASEERR + 58);
{$EXTERNALSYM WSAETOOMANYREFS}
  WSAETOOMANYREFS = (WSABASEERR + 59);
{$EXTERNALSYM WSAETIMEDOUT}
  WSAETIMEDOUT = (WSABASEERR + 60);
{$EXTERNALSYM WSAECONNREFUSED}
  WSAECONNREFUSED = (WSABASEERR + 61);
{$EXTERNALSYM WSAELOOP}
  WSAELOOP = (WSABASEERR + 62);
{$EXTERNALSYM WSAENAMETOOLONG}
  WSAENAMETOOLONG = (WSABASEERR + 63);
{$EXTERNALSYM WSAEHOSTDOWN}
  WSAEHOSTDOWN = (WSABASEERR + 64);
{$EXTERNALSYM WSAEHOSTUNREACH}
  WSAEHOSTUNREACH = (WSABASEERR + 65);
{$EXTERNALSYM WSAENOTEMPTY}
  WSAENOTEMPTY = (WSABASEERR + 66);
{$EXTERNALSYM WSAEPROCLIM}
  WSAEPROCLIM = (WSABASEERR + 67);
{$EXTERNALSYM WSAEUSERS}
  WSAEUSERS = (WSABASEERR + 68);
{$EXTERNALSYM WSAEDQUOT}
  WSAEDQUOT = (WSABASEERR + 69);
{$EXTERNALSYM WSAESTALE}
  WSAESTALE = (WSABASEERR + 70);
{$EXTERNALSYM WSAEREMOTE}
  WSAEREMOTE = (WSABASEERR + 71);

{$EXTERNALSYM WSAEDISCON}
  WSAEDISCON = (WSABASEERR + 101);

  { Extended Windows Sockets error constant definitions }

{$EXTERNALSYM WSASYSNOTREADY}
  WSASYSNOTREADY = (WSABASEERR + 91);
{$EXTERNALSYM WSAVERNOTSUPPORTED}
  WSAVERNOTSUPPORTED = (WSABASEERR + 92);
{$EXTERNALSYM WSANOTINITIALISED}
  WSANOTINITIALISED = (WSABASEERR + 93);

  { Error return codes from gethostbyname() and gethostbyaddr()
    (when using the resolver). Note that these errors are
    retrieved via WSAGetLastError() and must therefore follow
    the rules for avoiding clashes with error numbers from
    specific implementations or language run-time systems.
    For this reason the codes are based at WSABASEERR+1001.
    Note also that [WSA]NO_ADDRESS is defined only for
    compatibility purposes. }

  { Authoritative Answer: Host not found }

{$EXTERNALSYM WSAHOST_NOT_FOUND}
  WSAHOST_NOT_FOUND = (WSABASEERR + 1001);
{$EXTERNALSYM HOST_NOT_FOUND}
  HOST_NOT_FOUND = WSAHOST_NOT_FOUND;

  { Non-Authoritative: Host not found, or SERVERFAIL }

{$EXTERNALSYM WSATRY_AGAIN}
  WSATRY_AGAIN = (WSABASEERR + 1002);
{$EXTERNALSYM TRY_AGAIN}
  TRY_AGAIN = WSATRY_AGAIN;

  { Non recoverable errors, FORMERR, REFUSED, NOTIMP }

{$EXTERNALSYM WSANO_RECOVERY}
  WSANO_RECOVERY = (WSABASEERR + 1003);
{$EXTERNALSYM NO_RECOVERY}
  NO_RECOVERY = WSANO_RECOVERY;

  { Valid name, no data record of requested type }

{$EXTERNALSYM WSANO_DATA}
  WSANO_DATA = (WSABASEERR + 1004);
{$EXTERNALSYM NO_DATA}
  NO_DATA = WSANO_DATA;

  { no address, look for MX record }

{$EXTERNALSYM WSANO_ADDRESS}
  WSANO_ADDRESS = WSANO_DATA;
{$EXTERNALSYM NO_ADDRESS}
  NO_ADDRESS = WSANO_ADDRESS;

  { Windows Sockets errors redefined as regular Berkeley error constants.
    These are commented out in Windows NT to avoid conflicts with errno.h.
    Use the WSA constants instead. }

{$EXTERNALSYM EWOULDBLOCK}
  EWOULDBLOCK = WSAEWOULDBLOCK;
{$EXTERNALSYM EINPROGRESS}
  EINPROGRESS = WSAEINPROGRESS;
{$EXTERNALSYM EALREADY}
  EALREADY = WSAEALREADY;
{$EXTERNALSYM ENOTSOCK}
  ENOTSOCK = WSAENOTSOCK;
{$EXTERNALSYM EDESTADDRREQ}
  EDESTADDRREQ = WSAEDESTADDRREQ;
{$EXTERNALSYM EMSGSIZE}
  EMSGSIZE = WSAEMSGSIZE;
{$EXTERNALSYM EPROTOTYPE}
  EPROTOTYPE = WSAEPROTOTYPE;
{$EXTERNALSYM ENOPROTOOPT}
  ENOPROTOOPT = WSAENOPROTOOPT;
{$EXTERNALSYM EPROTONOSUPPORT}
  EPROTONOSUPPORT = WSAEPROTONOSUPPORT;
{$EXTERNALSYM ESOCKTNOSUPPORT}
  ESOCKTNOSUPPORT = WSAESOCKTNOSUPPORT;
{$EXTERNALSYM EOPNOTSUPP}
  EOPNOTSUPP = WSAEOPNOTSUPP;
{$EXTERNALSYM EPFNOSUPPORT}
  EPFNOSUPPORT = WSAEPFNOSUPPORT;
{$EXTERNALSYM EAFNOSUPPORT}
  EAFNOSUPPORT = WSAEAFNOSUPPORT;
{$EXTERNALSYM EADDRINUSE}
  EADDRINUSE = WSAEADDRINUSE;
{$EXTERNALSYM EADDRNOTAVAIL}
  EADDRNOTAVAIL = WSAEADDRNOTAVAIL;
{$EXTERNALSYM ENETDOWN}
  ENETDOWN = WSAENETDOWN;
{$EXTERNALSYM ENETUNREACH}
  ENETUNREACH = WSAENETUNREACH;
{$EXTERNALSYM ENETRESET}
  ENETRESET = WSAENETRESET;
{$EXTERNALSYM ECONNABORTED}
  ECONNABORTED = WSAECONNABORTED;
{$EXTERNALSYM ECONNRESET}
  ECONNRESET = WSAECONNRESET;
{$EXTERNALSYM ENOBUFS}
  ENOBUFS = WSAENOBUFS;
{$EXTERNALSYM EISCONN}
  EISCONN = WSAEISCONN;
{$EXTERNALSYM ENOTCONN}
  ENOTCONN = WSAENOTCONN;
{$EXTERNALSYM ESHUTDOWN}
  ESHUTDOWN = WSAESHUTDOWN;
{$EXTERNALSYM ETOOMANYREFS}
  ETOOMANYREFS = WSAETOOMANYREFS;
{$EXTERNALSYM ETIMEDOUT}
  ETIMEDOUT = WSAETIMEDOUT;
{$EXTERNALSYM ECONNREFUSED}
  ECONNREFUSED = WSAECONNREFUSED;
{$EXTERNALSYM ELOOP}
  ELOOP = WSAELOOP;
{$EXTERNALSYM ENAMETOOLONG}
  ENAMETOOLONG = WSAENAMETOOLONG;
{$EXTERNALSYM EHOSTDOWN}
  EHOSTDOWN = WSAEHOSTDOWN;
{$EXTERNALSYM EHOSTUNREACH}
  EHOSTUNREACH = WSAEHOSTUNREACH;
{$EXTERNALSYM ENOTEMPTY}
  ENOTEMPTY = WSAENOTEMPTY;
{$EXTERNALSYM EPROCLIM}
  EPROCLIM = WSAEPROCLIM;
{$EXTERNALSYM EUSERS}
  EUSERS = WSAEUSERS;
{$EXTERNALSYM EDQUOT}
  EDQUOT = WSAEDQUOT;
{$EXTERNALSYM ESTALE}
  ESTALE = WSAESTALE;
{$EXTERNALSYM EREMOTE}
  EREMOTE = WSAEREMOTE;

  (* -----------------------------------------------------------------------

  const
    {$EXTERNALSYM FD_SETSIZE}
    FD_SETSIZE     =   64;

  type
    {$EXTERNALSYM u_char}
    u_char = Char;
    {$EXTERNALSYM u_short}
    u_short = Word;
    {$EXTERNALSYM u_int}
    u_int = Integer;
    {$EXTERNALSYM u_long}
    u_long = Longint;

    {$NODEFINE PFDSet}
    PFDSet = ^TFDSet;
    {$NODEFINE TFDSet}
    TFDSet = record
      fd_count: u_int;
      fd_array: array[0..FD_SETSIZE-1] of TSocket;
    end;

  const
  { Maximum queue length specifiable by listen. }

    {$EXTERNALSYM SOMAXCONN}
    SOMAXCONN       = 5;

      { Protocol Types }
   IPPROTO_IP     =   0;             { dummy for IP }
   IPPROTO_ICMP   =   1;             { control message protocol }
   IPPROTO_IGMP   =   2;             { group management protocol }
   IPPROTO_GGP    =   3;             { gateway^2 (deprecated) }
   IPPROTO_TCP    =   6;             { tcp }
   IPPROTO_PUP    =  12;             { pup }
   IPPROTO_UDP    =  17;             { user datagram protocol }
   IPPROTO_IDP    =  22;             { xns idp }
   IPPROTO_ND     =  77;             { UNOFFICIAL net disk proto }
   IPPROTO_RAW    =  255;            { raw IP packet }
      IPPROTO_MAX    =  256;

      { Address Constants }
   INADDR_ANY: Cardinal 		= $00000000;
   INADDR_LOOPBACK: Cardinal	= $0100007F;
   INADDR_BROADCAST: Cardinal	= $FFFFFFFF;
   INADDR_NONE: Cardinal		= $FFFFFFFF;

   { Error codes }
   SOCKET_ERROR            	= -1;
   INVALID_SOCKET          	= SOCKET_ERROR;
   WSABASEERR              	= 10000;
   WSAEINTR                	= (WSABASEERR+4);
   WSAEBADF                	= (WSABASEERR+9);
   WSAEACCES               	= (WSABASEERR+13);
   WSAEFAULT               	= (WSABASEERR+14);
   WSAEINVAL               	= (WSABASEERR+22);
   WSAEMFILE               	= (WSABASEERR+24);
   WSAEWOULDBLOCK          	= (WSABASEERR+35);
   WSAEINPROGRESS          	= (WSABASEERR+36);
   WSAEALREADY             	= (WSABASEERR+37);
   WSAENOTSOCK             	= (WSABASEERR+38);
   WSAEDESTADDRREQ         	= (WSABASEERR+39);
   WSAEMSGSIZE             	= (WSABASEERR+40);
   WSAEPROTOTYPE           	= (WSABASEERR+41);
   WSAENOPROTOOPT          	= (WSABASEERR+42);
   WSAEPROTONOSUPPORT      	= (WSABASEERR+43);
   WSAESOCKTNOSUPPORT      	= (WSABASEERR+44);
   WSAEOPNOTSUPP           	= (WSABASEERR+45);
   WSAEPFNOSUPPORT         	= (WSABASEERR+46);
   WSAEAFNOSUPPORT         	= (WSABASEERR+47);
   WSAEADDRINUSE           	= (WSABASEERR+48);
   WSAEADDRNOTAVAIL        	= (WSABASEERR+49);
   WSAENETDOWN             	= (WSABASEERR+50);
   WSAENETUNREACH          	= (WSABASEERR+51);
   WSAENETRESET            	= (WSABASEERR+52);
   WSAECONNABORTED    	    	= (WSABASEERR+53);
   WSAECONNRESET       	    = (WSABASEERR+54);
   WSAENOBUFS   		        = (WSABASEERR+55);
   WSAEISCONN          	    = (WSABASEERR+56);
   WSAENOTCONN             	= (WSABASEERR+57);
   WSAESHUTDOWN   		        = (WSABASEERR+58);
   WSAETOOMANYREFS     	    = (WSABASEERR+59);
   WSAETIMEDOUT            	= (WSABASEERR+60);
   WSAECONNREFUSED         	= (WSABASEERR+61);
   WSAELOOP                	= (WSABASEERR+62);
   WSAENAMETOOLONG         	= (WSABASEERR+63);
   WSAEHOSTDOWN            	= (WSABASEERR+64);
   WSAEHOSTUNREACH         	= (WSABASEERR+65);
   WSAENOTEMPTY            	= (WSABASEERR+66);
   WSAEPROCLIM             	= (WSABASEERR+67);
   WSAEUSERS               	= (WSABASEERR+68);
   WSAEDQUOT               	= (WSABASEERR+69);
   WSAESTALE               	= (WSABASEERR+70);
   WSAEREMOTE              	= (WSABASEERR+71);
   WSAEDISCON              	= (WSABASEERR+101);
   WSASYSNOTREADY          	= (WSABASEERR+91);
   WSAVERNOTSUPPORTED      	= (WSABASEERR+92);
   WSANOTINITIALISED       	= (WSABASEERR+93);

   { Various }
   IOCPARM_MASK = $7f;
   IOC_VOID     = $20000000;
   IOC_OUT      = $40000000;
   IOC_IN       = $80000000;
   IOC_INOUT    = (IOC_IN or IOC_OUT);
   FIONREAD     = IOC_OUT or { get # bytes to read }
     ((Integer(SizeOf(Integer)) and IOCPARM_MASK) shl 16) or
     (Integer(Byte('f')) shl 8) or 127;
   FIONBIO      = IOC_IN or { set/clear non-blocking i/o }
     ((Integer(SizeOf(Integer)) and IOCPARM_MASK) shl 16) or
     (Integer(Byte('f')) shl 8) or 126;
   FIOASYNC     = IOC_IN or { set/clear async i/o }
     ((Integer(SizeOf(Integer)) and IOCPARM_MASK) shl 16) or
     (Integer(Byte('f')) shl 8) or 125;
   MSG_PEEK        = $2;             { peek at incoming message }

   { Socket Options }
   SOL_SOCKET  	= 1;
   SO_DEBUG    	= 1;
   SO_REUSEADDR	= 2;
   SO_TYPE     	= 3;
   SO_ERROR    	= 4;
   SO_DONTROUTE	= 5;
   SO_BROADCAST	= 6;
   SO_SNDBUF   	= 7;
   SO_RCVBUF   	= 8;
   SO_KEEPALIVE	= 9;
   SO_OOBINLINE	= 10;
   SO_NO_CHECK 	= 11;
   SO_PRIORITY 	= 12;
   SO_LINGER   	= 13;
   SO_BSDCOMPAT	= 14;
   SO_REUSEPORT	= 15;
   SO_PASSCRED		= 16;
   SO_PEERCRED		= 17;
   SO_RCVLOWAT		= 18;
   SO_SNDLOWAT		= 19;
   SO_RCVTIMEO		= 20;
   SO_SNDTIMEO		= 21;
   SO_CONNTIMEOUT  = 200;

   { Socket Types }
   SOCK_STREAM     = 1;               { stream (connection) socket   }
   SOCK_DGRAM      = 2;               { datagram (conn.less) socket  }
   SOCK_RAW        = 3;               { raw socket                   }
   SOCK_RDM        = 4;               { reliably-delivered message   }
   SOCK_SEQPACKET  = 5;               { sequential packet socket     }

  { Address families. }

    {$EXTERNALSYM AF_UNSPEC}
    AF_UNSPEC       = 0;               { unspecified }
    {$EXTERNALSYM AF_UNIX}
    AF_UNIX         = 1;               { local to host (pipes, portals) }
    {$EXTERNALSYM AF_INET}
    AF_INET         = 2;               { internetwork: UDP, TCP, etc. }
    {$EXTERNALSYM AF_IMPLINK}
    AF_IMPLINK      = 3;               { arpanet imp addresses }
    {$EXTERNALSYM AF_PUP}
    AF_PUP          = 4;               { pup protocols: e.g. BSP }
    {$EXTERNALSYM AF_CHAOS}
    AF_CHAOS        = 5;               { mit CHAOS protocols }
    {$EXTERNALSYM AF_IPX}
    AF_IPX          = 6;               { IPX and SPX }
    {$EXTERNALSYM AF_NS}
    AF_NS           = 6;               { XEROX NS protocols }
    {$EXTERNALSYM AF_ISO}
    AF_ISO          = 7;               { ISO protocols }
    {$EXTERNALSYM AF_OSI}
    AF_OSI          = AF_ISO;          { OSI is ISO }
    {$EXTERNALSYM AF_ECMA}
    AF_ECMA         = 8;               { european computer manufacturers }
    {$EXTERNALSYM AF_DATAKIT}
    AF_DATAKIT      = 9;               { datakit protocols }
    {$EXTERNALSYM AF_CCITT}
    AF_CCITT        = 10;              { CCITT protocols, X.25 etc }
    {$EXTERNALSYM AF_SNA}
    AF_SNA          = 11;              { IBM SNA }
    {$EXTERNALSYM AF_DECnet}
    AF_DECnet       = 12;              { DECnet }
    {$EXTERNALSYM AF_DLI}
    AF_DLI          = 13;              { Direct data link interface }
    {$EXTERNALSYM AF_LAT}
    AF_LAT          = 14;              { LAT }
    {$EXTERNALSYM AF_HYLINK}
    AF_HYLINK       = 15;              { NSC Hyperchannel }
    {$EXTERNALSYM AF_APPLETALK}
    AF_APPLETALK    = 16;              { AppleTalk }
    {$EXTERNALSYM AF_NETBIOS}
    AF_NETBIOS      = 17;              { NetBios-style addresses }
    {$EXTERNALSYM AF_VOICEVIEW}
    AF_VOICEVIEW    = 18;              { VoiceView }
    {$EXTERNALSYM AF_FIREFOX}
    AF_FIREFOX      = 19;              { FireFox }
    {$EXTERNALSYM AF_UNKNOWN1}
    AF_UNKNOWN1     = 20;              { Somebody is using this! }
    {$EXTERNALSYM AF_BAN}
    AF_BAN          = 21;              { Banyan }

    {$EXTERNALSYM AF_MAX}
    AF_MAX          = 22;

  { Protocol families, same as address families for now. }

    {$EXTERNALSYM PF_UNSPEC}
    PF_UNSPEC       = AF_UNSPEC;
    {$EXTERNALSYM PF_UNIX}
    PF_UNIX         = AF_UNIX;
    {$EXTERNALSYM PF_INET}
    PF_INET         = AF_INET;
    {$EXTERNALSYM PF_IMPLINK}
    PF_IMPLINK      = AF_IMPLINK;
    {$EXTERNALSYM PF_PUP}
    PF_PUP          = AF_PUP;
    {$EXTERNALSYM PF_CHAOS}
    PF_CHAOS        = AF_CHAOS;
    {$EXTERNALSYM PF_NS}
    PF_NS           = AF_NS;
    {$EXTERNALSYM PF_IPX}
    PF_IPX          = AF_IPX;
    {$EXTERNALSYM PF_ISO}
    PF_ISO          = AF_ISO;
    {$EXTERNALSYM PF_OSI}
    PF_OSI          = AF_OSI;
    {$EXTERNALSYM PF_ECMA}
    PF_ECMA         = AF_ECMA;
    {$EXTERNALSYM PF_DATAKIT}
    PF_DATAKIT      = AF_DATAKIT;
    {$EXTERNALSYM PF_CCITT}
    PF_CCITT        = AF_CCITT;
    {$EXTERNALSYM PF_SNA}
    PF_SNA          = AF_SNA;
    {$EXTERNALSYM PF_DECnet}
    PF_DECnet       = AF_DECnet;
    {$EXTERNALSYM PF_DLI}
    PF_DLI          = AF_DLI;
    {$EXTERNALSYM PF_LAT}
    PF_LAT          = AF_LAT;
    {$EXTERNALSYM PF_HYLINK}
    PF_HYLINK       = AF_HYLINK;
    {$EXTERNALSYM PF_APPLETALK}
    PF_APPLETALK    = AF_APPLETALK;
    {$EXTERNALSYM PF_VOICEVIEW}
    PF_VOICEVIEW    = AF_VOICEVIEW;
    {$EXTERNALSYM PF_FIREFOX}
    PF_FIREFOX      = AF_FIREFOX;
    {$EXTERNALSYM PF_UNKNOWN1}
    PF_UNKNOWN1     = AF_UNKNOWN1;
    {$EXTERNALSYM PF_BAN}
    PF_BAN          = AF_BAN;

    {$EXTERNALSYM PF_MAX}
    PF_MAX          = AF_MAX;

   { Security levels - as per NRL IPv6 - don't actually do anything }
   SO_SECURITY_AUTHENTICATION       = 22;
   SO_SECURITY_ENCRYPTION_TRANSPORT = 23;
   SO_SECURITY_ENCRYPTION_NETWORK   = 24;
   SO_BINDTODEVICE					 = 25;

   { Socket filtering }
   SO_ATTACH_FILTER=  26;
   SO_DETACH_FILTER=  27;

   { Shutdown flags }
   SHUT_RD         = 0;             { shut down the reading side }
   SHUT_WR         = 1;             { shut down the writing side }
   SHUT_RDWR       = 2;             { shut down both sides }

  Type
   {$EXTERNALSYM SunB}
   SunB = packed record
    s_b1, s_b2, s_b3, s_b4: BYTE;
   end;

   {$EXTERNALSYM SunW}
   SunW = packed record
    s_w1, s_w2: WORD;
   end;

   PInAddr = ^TInAddr;
   {$EXTERNALSYM in_addr}
   in_addr = record
    case integer of
     0: (S_un_b: SunB);
     1: (S_un_w: SunW);
     2: (S_addr: Cardinal);
   end;
   TInAddr = in_addr;

   PSockAddrIn = ^TSockAddrIn;
   {$EXTERNALSYM sockaddr_in}
   sockaddr_in = record
    case Integer of
     0: (sin_family: WORD;
      sin_port: WORD;
      sin_addr: TInAddr;
      sin_zero: array[0..7] of Char);
     1: (sa_family: WORD;
      sa_data: array[0..13] of Char)
   end;
   TSockAddrIn = sockaddr_in;

   { Structure used by kernel to store most addresses. }

   {$EXTERNALSYM PSOCKADDR}
   PSockAddr = ^TSockAddr;
   {$EXTERNALSYM TSockAddr}
   TSockAddr = sockaddr_in;
   {$EXTERNALSYM TSocket}
   TSocket = Integer;

  ----------------------------------------------------------------------- *)

type

  { This structure is used by "protocol developers" to add new protocols to }
  { this stack.                                                             }
  TProtocolStruct = record
    Domain: Integer;
    SocketType: Integer;
    Protocol: Integer;
    FNSocket: function(Domain, SocketType, Protocol: Integer): Pointer;
    FNClose: function(Sock: Pointer): Integer;
    FNGetLastError: function: Integer;
    FNSend: function(Sock: Pointer; var Addr; AddrLen, Flags: Integer): Integer;
    FNSendTo: function(Sock: Pointer; var Data; DataLen, Flags: Integer; var
      Addr; AddrLen: Integer): Integer;
    FNRecvFrom: function(Sock: Pointer; var Addr; AddrLen, Flags: Integer; var
      AddrFrom; var FromLen: Integer): Integer;
    FNRecv: function(Sock: Pointer; var Addr; AddrLen, Flags: Integer): Integer;
    FNBind: function(Sock: Pointer; var Addr; AddrLen: Integer): Integer;
    FNListen: function(Sock: Pointer; MaxConnect: Integer): Integer;
    FNAccept: function(Sock: Pointer; var Addr; var Addrlen: Integer): Pointer;
    FNConnect: function(Sock: Pointer; var Addr; Addrlen: Integer): Integer;
    FNShutdown: function(Sock: Pointer; How: Integer): Integer;
    FNGetSocketName: function(Sock: Pointer; var Addr; var Addrlen: Integer):
      Integer;
    FNGetPeerName: function(Sock: Pointer; var Addr; var Addrlen: Integer):
      Integer;
    FNSetSockOpt: function(Sock: Pointer; Level, OptName: Integer; var OptVal;
      OptLen: Integer): Integer;
    FNGetSockOpt: function(Sock: Pointer; Level, OptName: Integer; var OptVal;
      var OptLen: Integer): Integer;
    FNioctlsocket: function(Sock: Pointer; Option: Integer; d: Pointer):
      Integer;
    FNSelect: function(Sock: Pointer; var ReadReady, WriteReady,
      ExceptFlag: Boolean): Integer;
  end;

  { Basic Socket Functions (BCD compliant) }
function htons(w: Word): Word;
function ntohs(w: Word): Word;
function Socket(Domain, SocketType, Protocol: Integer): Integer;
function Send(Sock: Integer; var Data; DataLen, Flags: Integer): Integer;
function Recv(Sock: Integer; var Data; DataLen, Flags: Integer): Integer;
function Bind(Sock: Integer; var Addr; AddrLen: Integer): Integer;
function Listen(Sock, MaxConnect: Integer): Integer;
function Connect(Sock: Integer; var Addr; Addrlen: Integer): Integer;
function Accept(Sock: Integer; var Addr; var Addrlen: Integer): Integer;
function Shutdown(Sock: Integer; How: Integer): Integer;
function GetSocketName(Sock: Integer; var Addr; var Addrlen: Integer): Integer;
function GetPeerName(Sock: Integer; var Addr; var Addrlen: Integer): Integer;
function SetSockOpt(Sock, Level, OptName: Integer; var OptVal; OptLen: Integer):
  Integer;
function GetSockOpt(Sock, Level, OptName: Integer; var OptVal; var OptLen:
  Integer): Integer;

{ Functions specific to DOS-Sockets implementation }
function CloseSocket(Sock: Integer): Integer;
function SendTo(Sock: Integer; var Data; DataLen, Flags: Integer; var Addr;
  AddrLen: Integer): Integer;
function RecvFrom(Sock: Integer; var Addr; AddrLen, Flags: Integer; var
  AddrFrom; var FromLen: Integer): Integer;
function ioctlsocket(Sock: Integer; Option: Integer; d: Pointer): Integer;
function WSAGetLastError: Integer;

{ functions for compatibility with Delphi Sockets.pas unit }
function select(nfds: Integer; readfds, writefds, exceptfds: PFDSet; timeout: 
    PTimeVal): LongInt;
function inet_addr(cp: PChar): u_long;
function gethostbyaddr(addr: Pointer; len, Struct: Integer): PHostEnt;
function gethostbyname(name: PChar): PHostEnt;
function getservbyname(name, proto: PChar): PServEnt;
function getprotobyname(name: PChar): PProtoEnt;
function inet_ntoa(inaddr: TInAddr): PChar;
function gethostname(name: PChar; len: Integer): Integer;

{ Internal functions, used to manage protocols }
function AddProtocol(ProtocolStruct: TProtocolStruct): Boolean;

{******************************************************************************}

{$EXTERNALSYM FD_CLR}
procedure FD_CLR(Socket: TSocket; var FDSet: TFDSet);
{$EXTERNALSYM FD_ISSET}
function FD_ISSET(Socket: TSocket; var FDSet: TFDSet): Boolean;
{$EXTERNALSYM FD_SET}
procedure FD_SET(Socket: TSocket; var FDSet: TFDSet);
  // renamed due to conflict with fd_set (above)
{$EXTERNALSYM FD_ZERO}
procedure FD_ZERO(var FDSet: TFDSet);

procedure UpdateNetwork;

var
  NetworkAdapter: TNetworkAdapter;

{******************************************************************************}

implementation

uses
  WDosIP, WDosResolvers, WDosPktDrv, WDosDNS;

type
  PSockStruct = ^TSockStruct;
  TSockStruct = record
    Ptr: Pointer;
    ProtocolIndex: Integer;
  end;

var
  _LastError: Integer;
  Protocols: array[1..100] of TProtocolStruct;

  HostEntry: THostEnt;
  HostAddr: TInAddr;
  PHostAddr: PChar;
  ServiceEntry: TServEnt;
  NetAddr: array[0..15] of Char;

  {******************************************************************************
          Basic Socket Functions
  ******************************************************************************}

function htons(w: Word): Word;
begin
  Result := Swap(w);
end;

function ntohs(w: Word): Word;
begin
  Result := Swap(w);
end;

function socket(Domain, SocketType, Protocol: Integer): Integer;
var
  Ptr: Pointer;
  i, PIndex: Integer;
  PSock: PSockStruct;
begin
  Yield;
  Socket := INVALID_SOCKET;
  { Decide what protocol to use based on Domain, Socket Type and }
  { Protocol.                                                    }
  PIndex := -1;
  for i := Low(Protocols) to High(Protocols) do
  begin
    if (Protocols[i].Domain = Domain)
      and (Protocols[i].SocketType = SocketType)
      and (Protocols[i].Protocol = Protocol) then
    begin
      PIndex := i;
      Break;
    end;
  end;
  { If there's no protocol registered to satisfy the user's request, }
  { we exit with WSAEPROTONOSUPPORT.                                 }
  if PIndex = -1 then
  begin
    _LastError := WSAEPROTONOSUPPORT;
    Exit;
  end;
  { Request the selected protocol to create a new socket             }
  Ptr := Protocols[PIndex].FNSocket(Domain, SocketType, Protocol);
  if (Ptr = nil) then
  begin
    _LastError := Protocols[PIndex].FNGetLastError();
    Exit;
  end;
  { We've found a protocol to meet the user's request. Now it's time }
  { to allocate a handle for the new socket                          }
  New(PSock);
  PSock^.Ptr := Ptr;
  PSock^.ProtocolIndex := PIndex;
  Socket := Integer(PSock);
end;

function CloseSocket(Sock: Integer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    CloseSocket := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    CloseSocket := Protocols[PSock^.ProtocolIndex].FNClose(PSock^.Ptr);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    Dispose(PSock);
  end;
end;

function Send(Sock: Integer; var Data; DataLen, Flags: Integer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    Send := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    Send := Protocols[PSock^.ProtocolIndex].FNSend(PSock^.Ptr, Data, DataLen,
      Flags);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function SendTo(Sock: Integer; var Data; DataLen, Flags: Integer; var Addr;
  AddrLen: Integer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    Result := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    Result := Protocols[PSock^.ProtocolIndex].FNSendTo(PSock^.Ptr,
      Addr, AddrLen, Flags, Addr, AddrLen);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function RecvFrom(Sock: Integer; var Addr; AddrLen, Flags: Integer; var
  AddrFrom; var FromLen: Integer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    RecvFrom := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    RecvFrom := Protocols[PSock^.ProtocolIndex].FNRecvFrom(PSock^.Ptr,
      Addr, AddrLen, Flags, AddrFrom, FromLen);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function Recv(Sock: Integer; var Data; DataLen, Flags: Integer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    Recv := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    Recv := Protocols[PSock^.ProtocolIndex].FNRecv(PSock^.Ptr,
      Data, DataLen, Flags);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function Bind(Sock: Integer; var Addr; AddrLen: Integer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    Bind := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    Bind := Protocols[PSock^.ProtocolIndex].FNBind(PSock^.Ptr, Addr, AddrLen);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function Listen(Sock, MaxConnect: Integer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    Listen := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    Listen := Protocols[PSock^.ProtocolIndex].FNListen(PSock^.Ptr, MaxConnect);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function Accept(Sock: Integer; var Addr; var Addrlen: Integer): Integer;
var
  Ptr: Pointer;
  PSock, NewSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    Accept := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
    Exit;
  end;
  PSock := PSockStruct(Sock);
  Ptr := Protocols[PSock^.ProtocolIndex].FNAccept(PSock^.Ptr, Addr, AddrLen);
  if Ptr = nil then
  begin
    Accept := SOCKET_ERROR;
  end
  else
  begin
    New(NewSock);
    NewSock^.Ptr := Ptr;
    NewSock^.ProtocolIndex := PSock^.ProtocolIndex;
    Accept := Integer(NewSock);
  end;
  _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
end;

function Connect(Sock: Integer; var Addr; Addrlen: Integer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    Connect := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    Connect := Protocols[PSock^.ProtocolIndex].FNConnect(PSock^.Ptr, Addr,
      AddrLen);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function Shutdown(Sock: Integer; How: Integer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    Shutdown := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    Shutdown := Protocols[PSock^.ProtocolIndex].FNShutdown(PSock^.Ptr, How);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();

    // #iwa 2003-09-27: drop WSAENOTCONN "The socket is not connected" error
    if _LastError = WSAENOTCONN	then
    begin
      Shutdown := 0;
      _LastError := 0;
    end;
  end;
end;

function GetSocketName(Sock: Integer; var Addr; var Addrlen: Integer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    Result := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    GetSocketName := Protocols[PSock^.ProtocolIndex].FNGetSocketName(PSock^.Ptr,
      Addr, AddrLen);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function GetPeerName(Sock: Integer; var Addr; var Addrlen: Integer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    GetPeerName := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    GetPeerName := Protocols[PSock^.ProtocolIndex].FNGetPeerName(PSock^.Ptr,
      Addr, AddrLen);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function SetSockOpt(Sock, Level, OptName: Integer; var OptVal; OptLen: Integer):
  Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    Result := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    Result := Protocols[PSock^.ProtocolIndex].FNSetSockOpt(PSock^.Ptr,
      Level, OptName, OptVal, OptLen);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function GetSockOpt(Sock, Level, OptName: Integer; var OptVal; var OptLen:
  Integer): Integer;
var
  PSock: PSockStruct;
begin
  if (Sock <= 0) then
  begin
    Result := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    Result := Protocols[PSock^.ProtocolIndex].FNGetsockOpt(PSock^.Ptr,
      Level, OptName, OptVal, OptLen);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function ioctlsocket(Sock: Integer; Option: Integer; d: Pointer): Integer;
var
  PSock: PSockStruct;
begin
  Yield;
  if (Sock <= 0) then
  begin
    Result := SOCKET_ERROR;
    _LastError := WSAENOTSOCK;
  end
  else
  begin
    PSock := PSockStruct(Sock);
    ioctlsocket := Protocols[PSock^.ProtocolIndex].FNioctlsocket(PSock^.Ptr,
      Option, d);
    _LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
  end;
end;

function WSAGetLastError: Integer;
begin
  Result := _LastError;
end;

function AddProtocol(ProtocolStruct: TProtocolStruct): Boolean;
var
  i: Integer;
begin
  for i := Low(Protocols) to High(Protocols) do
  begin
    if Protocols[i].Domain < 0 then
    begin
      Protocols[i] := ProtocolStruct;
      Result := True;
      Exit;
    end;
  end;
  Result := False;
end;

function CheckReadSockets(readfds: PFDSet): Integer;
var
  I: Integer;
  Socket: TSocket;
  PSocket: PSockStruct;
  ReadReady, WriteReady, ExceptFlag: Boolean;
begin
  for I := readfds.fd_count - 1 downto 0 do
  begin
    Yield;
    Socket := readfds.fd_array[I];
(*
    if (Sock <= 0) then
    begin
      Result := SOCKET_ERROR;
      _LastError := WSAENOTSOCK;
      Exit;
    end;
*)
    PSocket := PSockStruct(Socket);

    ReadReady := False;
    WriteReady := False;
    ExceptFlag := False;

    Protocols[PSocket^.ProtocolIndex].FNSelect(PSocket^.Ptr,
        ReadReady, WriteReady, ExceptFlag);
    if not ReadReady then
      FD_CLR(Socket, readfds^);
  end;
  Result := readfds.fd_count;
end;

function CheckWriteSockets(writefds: PFDSet): Integer;
begin
  Result := 0;
end;

function CheckExceptSockets(exceptfds: PFDSet): Integer;
var
  I: Integer;
  Socket: TSocket;
  PSocket: PSockStruct;
  ReadReady, WriteReady, ExceptFlag: Boolean;
begin
  for I := exceptfds.fd_count - 1 downto 0 do
  begin
    Yield;
    Socket := exceptfds.fd_array[I];
    PSocket := PSockStruct(Socket);

    ReadReady := False;
    WriteReady := False;
    ExceptFlag := False;

    Protocols[PSocket^.ProtocolIndex].FNSelect(PSocket^.Ptr,
        ReadReady, WriteReady, ExceptFlag);
    if not ExceptFlag then
      FD_CLR(Socket, exceptfds^);
  end;
  Result := exceptfds.fd_count;
end;

function select(nfds: Integer; readfds, writefds, exceptfds: PFDSet; timeout:
    PTimeVal): LongInt;
var
  Handled: Integer;
begin
(*
Return Values
=============

select returns the total number of descriptors which are ready and contained
in the fd_set structures, zero if the time limit expired, or SOCKET_ERROR
if an error occurred. If the return value is SOCKET_ERROR, WSAGetLastError
can be used to retrieve a specific error code.


Error Codes
===========

WSANOTINITIALISED  A successful WSAStartup must occur before using this
                   function.
WSAEFAULT	   The Windows Sockets implementation was unable to allocated
                   needed resources for its internal operations, or the readfds,
                   writefds, exceptfds, or timeval parameters are not part of
                   the user address space.
WSAENETDOWN	   The network subsystem has failed.
WSAEINVAL	   The timeout value is not valid, or all three descriptor
                   parameters were NULL.
WSAEINTR	   The (blocking) call was canceled through
                   WSACancelBlockingCall.
WSAEINPROGRESS	   A blocking Windows Sockets 1.1 call is in progress, or the
                   service provider is still processing a callback function.
WSAENOTSOCK	   One of the descriptor sets contains an entry which is not
                   a socket.
*)

  if not Assigned(readfds) and not Assigned(writefds)
      and not Assigned(exceptfds) then
  begin
    Result := SOCKET_ERROR;
    _LastError := WSAEINVAL;
    Exit;
  end;

  Handled := 0;
  if Assigned(readfds) then
    Handled := CheckReadSockets(readfds);
  if Assigned(writefds) then
    Handled := Handled + CheckWriteSockets(writefds);
  if Assigned(exceptfds) then
    Handled := Handled + CheckExceptSockets(exceptfds);
  Result := Handled;
end;

function inet_addr(cp: PChar): u_long;
begin
  Result := IP_StringToAddr(cp);
end;

function gethostbyaddr(addr: Pointer; len, Struct: Integer): PHostEnt;
var
  S: string;
begin
  S := Resolver.GetHostByAddr(TInAddr(addr));
  HostEntry.h_name := PChar(S);
  Result := System.Addr(HostEntry);
end;

function gethostbyname(name: PChar): PHostEnt;
begin
  HostAddr := Resolver.GetHostByName(PChar(name));
  PHostAddr := Addr(HostAddr);
  HostEntry.h_addr := Addr(PHostAddr);
  Result := Addr(HostEntry);
end;

function getservbyname(name, proto: PChar): PServEnt;
var
  Port: Word;
begin
  Port := Resolver.GetServByName(name, proto);
  if Port <> 0 then
  begin
    ServiceEntry.s_port := Port;
    Result := Addr(ServiceEntry);
  end
  else Result := nil;
end;

function getprotobyname(name: PChar): PProtoEnt;
begin
  // TODO -cMM: default body inserted
  Result := nil;
end;

function inet_ntoa(inaddr: TInAddr): PChar;
var
  S: string;
begin
  S := IP_AddrToString(inaddr.S_addr);
  StrPCopy(NetAddr, S);
  Result := Addr(NetAddr);
end;

function gethostname(name: PChar; len: Integer): Integer;
begin
  Result := 0;
  StrPLCopy(name, Resolver.AppendDomain(Resolver.HostName), len);
end;

procedure UpdateNetwork;
begin
  NET_Work;
  Yield;
end;

{******************************************************************************}

procedure FD_CLR(Socket: TSocket; var FDSet: TFDSet);
var
  I: Integer;
begin
  I := 0;
  while I < FDSet.fd_count do
  begin
    if FDSet.fd_array[I] = Socket then
    begin
      while I < FDSet.fd_count - 1 do
      begin
        FDSet.fd_array[I] := FDSet.fd_array[I + 1];
        Inc(I);
      end;
      Dec(FDSet.fd_count);
      Break;
    end;
    Inc(I);
  end;
end;

function FD_ISSET(Socket: TSocket; var FDSet: TFDSet): Boolean;
var
  I: Integer;
begin
  // TODO: This function is not tested
  Result := False;
  for I := 0 to FDSet.fd_count - 1 do
  begin
    Result := (FDSet.fd_array[I] = Socket);
    if Result then Break;
  end;
end;

procedure FD_SET(Socket: TSocket; var FDSet: TFDSet);
begin
  if FDSet.fd_count < FD_SETSIZE then
  begin
    FDSet.fd_array[FDSet.fd_count] := Socket;
    Inc(FDSet.fd_count);
  end;
end;

procedure FD_ZERO(var FDSet: TFDSet);
begin
  FDSet.fd_count := 0;
end;

{******************************************************************************}

var
  IPHandle: Integer;
  IPAdapterConfig: TIPAdapterConfig;

procedure InitNetwork;
{$IFDEF DEBUG}
var
  F: TextFile;
  AE: TAddressEntry;
  RE: TRouteEntry;
  EthernetAddress: TEthernetAddress;
  I: Integer;
{$ENDIF DEBUG}
begin
  // Create an Adapter object representing out network adapter.
  NetworkAdapter := TPacketDriverAdapter.Create;
  // If the adapter works fine, we bind the TCP/IP protocol.
  if NetworkAdapter.Status <> asOk then
  begin
    WriteLn('Could not initialize network adapter.');
    Halt(2);
  end;
  IPAdapterConfig.DHCP := Resolver.Dhcp;
  IPAdapterConfig.IPAddress := Resolver.IpAddress;
  IPAdapterConfig.NetMask := Resolver.NetMask;
  IPAdapterConfig.Gateway := Resolver.Gateway;
  NET_SetHostName(Resolver.HostName);
  IPHandle := IP_Init(NetworkAdapter, IPAdapterConfig);
  if IPHandle = 0 then
  begin
    WriteLn('Could not initialize IP stack. Wrong Parameters.');
    Halt(3);
  end;
  SetDNSServer(Resolver.NameServer, Swap(WORD(53)));

{$IFDEF DEBUG}
  AssignFile(F, 'LOG.TXT');
  Rewrite(F);
  NetworkAdapter.GetAddress(EthernetAddress);
  IP_GetInfo(0, AE);
  WriteLn(F, 'IP Information');
  WriteLn(F, '--------------');
  WriteLn(F, 'Adapter   : ', NetworkAdapter.AdapterName, ' (',
    Ethernet_AddrToString(EthernetAddress), ')');
  WriteLn(F, 'IP Address: ', IP_AddrToString(AE.IPAddress));
  WriteLn(F, 'Net Mask  : ', IP_AddrToString(AE.NetMask));
  WriteLn(F, 'DHCP Used : ', AE.DHCPInUse);
  WriteLn(F);
  WriteLn(F, 'Routing Table');
  WriteLn(F, '-------------');
  WriteLn(F, 'Destination     Mask            Local IP        Gateway          Cost');
  WriteLn(F, '--------------- --------------- --------------- --------------- -----');
  I := 0;
  while IP_GetRoute(I, RE) do
  begin
    WriteLn(F,
      PadR(IP_AddrToString(RE.Destination), 16, ' ') +
      PadR(IP_AddrToString(RE.Mask), 16, ' ') +
      PadR(IP_AddrToString(RE.LocalIP), 16, ' ') +
      PadR(IP_AddrToString(RE.Gateway), 16, ' ') +
      IStr(RE.Cost, 5));
    inc(I);
  end;
  WriteLn(F);
  CloseFile(F);
{$ENDIF DEBUG}
end;

var
  I: Integer;

initialization
  _LastError := 0;
  for I := Low(Protocols) to High(Protocols) do
  begin
    Protocols[i].Domain := -1;
  end;
  WDosIPInitalization;
  InitNetwork;

finalization
  WDosIPFinalization;
  FreeAndNil(NetworkAdapter);

end.


