
{*******************************************************}
{                                                       }
{ DWPL Delphi WDosX and Cross Platform Library (rtl)    }
{ WDosSysUtils.pas                                      }
{                                                       }
{ for Delphi 3..6                                       }
{ Version 2.0                                           }
{ Copyright (c) 2002 by Team-DWP                        }
{ e-mail: team-dwp@dwp42.org                            }
{ www: http://www.dwp42.org                             }
{                                                       }
{ The initial developer of the code is                  }
{   Jason G Burgon, jay@jayman.demon.co.uk              }
{                                                       }
{ Portions created by Jason G Burgon are                }
{ Copyright (C) 1994-2002 Jason G Burgon.               }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache, immo.wache@dwp42.org                    }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ Licensees holding a valid Team-DWP No-Nonsense        }
{ License for this Software may use this file in        }
{ accordance with such license, which appears in        }
{ the file license.txt that came with this software.    }
{                                                       }
{ Licensees holding a valid Graphic Vision 32 License   }
{ may use this software in accordance with that         }
{ license, as sent to you by Jason G Burgon.            }
{ All other rights not expressely given up in this      }
{ statement are retained by the copyright holder.       }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Graphic Vision Unit                               }
{     Copyright (c) 1994-2001 Jason G Burgon            }
{     www: http://www.jayman.demon.co.uk                }
{     email: jay@jayman.demon.co.uk                     }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{-------------------------------------------------------}
{                                                       }
{ Compilation Notes:                                    }
{   Use 4-byte data alignment.                          }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosSysUtils.pas,v 1.7 2003/09/21 20:04:39 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosSysUtils.pas,v $
  Revision 1.7  2003/09/21 20:04:39  iwache
  Bug fixed in function IRQtoIntVec for IRQ >= 8.
  Bug fixed in procedure TPortW.SetPortW, wrong
  register was used.

  Revision 1.6  2003/05/13 19:43:51  iwache
  - TDosISR: constructor Create has a new parameter "Init: Boolean = True".
  If you dislike immediate Hook of the ISR/Callback set it to false and call
  new method InitHook to initiate.
  - TDosCallBack: constructor Create propagates the new parameter Init too.
  - Some typos fixed.

  Revision 1.5  2003/05/07 18:59:04  lagocs
  Header said WDosTypes instead of WDosSysUtils, fixed.

  Revision 1.4  2002/10/27 22:53:12  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{$A+,B-,E-,F-,G-,H-,I-,J-,K-,M+,N+,O+,P+,Q-,R-,S-,T+,U-,V+,W-,X+,Z1}
{$ifdef Debug}
{$C+}
{$endif Debug}

{$I WDOSX.INC}

{$IFNDEF COMPILER3_UP}
  $$ // This unit is not for use with this compiler
{$ENDIF}

unit WDosSysUtils;

interface

uses
  SysUtils, WDosTypes;

const

{ Flags register bit masks }

  fCarry           = $0001;
  fParity          = $0004;
  fAuxiliary       = $0010;
  fZero            = $0040;
  fSign            = $0080;
  fOverflow        = $0800;

const
{ Interrupt numbers }

  intDos           = $21;
  intDpmi          = $31;

{ DPMI function codes   }

  dpmiAllocSeg     = $0000;            { Allocate selector                  }
  dpmiFreeSeg      = $0001;            { Free selector                      }
  dpmiMapRealSeg   = $0002;            { Map real-mode segment to selector  }
  dpmiAllocSegs    = $0003;            { Allocate multiple selectors        }
  dpmiGetBaseAdr   = $0006;            { Get linear base address of segment }
  dpmiSetSegBase   = $0007;            { Set selector base address          }
  dpmiSetSegSize   = $0008;            { Set selector size                  }
  dpmiSetAccess    = $0009;            { Set selector access rights & type  }
  dpmiCloneSeg     = $000A;            { Create new selector with same props}
  dpmiGetDesc      = $000B;            { Copy selectors LDT into 8-byte buf }
  dpmiSetDesc      = $000C;            { Set selectors LDT from 8-byte buf  }

  dpmiAllocDosMem  = $0100;            { Allocate a block < 1MB address lim }
  dpmiFreeDosMem   = $0101;            { Free block of Dos memory           }

  dpmiGetRealInt   = $0200;            { Get real mode interrupt vector     }
  dpmiSetRealInt   = $0201;            { Set real mode interrupt vector     }
  dpmiGetExcpInt   = $0202;            { Get protected exception vector     }
  dmpiSetExcpInt   = $0203;            { Set protected exception vector     }
  dpmiGetProtInt   = $0204;            { Get protected mode interrupt vector}
  dpmiSetProtInt   = $0205;            { Set protected mode interrupt vector}

  dpmiCallRealInt  = $0300;            { Call real-mode interrupt           }
  dpmiCallRealFar  = $0301;            { Call far real-mode procedure       }
  dpmiCalliret     = $0302;            { Call real-mode with IRET frame     }
  dpmiAllocRMCB    = $0303;            { Allocate real mode call-back       }
  dpmiFreeRMCB     = $0304;            { Free real mode call-back           }

  dpmiGetInfo      = $0400;            { Get DPMI server information        }

  dpmiLockLinear   = $0600;            { Lock linear region of memory       }
  dpmiUnLockLinear = $0601;            { Unlock linear region of memory     }

  dpmiMapPhysical  = $0800;            { Map Physical memory block to linear}
  dpmiUnMapPhysical= $0801;            { Unmap physical memory block        }

type
  DosPtr    = LongWord;                { Seg:0fs 16:16 real mode far pointer}
  PDosPtr   = ^DosPtr;
  DosPtrRec = packed record
    Ofs: Word;
    Seg: Word;
  end;

{ Registers record used by Intr and MsDos, hooked Dos interrupts & RMCB's }

  PRealRegs = ^TRealRegs;
  TRealRegs = packed record
    case Integer of
      0: (EDI,ESI,EBP,EXX,EBX,EDX,ECX,EAX  : LongWord;
          Flags,ES,DS,FS,GS,IP,CS,SP,SS,YY : Word);
      1: (DI,DIH,SI,SIH,BP,BPH,XX,XXH      : Word;
          case Integer of
            0: (BX,BXH,DX,DXH,CX,CXH,AX,AXH: Word);
            1: (BL,BH,BLH,BHH,DL,DH,DLH,DHH,
                CL,CH,CLH,CHH,AL,AH,ALH,AHH: Byte));
  end;

{ Registers record used by IntrPM, MsDosPM, hooked p/m interrpts & callbacks}

  PRegisters = ^TRegisters;
  TRegisters = packed record
    case Integer of
      0: (EDI,ESI,EBP,EBX,
          EDX,ECX,EAX,EFlags  : LongWord;
          ES,DS,FS,GS         : Word;
          EIP: LongWord; CS   : Word;
          ESP: LongWord; SS   : Word);
      1: (DI,DIH,SI,SIH,BP,BPH: Word;
          case Integer of
            0: (BX,BXH,DX,DXH,CX,CXH,
                AX,AXH,Flags,FlagsH: Word);
            1: (BL,BH,BLH,BHH,DL,DH,DLH,DHH,
                CL,CH,CLH,CHH,AL,AH,ALH,AHH : Byte));
  end;

{ Structure used to allocate and access a DOS Memory block }

type
  TDosBuf = packed record
    case Integer of
      0: (Buf     : Pointer;    { Application near pointer                  }
          Selector: Word;       { DPMI-allocated selector                   }
          Size    : Word;       { Size of memory allocation in bytes        }
          RealBuf : DosPtr);    { Real mode far pointer equivalent of Buf   }
      1: (FarBuf  : FarPtr;     { Protected mode far pointer                }
          BufSize : Word;       { Size of memory allocation in bytes        }
          RealOfs : Word;       { Real mode offset - this is always zero    }
          RealSeg : Word);      { Real mode segment                         }
  end;

  TMemLock = (mlLock, mlUnLock);{ For Locking or Unlocking with MemLock     }

{--------- Protected Mode ISR, Real Mode ISR and Callback handlers ---------}

type
  EInterrupt = class(Exception);

  { Abstract base class for all interrupt, exception handlers and callbacks }

  TInterrupt = class(TObject)
   private
    Next     : TInterrupt;      { Pointer to next TInterrupt record in list }
    ID       : Cardinal;        { Interrupt/Callback number/ID and options  }
    SizeOfISR: Cardinal;        { Size of the ISR code - for memory locking }
    UserISR  : Pointer;         { Address of Self.ISR                       }
   protected
    procedure ISR; virtual; register; abstract; { User-defined ISR/Callback }
    function  ReHook(CodePtr: Pointer): Boolean; virtual; register; abstract;
    function  UnHook: Boolean; virtual; register; abstract;{ Unhook ISR/Clbk}
    function  MakeHook(Make: Boolean): Boolean; virtual; abstract;
    procedure LockMem(Lock: TMemLock); virtual; { Lock/UnLock memory areas  }
   public
    constructor Create(AID: Cardinal; ASizeOfISR: Cardinal);
    destructor Destroy; override;
    procedure Swap(Enable: Boolean);      { Enable/Disable hooking-out when }
  end;                                    { spawning a child process.       }

  { Abstract base class for real mode interrupts & callbacks.}

  TDosISR = class(TInterrupt)
   private
    RMCB   : DosPtr;            { Pointer to real mode callback             }
    RealPop: Cardinal;          { No. bytes to "pop" from r/m stack (4 or 6)}
    procedure CallBack; register; { The wrapper function that calls ISR     }
   protected
    OldVec : DosPtr;            { Original DOS vector ISR - read only       }
    Regs   : TRealRegs;         { DPMI structure used by the RMCB           }
    function  ReHook(CodePtr: Pointer): Boolean; override; register;
    function  UnHook: Boolean; override; register;        { Unhook the ISR  }
    function  MakeHook(Make: Boolean): Boolean; override; { Make/Unmake hook}
    procedure LockMem(Lock: TMemLock); override;  { Lock/UnLock memory areas}
    procedure InitHook;
   public
    constructor Create(IntNum, ASizeOfISR: Cardinal; Init: Boolean = True);
  end;

  { TDosSoftISR is used to hook a real mode software interrupt. }
  { ISR must be overriden by the user-defined descendant.       }

   TDosSoftISR = class(TDosISR)
    protected
     function  CallOldISR: Word; register;{ Call original interrupt handler }
   end;

  { TDosHardISR is used to hook a real mode hardware interrupt. }
  { ISR must be overriden by the user-defined descendant.       }

   TDosHardISR = class(TDosISR)
    protected
     procedure CallOldISR; register;      { Call original interrupt handler }
   end;

  { TDosCallBack is used to implement a r/m callback with a far return    }
  { frame. ISR, ReHook and UnHook must be overridden by the user object.  }

  TDosCallBack = class(TDosISR)
   protected
    function  ReHook(CodePtr: Pointer): Boolean; override; register; abstract;
    function  UnHook: Boolean; override; register; abstract;
    function  MakeHook(Make: Boolean): Boolean; override; { Make/Unmake hook}
   public
    constructor Create(ASizeOfISR: Cardinal; Init: Boolean = True);
    destructor  Destroy; override;
  end;

  { Abstract base class for all protected mode interrupts and callbacks }

   TAppISR = class(TInterrupt)
   private
    Glue    : PPtrArray;        { Dynamically allocated intitial glue code  }
    CallAddr: Pointer;          { Address of TAppISR.CallBack               }
    DataSeg : Word;             { Flat data segment                         }
    procedure CallBack; register; { The wrapper function that calls ISR     }
   protected
    OldVec  : FarPtr;           { Original p/m vector - read only           }
    Regs    : TRegisters;       { Saved machine state at time of interrupt  }
    function  ReHook(CodePtr: Pointer): Boolean; override; register;
    function  UnHook: Boolean; override; register; { Unhook the p/m int     }
    function  MakeHook(Make: Boolean): Boolean; override; { Create/Destroy  }
    procedure LockMem(Lock: TMemLock); override;   { [Un]Lock memory areas  }
   public
    constructor Create(IntNum: Byte; ASizeOfISR: Cardinal);
  end;

  { TAppSoftISR is used to hook a protected mode software interrupt }
  { ISR must be overriden by the user-defined descendant object.    }

   TAppSoftISR = class(TAppISR)
    protected
     function  CallOldISR: Cardinal; register;  { Call original int handler }
   end;

  { TAppHardISR is used to hook a real mode hardware interrupt  }
  { ISR must be overriden by the user-defined descendant object }

   TAppHardISR = class(TAppISR)
    protected
     procedure CallOldISR;                      { Call original int handler }
   end;

  { TAppCallBack is used to implement a p/m callback with far return frame. }
  { ISR, ReHook & UnHook must be overridden by the user.                    }

  TAppCallBack = class(TAppISR)
   protected
    procedure CallBack;         { The wrapper function that calls ISR       }
    function  ReHook(CodePtr: Pointer): Boolean; override; register; abstract;
    function  UnHook: Boolean; override; register; abstract;
    function  MakeHook(Make: Boolean): Boolean; override; { Make/Unmake hook}
    procedure LockMem(Lock: TMemLock); override;
   public
    constructor Create(ASizeOfISR: Cardinal);
    destructor  Destroy; override;
  end;

{ Initialize/Resize Application <-> Dos transfer memory for GetDosBuf }

procedure DosBufInit;

{ Release the Application <-> Dos transfer memory used by GetDosBuf }

procedure DosBufDone;

{ GetDosMem should be used to allocate a memory block suitable for passing  }
{ to a DOS or BIOS interrupt. Using GetDosMem guarantees that the memory    }
{ block has been allocated from the 1st megabyte of memory. Only use this   }
{ function when you need a buffer that is being updated frequently.         }

function GetDosMem(out Buf: TDosBuf; Size: Word): Boolean;

{ FreeDosMem disposes of a DOS buffer previously allocated with GetDosMem.  }

procedure FreeDosMem(var Buf: TDosBuf);

{ GetDosBuf sub-allocates a DOS/BIOS transfer buffer from the global DosBuf }
{ DOS-allocated memory block. Do not try to allocate more than DosBufSize!  }
{ Use this function for allocating a temporary App <-> DOS transfer buffer. }
{ The Regs TRealRegs structure is zero'ed. Passing 0 in Size means "give me }
{ everything (from DosBuf) you've got left". In this case an exception is   }
{ generated if there is less than 1K of DosBuf available. NOT THREAD-SAFE!  }

procedure GetDosBuf(out Buf: TDosBuf; Size: Cardinal; out Regs: TRealRegs);

{ FreeDosBuf releases a block of DOS memory allocated by GetDosBuf }

procedure FreeDosBuf(var Buf: TDosBuf);

{ Map a physical region of memory to a linear address }

function MapPhysicalMem(PhysicalBase, SizeInBytes: LongWord): Pointer;

{ Unmap a physical region of physical memory }

function UnMapPhysicalMem(LinearBase: Pointer): Boolean;

{ Create an application pointer that points to the given DOS/BIOS memory }

function MapDosPtr(RealPtr: DosPtr): Pointer;

{ Free a pointer previously allocated by MapDosPtr }

procedure FreeDosPtr(P: Pointer);

{ Locks or unlocks a linear region of memory to prevent swapping to disk }

function MemLock(BaseAdr: Pointer; Size: Cardinal;
                 LockFlag: TMemLock): Boolean; register;

{ Intr executes a specified software interrupt with a specified TRealsRegs  }
{ package. Returns the value of Regs.AX (as set by the real mode interrupt).}
{ Intr always call the real-mode (O/S) version of the interrupt.            }

function Intr(IntNo: Byte; var Regs: TRealRegs): Word;

{ IntrPM calls protected-mode interrupt "IntNo". This differs from Intr     }
{ in protected mode because Intr will call the real-mode version of the     }
{ given interrupt number. IntrPM is the same as Intr for real-mode programs }

function IntrPM(IntNo: Byte; var Regs: TRegisters): Cardinal;

{ MsDos invokes the DOS function call handler with a specified Registers    }
{ package. Returns the value of Regs.AX (as set by the real mode interrupt).}
{ MsDos always calls the real mode version of interrupt $21.                }

function MsDos(var Regs: TRealRegs): Word;

{ MsDosPM is only avaiable in protected mode. It invokes the Dos extender's }
{ Int $21 DOS function dispatch emulator.                                   }

function MsDosPM(var Regs: TRegisters): Word;

{ Call a real mode function with far call (16:16) return stack frame. The   }
{ real mode address of the function to be called must be in Regs.CS and     }
{ Regs.IP. Returns the value of Regs.AX (as set by the real mode function). }

function DosFarCall(var Regs: TRealRegs): Word;

{ Call a real mode function with an IRET return stack frame. The real mode  }
{ address of the function to be called must be in Regs.CS and Regs.IP.      }
{ Returns the value of Regs.AX (as set by the real mode function). This is  }
{ used to chain an original DOS/BIOS software interrupt (eg Video BIOS      }
{ interrupt $10) from inside a user-installed interrupt service routine.    }

function DosSoftIntr(var Regs: TRealRegs): Word; register;

{ Call a DOS/BIOS real mode function with an IRET return stack frame. This  }
{ is used to call hardware interrupts (eg system clock interrupt $08) from  }
{ inside a user-installed hardware interrupt service routine.               }

procedure DosHardIntr(RealAddr: DosPtr); register;

{ Call an application side function with an IRET return stack frame. This   }
{ is used to call software interrupts (eg video interrupt $10) from inside  }
{ the user-installed software interrupt service routine that replaced it.   }

function SoftIntr(var Regs: TRegisters): Cardinal; register;

{ Call an application side function with an IRET return stack frame. This   }
{ is used to call hardware interrupts (eg clock interrupt $08) from inside  }
{ the user-installed hardware interrupt service routine that replaced it.   }

procedure HardIntr(const ISR: FarPtr);

{ ClearRegs sets all register values to 0. This should be called before you }
{ set specific Regs fields prior a call to IntrPM or MsDosPM.               }

procedure ClearRegs(out Regs: TRegisters); register;

{ ClearRealRegs sets all register values to 0. This should be called before }
{ you set specific Regs fields prior a call to Intr, MsDos, DosFarCall etc. }

procedure ClearRealRegs(out Regs: TRealRegs); register;

{ GetIntVec returns the address of the given application mode interrupt.    }
{ This function returns the protected-mode interrupt vector in DMPI.        }

function GetIntVec(IntNo: Byte): FarPtr; register;

{ GetDosIntVec returns a real-mode (DOS/BIOS) vector.                       }

function GetDosIntVec(IntNo: Byte): DosPtr; register;

{ SetIntVec sets the address in the PM interrupt vector table to the        }
{ specified application Vector. Sets PM interrupt vector in DPMI apps.      }

function SetIntVec(IntNo: Byte; const Vector: FarPtr): Boolean;

{ Creates a far code pointer, given a flat near code pointer }

function MakeFarCodePtr(Offset: Pointer): FarPtr; register;

{ SetDosIntVec sets the address in the real mode interrupt vector table to  }
{ the specified real mode Vector. Vector must be a real mode ISR address    }

function SetDosIntVec(IntNo: Byte; DosISR: DosPtr): Boolean; register;

{ SwapVectors swaps the contents of the SaveIntXX pointers in the System    }
{ unit with the current contents of the interrupt vectors. SwapVectors is   }
{ typically called just before and just after a call to Exec. This insures  }
{ that the Exec'd process does not use any interrupt handlers installed by  }
{ the current process, and vice versa.                                      }

procedure SwapVectors;

{ UnHookAll unhooks all application-installed Dos/BIOS ISR's, callbacks &   }
{ exception handlers, and all Application ISR's and exception handlers.     }

procedure UnHookAll;

{ Enable hardware interrupts }

procedure EnableInterrupts; register;

{ Disable hardware interrupts }

procedure DisableInterrupts; register;

{ Convert an IRQ request (PIC) number to an interrupt vector table number }

function IRQtoIntVec(IRQ: Byte): Byte;

var
  MasterPicBase: Byte = $08;          { Primary PIC interrupt vector base   }
  SlavePicBase : Byte = $70;          { Secondary PIC interrupt vector base }
  DosBufSize   : Word = 1024 * 4;     { Default DosBuf size                 }
  DataSeg      : Word;                { Global flat data segment- Read only!}
  CodeSeg      : Word;                { Global flat code segment- Read only!}
  StackSeg     : Word;                { Global flat stack segment-Read only!}
  Dpmiflags    : Word;                { BX as returned by intDPMI-$0400     }

type

{ Pseudo Port[] arrays }

  TPort = class(TObject)
  private
    procedure SetPort(Index: Word; Data: Byte);
    function GetPort(Index: Word): Byte;
  public
    property Port[Index: Word]: Byte read GetPort write SetPort; default;
  end;

  TPortW = class(TObject)
  private
    procedure SetPortW(Index: Word; Value: Word);
    function GetPortW(Index: Word): Word;
  public
    property PortW[Index: Word]: Word read GetPortW write SetPortW; default;
  end;

  TPortL = class(TObject)
  private
    procedure SetPortL(Index: Word; Value: LongInt);
    function GetPortL(Index: Word): LongInt;
  public
    property PortL[Index: Word]: LongInt read GetPortL write SetPortL; default;
  end;

{ Pseudo Mem[] arrays }

{$IFDEF SUPPORTS_NO_ABSOLUTE}
  TMem = class(TObject)
  private
    function  GetMem(Index: LongInt): Byte;
    procedure SetMem(Index: LongInt; const Value: Byte);
  public
    property Mem[Index: LongInt]: Byte read GetMem write SetMem; default;
  end;
{$ENDIF SUPPORTS_NO_ABSOLUTE}

  TMemW = class(TObject)
  private
    function  GetMemW(Index: LongInt): Word;
    procedure SetMemW(Index: LongInt; const Value: Word);
  public
    property MemW[Index: LongInt]: Word read GetMemW write SetMemW; default;
  end;

  TMemL = class(TObject)
  private
    function  GetMemL(Index: LongInt): LongInt;
    procedure SetMemL(Index: LongInt; const Value: LongInt);
  public
    property MemL[Index: LongInt]: LongInt read GetMemL write SetMemL; default;
  end;

{ hexadecimal conversion function }

function WordToHex(X: Word): string;

function LongToHex(X: LongWord): string;

function AddrToHex(X: Pointer): string;

function FarToHex(X: FarPtr): string;

function DosPtrToHex(X: DosPtr): string;

function DecVal(C: Char): Byte;

function HexToWord(const S: string; var Value: Cardinal): Boolean;

var
  Port : TPort;
  PortW: TPortW;
  PortL: TPortL;
{$IFDEF SUPPORTS_NO_ABSOLUTE}
  Mem: TMem;
{$ELSE SUPPORTS_NO_ABSOLUTE}
  Mem: array[0..$7FFFFFFE] of Byte absolute 0;
{$ENDIF SUPPORTS_NO_ABSOLUTE}
  MemW : TMemW;
  MemL : TMemL;

{---------------------------------------------------------------------------}

implementation

var
  DosBuf    : TDosBuf;     { Global DOS memory used by GetDosBuf/FreeDosBuf }
  DosBufNext: NearPtr;     { Flat address of next available xfer buffer     }
  DosBufEnd : NearPtr;     { Last memory location + 1 of DosBuf             }

procedure DosBufDone;
begin
  if DosBufNext <> DosBuf.Buf
   then raise EInterrupt.Create('DosBuf is in use so cannot be freed');
  FreeDosMem(DosBuf);
end;

procedure DosBufInit;
begin
  if DosBuf.Size <> DosBufSize then
  begin
    DosBufDone;                       { Release previous block (if any)     }
    if not GetDosMem(DosBuf, DosBufSize)
     then raise EInterrupt.Create('Not enough DOS memory available');
    DosBufNext := DosBuf.Buf;
    DosBufEnd  := NearPtr(DosBuf.FarBuf.Ofs + DosBufSize);
{$ifdef Debug1}
    WriteLn('DosBuf.FarBuf  = ', FarToHex(DosBuf.FarBuf));
    WriteLn('DosBuf.RealBuf = ', DosPtrToHex(DosBuf.RealBuf));
    ReadLn;
{$endif Debug1}
  end;
end;

function MapDosPtr(RealPtr: DosPtr): Pointer;          { Tested by: MapDosPt}
begin                                      { Only works for True Flat memory}
  MapDosPtr := Pointer(LongWord((DosPtrRec(RealPtr).Seg) shl 4) +
                       DosPtrRec(RealPtr).Ofs);
end;

{ Free a pointer previously allocated by MapDosPtr }

procedure FreeDosPtr(P: Pointer);                      { Tested by: MapDosPt}
begin                                      { Nothing needs doing in a True  }
end;                                       { Flat Mode environment.         }

{ Locks or Unlocks a linear region of memory to prevent swapping to disk }

function MemLock(BaseAdr: Pointer; Size: Cardinal;
                 LockFlag: TMemLock): Boolean; register;
asm                                        { Tested by CallBack & HookDosi  }
            push  edi                      { EAX = BaseAdr;  EDX = Size     }
            push  esi                      { CL  = LockFlag (0 or 1)        }
            push  ebx
            xchg  ecx,eax                  { EAX = LockFlag  ECX = BaseAdr  }
            mov   edi,edx
            shld  ebx,ecx,16               { BX = BaseAdr.Hi CX = BaseAdr.Lo}
            add   eax,dpmiLockLinear       { EAX = dpmi[Un]LockLinear       }
            shld  esi,edi,16               { SI = Size.Hi    DI = Size.Lo   }
            int   intDPMI
            setnc al
            pop   ebx
            pop   esi
            pop   edi
end;

function GetDosMem(out Buf: TDosBuf; Size: Word): Boolean; register;
asm                                        { Tested by: MapDosPt            }
            add   dx,15                    { EAX = @Buf  DX = Size          }
            push  ebx
            push  eax
            shr   dx,4                     { DX = Allocation in paragraphs  }
            mov   ax,dpmiAllocDosMem
            movzx ebx,dx                   { BX = no paragraphs to allocate }
            int   intDpmi
            movzx ecx,dx                   { CX = p/m selector              }
            pop   edx                      { EDX = @Buf                     }
            jc    @@Error
            shl   eax,16                   { EAX = RealSeg * 2^16           }
            mov   [edx].TDosBuf.RealBuf,eax{ Only works in true flat mode...}
            shr   eax,16-4                 { EAX = RealSeg * 2^(16-12)      }
            mov   [edx].TDosBuf.Selector,cx{ DPMI-allocated selector is only}
            shl   ebx,4                    { EBX = Allocation size in bytes }
            mov   [edx].TDosBuf.Buf,eax    { used in FreeDosMem.            }
            mov   [edx].TDosBuf.Size,bx
            mov   al,true
            jmp   @@Exit

@@Error:    sub   eax,eax                  { Allocation failed, so return   }
            mov   [edx].TDosBuf.Buf,eax    { False, set Buf and RealBuf to  }
            mov   [edx].TDosBuf.RealBuf,eax{ nil, and Selector & Size to 0  }
            mov   dword ptr [edx].TDosBuf.Selector,eax { Selector & Size = 0}
@@Exit:     pop   ebx
end;

procedure FreeDosMem(var Buf: TDosBuf); register;     { Tested by: MapDosPt }
asm
            mov   dx,[eax].TDosBuf.Selector  { EAX = @Buf                   }
            mov   ecx,eax
            mov   ax,dpmiFreeDosMem
            int   intDpmi
            xor   edx,edx
            mov   [ecx].TDosBuf.Buf,eax      { Set Buf and RealBuf to nil,  }
            mov   [ecx].TDosBuf.RealBuf,eax  { and Selector & Size to 0     }
            mov   dword ptr [ecx].TDosBuf.Selector,edx { Selector & Size = 0}
end;

{ GetDosBuf implements a nested-safe way of allocating a temporary     }
{ block of DOS memory from the pool of DOS memory allocated at startup }

procedure GetDosBuf(out Buf: TDosBuf; Size: Cardinal; out Regs: TRealRegs);
begin                                                  { Tested by: VolInfo }
  Assert(Size <= DosBuf.Size, 'Illiegal [Size] passed to GetDosBuf - ' + LongToHex(Size));
  if Size = 0 then                 { Grab as much of DosBuf as is available }
    Size := Max(PtrRec(DosBufEnd).Ofs - PtrRec(DosBufNext).Ofs, 1024);
  Size := (Size + 15) and (not 15);    { All allocations are segment-aligned}
  if PtrRec(DosBufNext).Ofs + Size <= PtrRec(DosBufEnd).Ofs then
  begin
    Inc(PtrRec(DosBufNext).Ofs, Size);   { This must be an atomic operation }
    Buf.Buf     := Pointer(PtrRec(DosBufNext).Ofs - Size);
    Buf.Size    := Size;
    Buf.RealSeg := Buf.FarBuf.Ofs shr 4;    { Only works in True Flat model }
    Buf.RealOfs := 0;                       { This is why we segment align  }
    Buf.FarBuf.Seg := DosBuf.FarBuf.Seg;
    ClearRealRegs(Regs);                    { Zero all the registers        }
{$ifdef Debug1}
    WriteLn('Dos Buffer allocated at  ', FarToHex(Buf.FarBuf),
            ' of ', Buf.Size, '(', LongToHex(Buf.Size), ') Bytes');
    WriteLn('Buf Real mode address is ', DosPtrToHex(Buf.RealBuf));
    WriteLn('DosBufNext = ', AddrToHex(DosBufNext)); WriteLn;
{$endif Debug1}
  end else
  begin
{$ifdef Debug1}
    WriteLn;
    WriteLn('GetDosBuf has failed to allocate a buffer! - Size = ', Size);
    WriteLn('DosBufNext = ', AddrToHex(DosBufNext));
    WriteLn('DosBuf.Far = ', FarToHex(DosBuf.FarBuf));
{$endif Debug1}
    raise EInterrupt.Create('Not enough DOS sub-allocation memory available');
  end;
end;

procedure FreeDosBuf(var Buf: TDosBuf);                { Tested by: VolInfo }
begin
  Dec(PtrRec(DosBufNext).Ofs, Buf.Size);
  Buf.Buf := nil;
  Buf.Size:= 0;
  Buf.RealBuf := 0;
  Assert(PtrRec(DosBufNext).Ofs >= DosBuf.FarBuf.Ofs,
    'ERROR! FreeDosBuf has moved DosBufNext   (' + AddrToHex(DosBufNext) + ')' +
    '       below the start address of DosBuf (' + AddrToHex(DosBuf.Buf) + ')');
{$ifdef Debug1}
  WriteLn('FreeDosBuf moved DosBufNext down to ', AddrToHex(DosBufNext));
  WriteLn;
{$endif Debug1}
end;

function IntrDpmi(FuncNum: Cardinal; var Regs: TRegisters): Boolean; register;
begin                                                 { Tested by: HookDosI }
  Regs.EAX := FuncNum;
  IntrPM(intDPMI, Regs);
  Result := Regs.EFlags and fCarry = 0;
end;

function MapPhysicalMem(PhysicalBase, SizeInBytes: LongWord): Pointer;
asm                                                   { Tested by: DrvTest1 }
            mov   ecx,eax                  { ECX = EAX = Physical base addr }
            push  edi                      { EDX = Size of area in bytes    }
            push  esi
            mov   edi,edx                  { EDI = Size of region in bytes  }
            push  ebx
            shld  esi,edi,16               { SI:DI = Size of region in bytes}
            shld  ebx,ecx,16               { BX:CX = Physical base address  }
            mov   ax,dpmiMapPhysical
            int   intDPMI                  { DPMI - Map Physical to linear  }
            jnc   @@Ok                      { BX:CX = Linear base address    }
            xor   ecx,ecx                  { Function failed, so return nil }
            xor   ebx,ebx
@@Ok:       shl   ecx,16                   { ECX.Hi = LinearBase.Lo         }
            mov   eax,ebx                  { AX = LinearBase.Hi             }
            shld  eax,ecx,16               { EAX = base linear address      }
@@Exit:     pop   ebx
            pop   esi
            pop   edi
(*                                         { This is not needed in True_Flat}
            push  esi                      { Save SI:DI for later           }
            push  edi
            jnc   @@1                      { BX:CX = Linear base address    }
@@Failed:   add   esp,8
@@Failed1:  xor   eax,eax
            jmp   @@Exit

@@1:        mov   eax,ebx                  { AX = LinearBase.Hi             }
            shld  eax,ecx,16               { EAX = base linear address      }

 { Now we have the physical address we can allocate a selector to access it }

            xor   ecx,ecx
            mov   edi,eax                  { Save our function result in EDI}
            mov   ax,dpmiAllocSeg          { Allocate a sigle LDT descriptor}
            inc   ecx
            int   intDpmi
            jc    @@Failed
            mov   edx,edi                  { Set the segment base address of}
            mov   ebx,eax                  { the new descriptor to the LFB  }
            shld  ecx,edi,16               { CX:DX = Linear Base Address    }
            mov   ax,dpmiSetSegBase
            int   intDpmi
            jc    @@Failed
            pop   edx                      { Set the selector limit to      }
            pop   ecx                      { SizeInBytes                    }
            mov   ax,dpmiSetSegSize
            int   intDpmi
            jc    @@Failed1
            mov   eax,edi
@@Exit:     pop   ebx                      { EAX = nil if function failed   }
            pop   esi
            pop   edi
*)
end;

function UnMapPhysicalMem(LinearBase: Pointer): Boolean;
asm                                                   { Tested by: Untested }
            or    eax,eax
            push  ebx
            jz    @@1
            mov   ecx,eax
            shld  ebx,eax,16
            mov   ax,dpmiUnMapPhysical
            int   intDpmi
            setnc al
@@1:        pop   ebx
end;

function MsDos(var Regs: TRealRegs): Word; register;  { Tested by: TstMsDos }
asm
            mov   edx,eax
            mov   al,intDos
            jmp   Intr
end;

function Intr(IntNo: Byte; var Regs: TRealRegs): Word; register;
asm                                      { Tested by: MapDosPt              }
            push  ebx                    { AL = IntNo  EDX = @Regs          }
            push  edi
            mov   edi,edx                { ES:EDI = @Regs                   }
            movzx ebx,al                 { BL = interrupt number BH = 0     }
            xor   ecx,ecx                { No stack transfer                }
            mov   ax,dpmiCallRealInt     { Call real-mode interrupt         }
            int   intDPMI
            movzx eax,[edi].TRealRegs.&AX
            pop   edi
            pop   ebx
end;

function IntrPM(IntNo: Byte; var Regs: TRegisters): Cardinal; register;
asm       { AL = IntNo  EDX = @Regs }                 { Tested by: TestIntr }
            mov   byte ptr ds:[offset @@IntPatch],al  { Set self-modifying  }
            pushad                           { Save all the GP registers    }
            push  es
            push  fs
            push  gs
            push  edx                        { Save address of Regs         }
            mov   eax,[edx].TRegisters.&EAX  { Set all the registers from   }
            mov   ebx,[edx].TRegisters.&EBX  { the Regs parameter           }
            mov   ecx,[edx].TRegisters.&ECX
            mov   esi,[edx].TRegisters.&ESI
            mov   edi,[edx].TRegisters.&EDI
            mov   ebp,[edx].TRegisters.&EBP

  { The Regs.DS,ES and GS values must either be a null    }
  { selector (0-3) or they nust contain a valid selector. }
  { Arguments cannot be passed in FS for safety reasons.  }

            mov   gs,[edx].TRegisters.&GS
            mov   es,[edx].TRegisters.&ES
            mov   ds,[edx].TRegisters.&DS
            mov   edx,ss:[edx].TRegisters.&EDX

            db $CD; @@IntPatch: db 0   { Self-modified "INT xx" instruction }

            xchg  edx,[esp]                   { EDX = @Regs  [ESP] = EDX    }
            push  ss
            mov   ss:[edx].TRegisters.&DS,ds
            pop   ds                          { DS = SS = Flat_DS           }
            pop   [edx].TRegisters.&EDX       { Set Regs.EDX from pop       }
            mov   [edx].TRegisters.&EAX,eax   { Set the Rest of the Regs    }
            mov   [edx].TRegisters.&EBX,ebx   { fields from the registers   }
            mov   [edx].TRegisters.&ECX,ecx
            mov   [edx].TRegisters.&ESI,esi
            mov   [edx].TRegisters.&EDI,edi
            mov   [edx].TRegisters.&EBP,ebp
            mov   [edx].TRegisters.&ES,es
            pushfd
            mov   [edx].TRegisters.&FS,fs
            pop   eax
            mov   [edx].TRegisters.&GS,gs
            mov   [edx].TRegisters.EFlags,eax
            pop   gs                          { Restore the selector regs   }
            pop   fs
            pop   es
            popad                             { Restore the general regs    }
            mov   byte ptr ds:[offset @@IntPatch],0 { Restore self-modified }
            mov   eax,[edx].TRegisters.&EAX   { Set the function result     }
end;

function MsDosPM(var Regs: TRegisters): Word; register;
asm                                                   { Tested by: Untested }
            mov   edx,eax
            mov   al,intDos
            jmp   IntrPM
end;

function DosFarCall(var Regs: TRealRegs): Word; register;
asm                                                   { Tested by: Untested }
            push  edi                    { EAX = @Regs                      }
            mov   edi,eax                { EDX = @Regs                      }
            mov   ax,dpmiCallRealFar     { Simulate real-mode far call      }
            xor   ecx,ecx                { No stack transfer                }
            int   intDPMI
            movzx eax,[edi].TRealRegs.&AX
            pop   edi
end;

function DosSoftIntr(var Regs: TRealRegs): Word; register;
asm  { EAX = @Regs }                                  { Tested by: HookDosI }
            push  ebx
            xor   ebx,ebx                   { BH must be zero               }
            push  edi
            mov   edi,eax                   { EDI = @Regs                   }
            mov   ax,dpmiCallIret
            xor   ecx,ecx                   { No stack transfer             }
            int   intDPMI
            movzx eax,[edi].TRealRegs.&AX
            pop   edi
            pop   ebx
end;

procedure DosHardIntr(RealAddr: DosPtr); register;    { Tested by: Untested }
var
  Regs: TRealRegs;
asm                                         { EAX = RealAddr                }
            push  edi
            xor   ecx,ecx                   { No stack transfer             }
            lea   edi,[Regs]                { EDI = @Regs                   }
            mov   dword ptr [Regs].&FS,ecx  { Regs.FS = Regs.GS = 0         }
            mov   dword ptr [Regs].&SP,ecx  { Regs.SP & Regs.SS = 0         }
            mov   [Regs].EXX,ecx            { Regs.EXX = 0                  }
            mov   dword ptr [Regs].&IP,eax  { Set Regs.CS:IP to RealAddr    }
            mov   eax,dpmiCallIRet          { Call r/m procedure with       }
            int   intDPMI                   { @Regs.CS:IP with IRET stack   }
            pop   edi                       { frame.                        }
end;

function SoftIntr(var Regs: TRegisters): Cardinal; register;
asm  { EAX = @Regs }                                  { Tested by: HookIntr }
            push  gs
            push  fs
            pushad                          { Save all the GP registers     }
            pushfd                          { Push flags for the IRET       }
            push  cs                        { Push IRET return address Seg  }
            mov   esi,[eax].TRegisters.&ESI { Load the registers for intrpt }
            push  offset @@Retn             { Push IRET return address Ofs  }
            mov   ebx,[eax].TRegisters.&EBX
            push  dword ptr [eax].TRegisters.&CS{ Push ISR call addr-Regs.CS}
            mov   ecx,[eax].TRegisters.&ECX
            mov   edx,[eax].TRegisters.&EDX
            push  [eax].TRegisters.&EIP     { Push ISR call addr - Regs.EIP }
            mov   edi,[eax].TRegisters.&EDI
            mov   es,[eax].TRegisters.&ES
            push  [eax].TRegisters.&EAX     { Push Regs.EAX                 }
            mov   ebp,[eax].TRegisters.&EBP
            mov   fs,[eax].TRegisters.&FS
            mov   gs,[eax].TRegisters.&GS
            mov   ds,[eax].TRegisters.&DS   { DS = Regs.DS                  }
            pop   eax                       { EAX = Regs.EAX                }
            retf  0                         { Far Call the ISR              }

@@Retn:     pushfd                          { Push new Regs.EFlags          }
            push  ds                        { Push new Regs.DS              }
            push  esi                       { Push new Regs.ESI             }
            mov   ds,cs:[DataSeg]           { Set DS to Flat_DS             }
            mov   esi,[esp+(4*10)]          { DS:ESI = @Regs                }
            mov   [esi].TRegisters.&EAX,eax { Update the Regs fields...     }
            pop   [esi].TRegisters.&ESI     { Pop new Regs.ESI              }
            mov   [esi].TRegisters.&EBX,ebx
            pop   dword ptr [esi].TRegisters.&DS      { Pop new Regs.DS     }
            mov   [esi].TRegisters.&ECX,ecx
            mov   [esi].TRegisters.&EDX,edx
            mov   [esi].TRegisters.&EDI,edi
            mov   [esi].TRegisters.&ES,es
            pop   [esi].TRegisters.EFlags   { Pop new Regs.EFlags           }
            mov   [esi].TRegisters.&FS,fs
            mov   [esi].TRegisters.&EBP,ebp
            mov   [esi].TRegisters.&FS,fs
            mov   [esi].TRegisters.&GS,gs
            popad                           { Restore all the GP registers  }
            mov   es,[DataSeg]              { ES = DS = Flat_DS             }
            pop   fs                        { Restore GS and FS             }
            pop   gs
            mov   eax,[eax].TRegisters.&EAX { Return value is EAX           }
end;

procedure HardIntr(const ISR: FarPtr); register;      { Tested by: HookIntr }
asm                                         { EAX = @ISR                    }
            pop   ecx                       { ECX = User return address     }
            pushfd                          { For IRET return               }
            push  cs                        { IRET return address selector  }
            push  ecx                       { IRET return address offset    }
            push  dword ptr [eax].FarPtr.&Seg
            push  [eax].FarPtr.&Ofs
            retf  0                         { Call far OldVec               }
end;

procedure ClearRegs(out Regs: TRegisters); register; { Tested by: TestIntr  }
asm                                         { EAX = @Regs                   }
            xchg  eax,edi                   { ES:EDI = @Regs                }
            mov   ecx,type TRegisters/4     { ECX = SizeOf(Regs) in DWords  }
            mov   edx,eax                   { EDX = saved EDI               }
            xor   eax,eax
            rep   stosd                     { Fill Regs with zeros          }
            mov   edi,edx                   { Restore EDI                   }
end;

procedure ClearRealRegs(out Regs: TRealRegs); register;{ Tested by: TstMsDos}
asm
            xchg  eax,edi                   { ES:EDI = @Regs                }
            mov   ecx,type TRealRegs/4      { ECX = SizeOf(Regs) in DWords  }
            mov   edx,eax                   { EDX = saved EDI               }
            sub   eax,eax
            rep   stosd                     { Fill Regs with zeros          }
            mov   edi,edx                   { Restore EDI                   }
end;

function GetIntVec(IntNo: Byte): FarPtr; register;    { Tested by: GetIntVc }
asm                                         { AL = IntNo  EDX = @Result     }
            push  ebx                       { Use DPMI sever to return the  }
            movzx ebx,al                    { address of the protected mode }
            mov   ax,dpmiGetProtInt         { interrupt vector              }
            push  edx                       { Save pointer to Result        }
            int   intDPMI
            pop   eax                       { EAX = @Result                 }
            jnc   @@1                       { DPMI function was sucessful   }
            xor   edx,edx                   { Set result to nil             }
            mov   ecx,edx
@@1:        mov   [eax].FarPtr.&Ofs,edx     { Store FarPtr function result  }
            pop   ebx
            mov   [eax].FarPtr.&Seg,cx
end;

function GetDosIntVec(IntNo: Byte): DosPtr; register; { Tested by: HookDosI }
asm
            push  ebx                    { AL = IntNo                       }
            movzx ebx,al                 { Use the DPMI sever to return the }
            mov   ax,dpmiGetRealInt      { real mode address of a real mode }
            int   intDPMI                { interupt vector.                 }
            shl   ecx,16                 { ECX.Hi = Seg(Real_Vector)        }
            pop   ebx                    { DX = Ofs(Real_Vector)            }
            mov   eax,ecx                { AX.Hi = Seg(Real_Vector)         }
            mov   ax,dx                  { EAX = Seg:Ofs(Real_Vector)       }
end;

function SetIntVec(IntNo: Byte; const Vector: FarPtr): Boolean; register;
asm                                      { Tested by: GetIntVc              }
            push  ebx                    { AL =  IntNo  EDX = @Vector       }
            movzx ebx,al                 { EBX = IntNo                      }
            mov   ax,dpmiSetProtInt      { Set protected mode intrpt vector }
            movzx ecx,[edx].FarPtr.&Seg  { Use the DPMI server to set a     }
            mov   edx,[edx].FarPtr.&Ofs  { to a protected mode address.     }
            int   intDPMI                { protected mode interrupt vector  }
            setnc al                     { Return True if carry clear       }
            pop   ebx
end;

function SetDosIntVec(IntNo: Byte; DosISR: DosPtr): Boolean; register;
asm                                      { Tested by: HookDosI              }
            push  ebx                    { AL =  IntNo  EDX = DosISR        }
            movzx ebx,al                 { EBX = IntNo                      }
            shld  ecx,edx,16             { CX = Seg(DosISR) DX = Ofs(DosISR)}
            mov   ax,dpmiSetRealInt      { Use the DPMI server to set a     }
            int   intDPMI                { (BL = IntNo   CX:DX = DosISR)    }
            pop   ebx                    { a real mode interrupt vector     }
            setnc al                     { Return True if carry clear       }
end;

function MakeFarCodePtr(Offset: Pointer): FarPtr; register;
begin                                                 { Tested by: GetIntVc }
  Result.Ofs := PtrRec(Offset).Ofs;
  Result.Seg := CodeSeg;
end;

{----------- Linked-list of hooked p/m, r/m ISR's and callbacks ------------}

{ A list of user-installed ISR's, Exception handlers and callbacks are      }
{ kept so that:                                                             }
{  (1) No interrupt or exception is hooked more than once (RegisterHook).   }
{  (2) All ISR's/callbacks are removed before program terminatn (UnHookAll).}
{  (3) All ISR's/callbacks except those specified are unhooked before exec- }
{      cuting a child process; rehooked when child terminates (SwapVectors).}
{                                                                           }
{              ---------------- TInterrupt.ID ----------------              }
{              15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00              }
{               |  |        | [-Type-] [--- IntNum or ID ----]              }
{        itNoSwap  itUnHook                                                 }

const                             { Hook ID type identifiers                }
  itDosInterrupt  = $0000;        { Dos (real-mode) interrupt tag           }
  itAppInterrupt  = $0100;        { Application (Protected mode) intrpt tag }
  itDosException  = $0200;        { Dos (real mode) exception handler tag   }
  itAppException  = $0300;        { Application (Protected mode) Exception  }
  itDosCallBack   = $0400;        { Realmode far callback tag               }
  itAppCallBack   = $0500;        { Protected mode far callback tag         }

  itUnHook        = $8000;        { Toggled by SwapVectors to rehook/unhook }
  itNoSwap        = $4000;        { Set => Do not let SwapVectors unhook it }
  itIDMask        = $07FF;        { ID AND itIDMask just leaves the true ID }

var
  DosCallNum: Cardinal = itDosCallBack; { Used to uniquely identify callbcks}
  AppCallNum: Cardinal = itAppCallBack; { Same, except used for p/m callbcks}
var
  HookList   : TInterrupt;        { List of registered Dos/App/Callbacks    }
  UnitCleanUp: procedure;         { Unit cleanup procedure                  }

{ A contiguous block of (intStackCnt * intStackSize) memory is allocated to }
{ provide intStackCnt interrupt/callback stacks. The TInterrupt descendands }
{ use these in a First Used, Last Released manner to switch from the DPMI   }
{ server-allocated 1k stack (in use when the interrupt is invoked), to an   }
{ intStackSize sized Flat (SS = DS) stack. This is done before the user-    }
{ installed ISR or callback is invoked, and is done for two reasons:        }
{                                                                           }
{ (1) A 1k stack may not be big enough for some ISR's or callbacks.         }
{ (2) The Delphi code generator assumes SS = DS, so some functions, such as }
{     Move and FillChar would crash if used inside an application ISR.      }
{                                                                           }
{ This memory is allocated dynamically during the creation of the first     }
{ TInterrupt descendant. It is released only when there are no TInterrupt   }
{ descendant objects still in existance. Note that there is little point in }
{ allocating more than 16 interrupt stacks because the WDosX extender uses  }
{ a similar scheme for its 1k stacks, and it only allocates 16 stacks.      }

const
  intStackCnt  = 16;         { Provide 16 interrupt stacks.                 }
  intStackSize = 8 * 1024;   { Each interrupt stack is 8K in size           }
var
  IntStacks: PByteArray;     { Interrupt stacks memory allocation           }
  StackTop : FarPtr;         { Pointer to top of current interrupt stack    }

{ Unhook the specified Interrupt, Dos Interrupt or real mode callback from  }
{ the system and remove it from the HookList linked-list.                   }

function UnHookIt(ID: Cardinal): Boolean; register;   { Tested by: HookDosI }
var
  PCur: ^TInterrupt;
begin
  PCur := @HookList;
  ID := ID and itIdMask;               { Just leave the identification bits }
  while PCur^ <> nil do
  begin
    if ID <> (PCur^.ID and itIdMask)   { Compare the identification bits    }
     then PCur := @PCur^.Next          { Not this one, so keep looking.     }
     else if PCur^.MakeHook(false)     { Try to unhook the interrupt/callbck}
           then begin
                  PCur^ := PCur^.Next; { Remove it from the linked-list     }
                  Result:= true;
                  if HookList = nil then      { No more active TInterrupts  }
                  begin                       { so free resources           }
                    DosCallNum := itDosCallBack;    { Reset DosCallNum      }
                    AppCallNum := itDosCallBack;    { Reset AppCallNum      }
                    FreeMem(IntStacks, intStackSize * intStackCnt);
                    IntStacks := nil;         { Free the interrupt stacks   }
                  end;
                  Exit;
                end
           else raise EInterrupt.Create('Unable to unhook ' + PCur^.ClassName);
  end;
  Result := false;                     { The given ID was not found         }
end;

function RegisterHook(AInterrupt: TInterrupt): Boolean;
var                                                   { Tested by: HookDosI }
  List: TInterrupt;
begin
  Assert(AInterrupt <> nil, 'RegisterHook.AInterrupt cannot be nil!');
  List := HookList;
  while List <> nil do
  begin
    if List.ID and itIDMask = AInterrupt.ID and itIDMask then
    begin
      Result := false;
      Exit;
    end;
    List := List.Next;
  end;
  AInterrupt.Next := HookList;
  HookList:= AInterrupt;
  UnitCleanUp := UnHookAll; { Make sure all interrupts and callbacks will   }
  Result := true;           { be unhooked before program termination.       }
end;

{-------------------------------- TInterrupt -------------------------------}

const
  vmiTInterrupt_ISR = 0;     { ISR is the first user-defined virtual method }

constructor TInterrupt.Create(AID: Cardinal; ASizeOfISR: Cardinal);
begin
  inherited Create;
  ID := AID;                         { Set the unique interrupt identifier  }
  SizeOfISR := ASizeOfISR;
  UserISR := GetVirtualPtr(Self, vmiTInterrupt_ISR); { Address of ISR method}
  Assert(UserISR <> nil, 'GetVirtualPtr failed in TInterrupt.Create ');

  { Check that the interrupt stacks memory has been allocated }

  if IntStacks = nil then
    try
      GetMem(IntStacks, intStackSize * intStackCnt);
      MemLock(IntStacks, intStackSize * intStackCnt, mlLock);
      StackTop.Seg := DataSeg;
      StackTop.Ofs := PtrRec(IntStacks).Ofs + (intStackSize * intStackCnt);
    except
      on EOutOfMemory do
        raise EInterrupt.Create('Not enough memory for the interrupt stacks');
    end;
end;

destructor TInterrupt.Destroy;                        { Tested by: CallBack }
begin
  if UnHookIt(ID)
   then inherited Destroy
   else raise EInterrupt.Create('Unable to unhook ' + ClassName);
end;

procedure TInterrupt.LockMem(Lock: TMemLock);
begin                                            { Lock code & data memory  }
  MemLock(@Self, SizeOf(Self), Lock);            { touched by the ISR       }
  MemLock(GetVmt(Self), vmtHeaderSize + SizeOf(Pointer) * 20, Lock);
  MemLock(UserISR, SizeOfISR, Lock);             { Lock the user ISR code   }
end;

procedure TInterrupt.Swap(Enable: Boolean);           { Tested by: Untested }
begin
  ID := (ID and (not itNoSwap)) + (Cardinal(Ord(not Enable)) * itNoSwap);
end;

{--------------------------------- TDosISR ---------------------------------}

{ This procedure is called by the DPMI RMCB for Dos interrupts & callbacks. }
{ The code is fully re-enterable as long as hardware interrupts disabled on }
{ entry, and are not re-enabled except by the user ISR method.              }

procedure TDosISR.CallBack;                 { Tested by: HookDosI, Callback }
asm
{ DS:ESI = p/m pointer to top of r/m stack.      }
{ ES:EDI = @Self.Regs                            }
{ SS:ESP = Top of DPMI-allocated protected stack }

{ At this point, the stack frame contains the following data:               }
{                                                                           }
{       Return EFlags(RMCB)                    -----+ Pushed by the         }
{       Return CS    (RMCB)                         | Real mode callback    }
{       Return EIP   (RMCB)                    -----+                       }

            push  ds                        { Save the r/m Stack Selector   }
            push  es                        { Set DS to Flat_DS             }
            mov   eax,edi                   { EAX = @Self.RealRegs          }
            pop   ds
            sub   eax,offset Regs           { ES|DS:EAX = @Self             }

 { Switch to a larger Flat (SS = DS) stack }

            mov   edx,esp                   { Save the curent SS:ESP        }
            mov   ebx,ss
            db $0F,$B2,$25; dd StackTop     { lss   esp,fword ptr [StackTop]}
            sub   [StackTop].&Ofs,intStackSize
            push  ebx                       { Save DpmiStack.Seg            }
            push  edx                       { Save DpmiStack.Ofs            }

            mov   ebx,[eax].TDosISR.RealPop { EBX = 4 or 6                  }
           {mov   edx,[eax].TDosISR.OldVec  }{ Set RealRegs.CS:IP to OldVec }
           {mov   dword ptr [edi].TRealRegs.&IP,edx}{ for chaining 2 originl}
            cld
            call  dword ptr [eax].TInterrupt.UserISR { Call user-defined ISR}

 { Switch back to the DPMI-allocated stack }

            pop   edx
            pop   ss                        { The instruction immediately   }
            mov   esp,edx                   { following a "pop ss" cannot be}
            add   [StackTop].&Ofs,intStackSize { interrupted, so this is ok }

 { Set the callback return address CS:IP to the return address as is stored }
 { on the top of the realmode stack and "pop" the r/m IP, CS [and Flags].   }
 { ESI = RealStack.Ofs  ES:EDI = @Self.RealRegs  SS:[ESP] = RealStack.Seg   }
 { BX = No. bytes to "pop" off the r/m stack (Callback = 4, ISR = 6)        }

            pop   ds                        { DS:ESI = @RealStack           }
            mov   edx,[esi]                 { EDX = RM return CS:IP         }
            add   es:[edi].TRealRegs.&SP,bx { "Pop" r/m IP,CS [and Flags]   }
            mov   dword ptr es:[edi].TRealRegs.&IP,edx    { Set r/m Address }
            iretd                           { IRET always used with RMCB's  }
end; { ES:EDI = @RealReags }

constructor TDosISR.Create(IntNum, ASizeOfISR: Cardinal; Init: Boolean = True);
begin                                                  { Tested by HookDosI }
  inherited Create(IntNum + itDosInterrupt, ASizeOfISR);
  RealPop := SizeOf(DosPtr) + 2;            { 6 bytes "pop"d from r/m stack }
  if Init then InitHook;
end;

procedure TDosISR.LockMem(Lock: TMemLock);
begin
  inherited LockMem(Lock);
  MemLock(@TDosISR.CallBack,          { [Un]Lock the ISR callback glue code }
          PtrRec(@TDosISR.Create).Ofs - PtrRec(@TDosISR.CallBack).Ofs, Lock);
end;

{ Create and initialize all structures required to make the callback. }
{ This is used by both TDosISR and TDosCallback.                      }

function TDosISR.MakeHook(Make: Boolean): Boolean;    { Tested by: HookDosI }
var
  R: TRegisters;
begin
  ClearRegs(R);
  if Make then                           { Hook our ISR or r/m far callback }
  begin
    LockMem(mlLock);                     { Lock all memory touched by ISR   }
    Pointer(R.ESI) := @TDosISR.CallBack; { Try to allocate a DPMI callback..}
    Pointer(R.EDI) := @Regs;             { R.DS:ESI = FarPtr(@CallBack)     }
    R.ES := DataSeg;                     { R.ES:EDI = @Self.RealRegs        }
    R.DS := CodeSeg;
    if IntrDpmi(dpmiAllocRMCB, R) then
    begin
      DosPtrRec(RMCB).Ofs := R.DX;       { RMCB is the (r/m) address we     }
      DosPtrRec(RMCB).Seg := R.CX;       { will pass to Dos/BIOS            }
      Result := ReHook(Pointer(RMCB));   { Install the actual ISR/Callback  }
    end else
    begin
      RMCB := 0;
      Result := false;
    end
  end else
  begin                                  { Unhook our ISR or far callback   }
    Result := (RMCB = 0) or UnHook;
    if Result then                       { Set Int vector to OldVec or      }
    begin                                { unhook the real mode far callback}
      R.CX := DosPtrRec(RMCB).Seg;       { Free the DPMI-allocated RMCB     }
      R.DX := DosPtrRec(RMCB).Ofs;
      IntrDpmi(dpmiFreeRMCB, R);
    end;
  end;
end;

function TDosISR.ReHook(CodePtr: Pointer): Boolean;  { Tested by: HookDosI  }
begin                                                { Hook a r/m interrupt }
  OldVec := GetDosIntVec(ID);          { Save the previous value of vector  }
  Result := SetDosIntVec(ID, RMCB);    { Set the new r/m interrupt vector   }
end;

function TDosISR.UnHook: Boolean;                  { Tested by: HookDosI    }
begin                                  { Unhook a r/m interrupt             }
  UnHook := SetDosIntVec(ID, OldVec);  { Restore previous handler           }
end;

procedure TDosISR.InitHook;
begin
  if not (MakeHook(True) and RegisterHook(Self))
    then raise EInterrupt.Create('Dos Interrupt/Callback hook failed');
end;

{-------------------------------- TDosSoftISR ----------------------------- }

function TDosSoftISR.CallOldISR: Word;                { Tested by: HookDosI }
begin
  PDosPtr(@Regs.IP)^ := OldVec;               { Regs.IP:CS = orginal vector }
  Result := DosSoftIntr(Regs);                { Call original DOS handler   }
end;

{-------------------------------- TDosHardISR ----------------------------- }

procedure TDosHardISR.CallOldISR;
asm                                                   { Tested by: Untested }
            mov   eax,[eax].TDosHardISR.OldVec  { Set EAX to OldVec         }
            jmp   DosHardIntr                   { Call original DOS handler }
end;

{------------------------------- TDosCallback ------------------------------}

constructor TDosCallBack.Create(ASizeOfISR: Cardinal; Init: Boolean = True);
begin                                                 { Tested by: CallBack }
  inherited Create(DosCallNum, ASizeOfISR, Init);
  Inc(DosCallNum);
end;

destructor TDosCallBack.Destroy;                      { Tested by: CallBack }
begin
  if WordRec(LongRec(ID).Lo).Lo = WordRec(LongRec(DosCallNum).Lo).Lo - 1
   then Dec(DosCallNum);
  inherited Destroy;
end;

function TDosCallBack.MakeHook(Make: Boolean): Boolean;
begin                                                 { Tested by: CallBack }
  RealPop := SizeOf(DosPtr);                 { 4 bytes "pop"d from r/m stack}
  MakeHook := inherited MakeHook(Make);
end;

{--------------------------------- TAppISR ---------------------------------}

{ AppIsrGlue is the glue code required to hook p/m interrupts & callbacks.  }
{ Each ISR and callback gets its own copy of this code so that the @@Self   }
{ dword can be be patched to contain the correct Self value.                }
{ Note that the code below is itself never executed, only copies of it are. }
{ The copied code requires two "fixups" before it can be used:              }
{ (1) The @@Self dword must be set to the address of the TAppISR object     }
{     being used to handle the given interrupt, exception or callback.      }
{ (2) The dword address that is the @@Self part of the 3rd instruction      }
{     must be set to the absolute (linear) address of the @@Self dword.     }
{ These fixups are perfomed by TAppISR.MakeHook.                            }

const
  glueFixup = 9; { Position of the [@@Self] in mov edi,dword ptr cs:[@@Self]}

procedure AppIsrGlue;                                 { Tested by: HookIntr }
asm
@@Self:     dd    0                           { Patched to @Self at runtime }
@@GlueStrt: push  ds                          { First exectuted instruction }
            push  edi
            mov   edi,dword ptr cs:[@@Self]   { <-GlueFixup CS:EDI = @Self  }
            mov   ds,cs:[edi].TAppISR.DataSeg { DS:EDI = @Self              }
            xchg  eax,[esp]                   { EAX = EDI when called       }
            mov   [edi].TAppISR.Regs.&EDI,eax { [ESP] = EAX when called     }
            pop   eax                         { EAX = EAX when called       }
            jmp   [edi].TAppISR.CallAddr      { Jump to TAppISR.CallBack    }
end;

procedure AppIsrGlueEnd; { Dummy procedure used only to determine the size  }
begin                    { of AppIsrGlue.                                   }
end;

var
  GlueSize: Cardinal; { Size of the above glue code. Should really be const }

procedure TAppISR.CallBack; register;                 { Tested by: HookIntr }
{ DS:EDI = @Self }
{ At this point, the stack contains the following data:                     }
{                                                                           }
{  +08  Return EFlags                          -----+ Pushed by the IntXX   }
{  +04  Return CS                                   | or hardware IRQ       }
{  +00  Return EIP                             -----+                       }
{  -04  DS                                     -----+ pushed by AppIsrGlue  }
asm
            mov   [edi].TAppISR.Regs.&EAX,eax { Set Self.Regs from registers}
            pop   eax                         { AX = DS when called         }
            mov   [edi].TAppISR.Regs.&ESP,esp
            mov   [edi].TAppISR.Regs.&DS,ax
            mov   [edi].TAppISR.Regs.&EBX,ebx
            mov   [edi].TAppISR.Regs.&ECX,ecx
            mov   [edi].TAppISR.Regs.&EDX,edx
            mov   [edi].TAppISR.Regs.&ESI,esi
            mov   [edi].TAppISR.Regs.&EBP,ebp
         {  mov   eax,[edi].TAppISR.OldVec.Ofs }{ Set Regs.CS:EIP to OldVec }
            mov   [edi].TAppISR.Regs.&ES,es
         {  mov   [edi].TAppISR.Regs.&EIP,eax  }{ to allow chaining to the  }
            mov   [edi].TAppISR.Regs.&FS,fs
         {  mov   eax,dword ptr [edi].TAppISR.OldVec.Seg }{ original ISR    }
            mov   [edi].TAppISR.Regs.&GS,gs
         {  mov   [edi].TAppISR.Regs.&CS,ax    }
            mov   es,[edi].TAppISR.DataSeg    { ES = DS = Flat_DS           }

 { Switch to a larger Flat (SS = DS) stack }

            mov   edx,esp                   { Save the curent SS:ESP in     }
            mov   eax,ss                    { AX:EDX                        }
            db $0F,$B2,$25; dd StackTop     { lss   esp,fword ptr [StackTop]}
            sub   [StackTop].&Ofs,intStackSize { Claim this interrupt stack }
            push  eax                       { Save DpmiStack.Seg on Flat stk}
            push  edx                       { Save DpmiStack.Ofs on Flat stk}

            mov   eax,edi                   { EAX = @Self                   }
            cld
            call  dword ptr [edi].TInterrupt.UserISR { Call user-defined ISR}

 { Switch back to the DPMI-allocated stack }

            pop   edx                         { EDX = DpmiStack.Ofs         }
            pop   ss                          { SS  = DpmiStack.Seg         }
            mov   esp,edx                     { SS:ESP = DpmiStack          }
            add   [StackTop].&Ofs,intStackSize{ Release this interrupt stack}

            mov   eax,[edi].TAppISR.Regs.&EAX { Set registers from Self.Regs}
            mov   ebx,[edi].TAppISR.Regs.&EBX
            mov   edx,[edi].TAppISR.Regs.&EDX
            mov   ecx,[edi].TAppISR.Regs.&ECX
            mov   esi,[edi].TAppISR.Regs.&ESI
            push  [edi].TAppISR.Regs.&EDI     { Save Regs.EDI on the stack  }
            mov   ebp,[edi].TAppISR.Regs.&EBP
            mov   es,[edi].TAppISR.Regs.&ES
            mov   fs,[edi].TAppISR.Regs.&FS
            mov   gs,[edi].TAppISR.Regs.&GS
            mov   ds,[edi].TAppISR.Regs.&DS
            pop   edi                         { Pop the Regs.EDI value      }
            iretd
end;

constructor TAppISR.Create(IntNum: Byte; ASizeOfISR: Cardinal);
begin                                                 { Tested by: HookIntr }
  inherited Create(IntNum + itAppInterrupt, ASizeOfISR); { This is a hooked }
                                                         { p/m interrupt    }
  CallAddr := @TAppISR.CallBack;                 { Needed by the Glue code  }
  Self.DataSeg := WDosSysUtils.DataSeg;
  if not (MakeHook(true) and RegisterHook(Self))
   then raise EInterrupt.Create(ClassName + ' application ISR hook failed');
end;

procedure TAppISR.LockMem(Lock: TMemLock);            { Tested by: HookIntr }
begin
  inherited LockMem(Lock);
  MemLock(Glue, GlueSize, Lock);                 { Lock code & data memory  }
  MemLock(@TAppISR.CallBack,                     { Lock TAppISR.Callback    }
          PtrRec(@TAppISR.Create).Ofs - PtrRec(@TAppISR.CallBack).Ofs, Lock);
end;

function TAppISR.MakeHook(Make: Boolean): Boolean;    { Tested by: HookIntr }
begin
  if Make then
  begin
    GetMem(Glue, GlueSize);                      { Allocate unique glue code}
    Move(PByteArray(@AppIsrGlue)^, Glue^, GlueSize); { Copy AppIsrGlue to it}
    Glue^[0] := Self;                            { GlueCode@@Self = @Self   }
    PLongWord(@PByteArray(Glue)[glueFixup])^ := PtrRec(Glue).Ofs;{Fixup addr}
    LockMem(mlLock);
    Result := ReHook(@Glue^[1]);                 { Install the actual ISR   }
  end else
  begin
    Result := UnHook;                            { Unhook the actual ISR    }
    if Result then
    begin
      LockMem(mlUnLock);                         { Unlock the code & data   }
      FreeMem(Glue, GlueSize);                   { Free the unique glue code}
    end;
  end;
end;

function TAppISR.ReHook(CodePtr: Pointer): Boolean;   { Tested by: HookIntr }
begin                                                 { Hook a p/m interrupt}
  OldVec := GetIntVec(ID and $FF);
  ReHook := SetIntVec(ID and $FF, MakeFarCodePtr(CodePtr));
end;

function TAppISR.UnHook: Boolean;                     { Tested by: HookIntr }
begin                                            { Unhook the p/m Interrupt }
  UnHook := SetIntVec(ID and $FF, OldVec);
end;

{-------------------------------- TAppSoftISR ----------------------------- }

function TAppSoftISR.CallOldISR: Cardinal;            { Tested by: untested }
begin
  PFarPtr(@Regs.EIP)^ := OldVec;              { Regs.EIP:CS = orginal vector}
  Result := SoftIntr(Regs);                   { Call original handler       }
end;

{-------------------------------- TAppHardISR ----------------------------- }

procedure TAppHardISR.CallOldISR;                     { Tested by: untested }
asm
            pop   ecx                       { ECX = User return address     }
            pushfd                          { For IRET return               }
            push  cs                        { IRET return address selector  }
            push  ecx                       { IRET return address offset    }
            push  dword ptr [eax].TAppHardISR.OldVec.&Seg
            push  [eax].TAppHardISR.OldVec.&Ofs
            push  [eax].FarPtr.&Ofs
            retf  0                         { Call far OldVec               }
end;

{--------------------------------- TAppCallBack ----------------------------}

procedure TAppCallBack.CallBack; register;            { Tested by: CallBack }
{ DS:EDI = @Self }
{ At this point, the stack contains the following data:                     }
{                                                                           }
{  +04  Return CS                              -----+ Pushed by the callback}
{  +00  Return EIP                             -----+                       }
{  -04  DS                                     -----+ pushed by AppIsrGlue  }
asm  { Same code as TAppISR.Callback except for the last instruction (retf) }
{ DS:EDI = @Self }
{ At this point, the stack contains the following data:                     }
{                                                                           }
{  +08  Return EFlags                          -----+ Pushed by the IntXX   }
{  +04  Return CS                                   | or hardware IRQ       }
{  +00  Return EIP                             -----+                       }
{  -04  DS                                     -----+ pushed by AppIsrGlue  }
            mov   [edi].TAppISR.Regs.&EAX,eax { Set Self.Regs from registers}
            pop   eax                         { AX = DS when called         }
            mov   [edi].TAppISR.Regs.&ESP,esp
            mov   [edi].TAppISR.Regs.&DS,ax
            mov   [edi].TAppISR.Regs.&EBX,ebx
            mov   [edi].TAppISR.Regs.&ECX,ecx
            mov   [edi].TAppISR.Regs.&EDX,edx
            mov   [edi].TAppISR.Regs.&ESI,esi
            mov   [edi].TAppISR.Regs.&EBP,ebp
            mov   [edi].TAppISR.Regs.&ES,es
            mov   [edi].TAppISR.Regs.&FS,fs
            mov   [edi].TAppISR.Regs.&GS,gs
            mov   es,[edi].TAppISR.DataSeg    { ES = DS = Flat_DS           }

 { Switch to a larger Flat (SS = DS) stack }

            mov   edx,esp                   { Save the curent SS:ESP in     }
            mov   eax,ss                    { AX:EDX                        }
            db $0F,$B2,$25; dd StackTop     { lss   esp,fword ptr [StackTop]}
            sub   [StackTop].&Ofs,intStackSize { Claim this interrupt stack }
            push  eax                       { Save DpmiStack.Seg on Flat stk}
            push  edx                       { Save DpmiStack.Ofs on Flat stk}

            mov   eax,edi                   { EAX = @Self                   }
            cld
            call  dword ptr [edi].TInterrupt.UserISR { Call user-defined ISR}

 { Switch back to the DPMI-allocated stack }

            pop   edx                         { EDX = DpmiStack.Ofs         }
            pop   ss                          { SS  = DpmiStack.Seg         }
            mov   esp,edx                     { SS:ESP = DpmiStack          }
            add   [StackTop].&Ofs,intStackSize{ Release this interrupt stack}

            mov   eax,[edi].TAppISR.Regs.&EAX { Set registers from Self.Regs}
            mov   ebx,[edi].TAppISR.Regs.&EBX
            mov   edx,[edi].TAppISR.Regs.&EDX
            mov   ecx,[edi].TAppISR.Regs.&ECX
            mov   esi,[edi].TAppISR.Regs.&ESI
            push  [edi].TAppISR.Regs.&EDI     { Save Regs.EDI on the stack  }
            mov   ebp,[edi].TAppISR.Regs.&EBP
            mov   es,[edi].TAppISR.Regs.&ES
            mov   fs,[edi].TAppISR.Regs.&FS
            mov   gs,[edi].TAppISR.Regs.&GS
            mov   ds,[edi].TAppISR.Regs.&DS
            pop   edi                         { Pop the Regs.EDI value      }
            retf                              { Return to the callback      }
end;

constructor TAppCallBack.Create(ASizeOfISR: Cardinal);
begin                                                 { Tested by: CallBack }
  inherited Create(AppCallNum, ASizeOfISR);
  Inc(AppCallNum);
end;

procedure TAppCallBack.LockMem(Lock: TMemLock);       { Tested by: CallBack }
begin
  inherited LockMem(Lock);
  MemLock(@TAppCallBack.CallBack,              { Lock TAppCallBack.Callback }
   PtrRec(@TAppCallBack.Create).Ofs - PtrRec(@TAppCallBack.CallBack).Ofs, Lock);
end;

destructor TAppCallBack.Destroy;                      { Tested by: CallBack }
begin
  if WordRec(LongRec(ID).Lo).Lo = WordRec(LongRec(AppCallNum).Lo).Lo - 1
   then Dec(AppCallNum);
  inherited Destroy;
end;

function TAppCallBack.MakeHook(Make: Boolean): Boolean;
begin                                                 { Tested by: CallBack }
  CallAddr := @TAppCallBack.CallBack;             { Needed by the Glue code }
  MakeHook := inherited MakeHook(Make);
end;

{----------------- Unhook ISR/DosISR/DosCallback API functions -------------}

procedure SwapVectors;                                { Tested by: TestDpmi }
var                               { UnHook or ReHook all ISR's and callbacks}
  P: TInterrupt;                  { not tagged as unswapped.                }
begin
  P := HookList;
  while P <> nil do with P do
  begin
    if ID and itNoSwap = 0 then
    begin
      MakeHook(ID and itUnHook <> 0);
      ID := ID xor itUnHook;
    end;
    P := Next;
  end;
end;

procedure UnHookAll;                                  { Tested by: TestDPMI }
begin                                  { Unhook all interrupts and callbacks}
  while HookList <> nil do             { from the system and the linked-list}
    HookList.Free;
end;

{---- Translate a given IRQ number to its corresponding interrupt vector ---}

function IRQtoIntVec(IRQ: Byte): Byte;                { Tested by: HookIntr }
begin
  if IRQ < 8
   then IRQtoIntVec := IRQ + MasterPicBase
   else if IRQ < 16
         then IRQtoIntVec := IRQ - 8 + SlavePicBase
         else IRQtoIntVec := Byte(-1);
end;

{ Enable hardware interrupts }

procedure EnableInterrupts; register;
asm
            sti
end;

{ Disable hardware interrupts }

procedure DisableInterrupts; register;
asm
            cli
end;

{ hexadecimal conversion function }

function HexVal(X: Byte): Char;
begin
  case X of
    0..9:
      HexVal := Chr(X + Ord('0'));
    $A..$F:
      HexVal := Chr(X -10 + Ord('A'));
    else
      HexVal := '?';
  end;
end;

function WordToHex(X: Word): string;
var
  Place: Integer;
begin
  Result := '';
  Place := $1000;
  repeat
    Result := Result + HexVal(x div place);
    X := X mod Place;
    Place := Place div $10;
  until (place = $0000);
end;

function LongToHex(X: LongWord): string;
begin
  LongToHex := '$' + WordToHex(LongRec(X).Hi) + WordToHex(LongRec(X).Lo);
end;

function AddrToHex(X: Pointer): string;
begin
  AddrToHex := '$' + WordToHex(LongRec(PtrRec(X).Ofs).Hi) +
                     WordToHex(LongRec(PtrRec(X).Ofs).Lo);
end;

function FarToHex(X: FarPtr): string;
begin
  FarToHex := '$' + WordToHex(X.Seg) + ':$' +
              WordToHex(LongRec(X.Ofs).Hi) + WordToHex(LongRec(X.Ofs).Lo);
end;

function DosPtrToHex(X: DosPtr): string;
begin
  DosPtrToHex := '$' + WordToHex(DosPtrRec(X).Seg) +
                 ':$'+ WordToHex(DosPtrRec(X).Ofs);
end;

function DecVal(C: Char): Byte;
begin
  case C of
    '0'..'9':
       DecVal := Ord(C) - Ord('0');
    'A'..'F':
       DecVal := Ord(C) - Ord('A') + $0A;
    'a'..'f':
       DecVal := Ord(C) - Ord('a') + $0A;
    else
       DecVal := 0;
  end;
end;

function HexToWord(const S: string; var Value: Cardinal): Boolean;
var
  i    : Byte;
  Place: Word;
  IsOk : Boolean;
begin
  i     := Length(S);
  IsOk  := true;
  Place := 1;
  Value := 0;
  while (i > 0) and IsOk do
  begin
    IsOk := ((S[i]>='0') and (S[i] <= '9')) or
            ((S[i]>='a') and (s[i] <= 'f')) or
            ((S[i]>='A') and (S[i] <= 'F'));
    Inc(Value, Place * DecVal(S[i]));
    Dec(i);
    Place := place * $10;
  end;
  HexToWord := IsOk;
end;

{---------------------------------- TPort ----------------------------------}

function TPort.GetPort(Index: Word): Byte; register; assembler;
asm
            in    al,dx
end;

procedure TPort.SetPort(Index: Word; Data: Byte); register; assembler;
asm
            mov   eax,ecx
            out   dx,al
end;

{---------------------------------- TPortW ---------------------------------}

function TPortW.GetPortW(Index: Word): Word; register; assembler;
asm
            in    ax,dx
end;

procedure TPortW.SetPortW(Index: Word; Value: Word); register; assembler;
asm
            mov   eax,ecx
            out   dx,ax
end;

{---------------------------------- TPortL ---------------------------------}

function TPortL.GetPortL(Index: Word): LongInt; register; assembler;
asm
            in    eax,dx
end;

procedure TPortL.SetPortL(Index: Word; Value: Integer); register; assembler;
asm
            mov   eax,ecx
            out   dx,eax
end;


{---------------------------------- TMem -----------------------------------}

{$IFDEF SUPPORTS_NO_ABSOLUTE}

function TMem.GetMem(Index: LongInt): Byte; register; assembler;
asm
            mov   al,[edx]
end;

procedure TMem.SetMem(Index: LongInt; const Value: Byte); register; assembler;
asm
            mov   [edx],cl
end;

{$ENDIF SUPPORTS_NO_ABSOLUTE}

{---------------------------------- TMemW ----------------------------------}

function TMemW.GetMemW(Index: LongInt): Word; register; assembler;
asm
            mov   ax,[edx]
end;

procedure TMemW.SetMemW(Index: LongInt; const Value: Word); register; assembler;
asm
            mov   [edx],cx
end;

{---------------------------------- TMemL ----------------------------------}

function TMemL.GetMemL(Index: LongInt): LongInt; register; assembler;
asm
            mov   eax,[edx]
end;

procedure TMemL.SetMemL(Index: LongInt; const Value: LongInt); register;
    assembler;
asm
            mov   [edx],ecx
end;

{------------------------------ Unit initialization ------------------------}

initialization
begin
  asm
            mov   [DataSeg],ds
            mov   [CodeSeg],cs
            mov   [StackSeg],ss   { Initialize the hardware interrupt number}
            push  ebx             { to interrupt vector lookup table.       }
            mov   ax,dpmiGetInfo  { Get primary and secondary hardware      }
            int   intDPMI         { interrupt table offsets                 }
            xchg  dl,dh           { DL = Primary, DH = Secondary offset     }
            mov   [Dpmiflags],bx
            mov   word ptr [MasterPicBase],dx      { offset values.         }
            mov   eax,offset AppIsrGlueEnd         { Define GlueSize        }
            sub   eax,offset AppIsrGlue
            pop   ebx
            mov   [GlueSize],eax
  end;
  Port  := TPort.Create;           { Initialize the pseudo Port[] arrays.   }
  PortW := TPortW.Create;
  PortL := TPortL.Create;
{$IFDEF SUPPORTS_NO_ABSOLUTE}      { Initialize the pseudo Mem[] arrays.    }
  Mem   := TMem.Create;
{$ENDIF SUPPORTS_NO_ABSOLUTE}
  MemW  := TMemW.Create;
  MemL  := TMemL.Create;
end;

{------------------------------- Unit finalization -------------------------}

finalization
  PortL.Free;                { Free the pseudo Port[] arrays.               }
  PortW.Free;
  Port.Free;
  MemL.Free;                 { Free the pseudo Mem[] arrays.                }
  MemW.Free;
{$IFDEF SUPPORTS_NO_ABSOLUTE}
  Mem.Free;
{$ENDIF SUPPORTS_NO_ABSOLUTE}
  if Assigned(UnitCleanUp)   { Procedural var used to improve smart-linking }
   then UnitCleanUp;         { when no TInterrupt derivitives are used.     }
  DosBufDone;                { Release the DosBuf memory                    }
end.
